package wine.alert

import enums.*
import grails.converters.JSON
import grails.plugin.springsecurity.annotation.Secured
import grails.transaction.Transactional
import groovyx.gpars.GParsPool
import org.apache.commons.lang3.StringUtils
import utils.CharsetUtils
import wine.alert.enums.EliminationType
import wine.alert.enums.ReservedType
import wine.alert.enums.SubstitutionType
import wine.alert.model.DataResult
import wine.alert.model.WineAddition
import wine.alert.utils.HqlUtils
import wine.db.*
import wine.db.enums.Source
import wine.db.utils.WineKeywordUtils
import wine.db.utils.WineUtils

import javax.servlet.http.HttpServletResponse
import java.text.SimpleDateFormat
import java.util.concurrent.ConcurrentHashMap

@Secured(value=["hasRole('ROLE_ADMIN') or hasRole('ROLE_OPERATION')"])
@Transactional(readOnly = true)
class AppController {

    def springSecurityService

    def outputService
    def wineService
    def keywordsService
    def producerService
    def wineApiService

    @Secured(value=["hasRole('ROLE_ADMIN')"])
    def validation() {}

    @Secured(value=["hasRole('ROLE_ADMIN')"])
    def narrow() {}

    @Secured(value=["hasRole('ROLE_ADMIN')"])
    def addition() {}

    def wineSearcher() {}

    @Secured(value=["hasRole('ROLE_ADMIN')"])
    def noiseParser() {}

    @Secured(value=["hasRole('ROLE_ADMIN')"])
    def keywordGenerator() {}

    /**
     * This action will get all available {@link Retailer}
     * sorted by {@link Retailer#code}
     */
    def retailers() {
        def data = []
        SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd")
        outputService.getAvailableMatchedMap(MatchedType.KEYWORD)?.each { k, v ->
            data.push([code: k, date: sdf.format(v)])
        }
        data.sort { it.code }

        DataResult dataResult = new DataResult()
        dataResult.success = true
        dataResult.data = data
        render dataResult.result as JSON
    }

    /**
     * This action will get all {@link WorkGroup}
     * sorted by {@link WorkGroup#name} in ascending order
     */
    def workGroups() {
        def groups = WorkGroup.createCriteria().list {
            projections {
                property('name')
            }
            order("name", "asc")
        }

        DataResult dataResult = new DataResult()
        dataResult.success = true
        dataResult.data = groups?.collect {[code: it, date: 'Group']}
        render dataResult.result as JSON
    }

    /**
     * This action will get all {@link BottleSize}
     * sorted by {@link BottleSize#name} in ascending order
     */
    def bottleSizes() {
        def bottleList = BottleSize.wineDatabase.createCriteria().list {
            projections {
                property('name')
            }
            order("name", "asc")
        }

        DataResult dataResult = new DataResult()
        dataResult.success = true
        dataResult.data = bottleList

        render dataResult.result as JSON
    }

    /**
     * This action will get all {@link Vintage}
     * sorted by {@link Vintage#name} in ascending order
     */
    def vintages() {
        DataResult dataResult = new DataResult()

        def vintageList = Vintage.wineDatabase.createCriteria().list {
            projections {
                property('name')
            }
            order("name", "asc")
        }

        dataResult.success = true
        dataResult.data = vintageList

        render dataResult.result as JSON
    }

    /**
     * This action will get all reverse {@link WorkGroup}
     * sorted by {@link WorkGroup#name} in ascending order
     */
    def vintagesReverseMap() {
        DataResult dataResult = new DataResult()

        def vintageList = Vintage.wineDatabase.createCriteria().list {
            projections {
                property('id')
                property('name')
            }
            order("name", "asc")
        }

        def vtMap = new HashMap()
        vintageList?.each {
            vtMap.put(it.getAt(1), it.getAt(0))
        }

        dataResult.success = true
        dataResult.data = vtMap

        render dataResult.result as JSON
    }

    /**
     * This action will get all {@link WineDb#type}
     * sorted by {@link WineDb#type} in ascending order
     */
    def types() {
        DataResult dataResult = new DataResult()

        def typeList = WineDb.wineDatabase.createCriteria().list {
            projections {
                distinct('type')
            }
            and {
                isNotNull("type")
                ne("type", "")
            }
            order("type", "asc")
        }
        typeList?.removeAll([null])

        dataResult.success = true
        dataResult.data = [""] + typeList

        render dataResult.result as JSON
    }

    /**
     * This action will get all {@link WineDb#color}
     * sorted by {@link WineDb#color} in ascending order
     */
    def colors() {
        DataResult dataResult = new DataResult()

        def colorList = WineDb.wineDatabase.createCriteria().list {
            projections {
                distinct('color')
            }
            and {
                isNotNull("color")
                ne("color", "")
            }
            order("color", "asc")
        }
        colorList?.removeAll([null])

        dataResult.success = true
        dataResult.data = colorList

        render dataResult.result as JSON
    }

    /**
     * This action will get all {@link Variety}
     * sorted by {@link Variety#name} in ascending order
     */
    def varieties() {
        DataResult dataResult = new DataResult()
        def varietyList = []

        def varieties = Variety.wineDatabase.createCriteria().list {
            projections {
                distinct(['name', 'defaultColor'])
            }
            order("name", "asc")
        }

        varieties?.each {
            varietyList.push([name: it[0], defaultColor: it[1]])
        }

        dataResult.success = true
        dataResult.data = varietyList

        render dataResult.result as JSON
    }

    /**
     * This action will get all {@link WineDb#sweetness}
     * sorted by {@link WineDb#sweetness} in ascending order
     */
    def sweetness() {
        DataResult dataResult = new DataResult()

        def sweetnessList = WineDb.wineDatabase.createCriteria().list {
            projections {
                distinct('sweetness')
            }
            and {
                isNotNull("sweetness")
                ne("sweetness", "")
            }
            order("sweetness", "asc")
        }
        sweetnessList?.removeAll([null])

        dataResult.success = true
        dataResult.data = [""] + sweetnessList

        render dataResult.result as JSON
    }

    /**
     * This action will get all {@link RegionCode}
     * sorted by {@link RegionCode#id} in ascending order
     */
    def regionCodes() {
        DataResult dataResult = new DataResult()

        def codeList = []
        def rcs = RegionCode.wineDatabase.createCriteria().list {
            projections {
                property('id')
            }
            order("id", "asc")
        }
        rcs?.each {
            codeList.push(it)
        }

        dataResult.success = true
        dataResult.data = codeList

        render dataResult.result as JSON
    }

    def searchRegions() {
        DataResult dataResult = new DataResult()

        def jsonReq = request.JSON
        def words = jsonReq.words?.toList()
        def regions = jsonReq.regions?.toList()
        def partialSearch = jsonReq.searchPartial

        def regList = []
        if (words?.size() > 0 || regions?.size() > 0) {
            StringBuilder sb = new StringBuilder("SELECT DISTINCT r.id, w.region FROM WineAppellation w INNER JOIN w.regionCode r")
            List<String> rsRoot = new ArrayList<>()
            if (regions?.size() > 0) { // Locked regions
                rsRoot.add("r.id IN(:regions)")
            }
            if (words?.size() > 0) {
                List<String> ws = new ArrayList<>()
                words?.eachWithIndex { String word, int i ->
                    word = HqlUtils.escapeHQLParam(word)
                    StringBuilder sbw = new StringBuilder()
                    if (partialSearch) {
                        sbw.append(" w.region LIKE '%").append(word).append("%'")
                    } else {
                        sbw.append(" (")
                        sbw.append(" w.region = '").append(word).append("'")              // single word, eg: "Montana"
                        sbw.append(" OR w.region LIKE '").append(word).append(" %'")      // word at start, eg: "Riefle" in "Riefle Pinot Blanc"
                        sbw.append(" OR w.region LIKE '% ").append(word).append(" %'")    // word at middle, eg: "Pinot" in "Riefle Pinot Blanc"
                        sbw.append(" OR w.region LIKE '% ").append(word).append("'")      // word at end, eg: "Blanc" in "Riefle Pinot Blanc"
                        sbw.append(" OR w.region LIKE '").append(word).append("-%'")      // word at start (first word in hyphen concatenated string), eg: "Jean-" in "Jean-Baptiste Adam Pinot Blanc"
                        sbw.append(" OR w.region LIKE '% ").append(word).append("-%'")    // word at middle (first word in hyphen concatenated string), eg: "Domaine Barmes-" in "Domaine Barmes-Buecher Pinot Auxerrois"
                        sbw.append(" OR w.region LIKE '% (").append(word).append("-%'")   // word at middle (first word in hyphen concatenated string), eg: "Haut-" in "d'Arche (Haut-Medoc) Red"
                        sbw.append(" OR w.region LIKE '%-").append(word).append(" %'")    // word at middle (last word in hyphen concatenated string), eg: "-Buecher Pinot Auxerrois" in "Domaine Barmes-Buecher Pinot Auxerrois"
                        sbw.append(" OR w.region LIKE '%-").append(word).append(") %'")   // word at middle (last word in hyphen concatenated string), eg: "-Medoc" in "d'Arche (Haut-Medoc) Red"
                        sbw.append(" OR w.region LIKE '%-").append(word).append("'")      // word at end (last word in hyphen concatenated string), eg: "-Mendocino" in "Babcock Pinot Noir Tri-Counties Cuvee S. Barbara-Sonoma-Mendocino"
                        sbw.append(" OR w.region LIKE '%-").append(word).append("-%'")    // word at middle (middle word in hyphen concatenated string), eg: "-Sonoma-"
                        sbw.append(" OR w.region LIKE '").append(word).append("''%'")     // word at start, eg: "d" in "d'Alsace Reserve"
                        sbw.append(" OR w.region LIKE '% ").append(word).append("''%'")   // word at middle or end, eg: "d" in "Riesling d'Alsace Reserve", "Riesling d'Alsace"
                        sbw.append(" OR w.region LIKE '%''").append(word).append(" %'")   // word at start r middle, eg: "Alsace" in "d'Alsace Reserve", "Riesling d'Alsace Reserve"
                        sbw.append(" OR w.region LIKE '%''").append(word).append("'")     // word at end, eg: "Alsace" in "Riesling d'Alsace"
                        sbw.append(" )")
                    }
                    ws.push(sbw.toString())
                }
                StringBuilder sbws = new StringBuilder("(").append(ws.join(" AND ")).append(")")
                rsRoot.add(sbws.toString())
            }
            if (rsRoot.size() > 0) {
                sb.append(" WHERE ").append(rsRoot.join(" OR "))
            }
            sb.append(" ORDER BY w.region ASC")

            def queryParams = [:]
            if (regions?.size() > 0) queryParams.put("regions", regions)

            def regionList = WineAppellation.wineDatabase.executeQuery(sb.toString(), queryParams, [max: 300])
            regionList?.each {
                regList.push([id: it[0], name: it[1]])
            }
        }

        dataResult.success = true
        dataResult.data = regList

        render dataResult.result as JSON
    }

    def producers() {
        DataResult dataResult = new DataResult()

        def jsonReq = request.JSON
        def filterID = jsonReq.filterID
        def words = jsonReq.words?.toList()
        def producers = jsonReq.producers?.toList()
        String regCode = jsonReq.regionCode
        String producerId = jsonReq.producerId
        def regions = jsonReq.regions?.toList()
        def partialSearch = jsonReq.searchPartial

        def prodList = []
        if (words?.size() > 0 || regCode || regions?.size() > 0 || producerId || filterID || producers?.size() > 0) {
            StringBuilder sb = new StringBuilder("SELECT id, nameShow FROM Producer")
            List<String> rsRoot = new ArrayList<>()
            if (producerId) rsRoot.add("id = :producerId")
            if (filterID ) {
                List<String> rsIds = new ArrayList<>()
                rsIds.add("id LIKE :filterID")
                if (producers?.size() > 0) { // Locked producers
                    rsIds.add("id IN(:producers)")
                }
                rsRoot.add('(' + rsIds.join(' OR ') + ')')
            }
            if (regCode) { // From wine editor
                StringBuilder sbrc = new StringBuilder("id LIKE '")
                        .append(HqlUtils.escapeHQLParam(regCode)).append("%'")
                rsRoot.add(sbrc.toString())
            }
            if (regions?.size() > 0) { // Locked regions
                List<String> rcs = new ArrayList<>()
                regions?.each { String word ->
                    StringBuilder sbrc = new StringBuilder("id LIKE '")
                            .append(HqlUtils.escapeHQLParam(word)).append("%'")
                    rcs.push(sbrc.toString())
                }
                StringBuilder sbrc = new StringBuilder("(").append(rcs.join(" OR ")).append(")")
                rsRoot.add(sbrc.toString())
            }
            if (producers?.size() > 0 || words?.size() > 0) {
                List<String> rsPwr = new ArrayList<>()
                if (producers?.size() > 0) { // Locked producers
                    rsPwr.add("id IN(:producers)")
                }
                if (words?.size() > 0) {
                    List<String> ws = new ArrayList<>()
                    words?.each { String word ->
                        word = HqlUtils.escapeHQLParam(word)
                        StringBuilder sbw = new StringBuilder()
                        if (partialSearch) {
                            sbw.append(" nameShow LIKE '%").append(word).append("%'")
                        } else {
                            sbw.append(" (")
                            sbw.append(" nameShow = '").append(word).append("'")              // single word, eg: "Montana"
                            sbw.append(" OR nameShow LIKE '").append(word).append(" %'")      // word at start, eg: "Riefle" in "Riefle Pinot Blanc"
                            sbw.append(" OR nameShow LIKE '% ").append(word).append(" %'")    // word at middle, eg: "Pinot" in "Riefle Pinot Blanc"
                            sbw.append(" OR nameShow LIKE '% ").append(word).append("'")      // word at end, eg: "Blanc" in "Riefle Pinot Blanc"
                            sbw.append(" OR nameShow LIKE '").append(word).append("-%'")      // word at start (first word in hyphen concatenated string), eg: "Jean-" in "Jean-Baptiste Adam Pinot Blanc"
                            sbw.append(" OR nameShow LIKE '% ").append(word).append("-%'")    // word at middle (first word in hyphen concatenated string), eg: "Domaine Barmes-" in "Domaine Barmes-Buecher Pinot Auxerrois"
                            sbw.append(" OR nameShow LIKE '% (").append(word).append("-%'")   // word at middle (first word in hyphen concatenated string), eg: "Haut-" in "d'Arche (Haut-Medoc) Red"
                            sbw.append(" OR nameShow LIKE '%-").append(word).append(" %'")    // word at middle (last word in hyphen concatenated string), eg: "-Buecher Pinot Auxerrois" in "Domaine Barmes-Buecher Pinot Auxerrois"
                            sbw.append(" OR nameShow LIKE '%-").append(word).append(") %'")   // word at middle (last word in hyphen concatenated string), eg: "-Medoc" in "d'Arche (Haut-Medoc) Red"
                            sbw.append(" OR nameShow LIKE '%-").append(word).append("'")      // word at end (last word in hyphen concatenated string), eg: "-Mendocino" in "Babcock Pinot Noir Tri-Counties Cuvee S. Barbara-Sonoma-Mendocino"
                            sbw.append(" OR nameShow LIKE '%-").append(word).append("-%'")    // word at middle (middle word in hyphen concatenated string), eg: "-Sonoma-"
                            sbw.append(" OR nameShow LIKE '").append(word).append("''%'")     // word at start, eg: "d" in "d'Alsace Reserve"
                            sbw.append(" OR nameShow LIKE '% ").append(word).append("''%'")   // word at middle or end, eg: "d" in "Riesling d'Alsace Reserve", "Riesling d'Alsace"
                            sbw.append(" OR nameShow LIKE '%''").append(word).append(" %'")   // word at start r middle, eg: "Alsace" in "d'Alsace Reserve", "Riesling d'Alsace Reserve"
                            sbw.append(" OR nameShow LIKE '%''").append(word).append("'")     // word at end, eg: "Alsace" in "Riesling d'Alsace"
                            sbw.append(" )")
                        }
                        ws.push(sbw.toString())
                    }
                    StringBuilder sbws = new StringBuilder("(").append(ws.join(" AND ")).append(")")
                    rsPwr.add(sbws.toString())
                }
                if (rsPwr.size() > 0) {
                    rsRoot.add('(' + rsPwr.join(" OR ") + ')')
                }
            }
            if (rsRoot.size() > 0) {
                sb.append(" WHERE ").append(rsRoot.join(" AND "))
            }
            sb.append(" ORDER BY nameShow ASC")

            def queryParams = [:]
            if (producerId) queryParams.put('producerId', producerId)
            if (filterID) queryParams.put('filterID', filterID + '%')
            if (producers?.size() > 0) queryParams.put("producers", producers)

            def producerList = Producer.wineDatabase.executeQuery(sb.toString(), queryParams, [max: 300])
            producerList?.each {
                prodList.push([id: it[0], name: it[1]])
            }
        }

        dataResult.success = true
        dataResult.data = prodList

        render dataResult.result as JSON
    }

    def wineDbs() {
        DataResult dataResult = new DataResult()

        def jsonReq = request.JSON
        def filterID = jsonReq.filterID
        def words = jsonReq.words?.toList()
        def regions = jsonReq.regions?.toList()
        def producers = jsonReq.producers?.toList()
        def partialSearch = jsonReq.searchPartial
        def wineDBList = []

        if (words?.size() > 0 || regions?.size() > 0 || producers?.size() > 0 || filterID) {
            StringBuilder sb = new StringBuilder("""SELECT
                w.id,
                w.name,
                w.color,
                p.nameShow,
                k.pattern,
                w.variety,
                w.note,
                w.wineCountry,
                w.wineRegion,
                w.wineLocation,
                w.wineLocale,
                w.wineSite
                FROM WineDb w INNER JOIN w.producer p LEFT OUTER JOIN w.wineKeyword k""")
            List<String> rsRoot = new ArrayList<>()
            rsRoot.add("w.status=:status")
            if (filterID) {
                List<String> rsIds = new ArrayList<>()
                if (filterID) rsIds.add("w.id LIKE :filterID")
                if (producers?.size() > 0) { // Locked producecked producers
                    rsIds.add("p.id IN(:producers)")
                }
                rsRoot.add('(' + rsIds.join(' OR ') + ')')
            }
            if (regions?.size() > 0) { // Locked regions
                List<String> rcs = new ArrayList<>()
                regions?.each { String word ->
                    StringBuilder sbrc = new StringBuilder("p.id LIKE '")
                            .append(HqlUtils.escapeHQLParam(word)).append("%'")
                    rcs.push(sbrc.toString())
                }
                StringBuilder sbrc = new StringBuilder("(").append(rcs.join(" OR ")).append(")")
                rsRoot.add(sbrc.toString())
            }
            if (producers?.size() > 0) { // Locked producers
                rsRoot.add("p.id in(:producers)")
            }
            if (words?.size() > 0) {
                List<String> ws = new ArrayList<>()
                words?.each { String word ->
                    word = HqlUtils.escapeHQLParam(word)
                    StringBuilder sbw = new StringBuilder()
                    if (partialSearch) {
                        sbw.append(" concat_ws(' ', p.nameShow, w.name) LIKE '%" + word + "%' ")
                    } else {
                        sbw.append(" (")
                        sbw.append(" concat_ws(' ', p.nameShow, w.name) = '").append(word).append("'")              // single word, eg: "Montana"
                        sbw.append(" OR concat_ws(' ', p.nameShow, w.name) LIKE '").append(word).append(" %'")      // word at start, eg: "Riefle" in "Riefle Pinot Blanc"
                        sbw.append(" OR concat_ws(' ', p.nameShow, w.name) LIKE '% ").append(word).append(" %'")    // word at middle, eg: "Pinot" in "Riefle Pinot Blanc"
                        sbw.append(" OR concat_ws(' ', p.nameShow, w.name) LIKE '% ").append(word).append("'")      // word at end, eg: "Blanc" in "Riefle Pinot Blanc"
                        sbw.append(" OR concat_ws(' ', p.nameShow, w.name) LIKE '").append(word).append("-%'")      // word at start (first word in hyphen concatenated string), eg: "Jean-" in "Jean-Baptiste Adam Pinot Blanc"
                        sbw.append(" OR concat_ws(' ', p.nameShow, w.name) LIKE '% ").append(word).append("-%'")    // word at middle (first word in hyphen concatenated string), eg: "Domaine Barmes-" in "Domaine Barmes-Buecher Pinot Auxerrois"
                        sbw.append(" OR concat_ws(' ', p.nameShow, w.name) LIKE '% (").append(word).append("-%'")   // word at middle (first word in hyphen concatenated string), eg: "Haut-" in "d'Arche (Haut-Medoc) Red"
                        sbw.append(" OR concat_ws(' ', p.nameShow, w.name) LIKE '%-").append(word).append(" %'")    // word at middle (last word in hyphen concatenated string), eg: "-Buecher Pinot Auxerrois" in "Domaine Barmes-Buecher Pinot Auxerrois"
                        sbw.append(" OR concat_ws(' ', p.nameShow, w.name) LIKE '%-").append(word).append(") %'")   // word at middle (last word in hyphen concatenated string), eg: "-Medoc" in "d'Arche (Haut-Medoc) Red"
                        sbw.append(" OR concat_ws(' ', p.nameShow, w.name) LIKE '%-").append(word).append("'")      // word at end (last word in hyphen concatenated string), eg: "-Mendocino" in "Babcock Pinot Noir Tri-Counties Cuvee S. Barbara-Sonoma-Mendocino"
                        sbw.append(" OR concat_ws(' ', p.nameShow, w.name) LIKE '%-").append(word).append("-%'")    // word at middle (middle word in hyphen concatenated string), eg: "-Sonoma-"
                        sbw.append(" OR concat_ws(' ', p.nameShow, w.name) LIKE '").append(word).append("''%'")     // word at start, eg: "d" in "d'Alsace Reserve"
                        sbw.append(" OR concat_ws(' ', p.nameShow, w.name) LIKE '% ").append(word).append("''%'")   // word at middle or end, eg: "d" in "Riesling d'Alsace Reserve", "Riesling d'Alsace"
                        sbw.append(" OR concat_ws(' ', p.nameShow, w.name) LIKE '%''").append(word).append(" %'")   // word at start r middle, eg: "Alsace" in "d'Alsace Reserve", "Riesling d'Alsace Reserve"
                        sbw.append(" OR concat_ws(' ', p.nameShow, w.name) LIKE '%''").append(word).append("'")     // word at end, eg: "Alsace" in "Riesling d'Alsace"
                        sbw.append(" )")
                    }
                    ws.push(sbw.toString())
                }
                StringBuilder sbws = new StringBuilder("(").append(ws.join(" AND ")).append(")")
                rsRoot.add(sbws.toString())
            }
            if (rsRoot.size() > 0) {
                sb.append(" WHERE ").append(rsRoot.join(" AND "))
            }
            sb.append(" ORDER BY w.id")

            def queryParams = [:]
            queryParams.put("status", WineDb.Status.ACTIVE)
            if (filterID) queryParams.put('filterID', filterID + '%')
            if (producers?.size() > 0) queryParams.put("producers", producers)

            def wineList = WineDb.wineDatabase.executeQuery(sb.toString(), queryParams, [max: 300])
            wineList?.each {
                wineDBList.push([
                        id: it[0],
                        name: WineUtils.getDisplayName((String) it[3], (String) it[1]),
                        color: it[2],
                        pattern: it[4],
                        variety: it[5],
                        waNote: it[6],
                        country: it[7],
                        region: it[8],
                        location: it[9],
                        locale: it[10],
                        site: it[11]
                ])
            }
        }

        dataResult.success = true
        dataResult.data = wineDBList

        render dataResult.result as JSON
    }

    def wineDbDetails() {
        DataResult dataResult = new DataResult()

        WineDb wineDb = WineDb.wineDatabase.findById(params?.id)
        if (!wineDb) {
            dataResult = new DataResult(message: "Wine not found", success: false)
            if (!dataResult.success) response.status = 400
            render dataResult.result as JSON
            return
        }

        def data = [:]
        data.id = wineDb.id
        data.wineName = wineDb.name ?: ''
        data.wineDescription = wineDb.description ?: ''
        data.variety = wineDb.variety ?: ''
        data.color = wineDb.color ?: ''
        data.sweetness = wineDb.sweetness ?: ''
        data.wineType = wineDb.type ?: ''
        data.note = wineDb.note ?: ''
        data.producer = wineDb.producer?.id
        data.producerName = wineDb.producer?.name ?: ''
        data.producerShow = wineDb.producer?.nameShow ?: ''

        data.country = wineDb.wineCountry ?: ''
        data.region = wineDb.wineRegion ?: ''
        data.location = wineDb.wineLocation ?: ''
        data.locale = wineDb.wineLocale ?: ''
        data.site = wineDb.wineSite ?: ''

        if (wineDb.wineKeyword) {
            data.pattern = wineDb.wineKeyword.pattern
            data.producerKeyword = wineDb.wineKeyword.producerKeyword ?: ''
            data.baseKeyword = wineDb.wineKeyword.baseKeyword
            data.appellationKeyword = wineDb.wineKeyword.appellationKeyword
            data.colorKeyword = wineDb.wineKeyword.colorKeyword
            data.commonModifiersKeyword = wineDb.wineKeyword.commonModifiersKeyword
            data.modifiersKeyword = wineDb.wineKeyword.modifiersKeyword
            data.customNotKeyword = wineDb.wineKeyword.customNotKeyword
            data.globalNotKeyword = wineDb.wineKeyword.globalNotKeyword
        } else {
            data.producerKeyword = wineDb.producer?.producerKeyword?.keyword ?: ''
        }

        data.noises = wineDb.wineNoises?.collect { it.noise }?.sort { it.length() }?.reverse()
        data.producerNoises = wineDb.producer?.producerNoises?.collect { it.noise }?.sort { it.length() }?.reverse()

        dataResult.success = true
        dataResult.data = data

        render dataResult.result as JSON
    }

    def producerWines() {
        DataResult dataResult = new DataResult()

        def jsonReq = request.JSON
        String producerId = jsonReq.producer

        def wineDBList = []
        if (producerId) {

            String query = "select w.id, (SELECT pattern FROM WineKeyword k WHERE k.wineDb=w) from WineDb w join w.producer p where p.id=:producer AND w.status=:status order by w.id"
            def wineList = WineDb.wineDatabase.executeQuery(query, [producer: producerId, status: WineDb.Status.ACTIVE], [max: 300])

            wineList?.each {
                WineDb wineDb = WineDb.wineDatabase.findById(it?.getAt(0))
                wineDBList.push([id: it[0], description: wineDb?.getDescription(), pattern: it[1]])
            }
        }

        dataResult.success = true
        dataResult.data = wineDBList

        render dataResult.result as JSON
    }

    /**
     * Gets Country list.
     * @return the list of Country
     */
    def countries() {
        DataResult dataResult = new DataResult()

        def jsonReq = request.JSON
        String region = jsonReq.region

        def countryList = WineAppellation.wineDatabase.createCriteria().list {
            projections {
                distinct('country')
            }
            and {
                if (region) eq("region", region)
                isNotNull("country")
                ne("country", "")
            }
            order("country", "asc")
        }

        dataResult.success = true
        dataResult.data = countryList

        render dataResult.result as JSON
    }

    /**
     * Gets unique Region (location).<br>
     *     [{country: "", region: ""}]
     * @return the list of Region
     */
    def regions() {
        DataResult dataResult = new DataResult()

        def regionList = WineAppellation.wineDatabase.createCriteria().list {
            projections {
                distinct(['region','country'])
            }
            and {
                isNotNull("region")
                ne("region", "")
            }
            order("region", "asc")
        }

        dataResult.success = true
        dataResult.data = regionList?.collect{[region: it[0], country: it[1]]}

        render dataResult.result as JSON
    }

    /**
     * Gets unique Sub-region (location).<br>
     *     [{country: "", region: "", location: ""}]
     * @return the list of Sub-region
     */
    def locations() {
        DataResult dataResult = new DataResult()

        def locationList = WineAppellation.wineDatabase.createCriteria().list {
            projections {
                distinct(['location','region','country'])
            }
            and {
                isNotNull("location")
                ne("location", "")
            }
            order("location", "asc")
        }

        dataResult.success = true
        dataResult.data = locationList?.collect{[location: it[0], region: it[1], country: it[2]]}

        render dataResult.result as JSON
    }

    /**
     * Gets unique Appellation (locale).<br>
     *     [{country: "", region: "", location: "", locale: ""}]
     * @return the list of Appellation
     */
    def locales() {
        DataResult dataResult = new DataResult()

        def localeList = WineAppellation.wineDatabase.createCriteria().list {
            projections {
                distinct(['locale', 'location','region','country'])
            }
            and {
                isNotNull("locale")
                ne("locale", "")
            }
            order("locale", "asc")
        }

        dataResult.success = true
        dataResult.data = localeList?.collect{[
                locale: it[0], location: it[1], region: it[2], country: it[3]
        ]}

        render dataResult.result as JSON
    }

    /**
     * Gets unique Sub-appellation (site).<br>
     *     [{country: "", region: "", location: "", locale: "", site: ""}]
     * @return the list of Sub-appellation
     */
    def sites() {
        DataResult dataResult = new DataResult()

        def siteList = WineAppellation.wineDatabase.createCriteria().list {
            projections {
                distinct(['site', 'locale', 'location','region','country'])
            }
            and {
                isNotNull("site")
                ne("site", "")
            }
            order("site", "asc")
        }

        dataResult.success = true
        dataResult.data = siteList?.collect{[
                site: it[0], locale: it[1], location: it[2], region: it[3], country: it[4]
        ]}

        render dataResult.result as JSON
    }

    /**
     * Gets Appellation-Tree list. Format:<br>
     *     [{name: "", type: "", appellation: {country: "", region: "", location: "", locale: "", site: "", regionCode: ""}}]
     * @return the list of Appellation-Tree
     */
    def appellations() {
        DataResult dataResult = new DataResult()
        def appellationList = []

        List<WineAppellation> appList = WineAppellation.wineDatabase.createCriteria().list {
            order("country", "asc")
            order("region", "asc")
            order("location", "asc")
            order("locale", "asc")
            order("site", "asc")
        }
        appList?.each {
            def app = [country: it.country, region: it.region, location: it.location, locale: it.locale, site: it.site, regionCode: it.regionCode?.id]
            if (it.site) {
                appellationList.push([name: it.site, type: "Sub Appellation", appellation: app])
            } else if (it.locale) {
                appellationList.push([name: it.locale, type: "Appellation", appellation: app])
            } else if (it.location) {
                appellationList.push([name: it.location, type: "Sub Region", appellation: app])
            } else if (it.region) {
                appellationList.push([name: it.region, type: "Region", appellation: app])
            } else {
                appellationList.push([name: it.country, type: "Country", appellation: app])
            }
        }

        dataResult.success = true
        dataResult.data = appellationList

        render dataResult.result as JSON
    }

    /**
     * Gets producer's noise of the given Producer ID.
     * @return the list of related producer's noise
     */
    def producerNoise() {
        DataResult dataResult = new DataResult()

        def noiseList = ProducerNoise.wineDatabase.createCriteria().list {
            createAlias("producer", "p")
            projections {
                property('noise')
            }
            eq("p.id", params?.producerId)
        }

        dataResult.success = true
        dataResult.data = noiseList?.collect {it?.toLowerCase()}

        render dataResult.result as JSON
    }

    /**
     * Gets all global noises.
     * @return the list of global noise
     */
    def globalNoises() {
        DataResult dataResult = new DataResult()

        def globalNoiseList = GlobalNoise.createCriteria().list {
            projections {
                property('noise')
            }
        }

        dataResult.success = true
        dataResult.data = globalNoiseList?.collect {it?.toLowerCase()}

        render dataResult.result as JSON
    }

    /**
     * Gets wine's noises of the given WineDb ID.
     * @return the list of related wine's noise
     */
    def wineNoise() {
        DataResult dataResult = new DataResult()

        def wineDbId = params?.wineDbId
        def noiseList = WineNoise.wineDatabase.createCriteria().list {
            projections {
                property('noise')
            }
            isNotNull("noise")
            ne("noise", "")
            wineDb {
                eq("id", wineDbId)
            }
        }

        dataResult.success = true
        dataResult.data = noiseList?.collect {it?.toLowerCase()}

        render dataResult.result as JSON
    }

    def producer() {
        DataResult dataResult = new DataResult()

        Producer producer = Producer.wineDatabase.findById(params?.id)
        if (!producer) {
            dataResult = new DataResult(message: "Producer not found", success: false)
            if (!dataResult.success) response.status = 400
            render dataResult.result as JSON
            return
        }

        def data = [:]
        data.id = producer.id
        data.name = producer.name ?: ''
        data.nameShow = producer.nameShow ?: ''
        data.keyword = producer?.producerKeyword?.keyword ?: ''
        data.producerNoises = producer?.producerNoises?.collect { it.noise }?.sort { it.length() }?.reverse()

        dataResult.success = true
        dataResult.data = data

        render dataResult.result as JSON
    }

    def reverseAliases() {
        DataResult dataResult = new DataResult()

        List<WordAlias> wineAliases = WordAlias.createCriteria().list {
            eq("aliasType", AliasType.WINE)
        } as List<WordAlias>

        def aliases = []
        wineAliases?.each { wa ->
            wa.aliases?.each { alias ->
                aliases.add([(alias?.toLowerCase()): wa.word?.toLowerCase()])
            }
        }

        dataResult.success = true
        dataResult.data = aliases

        render dataResult.result as JSON
    }

    def eliminations() {
        DataResult dataResult = new DataResult()
        EliminationType eliminationType = null
        if (params?.type) {
            eliminationType = EliminationType.valueOf((String) params?.type)
        }

        def eliminations = WordElimination.createCriteria().list {
            projections {
                property('word')
            }
            eq("eliminationType", eliminationType)
        }

        dataResult.success = true
        dataResult.data = eliminations

        render dataResult.result as JSON
    }

    def wordChanges() {
        DataResult dataResult = new DataResult()
        AliasType aliasType = null
        if (params?.type) {
            aliasType = AliasType.valueOf((String) params?.type)
        }

        def aliases = WordAlias.createCriteria().list {
            projections {
                property('word')
                property('alias')
            }
            eq("aliasType", aliasType)
        }

        dataResult.success = true
        dataResult.data = aliases?.collect {[(it[0]): it[1]]}

        render dataResult.result as JSON
    }

    def spreadTables() {
        DataResult dataResult = new DataResult()
        def spreadList = KeywordSpread.list()
        def regionGroups = spreadList?.groupBy {it.regionCode}

        def spreadsMap = [:]
        regionGroups?.each { rg ->
            def typeMap = [:]
            def typeGroups = rg.value?.groupBy {it.keywordType}
            typeGroups?.each { tg ->
                typeMap.put(tg?.key, tg?.value?.collect{it.keyword}?.sort {it.length()}?.reverse(true))
            }
            spreadsMap.put(rg.key, typeMap)
        }

        dataResult.success = true
        dataResult.data = spreadsMap

        render dataResult.result as JSON
    }

    def substitutions() {
        DataResult dataResult = new DataResult()
        SubstitutionType substitutionType = null
        if (params?.type) {
            substitutionType = SubstitutionType.valueOf((String) params?.type)
        }

        def substitutes = KeywordSubstitution.createCriteria().list {
            projections {
                property('keyword')
                property('substitution')
            }
            eq("substitutionType", substitutionType)
        }
        def subsMap = [:]
        substitutes?.sort{it[0]?.length()}?.reverse(true)?.each {
            subsMap.put(it[0], it[1])
        }

        dataResult.success = true
        dataResult.data = subsMap

        render dataResult.result as JSON
    }

    def reservedWords() {
        DataResult dataResult = new DataResult()
        ReservedType reservedType = null
        if (params?.type) {
            reservedType = ReservedType.valueOf((String) params?.type)
        }

        def words = WordReserved.createCriteria().list {
            projections {
                property('word')
            }
            eq("reservedType", reservedType)
        }

        dataResult.success = true
        dataResult.data = words

        render dataResult.result as JSON
    }

    def nextWineDbId() {
        DataResult dataResult = new DataResult()

        String wineDbId = ""
        if (params?.producerId) {
            def lastWineDbId = WineDb.wineDatabase.createCriteria().get {
                createAlias("producer", "p")
                projections {
                    property("id")
                }
                and {
                    not { like("id", "%xxx") }
                    eq("p.id", params?.producerId)
                }
                ne("id", "z")
                order("id", "desc")
                maxResults(1)
            }

            if (lastWineDbId) {
                String lastId = lastWineDbId?.drop(6)
                wineDbId = params?.producerId + (Integer.valueOf(lastId) + 1)
            } else {
                wineDbId = params?.producerId + "100"
            }
        } else if (params?.regionCode) {
            def lastProducerId = Producer.wineDatabase.createCriteria().get {
                projections {
                    property("id")
                }
                like("id", params?.regionCode + "%")
                ne("id", "z")
                order("id", "desc")
                maxResults(1)
            }

            if (lastProducerId) {
                String lastId = lastProducerId?.drop(2)
                wineDbId = params?.regionCode + (Integer.valueOf(lastId) + 1) + "100"
            } else {
                wineDbId = params?.regionCode + "1100" + "100"
            }
        }

        dataResult.success = true
        dataResult.data = wineDbId

        render dataResult.result as JSON
    }

    def checkAdditionMatched() {
        String mongoId = params?.mongoId
        if (!mongoId) {
            render new DataResult(success: true, data: [
                    id: "",
                    description: ""
            ]).result as JSON
            return
        }

        Wine wine = Wine.wineDatabase.createCriteria().get {
            eq("mongoId", mongoId)
            maxResults(1)
        }
        if (!wine) {
            // Faster than left join to "wineDuplications"
            WineDuplication wineDuplication = WineDuplication.wineDatabase.createCriteria().get {
                eq("mongoId", mongoId)
                maxResults(1)
            }
            if (!wineDuplication) {
                render new DataResult(success: true, data: [
                        id: "",
                        description: ""
                ]).result as JSON
                return
            }

            render new DataResult(success: true, data: [
                    id: wineDuplication.wine.id,
                    description: wineDuplication.wine.wineDb.getDescription()
            ]).result as JSON
            return
        }

        render new DataResult(success: true, data: [
                id: wine.id,
                description: wine.wineDb.getDescription()
        ]).result as JSON
    }

    /**
     * Populates unique retailer codes from retailer codes and groups
     * @return the list of unique retailers
     */
    def populateRetailers() {
        DataResult dataResult = new DataResult()

        def jsonReq = request.JSON
        def codes = jsonReq.codes

        List<String> joinedList = new ArrayList<>()
        def groups = WorkGroup.createCriteria().list {
            projections {
                property('name')
            }
            order("name", "asc")
        } as List<String>

        codes?.remove("all")
        codes?.each { ret ->
            if (groups && !groups.isEmpty() && groups.contains(ret)) {
                def rts = WorkGroupWebCrawler.createCriteria().list {
                    projections {
                        webCrawler {
                            property('code')
                        }
                    }
                    workGroup {
                        eq("name", ret)
                    }
                } as List<String>
                if (rts && !rts.isEmpty()) joinedList.addAll(rts)
            } else {
                return joinedList.add((String) ret)
            }
        }
        joinedList?.unique()

        dataResult.success = true
        dataResult.data = joinedList

        render dataResult.result as JSON
    }

    /**
     * Loads keyword match data using {@link OutputService#getKeywordMatches{String,String,Boolean,Integer} of the given retailer code.
     * @return the list of keyword match
     */
    def loadMatched() {
        DataResult dataResult = new DataResult()

        def jsonReq = request.JSON
        String retailer = jsonReq.retailer
        String matchedFilter = jsonReq.matchedFilter
        Integer offset = jsonReq.offset

        Boolean isNarrow = params?.boolean("narrow", false)
        def data = outputService.getKeywordMatches(retailer, matchedFilter, isNarrow, offset)

        dataResult.success = true
        dataResult.data = data

        render dataResult.result as JSON
    }

    /**
     * Handle upload external keyword match file.
     * @return the list of keyword match
     */
    def uploadMatched() {
        DataResult dataResult = new DataResult()
        String matchedFilter = params.matchedFilter

        // Get Image File
        def file = request.getFile('file')
        if (file.empty) {
            dataResult = new DataResult(false, "error", "Please select file to upload.")
            response.setStatus(HttpServletResponse.SC_BAD_REQUEST)
            render dataResult.result as JSON
            return
        }

        def isNarrow = params?.boolean("narrow")

        String delimiter =  Separator.PIPE.delimiter()
        String fileEncoding = CharsetUtils.detectInputStream(file.getInputStream()) ?: "UTF-8"

        def kwMatched = []
        String line
        file.getInputStream().withReader(fileEncoding) { reader ->
            while ((line = reader.readLine())!=null) {
                List<String> wineInfo = line?.split(delimiter)?.toList()
                String retailerCode = wineInfo?.getAt(0)?.trim()
                def map = outputService.parseKeywordMatch(retailerCode, matchedFilter, isNarrow, wineInfo.drop(1))
                kwMatched.push(map)
            }
        }
        if (kwMatched) kwMatched.removeIf {!it}

        dataResult.success = true
        dataResult.data = kwMatched

        render dataResult.result as JSON
    }

    /**
     * This action will process the uploaded file into a {@link WineAddition} list
     * @return
     */
    def uploadWinesAddition() {
        DataResult dataResult = new DataResult()

        // Get Image File
        def file = request.getFile('file')
        if (!file || file.empty) {
            dataResult = new DataResult(false, "error", "Please select file to upload.")
            response.setStatus(HttpServletResponse.SC_BAD_REQUEST)
            render dataResult.result as JSON
            return
        }

        List<WineAddition> wineAdditions = new ArrayList<>()
        int c = 0
        def line
        String separator = "\t"
        String fileEncoding = CharsetUtils.detectInputStream(file.getInputStream()) ?: "UTF-8"
        file.getInputStream().withReader(fileEncoding) { reader ->
            while ((line = reader.readLine())!=null) {
                c++
                if (!line || c == 1) continue
                List<String> wineInfo = line?.split(separator)?.toList()

                String vintage = wineInfo?.getAt(3)?.trim()
                if (vintage?.equalsIgnoreCase("N.V.")) vintage = "NV"

                wineAdditions.add(new WineAddition(
                        mongoId: wineInfo?.getAt(0),
                        wineVinId: wineInfo?.getAt(1),
                        regionCode: wineInfo?.getAt(2),
                        vintage: vintage,
                        variety: StringUtils.stripAccents(wineInfo?.getAt(4)),
                        label: StringUtils.stripAccents(wineInfo?.getAt(5)),
                        producerName: StringUtils.stripAccents(wineInfo?.getAt(6)),
                        producerShow: StringUtils.stripAccents(wineInfo?.getAt(7)),
                        country: StringUtils.stripAccents(wineInfo?.getAt(8)),
                        region: StringUtils.stripAccents(wineInfo?.getAt(9)),
                        location: StringUtils.stripAccents(wineInfo?.getAt(10)),
                        locale: StringUtils.stripAccents(wineInfo?.getAt(11)),
                        site: StringUtils.stripAccents(wineInfo?.getAt(12)),
                        sweetness: wineInfo?.getAt(13),
                        color: StringUtils.stripAccents(wineInfo?.getAt(14)),
                        wineType: wineInfo?.getAt(15)
                ))
            }
        }

        dataResult.success = true
        dataResult.data = wineAdditions

        render dataResult.result as JSON
    }

    def uploadWinesForNoises() {
        DataResult dataResult = new DataResult()

        // Get file
        def file = request.getFile('file')
        if (!file || file.empty) {
            dataResult = new DataResult(false, "error", "Please select file to upload.")
            response.setStatus(HttpServletResponse.SC_BAD_REQUEST)
            render dataResult.result as JSON
            return
        }

        def outputList = []
        def line
        String separator = "\t"
        file.getInputStream().withReader("ISO-8859-1") { reader ->
            while ((line = reader.readLine())!=null) {
                if (line) outputList.add(line)
            }
        }

        List winesForNoises = new ArrayList<>()
        if (outputList && !outputList.isEmpty()) {

            def wks = WineKeyword.wineDatabase.createCriteria().list {
                projections {
                    wineDb {
                        property("id")
                    }
                    property("pattern")
                }
            }
            def wksMap = new ConcurrentHashMap()
            wks?.each {
                wksMap.put(it[0], it[1])
            }

            /*def noises = WineNoise.wineDatabase.createCriteria().list {
                projections {
                    wineDb {
                        property("id")
                    }
                    property('noise')
                }
            }
            def noisesMap = new ConcurrentHashMap()
            noises?.each {
                def list = noisesMap.get(it[0])
                if (!list) list = new ArrayList()
                list.add(it[1])
                noisesMap.put(it[0], list)
            }*/

            GParsPool.withPool {
                winesForNoises = outputList.collectParallel { String info ->
                    List<String> wineInfo = info?.split(separator)?.toList()

                    def wk = wksMap.get(wineInfo?.getAt(0) ?: "")

                    String wineKeyword = wineInfo?.getAt(1)
                    wineKeyword = Calculator.cleanupNumericHtml(wineKeyword)
                    wineKeyword = keywordsService.replaceAliases(wineKeyword) // replace with aliases before cleanup

                    wineKeyword = keywordsService.removeSizeInfo(wineKeyword)
                    try {
                        if (wk) {
                            wineKeyword = Calculator.removeVintages(wineKeyword, (String) wk)
                        } else {
                            wineKeyword = Calculator.removeVintages(wineKeyword)
                        }
                    } catch (Exception e) {
                        log.error(e?.message, e)
                    }
                    wineKeyword = Calculator.cleanup(wineKeyword)
                    wineKeyword = keywordsService.replaceAliases(wineKeyword)

                    return [
                            wineDbId: wineInfo?.getAt(0),
                            description: wineInfo?.getAt(1),
                            pattern: wk,
                            wineKeyword: wineKeyword,
                            //noises: noisesMap.get(wineInfo?.getAt(0)),
                            noises: []
                    ]
                }
            }
        }

        dataResult.success = true
        dataResult.data = winesForNoises

        render dataResult.result as JSON
    }

    def uploadForKeywordGenerator() {
        DataResult dataResult = new DataResult()

        // Get file
        def file = request.getFile('file')
        if (!file || file.empty) {
            dataResult = new DataResult(false, "error", "Please select file to upload.")
            response.setStatus(HttpServletResponse.SC_BAD_REQUEST)
            render dataResult.result as JSON
            return
        }

        List<String> outputList = []
        String encoding = CharsetUtils.detectInputStream(file.getInputStream()) ?: "UTF-8"
        def line
        file.getInputStream().withReader(encoding) { reader ->
            while ((line = reader.readLine()) != null) {
                if (line) outputList.add(line)
            }
        }
        outputList?.removeAll { !it }

        List winesData = new ArrayList<>()
        if (outputList && outputList.size() > 1) {
            outputList.removeAt(0)
            def wines = WineDb.wineDatabase.createCriteria().list {
                projections {
                    property("id")
                    property("name")
                    property("color")
                    producer {
                        property("nameShow")
                        producerKeyword {
                            property("keyword")
                        }
                    }
                }
                'in'('id', outputList)
            }
            if (wines) {
                def noises = WineNoise.wineDatabase.createCriteria().list {
                    projections {
                        property("noise")
                        wineDb {
                            property("id")
                        }
                    }
                    wineDb {
                        'in'('id', outputList)
                    }
                }

                wines?.each { w ->
                    def noise = noises?.findAll { it[1] == w[0] }?.collect { it[0] }?.sort { it.length() }?.reverse()
                    winesData.push([
                            wineDbId       : w[0],
                            wineDbName     : w[1],
                            color          : w[2],
                            producerShow   : w[3],
                            producerKeyword: w[4],
                            noises         : noise
                    ])
                }
            }
        }

        dataResult.success = true
        dataResult.data = winesData

        render dataResult.result as JSON
    }

    /**
     * This action will save or update the {@link ValidationLog}
     * working state when user left from validation match
     * @return
     */
    @Deprecated
    @Secured(value=["hasRole('ROLE_ADMIN')"])
    @Transactional
    def saveWorkingState() {
        DataResult dataResult = new DataResult()

        def jsonReq = request.JSON
        def retailers = jsonReq.retailers
        String source = jsonReq.source
        String dateString = jsonReq.date
        String matchedFilter = jsonReq.matchedFilter
        Integer currentRecord = Integer.valueOf(jsonReq.currentRecord as String)

        Date date = null
        if (StringUtils.isNotBlank(dateString)) {
            try {
                date = new SimpleDateFormat("yyyy-MM-dd").parse(dateString)
            } catch (Exception e) {
                log.error(e?.message, e)
            }
        }

        // Get Working History
        String sortedRetailers = retailers?.sort{ it }?.join(",")
        ValidationLog validationLog = ValidationLog.createCriteria().get {
            and {
                if ("internal".equalsIgnoreCase(source)) {
                    if (retailers?.contains("all")) {
                        eq("retailers", "all")
                    } else {
                        eq("retailers", sortedRetailers)
                    }
                    eq("matchedDate", date)
                }
                eq("source", source)
                eq("matchedFilter", matchedFilter)
                eq("user", springSecurityService.currentUser)
            }
            order("lastUpdated", "desc")
            maxResults(1)
        } as ValidationLog

        if (!validationLog) {
            validationLog = new ValidationLog()
        }

        if ("internal".equalsIgnoreCase(source)) {
            validationLog.retailers = retailers?.contains("all") ? "all" : sortedRetailers
            validationLog.matchedDate = date
        }
        validationLog.source = source
        validationLog.matchedFilter = matchedFilter
        validationLog.currentRecord = currentRecord
        validationLog.user = springSecurityService.currentUser

        validationLog.validate()
        if (validationLog.hasErrors()) {
            transactionStatus.setRollbackOnly()

            dataResult = new DataResult(false, "error", "")
            dataResult.errors = validationLog.errors.allErrors.collect{message([error: it])}

            response.setStatus(HttpServletResponse.SC_BAD_REQUEST)
            render dataResult.result as JSON
            return
        }

        try {
            validationLog.save()
            dataResult.success = true
        } catch (Exception e) {
            transactionStatus.setRollbackOnly()
            log.error(e?.message, e)

            dataResult = new DataResult(false, "error", e.message)
            response.setStatus(HttpServletResponse.SC_BAD_REQUEST)
        }

        render dataResult.result as JSON
    }

    /**
     * This action will save new {@link GlobalNoise}
     * @return
     */
    @Secured(value=["hasRole('ROLE_ADMIN')"])
    @Transactional
    def saveGlobalNoises() {
        DataResult dataResult = new DataResult()

        def failedList = []

        def jsonReq = request.JSON
        def gNoises = jsonReq.noises?.toList()

        gNoises?.each { String noise ->
            def g = GlobalNoise.findByNoise(noise)
            if (g == null) {
                g = new GlobalNoise()
                g.noise = noise
                g.validate()
                if (!g.hasErrors()) {
                    try {
                        g.save(flash: true)
                    } catch (Exception e) {
                        failedList.push(noise)
                        log.error(e?.message, e)
                    }
                } else {
                    failedList.push(noise)
                }
            }
        }

        dataResult.success = true
        dataResult.data = [
                failedList: failedList
        ]
        dataResult.message = failedList?.size() > 0 ? message(code: 'wineMatch.saveAllFailed.message', default: 'Some noises have not been saved.') : ''
        render dataResult.result as JSON
    }

    /**
     * This action will save new {@link ProducerNoise}
     * @return
     */
    @Secured(value=["hasRole('ROLE_ADMIN')"])
    @Transactional
    def saveProducerNoises() {
        DataResult dataResult = new DataResult()

        def failedList = []

        def jsonReq = request.JSON
        def producerId = jsonReq.producerId
        def noises = jsonReq.noises?.toList()

        def producerDb = Producer.wineDatabase.findById(producerId)

        noises?.each { String noise ->
            def c = ProducerNoise.wineDatabase.createCriteria().count {
                createAlias("producer", "p")
                and {
                    eq("noise", noise)
                    eq("p.id", producerId)
                }
            }

            if (c < 1) {
                def g = new ProducerNoise()
                g.noise = noise
                g.producer = producerDb
                g.validate()
                if (!g.hasErrors()) {
                    try {
                        g.save()
                    } catch (Exception e) {
                        failedList.push(noise)
                        log.error(e?.message, e)
                    }
                } else {
                    failedList.push(noise)
                }
            }
        }

        dataResult.success = true
        dataResult.data = [
                failedList: failedList
        ]
        dataResult.message = failedList?.size() > 0 ? message(code: 'wineMatch.saveAllFailed.message', default: 'Some noises have not been saved.') : ''
        render dataResult.result as JSON
    }

    /**
     * This action will save new {@link WineNoise}
     * @return
     */
    @Secured(value=["hasRole('ROLE_ADMIN')"])
    @Transactional
    def saveWineNoises() {
        DataResult dataResult = new DataResult()

        def failedList = []

        def jsonReq = request.JSON
        def wineDbId = jsonReq.wineDbId
        List<String> gNoises = jsonReq.noises?.toList()
        gNoises?.removeAll{!it}

        def wineDB = WineDb.wineDatabase.findById(wineDbId)

        gNoises?.each { String noise ->
            def c = WineNoise.wineDatabase.createCriteria().count {
                and {
                    eq("noise", noise)
                    wineDb {
                        eq("id", wineDbId)
                    }
                }
            }

            if (c < 1) {
                def g = new WineNoise()
                g.noise = noise
                g.wineDb = wineDB
                g.validate()
                if (g.hasErrors()) {
                    failedList.push(noise)
                }
                try {
                    g.save()
                } catch (Exception e) {
                    failedList.push(noise)
                    log.error(e?.message, e)
                }
            }
        }

        dataResult.success = true
        dataResult.data = [
                failedList: failedList
        ]
        dataResult.message = failedList?.size() > 0 ? message(code: 'wineMatch.saveAllFailed.message', default: 'Some noises have not been saved.') : ''
        render dataResult.result as JSON
    }

    @Secured(value=["hasRole('ROLE_ADMIN')"])
    @Transactional
    def updateWaNote() {
        DataResult dataResult = new DataResult()
        WineDb wineDb = WineDb.findById(params.wineDbId)
        if (wineDb == null) {
            String message = message(code: "wineMatch.wineDb.notFound", default: "Wine not found.")
            dataResult = new DataResult(false, message)
            response.setStatus(HttpServletResponse.SC_UNAUTHORIZED)
            render dataResult.result as JSON
            return
        }

        wineDb.note = params.note
        wineDb.validate()
        if (wineDb.hasErrors()) {
            transactionStatus.setRollbackOnly()

            dataResult = new DataResult(false, "error", "")
            dataResult.errors = wineDb.errors.allErrors.collect{message([error: it])}

            response.setStatus(HttpServletResponse.SC_BAD_REQUEST)
            render dataResult.result as JSON
            return
        }

        try {
            wineDb.save()
            dataResult.success = true
        } catch (Exception e) {
            transactionStatus.setRollbackOnly()
            log.error(e?.message, e)

            dataResult = new DataResult(false, "error", e?.message)
            response.setStatus(HttpServletResponse.SC_BAD_REQUEST)
        }

        render dataResult.result as JSON
    }

    /**
     * This action will save or update {@link Matched} from Validation Match and Narrow Match
     * @return
     */
    @Secured(value=["hasRole('ROLE_ADMIN')"])
    @Transactional
    def saveMatched() {
        DataResult dataResult = new DataResult()

        String wineDbId = params?.wineDbId
        String wineHistory = params?.wineHistory
        String wineOriginalKeyword = params?.wineOriginalKeyword
        String historyType = params?.historyType
        String wineVintage = params?.wineVintage
        String wineBottleSize = params?.wineBottleSize
        String wineRetailer = params?.wineRetailer
        String validationType = params?.validationType
        String wineAlertId = null
        Boolean isXorZ = 'Defer'.equalsIgnoreCase(historyType) || 'Insufficient'.equalsIgnoreCase(historyType)

        try {
            WineDb wineDb = WineDb.wineDatabase.findById(wineDbId)
            if (!wineDb && !isXorZ) {
                response.status = 400
                render new DataResult(success: false, message: "WineDB not found.").result as JSON
                return
            }

            WebCrawler webCrawler = WebCrawler.findByCode(wineRetailer)
            if (!webCrawler) {
                response.status = 400
                render new DataResult(success: false, message: "Retailer not found.").result as JSON
                return
            }

            Vintage vintage = Vintage.wineDatabase.findByName(wineVintage)
            if (!vintage && !isXorZ) {
                response.status = 400
                render new DataResult(success: false, message: "Vintage not found.").result as JSON
                return
            }

            BottleSize bottleSize = BottleSize.wineDatabase.findByName(wineBottleSize)
            if (!bottleSize && !isXorZ) {
                response.status = 400
                render new DataResult(success: false, message: "Bottle Size not found.").result as JSON
                return
            }

            if (vintage && bottleSize) {
                wineAlertId = wineDbId + vintage.id + bottleSize.id
            }

            Matched matched = Matched.findOrCreateByWineAlertIdAndDescriptionAndWebCrawler(wineAlertId, wineHistory, webCrawler)
            matched.status = Status.DRAFT
            if (StringUtils.isNotBlank(validationType)) {
                matched.validationType = ValidationType.valueOf(validationType)
            }
            matched.keyword = wineOriginalKeyword
            if ('Defer'.equalsIgnoreCase(historyType)) {
                matched.wineAlertId = 'x'
            } else if ('Insufficient'.equalsIgnoreCase(historyType)) {
                matched.wineAlertId = 'z'
            }

            matched.validate()
            if (matched.hasErrors()) {
                transactionStatus.setRollbackOnly()

                dataResult = new DataResult(false, "error", "")
                dataResult.errors = matched.errors.allErrors.collect{message([error: it])}

                response.setStatus(HttpServletResponse.SC_BAD_REQUEST)
                render dataResult.result as JSON
                return
            }

            matched.save()

            dataResult.success = true
        } catch (Exception e) {
            log.error(e?.message, e)

            dataResult = new DataResult(false, "error", e?.message)
            response.setStatus(HttpServletResponse.SC_BAD_REQUEST)
        }

        render dataResult.result as JSON
    }

    /**
     * This action will update new Mongo Id the {@link Wine}
     * @return
     */
    @Secured(value=["hasRole('ROLE_ADMIN')"])
    @Transactional
    def updateMongoMap() {
        String wineDbId = params?.wineDbId
        String mongoId = params?.mongoId
        String wineVinId = params?.wineVinId
        String vintageName = params?.vintage

        WineDb wineDb = WineDb.wineDatabase.findById(wineDbId)
        if (!wineDb) {
            response.status = 400
            render new DataResult(success: false, message: "Wine not found.").result as JSON
            return
        }

        Wine wine = Wine.wineDatabase.createCriteria().get {
            eq("wineDb", wineDb)
            eq("vintage", vintageName)
            maxResults(1)
        }

        String wineId
        String name
        String color
        String pattern
        try {
            Source source = Source.valueOf((String) params?.source)

            if (wine) {
                if (wine.mongoId && wine.mongoId != mongoId) {
                    WineDuplication wineDuplication = new WineDuplication(
                            wine: wine,
                            mongoId: mongoId,
                            source: source
                    )
                    wineDuplication.validate()
                    if (wineDuplication.hasErrors()) {
                        DataResult dataResult = new DataResult(false, "")
                        dataResult.errors = wineDuplication.errors.allErrors.collect{message([error: it])}
                        response.status = 400
                        render dataResult.result as JSON
                        return
                    }
                    wineDuplication.save()
                } else {
                    wine.mongoId = mongoId
                    wine.wineNId = wineVinId
                    wine.save()
                }

                wineId = wine.id
                name = wine.wineDb.getDisplayName()
                color = wine.wineDb.color
                pattern = wine.wineDb.wineKeyword?.pattern
            } else {
                Vintage vintage = Vintage.wineDatabase.findByName(vintageName)
                if (!vintage) {
                    response.status = 400
                    render new DataResult(success: false, message: "Vintage not found.").result as JSON
                    return
                }

                Wine wine1 = new Wine(
                        wineDb: wineDb,
                        vintage: vintageName,
                        wineNId: wineVinId,
                        mongoId: mongoId,
                        source: source
                )
                wine1.id = wineDbId + vintage.id
                wine1.validate()
                if (wine1.hasErrors()) {
                    DataResult dataResult = new DataResult(false, "")
                    dataResult.errors = wine1.errors.allErrors.collect{message([error: it])}
                    response.status = 400
                    render dataResult.result as JSON
                    return
                }
                wine1.save()

                wineId = wine1.id
                name = wine1.wineDb.getDisplayName()
                color = wine1.wineDb.color
                pattern = wine1.wineDb.wineKeyword?.pattern
            }
        } catch(Exception e) {
            log.error(e?.message, e)

            response.status = 400
            render new DataResult(success: false, message: e?.message).result as JSON
            return
        }

        render new DataResult(success: true, data: [
                id: wineId,
                name: name,
                color: color,
                pattern: pattern
        ]).result as JSON
    }

    /**
    * This action will revert match of {@link Wine}
    * @return
    */
    @Secured(value=["hasRole('ROLE_ADMIN')"])
    @Transactional
    def undoMongoMap() {

        String id = params?.id
        String mongoId = params?.mongoId

        Wine wine = Wine.wineDatabase.createCriteria().get {
            eq("id", id)
            eq("mongoId", mongoId)
            maxResults(1)
        }

        if (!wine) {

            WineDuplication wineDuplication = WineDuplication.wineDatabase.createCriteria().get {
                eq("mongoId", mongoId)
                maxResults(1)
            }

            if (!wineDuplication) {
                response.status = 400
                render new DataResult(success: false, message: "Wine not found.").result as JSON
                return
            }

            wineDuplication.delete()
        } else {

            try {
                wine.mongoId = null
                wine.wineNId = null

                wine.validate()
                if (wine.hasErrors()) {
                    DataResult dataResult = new DataResult(false, "")
                    dataResult.errors = wine.errors.allErrors.collect { message([error: it]) }
                    response.status = 400
                    render dataResult.result as JSON
                    return
                }
                wine.save()
            } catch (Exception e) {
                log.error(e?.message, e)

                response.status = 400
                render new DataResult(success: false, message: e?.message).result as JSON
                return
            }
        }

        render new DataResult(success: true, data: [
                success: true
        ]).result as JSON
    }

    /**
     * This action will save new {@link Wine}
     * @return
     */
    @Transactional
    def updateProducer() {
        //String regionId = params?.regionId

        String producerId = params?.producerId
        String name = params?.name
        String nameShow = params?.nameShow

        Producer producer = Producer.wineDatabase.findById(producerId)
        if (!producer) {
            response.status = 400
            render new DataResult(success: false, message: "Producer not found.").result as JSON
            return
        }

        Boolean wineApiReindexRequired
        try {
            Source source = Source.valueOf((String) params?.source)

            producer.name = name
            producer.nameShow = nameShow

            producer.validate()
            if (producer.hasErrors()) {
                transactionStatus.setRollbackOnly()

                DataResult dataResult = new DataResult(false, "")
                dataResult.errors = producer.errors.allErrors.collect{message([error: it])}
                response.status = 400
                render dataResult.result as JSON
                return
            }

            wineApiReindexRequired = producerService.isWineApiReindexRequired(producer)

            producer.save()

            if (params.keyword) {
                ProducerKeyword producerKeyword
                if (producer.producerKeyword) {
                    producerKeyword = producer.producerKeyword
                } else {
                    producerKeyword = new ProducerKeyword()
                    producerKeyword.producer = producer
                    producerKeyword.source = source
                }
                producerKeyword.keyword = params.keyword

                producerKeyword.validate()
                if (producerKeyword.hasErrors()) {
                    transactionStatus.setRollbackOnly()

                    DataResult dataResult = new DataResult(false, "")
                    dataResult.errors = producerKeyword.errors.allErrors.collect{message([error: it])}
                    response.status = 400
                    render dataResult.result as JSON
                    return
                }
                producerKeyword.save()
            }
        } catch (Exception e) {
            log.error(e?.message, e)

            response.status = 400
            render new DataResult(success: false, message: e?.message).result as JSON
            return
        }

        if (wineApiReindexRequired) {
            //wineApiService.indexingWineDb(producer.wineDbs?.id)
        }

        render new DataResult(success: true, data: [
                producerId: producer.id
        ]).result as JSON
    }

    @Transactional
    def saveWine() {
        String regionId = params?.regionId

        String producerId = params?.producerId
        String nextProducerId = params?.nextProducerId
        String producerName = params?.producerName
        String producerShow = params?.producerShow

        String wineDbId = params?.wineDbId
        String nextWineDbId = params?.nextWineDbId
        String name = params?.name
        String color = params?.color
        String wineType = params?.wineType
        String sweetness = params?.sweetness
        String variety = params?.variety
        String wineCountry = params?.wineCountry
        String wineRegion = params?.wineRegion
        String wineLocation = params?.wineLocation
        String wineLocale = params?.wineLocale
        String wineSite = params?.wineSite
        String wineNId = params?.wineNId

        String nextWineId = params?.nextWineId
        String vintage = params?.vintage
        String mongoId = params?.mongoId
        List<String> noises = params?.list('noises')

        Boolean wineApiReindexRequired
        try {
            Source source = Source.valueOf((String) params?.source)

            Producer producer
            if (producerId) {
                producer = Producer.wineDatabase.findById(producerId)
            } else {
                if (RegionCode.wineDatabase.countById(regionId) == 0) {
                    RegionCode regionCode = new RegionCode(
                            source: source
                    )
                    regionCode.id = regionId
                    regionCode.validate()
                    if (regionCode.hasErrors()) {
                        transactionStatus.setRollbackOnly()

                        DataResult dataResult = new DataResult(false, "")
                        dataResult.errors = regionCode.errors.allErrors.collect{message([error: it])}
                        response.status = 400
                        render dataResult.result as JSON
                        return
                    }
                    regionCode.save()
                }

                producer = new Producer(
                        name: producerName,
                        nameShow: producerShow,
                        source: source
                )
                producer.id = nextProducerId
                producer.validate()
                if (producer.hasErrors()) {
                    transactionStatus.setRollbackOnly()

                    DataResult dataResult = new DataResult(false, "")
                    dataResult.errors = producer.errors.allErrors.collect{message([error: it])}
                    response.status = 400
                    render dataResult.result as JSON
                    return
                }
                producer.save()
            }

            WineDb wineDb
            if (wineDbId) {
                wineDb = WineDb.wineDatabase.findById(wineDbId)
            } else {
                wineDb = new WineDb(
                        name: name,
                        color: color,
                        type: wineType,
                        sweetness: sweetness,
                        variety: variety,
                        producer: producer,
                        wineCountry: wineCountry,
                        wineRegion: wineRegion,
                        wineLocation: wineLocation,
                        wineLocale: wineLocale,
                        wineSite: wineSite,
                        wineNId: wineNId,
                        source: source,
                        status: WineDb.Status.ACTIVE
                )
                wineDb.id = nextWineDbId

                wineDb.validate()
                if (wineDb.hasErrors()) {
                    transactionStatus.setRollbackOnly()

                    DataResult dataResult = new DataResult(false, "")
                    dataResult.errors = wineDb.errors.allErrors.collect{message([error: it])}
                    response.status = 400
                    render dataResult.result as JSON
                    return
                }

                wineApiReindexRequired = true

                wineDb.save()

                if (params.producerKeyword || params.baseKeyword || params.appellationKeyword || params.colorKeyword ||
                        params.commonModifiersKeyword || params.modifiersKeyword || params.customNotKeyword || params.globalNotKeyword) {

                    if (!producer.producerKeyword) {
                        ProducerKeyword producerKeyword = new ProducerKeyword(
                                producer: producer,
                                keyword: params.producerKeyword,
                                source: source
                        )
                        producerKeyword.validate()
                        if (producerKeyword.hasErrors()) {
                            transactionStatus.setRollbackOnly()

                            DataResult dataResult = new DataResult(false, "")
                            dataResult.errors = producerKeyword.errors.allErrors.collect{message([error: it])}
                            response.status = 400
                            render dataResult.result as JSON
                            return
                        }
                        producerKeyword.save()
                    }

                    WineKeyword wineKeyword = new WineKeyword()
                    wineKeyword.wineDb = wineDb
                    wineKeyword.producerKeyword = params.producerKeyword
                    wineKeyword.baseKeyword = params.baseKeyword
                    wineKeyword.appellationKeyword = params.appellationKeyword
                    wineKeyword.colorKeyword = params.colorKeyword
                    wineKeyword.commonModifiersKeyword = params.commonModifiersKeyword
                    wineKeyword.modifiersKeyword = params.modifiersKeyword
                    wineKeyword.customNotKeyword = params.customNotKeyword
                    wineKeyword.globalNotKeyword = params.globalNotKeyword
                    wineKeyword.pattern = WineKeywordUtils.joinPattern(wineKeyword)
                    wineKeyword.source = source

                    wineKeyword.validate()
                    if (wineKeyword.hasErrors()) {
                        transactionStatus.setRollbackOnly()

                        DataResult dataResult = new DataResult(false, "")
                        dataResult.errors = wineKeyword.errors.allErrors.collect{message([error: it])}
                        response.status = 400
                        render dataResult.result as JSON
                        return
                    }
                    wineKeyword.save()
                }

                noises?.each { n ->
                    new WineNoise(wineDb: wineDb, noise: n).save()
                }
            }

            if (vintage) {
                Wine wine = new Wine(
                        wineDb: wineDb,
                        vintage: vintage,
                        wineNId: wineNId,
                        mongoId: mongoId,
                        source: source
                )
                wine.id = nextWineId
                wine.validate()
                if (wine.hasErrors()) {
                    transactionStatus.setRollbackOnly()

                    DataResult dataResult = new DataResult(false, "")
                    dataResult.errors = wine.errors.allErrors.collect{message([error: it])}
                    response.status = 400
                    render dataResult.result as JSON
                    return
                }
                wine.save()
            }

        } catch (Exception e) {
            log.error(e?.message, e)

            response.status = 400
            render new DataResult(success: false, message: e?.message).result as JSON
            return
        }

        if (wineApiReindexRequired) {
            //wineApiService.indexingWineDb([nextWineDbId])
        }

        render new DataResult(success: true, data: [
                id: nextWineId
        ]).result as JSON
    }

    /**
     * This action will update detail of the {@link WineDb}
     * and the {@link WineKeyword} of (@link WineDB)
     * @return
     */
    @Transactional
    def updateWineDb() {
        //String regionId = params?.regionId

        //String producerId = params?.producerId
        //String nextProducerId = params?.nextProducerId
        //String producerName = params?.producerName
        //String producerShow = params?.producerShow

        String wineDbId = params?.wineDbId
        //String nextWineDbId = params?.nextWineDbId
        String name = params?.name
        String color = params?.color
        String wineType = params?.wineType
        String sweetness = params?.sweetness
        String variety = params?.variety
        String wineCountry = params?.wineCountry
        String wineRegion = params?.wineRegion
        String wineLocation = params?.wineLocation
        String wineLocale = params?.wineLocale
        String wineSite = params?.wineSite
        //String wineNId = params?.wineNId
        List<String> noises = params?.list('noises')

        WineDb wineDb = WineDb.wineDatabase.findById(wineDbId)
        if (!wineDb) {
            response.status = 400
            render new DataResult(success: false, message: "Wine not found.").result as JSON
            return
        }

        Boolean wineApiReindexRequired
        try {
            Source source = Source.valueOf((String) params?.source)

            wineDb.name = name
            wineDb.color = color
            wineDb.type = wineType
            wineDb.sweetness = sweetness
            wineDb.variety = variety
            wineDb.wineCountry = wineCountry
            wineDb.wineRegion = wineRegion
            wineDb.wineLocation = wineLocation
            wineDb.wineLocale = wineLocale
            wineDb.wineSite = wineSite

            wineDb.validate()
            if (wineDb.hasErrors()) {
                transactionStatus.setRollbackOnly()

                DataResult dataResult = new DataResult(false, "")
                dataResult.errors = wineDb.errors.allErrors.collect{message([error: it])}
                response.status = 400
                render dataResult.result as JSON
                return
            }

            wineApiReindexRequired = wineService.isWineApiReindexRequired(wineDb)

            wineDb.save()

            if (params.producerKeyword || params.baseKeyword || params.appellationKeyword || params.colorKeyword ||
                    params.commonModifiersKeyword || params.modifiersKeyword || params.customNotKeyword || params.globalNotKeyword) {

                if (!wineDb.producer.producerKeyword) {
                    ProducerKeyword producerKeyword = new ProducerKeyword(
                            producer: wineDb.producer,
                            keyword: params.producerKeyword,
                            source: source
                    )
                    producerKeyword.validate()
                    if (producerKeyword.hasErrors()) {
                        transactionStatus.setRollbackOnly()

                        DataResult dataResult = new DataResult(false, "")
                        dataResult.errors = producerKeyword.errors.allErrors.collect{message([error: it])}
                        response.status = 400
                        render dataResult.result as JSON
                        return
                    }
                    producerKeyword.save()
                }

                WineKeyword wineKeyword
                if (wineDb.wineKeyword) {
                    wineKeyword = wineDb.wineKeyword
                } else {
                    wineKeyword = new WineKeyword()
                    wineKeyword.wineDb = wineDb
                    wineKeyword.source = source
                }
                wineKeyword.producerKeyword = params.producerKeyword
                wineKeyword.baseKeyword = params.baseKeyword
                wineKeyword.appellationKeyword = params.appellationKeyword
                wineKeyword.colorKeyword = params.colorKeyword
                wineKeyword.commonModifiersKeyword = params.commonModifiersKeyword
                wineKeyword.modifiersKeyword = params.modifiersKeyword
                wineKeyword.customNotKeyword = params.customNotKeyword
                wineKeyword.globalNotKeyword = params.globalNotKeyword
                wineKeyword.pattern = WineKeywordUtils.joinPattern(wineKeyword)

                wineKeyword.validate()
                if (wineKeyword.hasErrors()) {
                    transactionStatus.setRollbackOnly()

                    DataResult dataResult = new DataResult(false, "")
                    dataResult.errors = wineKeyword.errors.allErrors.collect{message([error: it])}
                    response.status = 400
                    render dataResult.result as JSON
                    return
                }
                wineKeyword.save()
            }

            noises?.each { n ->
                def exists = wineDb.wineNoises?.any { it.noise.toLowerCase() == n.toLowerCase() }
                if (!exists) {
                    new WineNoise(wineDb: wineDb, noise: n).save()
                }
            }

            def lnoises = noises?.collect { it.toLowerCase() }
            wineDb.wineNoises?.findAll()?.each {
                if (!lnoises.contains(it.noise.toLowerCase())) {
                    wineDb.removeFromWineNoises(it)
                    it.delete()
                }
            }
        } catch (Exception e) {
            log.error(e?.message, e)

            response.status = 400
            render new DataResult(success: false, message: e?.message).result as JSON
            return
        }

        if (wineApiReindexRequired) {
            //wineApiService.indexingWineDb([wineDbId])
        }

        render new DataResult(success: true, data: [
                wineDbId: wineDb.id
        ]).result as JSON
    }

    /**
     * Merges the given wine ids (params.list("wineMerge")) to the wineDb destination.
     * @param wineDb the wineDb to be kept ({@link WineDb#id})
     * @return the status of the merge
     */
    @Transactional
    def mergeWines() {
        DataResult dataResult = new DataResult()

        //List<String> modifiedWines = new ArrayList<>()
        try {
            String id = params?.id
            String producerId = params?.producer
            String producerName = params?.producerName
            String nameShow = params?.nameShow
            if (!id) throw new RuntimeException('WineDb ID is required')
            if (!producerId) producerId = id.take(6)

            Producer producer = Producer.wineDatabase.findById(producerId)
            if (!producer) {
                producer = new Producer(id: producerId, name: producerName, nameShow: nameShow)

                producer.validate()
                if (producer.hasErrors()) {
                    transactionStatus.setRollbackOnly()

                    dataResult = new DataResult(false, "error", "")
                    dataResult.errors = producer.errors.allErrors.collect { message([error: it]) }

                    response.setStatus(HttpServletResponse.SC_BAD_REQUEST)
                    render dataResult.result as JSON
                    return
                }
                producer.save()
            }

            WineDb wineDb = WineDb.wineDatabase.findOrCreateById(id)
            bindData(wineDb, params, [exclude: ['id', 'producer', 'status']])
            wineDb.producer = producer
            if (!wineDb.status) wineDb.status = WineDb.Status.ACTIVE

            wineDb.validate()
            if (wineDb.hasErrors()) {
                transactionStatus.setRollbackOnly()

                dataResult = new DataResult(false, "error", "")
                dataResult.errors = wineDb.errors.allErrors.collect { message([error: it]) }

                response.setStatus(HttpServletResponse.SC_BAD_REQUEST)
                render dataResult.result as JSON
                return
            }

            wineDb.save()
            wineService.mergeWineDb(wineDb, params)

            List<WineDb> deletedWineDbs = []
            List<String> wineMerges = params?.list("wineMerge")
            wineMerges?.each { mergeId ->
                if (mergeId != wineDb.id) {
                    WineDb merging = WineDb.wineDatabase.findById(mergeId)
                    if (merging) {
                        deletedWineDbs.push(merging)
                    }
                }
            }

            deletedWineDbs.each {
                wineService.deleteWineDb(it)
            }

            //modifiedWines = deletedWineDbs.collect {it.id}
            //modifiedWines.add(wineDb.id)

            dataResult.success = true
        } catch (Exception e) {
            transactionStatus.setRollbackOnly()
            log.error(e?.message, e)

            dataResult = new DataResult(false, "error", e.message)
            response.setStatus(HttpServletResponse.SC_BAD_REQUEST)
        }

        //wineApiService.indexingWineDb(modifiedWines)

        render dataResult.result as JSON
    }

    /**
     * SAves wine keywords
     * @return
     */
    @Transactional
    def saveWineKeywords() {
        DataResult dataResult = new DataResult()
        def jsonReq = request.JSON

        wineService.saveKeywords(jsonReq?.keywords, Source.KW_GENERATOR)

        dataResult.success = true
        render dataResult.result as JSON
    }
}
