package wine.alert

import grails.plugin.springsecurity.annotation.Secured

import static org.springframework.http.HttpStatus.*
import grails.transaction.Transactional

@Secured(value=["hasRole('ROLE_ADMIN')"])
@Transactional(readOnly = true)
class CrawlPropertyController {

    static allowedMethods = [save: "POST", update: "PUT", delete: "DELETE"]

    /**
     * To get list of {@link CrawlProperty}
     * @param max the maximum offset
     * @return crawlPropertyList List of {@link CrawlProperty}
     * from minimum offset to maximum offset
     * @return model crawlPropertyCount length of crawlPropertyList
     */
    def index(Integer max) {
        params.max = Math.min(max ?: 20, 100)
        respond CrawlProperty.list(params), model:[crawlPropertyCount: CrawlProperty.count()]
    }

    /**
     * To get detail {@link CrawlProperty}
     * @param crawlProperty object of current {@link CrawlProperty}
     * @return crawlProperty current {@link CrawlProperty}
     */
    def show(CrawlProperty crawlProperty) {
        respond crawlProperty
    }

    /**
     * To render create page of {@link CrawlProperty}
     */
    def create() {
        def crawlProperty = new CrawlProperty(params)
        crawlProperty.dataOutput = true

        if (params.webCrawler) {
            def lastIndex = CrawlProperty.createCriteria().get {
                projections {
                    max "propertyIndex"
                }
                eq("webCrawler.id", params.webCrawler.id.toLong())
            }

            crawlProperty.propertyIndex = lastIndex == null ? 0 : lastIndex + 1
        }

        respond crawlProperty
    }

    /**
     * To save new {@link CrawlProperty}
     * @param crawlProperty object of {@link CrawlProperty}
     */
    @Transactional
    def save(CrawlProperty crawlProperty) {
        if (crawlProperty == null) {
            transactionStatus.setRollbackOnly()
            notFound()
            return
        }

        if (crawlProperty.hasErrors()) {
            transactionStatus.setRollbackOnly()
            respond crawlProperty.errors, view:'create'
            return
        }

        crawlProperty.save flush:true

        request.withFormat {
            form multipartForm {
                flash.message = message(code: 'default.created.message', args: [message(code: 'crawlProperty.label', default: 'CrawlProperty'), crawlProperty.id])
                redirect crawlProperty
            }
            '*' { respond crawlProperty, [status: CREATED] }
        }
    }

    /**
     * To render edit page of {@link CrawlProperty}
     * @param crawlProperty The current {@link CrawlProperty}
     * object that you want to edit
     */
    def edit(CrawlProperty crawlProperty) {
        respond crawlProperty
    }

    /**
     * To update current crawl property
     * @param crawlProperty The current {@link CrawlProperty}
     * object that you want to edit
     */
    @Transactional
    def update(CrawlProperty crawlProperty) {
        if (crawlProperty == null) {
            transactionStatus.setRollbackOnly()
            notFound()
            return
        }

        if (crawlProperty.hasErrors()) {
            transactionStatus.setRollbackOnly()
            respond crawlProperty.errors, view:'edit'
            return
        }

        crawlProperty.save flush:true

        request.withFormat {
            form multipartForm {
                flash.message = message(code: 'default.updated.message', args: [message(code: 'crawlProperty.label', default: 'CrawlProperty'), crawlProperty.id])
                redirect crawlProperty
            }
            '*'{ respond crawlProperty, [status: OK] }
        }
    }

    /**
     * To Delete the current {@link CrawlProperty}
     * that you want to delete
     * @param crawlProperty The current {@link CrawlProperty}
     * to delete
     */
    @Transactional
    def delete(CrawlProperty crawlProperty) {

        if (crawlProperty == null) {
            transactionStatus.setRollbackOnly()
            notFound()
            return
        }

        crawlProperty.delete flush:true

        request.withFormat {
            form multipartForm {
                flash.message = message(code: 'default.deleted.message', args: [message(code: 'crawlProperty.label', default: 'CrawlProperty'), crawlProperty.id])
                redirect controller: "webCrawler", action: "edit", id: crawlProperty.webCrawler.id, method: "GET"
            }
            '*'{ render status: NO_CONTENT }
        }
    }

    /**
     * To redirect not found page
     */
    protected void notFound() {
        request.withFormat {
            form multipartForm {
                flash.message = message(code: 'default.not.found.message', args: [message(code: 'crawlProperty.label', default: 'CrawlProperty'), params.id])
                redirect action: "index", method: "GET"
            }
            '*'{ render status: NOT_FOUND }
        }
    }
}
