package wine.alert

import grails.plugin.springsecurity.annotation.Secured

import static org.springframework.http.HttpStatus.*
import grails.transaction.Transactional

@Secured(value=["hasRole('ROLE_ADMIN')"])
@Transactional(readOnly = true)
class CrawlUrlController {

    static allowedMethods = [save: "POST", update: "PUT", delete: "DELETE"]

    /**
     * To get list of {@link CrawlUrl}
     * @param max the maximum offset
     * @return crawlUrlList List of {@link CrawlUrl}
     * from minimum offset to maximum offset
     * @return model crawlUrlCount length of crawlUrlList
     */
    def index(Integer max) {
        params.max = Math.min(max ?: 10, 100)
        respond CrawlUrl.list(params), model:[crawlUrlCount: CrawlUrl.count()]
    }

    /**
     * To get detail {@link CrawlUrl}
     * @param crawlUrl object of current {@link CrawlUrl}
     * @return crawlUrl current {@link CrawlUrl}
     */
    def show(CrawlUrl crawlUrl) {
        respond crawlUrl
    }

    /**
     * To render create page of {@link CrawlUrl}
     */
    def create() {
        respond new CrawlUrl(params)
    }

    /**
     * To save new {@link CrawlUrl}
     * @param crawlUrl object of {@link CrawlUrl}
     */
    @Transactional
    def save(CrawlUrl crawlUrl) {
        if (crawlUrl == null) {
            transactionStatus.setRollbackOnly()
            notFound()
            return
        }

        if (crawlUrl.hasErrors()) {
            transactionStatus.setRollbackOnly()
            respond crawlUrl.errors, view:'create'
            return
        }

        crawlUrl.save flush:true

        request.withFormat {
            form multipartForm {
                flash.message = message(code: 'default.created.message', args: [message(code: 'crawlUrl.label', default: 'CrawlUrl'), crawlUrl.id])
                redirect crawlUrl
            }
            '*' { respond crawlUrl, [status: CREATED] }
        }
    }

    /**
     * To render edit page of {@link CrawlUrl}
     * @param crawlUrl The current {@link CrawlUrl}
     * object that you want to edit
     */
    def edit(CrawlUrl crawlUrl) {
        respond crawlUrl
    }

    /**
     * To update current crawl url
     * @param crawlUrl The current {@link CrawlUrl}
     * object that you want to edit
     */
    @Transactional
    def update(CrawlUrl crawlUrl) {
        if (crawlUrl == null) {
            transactionStatus.setRollbackOnly()
            notFound()
            return
        }

        if (crawlUrl.hasErrors()) {
            transactionStatus.setRollbackOnly()
            respond crawlUrl.errors, view:'edit'
            return
        }

        crawlUrl.save flush:true

        request.withFormat {
            form multipartForm {
                flash.message = message(code: 'default.updated.message', args: [message(code: 'crawlUrl.label', default: 'CrawlUrl'), crawlUrl.id])
                redirect crawlUrl
            }
            '*'{ respond crawlUrl, [status: OK] }
        }
    }

    /**
     * To Delete the current {@link CrawlUrl}
     * that you want to delete
     * @param crawlUrl The current {@link CrawlUrl}
     * to delete
     */
    @Transactional
    def delete(CrawlUrl crawlUrl) {

        if (crawlUrl == null) {
            transactionStatus.setRollbackOnly()
            notFound()
            return
        }

        crawlUrl.delete flush:true

        request.withFormat {
            form multipartForm {
                flash.message = message(code: 'default.deleted.message', args: [message(code: 'crawlUrl.label', default: 'CrawlUrl'), crawlUrl.id])
                redirect action:"index", method:"GET"
            }
            '*'{ render status: NO_CONTENT }
        }
    }

    /**
     * To redirect not found page
     */
    protected void notFound() {
        request.withFormat {
            form multipartForm {
                flash.message = message(code: 'default.not.found.message', args: [message(code: 'crawlUrl.label', default: 'CrawlUrl'), params.id])
                redirect action: "index", method: "GET"
            }
            '*'{ render status: NOT_FOUND }
        }
    }
}
