package wine.alert

import grails.plugin.springsecurity.annotation.Secured
import grails.transaction.Transactional
import groovy.time.Duration
import groovy.time.TimeCategory
import org.springframework.beans.factory.annotation.Value

import java.text.SimpleDateFormat

import static org.springframework.http.HttpStatus.*

@Secured(value=["hasRole('ROLE_ADMIN')"])
@Transactional(readOnly = true)
class GlobalNoiseController {

    static allowedMethods = [save: "POST", update: "PUT", delete: ["DELETE", "GET"]]

    @Value('${application.download.location}')
    private String downloadLocation

    @Value('${application.output.encoding}')
    private String outputEncoding

    /**
     * To get list of {@link GlobalNoise}
     * @param max the maximum offset
     * @return globalNoise List of {@link GlobalNoise}
     * from minimum offset to maximum offset
     * @return model globalNoise length of globalNoiseList
     */
    def index(Integer max) {
        params.max = Math.min(max ?: 30, 100)
        SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd")
        Date from
        Date to
        try {
            from = params.dateFrom ? sdf.parse((String) params.dateFrom) : null
            to = params.dateTo ? sdf.parse((String) params.dateTo) : null
            if (to) to = TimeCategory.plus(to, new Duration(0, 23, 59, 59, 99))
        } catch (Exception e) {
            log.warn(e?.message)
        }

        def globalNoises = GlobalNoise.createCriteria().list (params) {
            if (params?.q) {
                ilike("noise", "%" + params?.q + "%")
            }
            if (from) ge("lastUpdated", from)
            if (to) le("lastUpdated", to)
            if (!params.sort) order("lastUpdated", "desc")
        }

        def globalNoiseCount = GlobalNoise.createCriteria().count {
            if (params?.q) {
                ilike("noise", "%" + params?.q + "%")
            }
            if (from) ge("lastUpdated", from)
            if (to) le("lastUpdated", to)
        }

        respond globalNoises, model:[globalNoiseCount: globalNoiseCount]
    }

    /**
     * Shows the global noise {@link GlobalNoise} information.
     * @param globalNoise the global noise to be shown {@link GlobalNoise#id}
     * @return the wine noise details page
     */
    def show(GlobalNoise globalNoise) {
        respond globalNoise
    }

    /**
     * Creator page of {@link GlobalNoise}.
     * @return the creator page
     */
    def create() {
        respond new GlobalNoise(params)
    }

    /**
     * Saves the global noise {@link GlobalNoise}.
     * @param globalNoise the properties of {@link GlobalNoise}
     * @return the redirection status to {@link GlobalNoiseController#show(GlobalNoise)}
     */
    @Transactional
    def save(GlobalNoise globalNoise) {
        if (globalNoise == null) {
            transactionStatus.setRollbackOnly()
            notFound()
            return
        }

        if (globalNoise.hasErrors()) {
            transactionStatus.setRollbackOnly()
            respond globalNoise.errors, view:'create'
            return
        }

        globalNoise.save flush:true

        request.withFormat {
            form multipartForm {
                flash.message = message(code: 'default.created.message', args: [message(code: 'globalNoise.label', default: 'GlobalNoise'), globalNoise.id])
                redirect globalNoise
            }
            '*' { respond globalNoise, [status: CREATED] }
        }
    }

    /**
     * Editor page of global noise {@link GlobalNoise}.
     * @param globalNoise the global noise to be edited {@link GlobalNoise#id}
     * @return the editor page
     */
    def edit(GlobalNoise globalNoise) {
        respond globalNoise
    }

    /**
     * Updates particular global noise.
     * @param globalNoise The global noise {@link GlobalNoise#id} to be updated
     * @return the redirection status to {@link GlobalNoiseController#edit(GlobalNoise)}
     */
    @Transactional
    def update(GlobalNoise globalNoise) {
        if (globalNoise == null) {
            transactionStatus.setRollbackOnly()
            notFound()
            return
        }

        if (globalNoise.hasErrors()) {
            transactionStatus.setRollbackOnly()
            respond globalNoise.errors, view:'edit'
            return
        }

        globalNoise.save flush:true

        request.withFormat {
            form multipartForm {
                flash.message = message(code: 'default.updated.message', args: [message(code: 'globalNoise.label', default: 'GlobalNoise'), globalNoise.id])
                redirect globalNoise
            }
            '*'{ respond globalNoise, [status: OK] }
        }
    }

    /**
     * Deletes global noise {@link GlobalNoise} by the given id.
     * @param globalNoise the global noise {@link GlobalNoise#id} to be deleted
     * @return the redirection status to {@link GlobalNoiseController#index(Integer)}
     */
    @Transactional
    def delete(GlobalNoise globalNoise) {

        if (globalNoise == null) {
            transactionStatus.setRollbackOnly()
            notFound()
            return
        }

        globalNoise.delete flush:true

        /*request.withFormat {
            form multipartForm {
                flash.message = message(code: 'default.deleted.message', args: [message(code: 'globalNoise.label', default: 'GlobalNoise'), globalNoise.id])
                redirect action:"index", method:"GET"
            }
            '*'{ render status: NO_CONTENT }
        }*/

        flash.message = message(code: 'default.deleted.message', args: [message(code: 'globalNoise.label', default: 'GlobalNoise'), globalNoise.id])
        redirect action:"index", method:"GET"
    }

    /**
     * Downloads global noises based on the given filters.
     * @return the file
     */
    def download(){
        SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd")
        Date from
        Date to
        try {
            from = params.dateFrom ? sdf.parse((String) params.dateFrom) : null
            to = params.dateTo ? sdf.parse((String) params.dateTo) : null
            if (to) to = TimeCategory.plus(to, new Duration(0, 23, 59, 59, 99))
        } catch (Exception e) {
            log.warn(e?.message)
        }

        try {
            def noiseList = GlobalNoise.createCriteria().list(params) {
                projections {
                    property("noise")
                }
                if (params?.q) {
                    ilike("noise", "%" + params?.q + "%")
                }
                if (from) ge("dateCreated", from)
                if (to) le("dateCreated", to)
                if (!params.sort) order("lastUpdated", "desc")
            }

            StringBuilder sb = new StringBuilder()
            noiseList?.each { noise ->
                sb.append(noise).append("\n")
            }

            Date date = new Date()
            String fileName = "global-noise_${date.format('yyyy-MM-dd-HHmmss')}.txt"
            File file = new File(downloadLocation + File.separator + fileName)
            file.getParentFile().mkdirs()
            file.createNewFile()
            file.append(sb.toString(), outputEncoding)

            if (file) {
                render(file: file, contentType: 'text/plain', fileName: file.name)
            } else {
                response.status = 404
            }
        } catch(Exception e) {
            log.error(e?.message, e)
            response.status = 500
        }
    }

    /**
     * Page Not Found
     */
    protected void notFound() {
        request.withFormat {
            form multipartForm {
                flash.message = message(code: 'default.not.found.message', args: [message(code: 'globalNoise.label', default: 'GlobalNoise'), params.id])
                redirect action: "index", method: "GET"
            }
            '*'{ render status: NOT_FOUND }
        }
    }
}
