package wine.alert

import grails.plugin.springsecurity.annotation.Secured
import grails.transaction.Transactional
import pl.touk.excel.export.WebXlsxExporter

import static org.springframework.http.HttpStatus.NOT_FOUND

@Secured(value=["hasRole('ROLE_ADMIN') or hasRole('ROLE_OPERATION')"])
@Transactional(readOnly = true)
class JobLogController {

    static allowedMethods = [save: "POST", update: "PUT", delete: "DELETE"]

    def crawlerReportService

    /**
     * To get list of {@link CrawlProperty}
     * @param max the maximum offset
     * @return crawlPropertyList List of {@link CrawlProperty}
     * from minimum offset to maximum offset
     * @return model crawlPropertyCount length of crawlPropertyList
     */
    def index(Integer max) {
        if (request.isUserInRole("ROLE_OPERATION") && !request.isUserInRole("ROLE_ADMIN")) {
            redirect(controller: "app", action: "wineSearcher")
            return
        }

        if (!max) {
            max = WebCrawler.createCriteria().count {
                and {
                    eq("disabled", false)
                }
            }
        }
        params.max = Math.min(max, 20)
        params.sort = params?.sort ?: "id"
        params.order = params?.order ?: "desc"

        def count = JobLog.createCriteria().get() {
            projections {
                count()
            }
            if (params.code) {
                eq('code', params.code)
            }
            if (params.q) {
                ilike("code", "%" + params.q + "%")
            }
        }

        respond JobLog.createCriteria().list(params) {
            if (params.code) {
                eq('code', params.code)
            }
            if (params.q) {
                ilike("code", "%" + params.q + "%")
            }
        }, model:[jobLogCount: count]
    }

    /**
     * To get detail {@link CrawlProperty}
     * @param crawlProperty object of current {@link CrawlProperty}
     * @return crawlProperty current {@link CrawlProperty}
     */
    @Secured(value=["hasRole('ROLE_ADMIN')"])
    def show(JobLog jobLog) {
        respond jobLog
    }

    @Secured(value=["hasRole('ROLE_ADMIN')"])
    def report() {
        def failedCrawlers = crawlerReportService.failedCrawlers()
        def crawlersWithChange = crawlerReportService.crawlersWithChange()
        def crawlersWithRecordChange = crawlersWithChange.findAll { Math.abs(it["recordChange"]) >= 0.2 }
        def crawlersWithPriceChange = crawlersWithChange.findAll { Math.abs(it["priceChange"]) >= 0.15 }

        def m = [
            crawlerCount: crawlerReportService.activeCrawlerCount(),
            successfulCrawlerCount: crawlerReportService.successfulCrawlerCount(),
            failedCrawlerCount: failedCrawlers.size(),
            failedList: failedCrawlers,
            externalCrawlerCount: crawlerReportService.externalCrawlerCount(),
            crawlersWithNoChange: crawlerReportService.crawlersWithNoChange(),
            htmlCrawlers: crawlerReportService.htmlCrawlers(),
            htmlAsyncCrawlers: crawlerReportService.htmlAsyncCrawlers(),
            crawlersWithRecordChange: crawlersWithRecordChange,
            crawlersWithPriceChange: crawlersWithPriceChange,
        ]

        withFormat {
            html { m }
            xlsx { exportExcel(response, m) }
        }
    }

//    @Transactional
//    def delete(JobLog jobLog) {
//
//        if (jobLog == null) {
//            transactionStatus.setRollbackOnly()
//            notFound()
//            return
//        }
//
//        jobLog.delete flush:true
//
//        request.withFormat {
//            form multipartForm {
//                flash.message = message(code: 'default.deleted.message', args: [message(code: 'jobLog.label', default: 'JobLog'), jobLog.id])
//                redirect action:"index", method:"GET"
//            }
//            '*'{ render status: NO_CONTENT }
//        }
//    }

    /**
     * To redirect not found page
     */
    protected void notFound() {
        request.withFormat {
            form multipartForm {
                flash.message = message(code: 'default.not.found.message', args: [message(code: 'jobLog.label', default: 'JobLog'), params.id])
                redirect action: "index", method: "GET"
            }
            '*'{ render status: NOT_FOUND }
        }
    }

    private def exportExcel(response, m) {
        new WebXlsxExporter().with {
            setResponseHeaders(response)

            sheet('crawler count').with {
                fillHeader(['crawler type', 'count'])
                fillRow(['active', m['crawlerCount']], 1)
                fillRow(['successful', m['successfulCrawlerCount']], 2)
                fillRow(['failed', m['failedCrawlerCount']], 3)
                fillRow(['external', m['externalCrawlerCount']], 4)
            }

            sheet('failed crawlers').with {
                fillHeader(['code', 'last success'])
                m['failedList'].eachWithIndex { c, i ->
                    fillRow((List) c, i + 1)
                }
            }

            sheet('no change crawlers').with {
                fillHeader(['code', 'days with no change', 'wine found', 'since'])
                m['crawlersWithNoChange'].eachWithIndex { c, i ->
                    fillRow([c['code'], c['days'], c['totalRecords'], c['since']], i + 1)
                }
            }

            sheet('html crawlers').with {
                fillHeader(['code', 'total records', 'daily change', 'average change'])
                m['htmlCrawlers'].eachWithIndex { c, i ->
                    fillRow((List) c, i + 1)
                }
            }

            sheet('html async crawlers').with {
                fillHeader(['code', 'total records', 'daily change', 'average change'])
                m['htmlAsyncCrawlers'].eachWithIndex { c, i ->
                    fillRow((List) c, i + 1)
                }
            }

            sheet('>= 20% wine found change').with {
                fillHeader(['code', 'change', 'wine found (today)', 'wine found (previous)'])
                m['crawlersWithRecordChange'].eachWithIndex { c, i ->
                    fillRow([c['code'], c['recordChange'], c['totalRecords'], c['lastTotalRecords']], i + 1)
                }
            }

            sheet('>= 15% total price change').with {
                fillHeader(['code', 'change', 'currency', 'total prices (today)', 'total prices (previous)'])
                m['crawlersWithPriceChange'].eachWithIndex { c, i ->
                    fillRow([c['code'], c['priceChange'], c['currency'], c['totalPrices'], c['lastTotalPrices']], i + 1)
                }
            }

            save(response.outputStream)
        }
    }
}
