package wine.alert

import enums.Status
import enums.ValidationType
import grails.plugin.springsecurity.annotation.Secured
import grails.transaction.Transactional
import groovy.time.Duration
import groovy.time.TimeCategory
import org.springframework.beans.factory.annotation.Value
import wine.db.WineDb

import java.text.SimpleDateFormat

import static org.springframework.http.HttpStatus.*

@Secured(value=["hasRole('ROLE_ADMIN')"])
@Transactional(readOnly = true)
class MatchedController {

    static allowedMethods = [save: "POST", update: "PUT", delete: ["DELETE", "GET"]]

    @Value('${application.download.location}')
    private String downloadLocation

    /**
     * To get list of {@link Matched}
     * @param max the maximum offset
     * @return matchedList List of {@link Matched}
     * from minimum offset to maximum offset
     * @return model matchedCount length of matchedList
     */
    def index(Integer max) {
        params.max = Math.min(max ?: 30, 100)
        SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd")
        Date from = null
        Date to = null
        try {
            from = params.dateFrom ? sdf.parse((String) params.dateFrom) : null
            to = params.dateTo ? sdf.parse((String) params.dateTo) : null
            if (to) to = TimeCategory.plus(to, new Duration(0, 23, 59, 59, 99))
        } catch (Exception e) {
            log.warn(e?.message)
        }

        List<Matched> matchedList = (List<Matched>) Matched.createCriteria().list (params) {
            if (params.code) {
                webCrawler {
                    like("code", "%${params.code}%")
                }
            }
            if (params.validationType) eq("validationType", params.validationType as ValidationType)
            if (params.status) eq("status", params.status as Status)
            if (from) ge("lastUpdated", from)
            if (to) le("lastUpdated", to)
            if (!params.sort) order("lastUpdated", "desc")
        }

        def matchedCount = Matched.createCriteria().count {
            if (params.code) {
                webCrawler {
                    like("code", "%${params.code}%")
                }
            }
            if (params.validationType) eq("validationType", params.validationType as ValidationType)
            if (params.status) eq("status", params.status as Status)
            if (from) ge("lastUpdated", from)
            if (to) le("lastUpdated", to)
        }

        Map<String, String> wineDbMap = new HashMap()
        matchedList.each { matched ->
            String wineDbId = matched.wineAlertId?.take(9)
            wineDbMap.put(matched.wineAlertId, WineDb.wineDatabase.findById(wineDbId)?.getDescription() ?: "")
        }

        respond matchedList, model:[matchedCount: matchedCount, wineDbMap: wineDbMap]
    }

    /**
     * To get detail {@link Matched}
     * @param matched object of current {@link Matched}
     * @return matched current {@link Matched}
     */
    def show(Matched matched) {
        respond matched
    }

    /**
     * To render create page of {@link Matched}
     */
    def create() {
        respond new Matched(params)
    }

    /**
     * To save new {@link Matched}
     * @param matched object of {@link Matched}
     */
    @Transactional
    def save(Matched matched) {
        if (matched == null) {
            transactionStatus.setRollbackOnly()
            notFound()
            return
        }

        if (matched.hasErrors()) {
            transactionStatus.setRollbackOnly()
            respond matched.errors, view:'create'
            return
        }

        matched.save flush:true

        request.withFormat {
            form multipartForm {
                flash.message = message(code: 'default.created.message', args: [message(code: 'matched.label', default: 'Matched'), matched.id])
                redirect matched
            }
            '*' { respond matched, [status: CREATED] }
        }
    }

    /**
     * To render edit page of {@link Matched}
     * @param matched The current {@link Matched}
     * object that you want to edit
     */
    def edit(Matched matched) {
        respond matched
    }

    /**
     * To update current matched
     * @param matched The current {@link Matched}
     * object that you want to edit
     */
    @Transactional
    def update(Matched matched) {
        if (matched == null) {
            transactionStatus.setRollbackOnly()
            notFound()
            return
        }

        if (matched.hasErrors()) {
            transactionStatus.setRollbackOnly()
            respond matched.errors, view:'edit'
            return
        }

        matched.save flush:true

        request.withFormat {
            form multipartForm {
                flash.message = message(code: 'default.updated.message', args: [message(code: 'matched.label', default: 'Matched'), matched.id])
                redirect matched
            }
            '*'{ respond matched, [status: OK] }
        }
    }

    /**
     * To Delete the current {@link Matched}
     * that you want to delete
     * @param matched The current {@link Matched}
     * to delete
     */
    @Transactional
    def delete(Matched matched) {

        if (matched == null) {
            transactionStatus.setRollbackOnly()
            notFound()
            return
        }

        matched.delete flush:true

        /*request.withFormat {
            form multipartForm {
                flash.message = message(code: 'default.deleted.message', args: [message(code: 'matched.label', default: 'Matched'), matched.id])
                redirect action:"index", method:"GET"
            }
            '*'{ render status: NO_CONTENT }
        }*/

        flash.message = message(code: 'default.deleted.message', args: [message(code: 'matched.label', default: 'Matched'), matched.id])
        redirect action:"index", method:"GET"
    }

    /**
     * This action will create file list of {@link Matched} list
     * with current filters
     * @return
     */
    def download(){
        SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd")
        Date from
        Date to
        try {
            from = params.dateFrom ? sdf.parse((String) params.dateFrom) : null
            to = params.dateTo ? sdf.parse((String) params.dateTo) : null
            if (to) to = TimeCategory.plus(to, new Duration(0, 23, 59, 59, 99))
        } catch (Exception e) {
            log.warn(e?.message)
        }

        try {
            String separator = "\t"
            StringBuilder sb = new StringBuilder("Retailer Code\tWine Alert Id\tWine Description\tKeyword String\tHistory String\tValidation Type\tStatus\n")
            List<Matched> matchedList = (List<Matched>) Matched.createCriteria().list () {
                if (params.code) {
                    webCrawler {
                        like("code", "%${params.code}%")
                    }
                }
                if (params.validationType) eq("validationType", params.validationType as ValidationType)
                if (params.status) eq("status", params.status as Status)
                if (from) ge("lastUpdated", from)
                if (to) le("lastUpdated", to)
                order("lastUpdated", "desc")
            }

            matchedList.eachWithIndex{ matched, i ->
                String wineDbId = matched.wineAlertId?.take(9)
                WineDb wineDb = WineDb.wineDatabase.findById(wineDbId)
                sb.append(matched?.webCrawler?.code).append(separator)
                sb.append(matched?.wineAlertId).append(separator)
                sb.append(wineDb?.getDescription() ?: "").append(separator)
                sb.append(matched?.keyword).append(separator)
                sb.append(matched?.description).append(separator)
                sb.append(matched?.validationType?.name() ?: "").append(separator)
                sb.append(matched?.status?.name() ?: "")
                sb.append("\n")
            }

            Date date = new Date()
            String fileName = "Matched_${date.format('yyyy-MM-dd-HHmmss')}.txt"
            File file = new File(downloadLocation + File.separator + fileName)
            file.getParentFile().mkdirs()
            file.createNewFile()
            file.append(sb.toString(), "ISO-8859-1")

            if (file) {
                render(file: file, contentType: 'text/plain', fileName: file.name)
            } else {
                response.status = 404
            }
        } catch(Exception e) {
            log.error(e?.message, e)
            response.status = 500
        }
    }

    /**
     * To redirect not found page
     */
    protected void notFound() {
        request.withFormat {
            form multipartForm {
                flash.message = message(code: 'default.not.found.message', args: [message(code: 'matched.label', default: 'Matched'), params.id])
                redirect action: "index", method: "GET"
            }
            '*'{ render status: NOT_FOUND }
        }
    }
}
