package wine.alert

import enums.PageType
import enums.Status
import grails.converters.JSON
import grails.plugin.springsecurity.annotation.Secured
import grails.transaction.Transactional
import org.apache.commons.collections.CollectionUtils
import org.jsoup.HttpStatusException
import org.springframework.beans.factory.DisposableBean
import org.springframework.beans.factory.InitializingBean
import wine.alert.model.DataResult

import java.util.concurrent.LinkedBlockingQueue
import java.util.concurrent.ThreadPoolExecutor
import java.util.concurrent.TimeUnit

import static org.springframework.http.HttpStatus.*

@Secured(value = ["hasRole('ROLE_ADMIN')"])
@Transactional(readOnly = true)
class WebCrawlerController implements InitializingBean, DisposableBean {

    static allowedMethods = [save: "POST", update: "PUT", delete: "DELETE"]

    ThreadPoolExecutor threadPoolExecutor
    def webCrawlerService
    def logsService
    def crawlerService

    /**
     * To get list of {@link WebCrawler}
     * @param max the maximum offset
     * @return webCrawlerList List of {@link WebCrawler}
     * from minimum offset to maximum offset
     * @return model webCrawlerCount length of webCrawlerList
     */
    def index(Integer max) {
        params.max = Math.min(max ?: 20, 2000)
        params.sort = params?.sort ?: "code"
        params.order = params?.order ?: "asc"

        def webCrawlerList = WebCrawler.createCriteria().list(params) {
            if (params?.q) {
                or {
                    ilike("code", "%" + params?.q + "%")
                    ilike("name", "%" + params?.q + "%")
                    eq("historyName", params?.q)
                }
            }
            ne("status", Status.DELETED)
        }

        def webCrawlerCount = WebCrawler.createCriteria().count {
            if (params?.q) {
                or {
                    ilike("code", "%" + params?.q + "%")
                    ilike("name", "%" + params?.q + "%")
                    eq("historyName", params?.q)
                }
            }
            ne("status", Status.DELETED)
        }

        respond webCrawlerList, model: [webCrawlerCount: webCrawlerCount]
    }

    /**
     * To get detail {@link WebCrawler}
     * @param webCrawler object of current {@link WebCrawler}
     * @return webCrawler current {@link WebCrawler}
     */
    def show(WebCrawler webCrawler) {
        respond webCrawler
    }

    /**
     * To render create page of {@link WebCrawler}
     */
    def create() {
        respond new WebCrawler(params)
    }

    /**
     * To save new {@link WebCrawler}
     * @param webCrawler object of {@link WebCrawler}
     */
    @Transactional
    def save(WebCrawler webCrawler) {
        if (webCrawler == null) {
            transactionStatus.setRollbackOnly()
            notFound()
            return
        }

        if (webCrawler.hasErrors()) {
            transactionStatus.setRollbackOnly()
            respond webCrawler.errors, view: 'create'
            return
        }

        webCrawler.save flush: true

        request.withFormat {
            form multipartForm {
                flash.message = message(code: 'default.created.message', args: [message(code: 'webCrawler.label', default: 'WebCrawler'), webCrawler.id])
                redirect webCrawler
            }
            '*' { respond webCrawler, [status: CREATED] }
        }
    }

    /**
     * To render edit page of {@link WebCrawler}
     * @param webCrawler The current {@link WebCrawler}
     * object that you want to edit
     */
    def edit(WebCrawler webCrawler) {
        respond webCrawler
    }

    /**
     * To update current web crawler
     * @param webCrawler The current {@link WebCrawler}
     * object that you want to edit
     */
    @Transactional
    def update(WebCrawler webCrawler) {
        if (webCrawler == null) {
            transactionStatus.setRollbackOnly()
            notFound()
            return
        }

        if (webCrawler.hasErrors()) {
            transactionStatus.setRollbackOnly()
            respond webCrawler.errors, view: 'edit'
            return
        }

        webCrawler.save flush: true

        request.withFormat {
            form multipartForm {
                flash.message = message(code: 'default.updated.message', args: [message(code: 'webCrawler.label', default: 'WebCrawler'), webCrawler.id])
                redirect webCrawler
            }
            '*' { respond webCrawler, [status: OK] }
        }
    }

    /**
     * To Delete the current {@link WebCrawler}
     * that you want to delete
     * @param webCrawler The current {@link WebCrawler}
     * to delete
     */
    @Transactional
    def delete(WebCrawler webCrawler) {

        if (webCrawler == null) {
            transactionStatus.setRollbackOnly()
            notFound()
            return
        }

        webCrawler.status = Status.DELETED
        webCrawler.save flush: true

        request.withFormat {
            form multipartForm {
                flash.message = message(code: 'default.deleted.message', args: [message(code: 'webCrawler.label', default: 'WebCrawler'), webCrawler.id])
                redirect action: "index", method: "GET"
            }
            '*' { render status: NO_CONTENT }
        }
    }

    /**
     * This action will update {@link WebCrawler#disabled} to true
     * @param webCrawler
     */
    @Transactional
    def disable(WebCrawler webCrawler) {

        if (webCrawler == null) {

            response.status = 400
            render false
            return
        }

        webCrawler.disabled = true
        webCrawler.save flush: true

        render true
    }

    /**
     * This action will update {@link WebCrawler#disabled} to false
     * @param webCrawler
     */
    @Transactional
    def enable(WebCrawler webCrawler) {

        if (webCrawler == null) {
            transactionStatus.setRollbackOnly()

            response.status = 400
            render false
            return
        }

        webCrawler.disabled = false
        webCrawler.save flush: true

        render true
    }

    /**
     * This action will run {@link WebCrawlerService#parse} and add
     * {@link JobLog}
     */
    @Transactional
    def proceed() {
        Long id = params?.long("id")
        boolean success = true
        WebCrawler webCrawler = WebCrawler.findByIdAndDisabledAndStatusNotEqual(id, Boolean.FALSE, Status.DELETED)
        if (webCrawler && !webCrawler.getProceed()) {
            threadPoolExecutor.execute({
                def start = new Date()
                logsService.saveCrawlerLog(webCrawler, Status.RUNNING, "", 0, BigDecimal.ZERO, 0, 0, null, "", start, null)
                log.info("Running crawler [${webCrawler.code}]...")
                if (webCrawler.pageType == PageType.AUTO_DETECT) {
                    String msg = ""
                    Status status = Status.SUCCESS
                    Integer total = 0
                    try {
                        total = crawlerService.runWithRetry(webCrawler)
                    } catch (MalformedURLException | SocketTimeoutException | HttpStatusException e) {
                        msg = e.toString()
                        status = Status.FAILED
                        log.error(msg)
                    } catch (Exception e) {
                        msg = e.toString()
                        status = Status.FAILED
                        log.error(msg, e)
                    }
                    logsService.saveCrawlerLog(webCrawler, status, msg, total, BigDecimal.ZERO, 0, 0, null, "", start, new Date())
                    log.info("Crawler completed [${webCrawler.code}]. Record(s): [$total]")
                } else {
                    webCrawlerService.parse(webCrawler)
                }
            } as Runnable)
        } else {
            success = false
        }

        if (!success) response.status = 400
        render success
    }

    /**
     * Terminates crawler: This method only update the latest status of the job
     */
    @Transactional
    def terminate() {
        List<Long> ids = params?.list("id[]")?.collect { it as Long }
        if (CollectionUtils.isNotEmpty(ids)) {
            def codes = WebCrawler.where {
                id in ids
            }.projections {
                property('code')
            }.list()
            codes?.each { code ->
                def job = JobLog.where {
                    code == code && status == Status.RUNNING
                }.get()
                if (job) {
                    job.status = Status.FAILED
                    job.save()
                }
            }
        }

        DataResult dataResult = new DataResult()
        dataResult.success = true
        render dataResult.result as JSON
    }

    @Override
    public void destroy() throws Exception {
        threadPoolExecutor.shutdown()
        threadPoolExecutor.awaitTermination(10, TimeUnit.SECONDS)
    }

    @Override
    public void afterPropertiesSet() throws Exception {
        threadPoolExecutor = new ThreadPoolExecutor(1, 1, 60, TimeUnit.SECONDS, new LinkedBlockingQueue<Runnable>())
    }

    /**
     * To redirect not found page
     */
    protected void notFound() {
        request.withFormat {
            form multipartForm {
                flash.message = message(code: 'default.not.found.message', args: [message(code: 'webCrawler.label', default: 'WebCrawler'), params.id])
                redirect action: "index", method: "GET"
            }
            '*' { render status: NOT_FOUND }
        }
    }
}
