package wine.alert

import enums.Status
import grails.plugin.springsecurity.annotation.Secured
import grails.transaction.Transactional

import static org.springframework.http.HttpStatus.CREATED
import static org.springframework.http.HttpStatus.NOT_FOUND
import static org.springframework.http.HttpStatus.NO_CONTENT
import static org.springframework.http.HttpStatus.OK

@Secured(value=["hasRole('ROLE_ADMIN')"])
@Transactional(readOnly = true)
class WineMatchConfigController {

    /**
     * To get list of {@link WebCrawler}
     * @param max the maximum offset
     * @return webCrawlerList List of {@link WebCrawler}
     * from minimum offset to maximum offset
     * @return model webCrawlerCount length of webCrawlerList
     */
    def index(Integer max) {
        params.max = Math.min(max ?: 1000, 2000)
        params.sort = params?.sort ?: "code"
        params.order = params?.order ?: "asc"

        def webCrawlerList = WebCrawler.createCriteria().list(params) {
            if (params?.q) {
                or {
                    ilike("code", "%" + params?.q + "%")
                    ilike("name", "%" + params?.q + "%")
                }
            }
            ne("status", Status.DELETED)
        }

        def webCrawlerCount = WebCrawler.createCriteria().count {
            if (params?.q) {
                or {
                    ilike("code", "%" + params?.q + "%")
                    ilike("name", "%" + params?.q + "%")
                }
            }
            ne("status", Status.DELETED)
        }

        respond webCrawlerList, model: [webCrawlerCount: webCrawlerCount]
    }

    /**
     * To get detail {@link WebCrawler}
     * @param webCrawler object of current {@link WebCrawler}
     * @return webCrawler current {@link WebCrawler}
     */
    def show(WebCrawler webCrawler) {
        respond webCrawler
    }

    /**
     * To render create page of {@link WebCrawler}
     */
    def create() {
        respond new WebCrawler(params)
    }

    /**
     * To save new {@link WebCrawler}
     * @param webCrawler object of {@link WebCrawler}
     */
    @Transactional
    def save(WebCrawler webCrawler) {
        if (webCrawler == null) {
            transactionStatus.setRollbackOnly()
            notFound()
            return
        }

        if (webCrawler.hasErrors()) {
            transactionStatus.setRollbackOnly()
            respond webCrawler.errors, view:'create'
            return
        }

        webCrawler.save flush:true

        request.withFormat {
            form multipartForm {
                flash.message = message(code: 'default.created.message', args: [message(code: 'wineMatchConfig.label', default: 'WineMatchConfig'), webCrawler.id])
                redirect action: 'show', id: webCrawler.id
            }
            '*' { respond webCrawler, [status: CREATED] }
        }
    }

    /**
     * To render edit page of {@link WebCrawler}
     * @param webCrawler The current {@link WebCrawler}
     * object that you want to edit
     */
    def edit(WebCrawler webCrawler) {
        respond webCrawler
    }

    /**
     * To update current web crawler
     * @param webCrawler The current {@link WebCrawler}
     * object that you want to edit
     */
    @Transactional
    def update(WebCrawler webCrawler) {
        if (webCrawler == null) {
            transactionStatus.setRollbackOnly()
            notFound()
            return
        }

        if (webCrawler.hasErrors()) {
            transactionStatus.setRollbackOnly()
            respond webCrawler.errors, view:'edit'
            return
        }

        webCrawler.save flush:true

        request.withFormat {
            form multipartForm {
                flash.message = message(code: 'default.updated.message', args: [message(code: 'wineMatchConfig.label', default: 'WineMatchConfig'), webCrawler.id])
                redirect action: 'show', id: webCrawler.id
            }
            '*'{ respond webCrawler, [status: OK] }
        }
    }


    /**
     * To Delete the current {@link WebCrawler}
     * that you want to delete
     * @param webCrawler The current {@link WebCrawler}
     * to delete
     */
    @Transactional
    def delete(WebCrawler webCrawler) {

        if (webCrawler == null) {
            transactionStatus.setRollbackOnly()
            notFound()
            return
        }

        webCrawler.status = Status.DELETED
        webCrawler.save flush:true

        request.withFormat {
            form multipartForm {
                flash.message = message(code: 'default.deleted.message', args: [message(code: 'wineMatchConfig.label', default: 'WineMatchConfig'), webCrawler.id])
                redirect action:"index", method:"GET"
            }
            '*'{ render status: NO_CONTENT }
        }
    }

    /**
     * To redirect not found page
     */
    protected void notFound() {
        request.withFormat {
            form multipartForm {
                flash.message = message(code: 'default.not.found.message', args: [message(code: 'wineMatchConfig.label', default: 'WineMatchConfig'), params.id])
                redirect action: "index", method: "GET"
            }
            '*'{ render status: NOT_FOUND }
        }
    }
}
