package wine.alert

import enums.MatchedType
import enums.Status
import grails.converters.JSON
import grails.plugin.springsecurity.annotation.Secured
import grails.transaction.Transactional
import org.apache.commons.lang3.StringUtils
import wine.alert.model.DataResult

import java.text.SimpleDateFormat

import static org.springframework.http.HttpStatus.NOT_FOUND

@Secured(value=["hasRole('ROLE_ADMIN')"])
@Transactional(readOnly = true)
class WineMatchController {

    def outputService
    def matcherService

    /**
     * To get list of {@link WebCrawler}
     * @param max the maximum offset
     * @return webCrawlerList List of {@link WebCrawler}
     * from minimum offset to maximum offset
     * @return model webCrawlerCount length of webCrawlerList
     */
    def index(Integer max) {
        params.max = Math.min(max ?: 1000, 2000)
        params.sort = params?.sort ?: "code"
        params.order = params?.order ?: "asc"
        List<WebCrawler> webCrawlerList = WebCrawler.createCriteria().list(params) {
            if (params?.q) {
                or {
                    ilike("code", "%" + params?.q + "%")
                    ilike("name", "%" + params?.q + "%")
                    eq("historyName", params?.q)
                }
            }
            ne("status", Status.DELETED)
        }

        if (webCrawlerList && !webCrawlerList.isEmpty()) {
            List<String> codes = webCrawlerList?.collect {it.code}
            outputService.cleanupCrawlOutputAndMatchOutput(codes, params?.matchedType == MatchedType.KEYWORD.name() ? MatchedType.KEYWORD : MatchedType.HISTORY)
        }

        def inputDates = [:]
        def historyMatchedDates = [:]
        def keywordMatchedDates = [:]
        outputService.populateWineMatchDates(webCrawlerList, inputDates, historyMatchedDates, keywordMatchedDates)

        def webCrawlerCount = WebCrawler.createCriteria().count {
            if (params?.q) {
                or {
                    ilike("code", "%" + params?.q + "%")
                    ilike("name", "%" + params?.q + "%")
                    eq("historyName", params?.q)
                }
            }
            ne("status", Status.DELETED)
        }

        def runningKeywordMatches = matcherService.getRunningMatches(MatchedType.KEYWORD)
        def runningHistoryMatches = matcherService.getRunningMatches(MatchedType.HISTORY)

        respond webCrawlerList, model: [
                webCrawlerCount: webCrawlerCount,
                inputDates: inputDates,
                historyMatchedDates: historyMatchedDates,
                keywordMatchedDates: keywordMatchedDates,
                runningKeywordMatches: runningKeywordMatches,
                runningHistoryMatches: runningHistoryMatches
        ]
    }

    /**
     * This action will run {@link WineMatchJob}
     */
    @Transactional
    def proceed() {
        DataResult dataResult = new DataResult()

        List<String> codes = []
        List<Boolean> runKeywords = []
        List<Date> dates = []
        def sdf = new SimpleDateFormat("yyyy-MM-dd")

        Boolean runKeyword = MatchedType.KEYWORD.name().equalsIgnoreCase(params?.matchedType as String)

        if (params.retailers != null && params.retailers != "") {
            params.retailers.findAll { k, v -> !k.contains(".") }.each { k, v ->
                codes.push(v?.code)
                //runKeywords.push(Boolean.parseBoolean(v?.runKeyword))
                runKeywords.push(runKeyword)

                Date date = null
                String dateString = v?.date
                if (StringUtils.isNotBlank(dateString)) {
                    try {
                        date = sdf.parse(dateString)
                    } catch (Exception e) {
                        log.error(e?.message, e)
                    }
                }
                dates.push(date)
            }
        }

        dataResult.errors = new ArrayList<String>()
        codes?.eachWithIndex { String code, int i ->
            Date d = dates?.getAt(i)
            if (d != null) {
                if (!outputService.isValidCrawlOutput(code, d)) {
                    dataResult.errors.push("Input file not found for ["+code+"] at date ["+sdf.format(d)+"]")
                }
            } else {
                dataResult.errors.push("Input file not found for ["+code+"] at date [null]")
            }
        }

        dataResult.success = true

        WineMatchJob.triggerNow([
                codes: codes,
                runKeywords: runKeywords,
                dates: dates,
                mode: params?.mode
        ])

        if (!dataResult.success) response.status = 400
        render dataResult.result as JSON
    }

    /**
     * This action will terminate {@link WineMatchJob} using
     * {@link MatcherService#removeFromRunningMatches(String,MatchedType)}
     * @return
     */
    def terminate() {
        DataResult dataResult = new DataResult()
        List<String> codes = []
        MatchedType matchedType = MatchedType.HISTORY
        if (StringUtils.isNotBlank((String) params?.matchedType)) {
            try { matchedType = MatchedType.valueOf((String) params?.matchedType)} catch(Exception e) { log.warn(e?.message) }
        }

        if (params.retailers != null && params.retailers != "") {
            params.retailers.findAll { k, v -> !k.contains(".") }.each { k, v ->
                codes.push((String) v?.code)
            }
        }

        dataResult.errors = new ArrayList<String>()
        codes?.eachWithIndex { String code, int i ->
            matcherService.removeFromRunningMatches(code, matchedType)
        }
        dataResult.success = true

        if (!dataResult.success) response.status = 400
        render dataResult.result as JSON
    }

    /**
     * This action will send the last dump file use
     * {@link OutputService#dumpWineMatch(String)}
     */
    def download() {
        try {
            String matchedType = params?.matchedType
            File file = outputService.dumpWineMatch(matchedType)
            if (file) {
                render(file: file, contentType: 'text/plain', fileName: file.name)
            } else {
                response.status = 404
            }
        } catch (Exception e) {
            log.error(e?.message, e)
            response.status = 500
        }
    }

    /**
     * To redirect not found page
     */
    protected void notFound() {
        request.withFormat {
            form multipartForm {
                flash.message = message(code: 'default.not.found.message', args: [message(code: 'webCrawler.label', default: 'WebCrawler'), params.id])
                redirect action: "index", method: "GET"
            }
            '*'{ render status: NOT_FOUND }
        }
    }
}