package wine.alert

import grails.plugin.springsecurity.annotation.Secured
import org.apache.commons.lang.StringUtils

import static org.springframework.http.HttpStatus.*
import grails.transaction.Transactional

@Transactional(readOnly = true)
@Secured(value=["hasRole('ROLE_ADMIN')"])
class WorkGroupController {

    static allowedMethods = [save: "POST", update: "PUT", delete: ["DELETE", "GET"]]

    /**
     * To get list of {@link WorkGroup}
     * @param max the maximum offset
     * @return workGroupList List of {@link WorkGroup}
     * from minimum offset to maximum offset
     * @return model workGroupCount length of workGroupList
     */
    def index(Integer max) {
        params.max = Math.min(max ?: 10, 100)
        respond WorkGroup.list(params), model:[workGroupCount: WorkGroup.count()]
    }

    /**
     * To get detail {@link WorkGroup}
     * @param workGroup object of current {@link WorkGroup}
     * @return workGroup current {@link WorkGroup}
     */
    def show(WorkGroup workGroup) {
        respond workGroup
    }

    /**
     * To render create page of {@link WorkGroup}
     */
    def create() {
        respond new WorkGroup(params)
    }

    /**
     * To save new {@link WorkGroup}
     * @param workGroup object of {@link WorkGroup}
     */
    @Transactional
    def save(WorkGroup workGroup) {
        if (workGroup == null) {
            transactionStatus.setRollbackOnly()
            notFound()
            return
        }

        def retailers = ((String) params?.retailers)?.tokenize(",")?.unique()?.collect {StringUtils.trim(it)}
        retailers?.each { code ->
            WebCrawler webCrawler = WebCrawler.findByCode(code)
            if (webCrawler) {
                workGroup.addToWorkGroupWebCrawlers(new WorkGroupWebCrawler(
                        webCrawler: webCrawler,
                        workGroup: workGroup
                ))
            }
        }

        if (workGroup.hasErrors()) {
            transactionStatus.setRollbackOnly()
            respond workGroup.errors, view:'create'
            return
        }

        workGroup.save flush:true

        request.withFormat {
            form multipartForm {
                flash.message = message(code: 'default.created.message', args: [message(code: 'workGroup.label', default: 'WorkGroup'), workGroup.id])
                redirect workGroup
            }
            '*' { respond workGroup, [status: CREATED] }
        }
    }

    /**
     * To render edit page of {@link WorkGroup}
     * @param workGroup The current {@link WorkGroup}
     * object that you want to edit
     */
    def edit(WorkGroup workGroup) {
        respond workGroup
    }

    /**
     * To update current work group
     * @param workGroup The current {@link WorkGroup}
     * object that you want to edit
     */
    @Transactional
    def update(WorkGroup workGroup) {
        if (workGroup == null) {
            transactionStatus.setRollbackOnly()
            notFound()
            return
        }

        def retailers = ((String) params?.retailers)?.tokenize(",")?.unique()?.collect {StringUtils.trim(it)}
        def existings = workGroup?.workGroupWebCrawlers?.collect {it.webCrawler?.code}
        if (existings && !existings.isEmpty()) {
            if (retailers && !retailers.isEmpty()) {
                def deletedList = workGroup?.workGroupWebCrawlers?.findAll {!existings.contains(it.webCrawler?.code)}
                deletedList?.each {
                    workGroup.removeFromWorkGroupWebCrawlers(it)
                    it.delete()
                }
            }
            retailers = retailers?.findAll {!existings.contains(it)}
        }

        retailers?.each { code ->
            code = StringUtils.trim((String) code)
            WebCrawler webCrawler = WebCrawler.findByCode(code)
            if (webCrawler) {
                workGroup.addToWorkGroupWebCrawlers(new WorkGroupWebCrawler(
                        webCrawler: webCrawler,
                        workGroup: workGroup
                ))
            }
        }

        if (workGroup.hasErrors()) {
            transactionStatus.setRollbackOnly()
            respond workGroup.errors, view:'edit'
            return
        }

        workGroup.save flush:true

        request.withFormat {
            form multipartForm {
                flash.message = message(code: 'default.updated.message', args: [message(code: 'workGroup.label', default: 'WorkGroup'), workGroup.id])
                redirect workGroup
            }
            '*'{ respond workGroup, [status: OK] }
        }
    }

    /**
     * To Delete the current {@link WorkGroup}
     * that you want to delete
     * @param workGroup The current {@link WorkGroup}
     * to delete
     */
    @Transactional
    def delete(WorkGroup workGroup) {

        if (workGroup == null) {
            transactionStatus.setRollbackOnly()
            notFound()
            return
        }

        log.debug("Deleting workGroup: " + workGroup.id)
        WorkGroupWebCrawler.executeUpdate("DELETE FROM WorkGroupWebCrawler WHERE workGroup=:workGroup", [workGroup: workGroup])
        workGroup.delete flush:true

        /*request.withFormat {
            form multipartForm {
                flash.message = message(code: 'default.deleted.message', args: [message(code: 'workGroup.label', default: 'WorkGroup'), workGroupName])
                redirect action:"index", method:"GET"
            }
            '*'{ render status: NO_CONTENT }
        }*/

        flash.message = message(code: 'default.deleted.message', args: [message(code: 'workGroup.label', default: 'WorkGroup'), workGroup.id])
        redirect action:"index", method:"GET"
    }

    /**
     * To redirect not found page
     */
    protected void notFound() {
        request.withFormat {
            form multipartForm {
                flash.message = message(code: 'default.not.found.message', args: [message(code: 'workGroup.label', default: 'WorkGroup'), params.id])
                redirect action: "index", method: "GET"
            }
            '*'{ render status: NOT_FOUND }
        }
    }
}
