package wine.db

import grails.plugin.springsecurity.annotation.Secured
import groovy.time.Duration
import groovy.time.TimeCategory
import org.springframework.beans.factory.annotation.Value

import java.text.SimpleDateFormat

import static org.springframework.http.HttpStatus.*
import grails.transaction.Transactional

@Secured(value=["hasRole('ROLE_ADMIN')"])
@Transactional(readOnly = true)
class ProducerNoiseController {

    static allowedMethods = [save: "POST", update: "PUT", delete: ["DELETE", "GET"]]

    @Value('${application.download.location}')
    private String downloadLocation

    @Value('${application.output.encoding}')
    private String outputEncoding

    /**
     * To get list of {@link ProducerNoise}
     * @param max the maximum offset
     * @return producerNoiseList List of {@link ProducerNoise} from minimum offset to maximum
     * offset
     * @return model producerCount length of producerNoiseList
     */
    def index(Integer max) {

        params.max = Math.min(max ?: 30, 100)
        SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd")
        Date from
        Date to
        try {
            from = params.dateFrom ? sdf.parse((String) params.dateFrom) : null
            to = params.dateTo ? sdf.parse((String) params.dateTo) : null
            if (to) to = TimeCategory.plus(to, new Duration(0, 23, 59, 59, 99))
        } catch (Exception e) {
            log.warn(e?.message)
        }

        def producerNoiseList = ProducerNoise.wineDatabase.createCriteria().list(params) {
            if (params?.q) {
                ilike("noise", "%" + params?.q + "%")
            }
            if (params.get('producer.id')) {
                producer {
                    like("id", "%" + params.get('producer.id') + "%")
                }
            }
            if (from) ge("lastUpdated", from)
            if (to) le("lastUpdated", to)
            if (!params.sort) order("lastUpdated", "desc")
        }
        def producerNoiseCount = ProducerNoise.wineDatabase.createCriteria().count {
            if (params?.q) {
                ilike("noise", "%" + params?.q + "%")
            }
            if (params.get('producer.id')) {
                producer {
                    like("id", "%" + params.get('producer.id') + "%")
                }
            }
            if (from) ge("lastUpdated", from)
            if (to) le("lastUpdated", to)
        }
        respond producerNoiseList, model:[producerNoiseCount: producerNoiseCount]
    }

    /**
     * Shows the producer noise {@link ProducerNoise} information.
     * @param producerNoise the producer noise to be shown {@link ProducerNoise#id}
     * @return the producer noise details page
     */
    def show(ProducerNoise producerNoise) {
        respond producerNoise
    }

    /**
     * Creator page of {@link ProducerNoise}.
     * @return the creator page
     */
    def create() {
        respond new ProducerNoise(params)
    }

    /**
     * Saves the producer noise {@link ProducerNoise}.
     * @param producerNoise the properties of {@link ProducerNoise}
     * @return the redirection status to {@link ProducerNoiseController#show(ProducerNoise)}
     */
    @Transactional
    def save(ProducerNoise producerNoise) {
        if (producerNoise == null) {
            transactionStatus.setRollbackOnly()
            notFound()
            return
        }

        if (producerNoise.hasErrors()) {
            transactionStatus.setRollbackOnly()
            respond producerNoise.errors, view:'create'
            return
        }

        producerNoise.save flush:true

        request.withFormat {
            form multipartForm {
                flash.message = message(code: 'default.created.message', args: [message(code: 'producerNoise.label', default: 'ProducerNoise'), producerNoise.id])
                redirect producerNoise
            }
            '*' { respond producerNoise, [status: CREATED] }
        }
    }

    /**
     * Editor page of producer noise {@link ProducerNoise}.
     * @param producerNoise the producer noise to be edited {@link ProducerNoise#id}
     * @return the editor page
     */
    def edit(ProducerNoise producerNoise) {
        respond producerNoise
    }

    /**
     * Updates particular producer noise.
     * @param producerNoise The producer noise {@link ProducerNoise#id} to be updated
     * @return the redirection status to {@link ProducerNoiseController#edit(ProducerNoise)}
     */
    @Transactional
    def update(ProducerNoise producerNoise) {
        if (producerNoise == null) {
            transactionStatus.setRollbackOnly()
            notFound()
            return
        }

        if (producerNoise.hasErrors()) {
            transactionStatus.setRollbackOnly()
            respond producerNoise.errors, view:'edit'
            return
        }

        producerNoise.save flush:true

        request.withFormat {
            form multipartForm {
                flash.message = message(code: 'default.updated.message', args: [message(code: 'producerNoise.label', default: 'ProducerNoise'), producerNoise.id])
                redirect producerNoise
            }
            '*'{ respond producerNoise, [status: OK] }
        }
    }

    /**
     * Deletes producer noise {@link ProducerNoise} by the given id.
     * @param producerNoise the producer noise {@link ProducerNoise#id} to be deleted
     * @return the redirection status to {@link ProducerNoiseController#index(Integer)}
     */
    @Transactional
    def delete(ProducerNoise producerNoise) {

        if (producerNoise == null) {
            transactionStatus.setRollbackOnly()
            notFound()
            return
        }

        producerNoise.delete flush:true

        flash.message = message(code: 'default.deleted.message', args: [message(code: 'producerNoise.label', default: 'ProducerNoise'), producerNoise.id])
        redirect action:"index", method:"GET"
    }

    /**
     * Downloads producer noises based on the given filters.
     * @return the file
     */
    def download(){
        SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd")
        Date from
        Date to
        try {
            from = params.dateFrom ? sdf.parse((String) params.dateFrom) : null
            to = params.dateTo ? sdf.parse((String) params.dateTo) : null
            if (to) to = TimeCategory.plus(to, new Duration(0, 23, 59, 59, 99))
        } catch (Exception e) {
            log.warn(e?.message)
        }

        try {
            String separator = "\t"
            StringBuilder sb = new StringBuilder("Producer Id\tNoise\n")
            def producerNoiseList = ProducerNoise.wineDatabase.createCriteria().list(params) {
                if (params?.q) {
                    ilike("noise", "%" + params?.q + "%")
                }
                if (params.get('producer.id')) {
                    producer {
                        like("id", "%" + params.get('producer.id') + "%")
                    }
                }
                if (from) ge("dateCreated", from)
                if (to) le("dateCreated", to)
            } as List<ProducerNoise>

            producerNoiseList.eachWithIndex{ producerNoise, i ->
                sb.append(producerNoise?.producer?.id).append(separator)
                sb.append(producerNoise?.noise)
                sb.append("\n")
            }

            Date date = new Date()
            String fileName = "producer-noise_${date.format('yyyy-MM-dd-HHmmss')}.txt"
            File file = new File(downloadLocation + File.separator + fileName)
            file.getParentFile().mkdirs()
            file.createNewFile()
            file.append(sb.toString(), outputEncoding)

            if (file) {
                render(file: file, contentType: 'text/plain', fileName: file.name)
            } else {
                response.status = 404
            }
        } catch(Exception e) {
            log.error(e?.message, e)
            response.status = 500
        }
    }

    /**
     * Page Not Found
     */
    protected void notFound() {
        request.withFormat {
            form multipartForm {
                flash.message = message(code: 'default.not.found.message', args: [message(code: 'producerNoise.label', default: 'ProducerNoise'), params.id])
                redirect action: "index", method: "GET"
            }
            '*'{ render status: NOT_FOUND }
        }
    }
}
