package wine.db

import grails.plugin.springsecurity.annotation.Secured
import grails.transaction.Transactional

import static org.springframework.http.HttpStatus.*

@Transactional(readOnly = true)
@Secured(value=["hasRole('ROLE_ADMIN')"])
class WineKeywordController {

    static allowedMethods = [save: "POST", update: "PUT", delete: "DELETE"]

    /**
     * To get list of {@link WineKeyword}
     * @param max the maximum offset
     * @return wineKeywordList List of {@link WineKeyword}
     * from minimum offset to maximum offset
     * @return model wineKeywordCount length of wineKeywordList
     */
    def index(Integer max) {
        params.max = Math.min(max ?: 20, 100)
        params.sort = params?.sort ?: "wd.id"
        params.order = params?.order ?: "asc"

        List<WineKeyword> wineKeywords = WineKeyword.wineDatabase.createCriteria().list(params) {
            createAlias("wineDb", "wd")
            if (params?.q) {
                or {
                    ilike("pattern", "%" + params?.q + "%")
                    ilike("wd.id", "%" + params?.q + "%")
                    ilike("wd.name", "%" + params?.q + "%")
                }
            }
        }

        def count = WineKeyword.wineDatabase.createCriteria().count {
            createAlias("wineDb", "wd")
            if (params?.q) {
                or {
                    ilike("pattern", "%" + params?.q + "%")
                    ilike("wd.id", "%" + params?.q + "%")
                    ilike("wd.name", "%" + params?.q + "%")
                }
            }
        }
        respond wineKeywords, model: [wineKeywordCount: count]
    }

    /**
     * To get detail {@link WineKeyword}
     * @param wineKeyword object of current {@link WineKeyword}
     * @return wineKeyword current {@link WineKeyword}
     */
    def show(WineKeyword wineKeyword) {
        respond wineKeyword
    }

    /**
     * To render create page of {@link WineKeyword}
     */
    def create() {
        respond new WineKeyword(params)
    }

    /**
     * To save new {@link WineKeyword}
     * @param wineKeyword object of {@link WineKeyword}
     */
    @Transactional
    def save(WineKeyword wineKeyword) {
        if (wineKeyword == null) {
            transactionStatus.setRollbackOnly()
            notFound()
            return
        }

        if (wineKeyword.hasErrors()) {
            transactionStatus.setRollbackOnly()
            respond wineKeyword.errors, view:'create'
            return
        }

        wineKeyword.save flush:true

        request.withFormat {
            form multipartForm {
                flash.message = message(code: 'default.created.message', args: [message(code: 'wineKeyword.label', default: 'WineKeyword'), wineKeyword.id])
                redirect wineKeyword
            }
            '*' { respond wineKeyword, [status: CREATED] }
        }
    }

    /**
     * To render edit page of {@link WineKeyword}
     * @param wineKeyword The current {@link WineKeyword}
     * object that you want to edit
     */
    def edit(WineKeyword wineKeyword) {
        respond wineKeyword
    }

    /**
     * To update current wine keyword
     * @param wineKeyword The current {@link WineKeyword}
     * object that you want to edit
     */
    @Transactional
    def update(WineKeyword wineKeyword) {
        if (wineKeyword == null) {
            transactionStatus.setRollbackOnly()
            notFound()
            return
        }

        if (wineKeyword.hasErrors()) {
            transactionStatus.setRollbackOnly()
            respond wineKeyword.errors, view:'edit'
            return
        }

        wineKeyword.save flush:true

        request.withFormat {
            form multipartForm {
                flash.message = message(code: 'default.updated.message', args: [message(code: 'wineKeyword.label', default: 'WineKeyword'), wineKeyword.id])
                redirect wineKeyword
            }
            '*'{ respond wineKeyword, [status: OK] }
        }
    }

    /**
     * To Delete the current {@link WineKeyword}
     * that you want to delete
     * @param wineKeyword The current {@link WineKeyword}
     * to delete
     */
    @Transactional
    def delete(WineKeyword wineKeyword) {

        if (wineKeyword == null) {
            transactionStatus.setRollbackOnly()
            notFound()
            return
        }

        wineKeyword.delete flush:true

        request.withFormat {
            form multipartForm {
                flash.message = message(code: 'default.deleted.message', args: [message(code: 'wineKeyword.label', default: 'WineKeyword'), wineKeyword.id])
                redirect action:"index", method:"GET"
            }
            '*'{ render status: NO_CONTENT }
        }
    }

    /**
     * To redirect not found page
     */
    protected void notFound() {
        request.withFormat {
            form multipartForm {
                flash.message = message(code: 'default.not.found.message', args: [message(code: 'wineKeyword.label', default: 'WineKeyword'), params.id])
                redirect action: "index", method: "GET"
            }
            '*'{ render status: NOT_FOUND }
        }
    }
}
