package wine.db

import enums.Status
import grails.plugin.springsecurity.annotation.Secured
import grails.plugins.rest.client.RestResponse
import org.springframework.http.HttpStatus
import org.springframework.util.LinkedMultiValueMap
import org.springframework.util.MultiValueMap

import java.text.SimpleDateFormat

import static org.springframework.http.HttpStatus.*
import grails.transaction.Transactional

@Secured(value=["hasRole('ROLE_ADMIN')"])
@Transactional
class WineLabelController {

    static allowedMethods = [save: "POST", update: "PUT", delete: ["DELETE", "GET"]]

    def wineRecognitionApiService
    def wineLabelService

    /**
     * To get list of {@link WineLabel}
     * @param max the maximum offset
     * @return wineLabelList List of {@link WineLabel}
     * from minimum offset to maximum offset
     * @return model wineLabelCount length of wineLabelList
     */
    def index(Integer max) {
        params.max = Math.min(max ?: 10, 100)

        List<WineLabel> wineLabelList = WineLabel.list(params)

        def imageMap = [:]
        wineLabelList.each {WineLabel wl ->
            imageMap.put(wl.id, wineLabelService.getImageUrl(wl.getImageDir(), wl.status?.value()))
        }

        respond wineLabelList, model:[wineLabelCount: WineLabel.count(), imageMap: imageMap]
    }

    /**
     * To get detail {@link WineLabel}
     * @param wineLabel object of current {@link WineLabel}
     * @return wineLabel current {@link WineLabel}
     */
    def show(WineLabel wineLabel) {
        respond wineLabel
    }

    /**
     * To render create page of {@link WineLabel}
     */
    def create() {
        respond new WineLabel(params)
    }

    /**
     * To save new {@link WineLabel}
     * @param wineLabel object of {@link WineLabel}
     */
    @Transactional
    def save(WineLabel wineLabel) {
        if (wineLabel == null) {
            transactionStatus.setRollbackOnly()
            notFound()
            return
        }

        if (wineLabel.hasErrors()) {
            transactionStatus.setRollbackOnly()
            respond wineLabel.errors, view:'create'
            return
        }

        wineLabel.save flush:true

        request.withFormat {
            form multipartForm {
                flash.message = message(code: 'default.created.message', args: [message(code: 'wineLabel.label', default: 'WineLabel'), wineLabel.id])
                redirect wineLabel
            }
            '*' { respond wineLabel, [status: CREATED] }
        }
    }
    
    /**
     * To render edit page of {@link WineLabel}
     * @param wineLabel The current {@link WineLabel}
     * object that you want to edit
     */
    def edit(WineLabel wineLabel) {
        respond wineLabel
    }

    /**
     * To update current wine label
     * @param wineLabel The current {@link WineLabel}
     * object that you want to edit
     */
    @Transactional
    def update(WineLabel wineLabel) {
        if (wineLabel == null) {
            transactionStatus.setRollbackOnly()
            notFound()
            return
        }

        if (wineLabel.hasErrors()) {
            transactionStatus.setRollbackOnly()
            respond wineLabel.errors, view:'edit'
            return
        }

        wineLabel.save flush:true

        request.withFormat {
            form multipartForm {
                flash.message = message(code: 'default.updated.message', args: [message(code: 'wineLabel.label', default: 'WineLabel'), wineLabel.id])
                redirect wineLabel
            }
            '*'{ respond wineLabel, [status: OK] }
        }
    }

    /**
     * To Delete the current {@link WineLabel}
     * that you want to delete
     * @param wineLabel The current {@link WineLabel}
     * to delete
     */
    @Transactional
    def delete(WineLabel wineLabel) {

        if (wineLabel == null) {
            transactionStatus.setRollbackOnly()
            notFound()
            return
        }

        wineLabel.delete flush:true

        flash.message = message(code: 'wineLabel.delete.message', args: [wineLabel.vintage, wineLabel.name], default: "Label for {0} {1} deleted")
        redirect action:"index", method:"GET"
    }

    /**
     * To redirect not found page
     */
    protected void notFound() {
        request.withFormat {
            form multipartForm {
                flash.message = message(code: 'default.not.found.message', args: [message(code: 'wineLabel.label', default: 'WineLabel'), params.id])
                redirect action: "index", method: "GET"
            }
            '*'{ render status: NOT_FOUND }
        }
    }

    def approve(WineLabel wineLabel) {

        SimpleDateFormat sdf = new SimpleDateFormat("ddMMyyyy")

        MultiValueMap<String, Object> apiParams = new LinkedMultiValueMap<>()
        apiParams.add("imageName", wineLabel.imageId)
        apiParams.add("imageExt", wineLabel.imageExt)
        apiParams.add("wineId", wineLabel.wineId)
        apiParams.add("submittedDate", sdf.format(wineLabel.submittedDate))

        Map<String, Object> apiResponse = wineRecognitionApiService.approveWineLabel(apiParams)

        String msg = ""
        if(((int) apiResponse.get('status')) == HttpStatus.OK.value()) {
            wineLabel.setImageId(wineLabel.wineId)
            wineLabel.setStatus(Status.APPROVED)
            wineLabel.setStatusUpdated(new Date())

            wineLabel.save flush:true

            msg = message(code: 'wineLabel.approved.success.message', args: [wineLabel.vintage, wineLabel.name], default: "Label for {0} {1} has assigned")
        } else {
            msg = message(code: 'wineLabel.approved.fail.message', args: [wineLabel.vintage, wineLabel.name], default: "Label for {0} {1} cannot be assigned")
        }

        flash.message = msg
        redirect action:"index", method:"GET"

    }

    def reject(WineLabel wineLabel) {

        wineLabel.setStatus(Status.REJECTED)
        wineLabel.setStatusUpdated(new Date())

        if (wineLabel.hasErrors()) {
            transactionStatus.setRollbackOnly()
            respond wineLabel.errors, view:'index'
            return
        }

        wineLabel.save flush:true

        flash.message = message(code: 'wineLabel.reject.message', args: [wineLabel.vintage, wineLabel.name], default: "Label for {0} {1} has rejected")
        redirect action:"index", method:"GET"

    }

    def undo(WineLabel wineLabel) {

        SimpleDateFormat sdf = new SimpleDateFormat("ddMMyyyy")

        MultiValueMap<String, Object> apiParams = new LinkedMultiValueMap<>()
        apiParams.add("imageExt", wineLabel.imageExt)
        apiParams.add("wineId", wineLabel.wineId)
        apiParams.add("submittedDate", sdf.format(wineLabel.submittedDate))


        int stat = 0
        String newWineName = null
        if(wineLabel.status == Status.APPROVED){
            Map<String, Object> apiResponse = wineRecognitionApiService.undoWineLabel(apiParams)

            stat = ((int) apiResponse.get('status'))
            newWineName = apiResponse.get('data')
        }

        String msg = ""
        if(wineLabel.status == Status.REJECTED || stat == HttpStatus.OK.value()) {
            if(newWineName) {
                wineLabel.setImageId(newWineName)
            }
            wineLabel.setStatus(Status.WAIT_FOR_APPROVAL)
            wineLabel.setStatusUpdated(new Date())

            wineLabel.save flush:true

            msg = message(code: 'wineLabel.undo.success.message', args: [wineLabel.vintage, wineLabel.name], default: "Label for {0} {1} has been restored its status")
        } else {
            msg = message(code: 'wineLabel.undo.fail.message', args: [wineLabel.vintage, wineLabel.name], default: "Label for {0} {1} cannot be restored its status")
        }

        flash.message = msg
        redirect action:"index", method:"GET"
    }
}
