package wine.db

import grails.plugin.springsecurity.annotation.Secured
import grails.transaction.Transactional
import groovy.time.Duration
import groovy.time.TimeCategory
import org.springframework.beans.factory.annotation.Value

import java.text.SimpleDateFormat

import static org.springframework.http.HttpStatus.*

@Secured(value=["hasRole('ROLE_ADMIN')"])
@Transactional(readOnly = true)
class WineNoiseController {

    static allowedMethods = [save: "POST", update: "PUT", delete: ["DELETE", "GET"]]

    @Value('${application.download.location}')
    private String downloadLocation

    @Value('${application.output.encoding}')
    private String outputEncoding

    /**
     * To get list of {@link WineNoise}
     * @param max the maximum offset
     * @return wineNoiseList List of {@link WineNoise}
     * from minimum offset to maximum offset
     * @return model wineNoiseCount length of wineNoiseList
     */
    def index(Integer max) {
        params.max = Math.min(max ?: 30, 100)
        SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd")
        Date from
        Date to
        try {
            from = params.dateFrom ? sdf.parse((String) params.dateFrom) : null
            to = params.dateTo ? sdf.parse((String) params.dateTo) : null
            if (to) to = TimeCategory.plus(to, new Duration(0, 23, 59, 59, 99))
        } catch (Exception e) {
            log.warn(e?.message)
        }

        def wineNoiseList = WineNoise.wineDatabase.createCriteria().list(params) {
            if (params?.q) {
                ilike("noise", "%" + params?.q + "%")
            }
            if (params.get('wineDb.id')) {
                wineDb {
                    like("id", "%" + params.get('wineDb.id') + "%")
                }
            }
            if (from) ge("lastUpdated", from)
            if (to) le("lastUpdated", to)
            if (!params.sort) order("lastUpdated", "desc")
        }
        def wineNoiseCount = WineNoise.wineDatabase.createCriteria().count {
            if (params?.q) {
                ilike("noise", "%" + params?.q + "%")
            }
            if (params.get('wineDb.id')) {
                wineDb {
                    like("id", "%" + params.get('wineDb.id') + "%")
                }
            }
            if (from) ge("lastUpdated", from)
            if (to) le("lastUpdated", to)
        }
        respond wineNoiseList, model:[wineNoiseCount: wineNoiseCount]
    }

    /**
     * Shows the wine noise {@link WineNoise} information.
     * @param wineNoise the wine noise to be shown {@link WineNoise#id}
     * @return the wine noise details page
     */
    def show(WineNoise wineNoise) {
        respond wineNoise
    }

    /**
     * Creator page of {@link WineNoise}.
     * @return the creator page
     */
    def create() {
        respond new WineNoise(params)
    }

    /**
     * Saves the wine noise {@link WineNoise}.
     * @param wineNoise the properties of {@link WineNoise}
     * @return the redirection status to {@link WineNoiseController#show(WineNoise)}
     */
    @Transactional
    def save(WineNoise wineNoise) {
        if (wineNoise == null) {
            transactionStatus.setRollbackOnly()
            notFound()
            return
        }

        if (wineNoise.hasErrors()) {
            transactionStatus.setRollbackOnly()
            respond wineNoise.errors, view:'create'
            return
        }

        wineNoise.save flush:true

        request.withFormat {
            form multipartForm {
                flash.message = message(code: 'default.created.message', args: [message(code: 'wineNoise.label', default: 'WineNoise'), wineNoise.id])
                redirect wineNoise
            }
            '*' { respond wineNoise, [status: CREATED] }
        }
    }

    /**
     * Editor page of wine noise {@link WineNoise}.
     * @param wineNoise the wine noise to be edited {@link WineNoise#id}
     * @return the editor page
     */
    def edit(WineNoise wineNoise) {
        respond wineNoise
    }

    /**
     * Updates particular wine noise.
     * @param wineNoise The wine noise {@link WineNoise#id} to be updated
     * @return the redirection status to {@link WineNoiseController#edit(WineNoise)}
     */
    @Transactional
    def update(WineNoise wineNoise) {
        if (wineNoise == null) {
            transactionStatus.setRollbackOnly()
            notFound()
            return
        }

        if (wineNoise.hasErrors()) {
            transactionStatus.setRollbackOnly()
            respond wineNoise.errors, view:'edit'
            return
        }

        wineNoise.save flush:true

        request.withFormat {
            form multipartForm {
                flash.message = message(code: 'default.updated.message', args: [message(code: 'wineNoise.label', default: 'WineNoise'), wineNoise.id])
                redirect wineNoise
            }
            '*'{ respond wineNoise, [status: OK] }
        }
    }

    /**
     * Deletes wine noise {@link WineNoise} by the given id.
     * @param wineNoise the producer noise {@link WineNoise#id} to be deleted
     * @return the redirection status to {@link WineNoiseController#index(Integer)}
     */
    @Transactional
    def delete(WineNoise wineNoise) {

        if (wineNoise == null) {
            transactionStatus.setRollbackOnly()
            notFound()
            return
        }

        wineNoise.delete flush:true

        /*request.withFormat {
            form multipartForm {
                flash.message = message(code: 'default.deleted.message', args: [message(code: 'wineNoise.label', default: 'WineNoise'), wineNoise.id])
                redirect action:"index", method:"GET", params: ['wineDb.id': params?.get('wineDb.id')]
            }
            '*'{ render status: NO_CONTENT }
        }*/

        flash.message = message(code: 'default.deleted.message', args: [message(code: 'wineNoise.label', default: 'WineNoise'), wineNoise.id])
        redirect action:"index", method:"GET"
    }

    /**
     * Downloads wine noises based on the given filters.
     * @return the file
     */
    def download(){
        SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd")
        Date from
        Date to
        try {
            from = params.dateFrom ? sdf.parse((String) params.dateFrom) : null
            to = params.dateTo ? sdf.parse((String) params.dateTo) : null
            if (to) to = TimeCategory.plus(to, new Duration(0, 23, 59, 59, 99))
        } catch (Exception e) {
            log.warn(e?.message)
        }

        try {
            String separator = "\t"
            StringBuilder sb = new StringBuilder("Wine Alert Id\tWine Description\tNoise\n")
            def wineNoiseList = WineNoise.wineDatabase.createCriteria().list(params) {
                if (params?.q) {
                    ilike("noise", "%" + params?.q + "%")
                }
                if (params.get('wineDb.id')) {
                    wineDb {
                        like("id", "%" + params.get('wineDb.id') + "%")
                    }
                }
                if (from) ge("dateCreated", from)
                if (to) le("dateCreated", to)
            } as List<WineNoise>

            wineNoiseList.eachWithIndex{ wineNoise, i ->
                sb.append(wineNoise?.wineDb?.id).append(separator)
                sb.append(wineNoise?.wineDb?.description).append(separator)
                sb.append(wineNoise?.noise)
                sb.append("\n")
            }

            Date date = new Date()
            String fileName = "wid-noise_${date.format('yyyy-MM-dd-HHmmss')}.txt"
            File file = new File(downloadLocation + File.separator + fileName)
            file.getParentFile().mkdirs()
            file.createNewFile()
            file.append(sb.toString(), outputEncoding)

            if (file) {
                render(file: file, contentType: 'text/plain', fileName: file.name)
            } else {
                response.status = 404
            }
        } catch(Exception e) {
            log.error(e?.message, e)
            response.status = 500
        }
    }

    /**
     * Page Not Found
     */
    protected void notFound() {
        request.withFormat {
            form multipartForm {
                flash.message = message(code: 'default.not.found.message', args: [message(code: 'wineNoise.label', default: 'WineNoise'), params.id])
                redirect action: "index", method: "GET"
            }
            '*'{ render status: NOT_FOUND }
        }
    }
}
