package wine.alert

import enums.AuthType
import enums.CsvQuoteMode
import enums.DataPlacement
import enums.FileSource
import enums.PageType
import enums.PaginationType
import enums.RequestType
import enums.Separator
import enums.Status

class WebCrawler {

    static hasMany = [
            crawlProperties: CrawlProperty,
            crawlUrls      : CrawlUrl
    ]

    static transients = ['latestJob', 'foundRecords', 'latestRun', 'proceed', 'averageRecords', 'averagePrices',
                         'historyFileName', 'matchInputSeparator', 'latestHistoryMatched', 'latestKeywordMatched']

    RetailerGroup group
    String clientId

    String code
    String name
    String currencyCode
    String baseUrl
    String address
    String city
    String state
    String zip
    String country
    String shipToCountry
    String phone
    String fax
    String email
    Boolean listing = false
    Boolean download = false
    Status status = Status.PUBLISHED
    Boolean disabled = false
    Boolean removeDuplicate = Boolean.FALSE
    Integer duplicateIndex

    FileSource fileSource = FileSource.INTERNAL
    Separator outputDelimiter = Separator.PIPE
    String outputTitle // Override title

    String selector
    Boolean ignoreSslError = false
    Boolean saveAsIs = false
    PageType pageType = PageType.AUTO_DETECT
    RequestType requestType = RequestType.GET

    AuthType authType = AuthType.NONE
    String authUsername
    String authPassword

    // URL & Pagination
    String url
    Boolean multipleUrl = false
    Boolean usePagination
    PaginationType paginationType
    String paginationParams // additional url params

    // PaginationType.PAGE_NUMBER
    String pageNumberName
    Integer firstPage
    Integer lastPage

    // PaginationType.LIMIT_OFFSET
    String pageLimitName
    Integer itemsPerPage
    String pageOffsetName
    Integer lastOffset

    // html details page
    Boolean digFurther = false
    String detailsSelectors
    String detailsLinkSelector
    DataPlacement detailsLinkPlacement
    String detailsLinkPropertyName

    // CSV
    CsvQuoteMode csvQuoteMode

    // Separated Values
    String dataBorder

    // CSV & Separated Values
    Integer rowStart
    Separator delimiter

    // Async type
    Boolean asyncDownload = false
    Integer asyncWaitTimeOut
    String asyncWaitSelector
    Boolean detailsIsAsync = false
    String detailsSelectorToWait

    // Async type
    // PaginationType.FOLLOW
    String paginationSelector

    // Validations
    BigDecimal allowedMargin // Price global validation
    Integer minValuesExist // Row data validation :: Min Crawl Properties that have values

    // Encoding
    String charsetEncoding
    String outputEncoding
    String loaderEncoding

    // Thread settings
    Integer maxThreads
    Long threadIntervals
    Boolean crawlSeparately = Boolean.FALSE

    // Server settings
    String serverIdentifier

    // Wine matcher
    Integer priceIndex
    String historyIndexes
    String keywordIndexes
    String descriptionIndexes
    Integer vintageIndex
    Boolean searchVintage = true
    Boolean vintageDefaultNv = false
    Integer bottleSizeIndex
    Boolean searchBottleSize = true
    Boolean bottleSizeDefaultBlank = false
    Integer skuIndex
    Integer taxStatusIndex
    Integer urlIndex
    String historyName
    Boolean disableHistory = Boolean.FALSE
    Boolean disableKeyword = Boolean.FALSE

    // Upload Tasting Note
    Boolean uploadTastingNote = false
    String tastingNoteUsername
    String tastingNotePassword

    Date dateCreated
    Date lastUpdated

    JobLog latestJob
    MatchedLog latestHistoryMatched
    MatchedLog latestKeywordMatched

    static constraints = {
        group nullable: true
        clientId nullable: true

        code unique: true
        name nullable: true, blank: true
        currencyCode nullable: true, blank: true
        baseUrl nullable: true, blank: true
        address nullable: true, blank: true
        city nullable: true, blank: true
        state nullable: true, blank: true
        zip nullable: true, blank: true
        country nullable: true, blank: true
        shipToCountry nullable: true, blank: true
        phone nullable: true, blank: true
        fax nullable: true, blank: true
        email nullable: true, blank: true
        status nullable: true
        duplicateIndex nullable: true

        fileSource nullable: true
        outputTitle nullable: true, blank: true

        selector nullable: true, blank: true
        saveAsIs nullable: true
        requestType nullable: true

        authType nullable: true
        authUsername nullable: true, blank: true
        authPassword nullable: true, blank: true

        // URL & Pagination
        url nullable: true
        multipleUrl nullable: true
        usePagination nullable: true
        paginationType nullable: true
        paginationParams nullable: true, blank: true

        pageNumberName nullable: true, blank: true
        firstPage nullable: true
        lastPage nullable: true

        pageLimitName nullable: true, blank: true
        itemsPerPage nullable: true
        pageOffsetName nullable: true, blank: true
        lastOffset nullable: true

        // html details page
        detailsSelectors nullable: true, blank: true
        detailsLinkSelector nullable: true, blank: true
        detailsLinkPlacement nullable: true
        detailsLinkPropertyName nullable: true, blank: true

        // CSV
        csvQuoteMode nullable: true

        // Separated Values
        rowStart nullable: true
        delimiter nullable: true
        dataBorder nullable: true, blank: true

        // Async type
        asyncDownload nullable: true
        asyncWaitTimeOut nullable: true
        asyncWaitSelector nullable: true
        paginationSelector nullable: true
        detailsIsAsync nullable: true
        detailsSelectorToWait nullable: true

        // Validations
        allowedMargin nullable: true
        minValuesExist nullable: true

        // Encoding
        charsetEncoding nullable: true, blank: true
        outputEncoding nullable: true, blank: true
        loaderEncoding nullable: true, blank: true

        // Thread settings
        maxThreads nullable: true
        threadIntervals nullable: true
        crawlSeparately nullable: true

        // Server settings
        serverIdentifier nullable: true, blank: true

        // Wine matcher
        priceIndex nullable: true
        historyIndexes nullable: true, blank: true
        keywordIndexes nullable: true, blank: true
        descriptionIndexes nullable: true, blank: true
        vintageIndex nullable: true
        bottleSizeIndex nullable: true
        skuIndex nullable: true
        taxStatusIndex nullable: true
        urlIndex nullable: true
        historyName nullable: true, blank: true
        disableHistory nullable: true
        disableKeyword nullable: true

        // Upload Tasting Note
        tastingNoteUsername nullable: true, blank: true
        tastingNotePassword nullable: true, blank: true
    }

    static mapping = {
        ignoreSslError defaultValue: "0"
        outputTitle sqlType: "text"
        paginationParams sqlType: "text"
        url sqlType: "text"
        multipleUrl defaultValue: "0"
        outputDelimiter defaultValue: Separator.PIPE
        requestType defaultValue: RequestType.GET
        disabled defaultValue: "0"
        removeDuplicate defaultValue: "0"
        saveAsIs defaultValue: "0"
        digFurther defaultValue: "0"
        searchVintage defaultValue: "1"
        vintageDefaultNv defaultValue: "0"
        searchBottleSize defaultValue: "1"
        bottleSizeDefaultBlank defaultValue: "0"
        uploadTastingNote defaultValue: "0"
        crawlProperties sort: "id"
        crawlUrls sort: "id"
    }

    Separator getMatchInputSeparator() {
        return outputDelimiter
    }

    JobLog getLatestJob() {
        if (disabled) return null
        if (!latestJob) {
            latestJob = JobLog.createCriteria().get() {
                and {
                    eq("code", code)
                }
                order("dateCreated", "desc")
                maxResults(1)
            } as JobLog
        }
        return latestJob
    }

    /**
     *
     * @return totalRecords of the last jobLog for selected retailer
     */
    Integer getFoundRecords() {
        if (disabled) return 0

        return getLatestJob()?.totalRecords ?: 0
    }

    /**
     * Get date start of the latest job
     * @return Date
     */
    Date getLatestRun() {
        if (disabled) return null

        getLatestJob()?.dateStart
    }

    Boolean getProceed() {
        return getLatestJob()?.status == Status.RUNNING
    }

    /**
     *
     * @return average of the totalRecords of last 30 success jobLog for selected retailer
     */
    def getAverageRecords() {
        def lastLogs = JobLog.createCriteria().list(sort: "dateCreated", order: "desc") {
            projections {
                property('totalRecords')
            }
            and {
                eq("code", code)
                eq("status", Status.SUCCESS)
                eq("validationStatus", Status.SUCCESS)
                ne("totalRecords", 0)
            }
            order("dateCreated", "desc")
            maxResults(20)
        }
        return lastLogs && lastLogs.size() > 0 ? lastLogs?.sum() / lastLogs?.size() : 0
    }

    /**
     *
     * @return average of the totalPrices of last 30 success jobLog for selected retailer
     */
    def getAveragePrices() {
        def avr = BigDecimal.ZERO
        def lastLogs = JobLog.createCriteria().list(sort: "dateCreated", order: "desc") {
            projections {
                property('totalPrices')
            }
            and {
                eq("code", code)
                eq("status", Status.SUCCESS)
                eq("validationStatus", Status.SUCCESS)
                ne("totalPrices", BigDecimal.ZERO)
                ne("totalRecords", 0)
            }
            maxResults(20)
        }

        if (lastLogs && lastLogs.size() > 0) {
            avr = lastLogs?.sum() / lastLogs?.size()
        }
        return avr
    }

    def getListHistoryIndexes() {
        def iList = historyIndexes?.tokenize(",")
        iList?.removeAll {
            !it.trim()?.isInteger()
        }
        return iList*.toInteger() ?: []
    }

    def getListKeywordIndexes() {
        def iList = keywordIndexes?.tokenize(",")
        iList?.removeAll {
            !it.trim()?.isInteger()
        }
        return iList*.toInteger() ?: []
    }

    def getListDescriptionIndexes() {
        def iList = descriptionIndexes?.tokenize(",")
        iList?.removeAll {
            !it.trim()?.isInteger()
        }
        return iList*.toInteger() ?: []
    }

    def getHistoryFileName() {
        if (!historyName || "-".equalsIgnoreCase(historyName)) {
            return code
        }
        return historyName
    }

    def getLatestHistoryMatched() {
        if (!latestHistoryMatched) {
            MatchedLog matchedLog = MatchedLog.createCriteria().get {
                eq("code", code)
                eq("processKeyword", Boolean.FALSE)
                order("dateCreated", "desc")
                maxResults(1)
            } as MatchedLog
            latestHistoryMatched = matchedLog
        }
        return latestHistoryMatched
    }

    def getLatestKeywordMatched() {
        if (!latestKeywordMatched) {
            MatchedLog matchedLog = MatchedLog.createCriteria().get {
                eq("code", code)
                eq("processKeyword", Boolean.TRUE)
                order("dateCreated", "desc")
                maxResults(1)
            } as MatchedLog
            latestKeywordMatched = matchedLog
        }
        return latestKeywordMatched
    }
}
