package wine.db

import org.apache.commons.lang.StringUtils
import wine.db.enums.Source

class Wine {

    static hasMany = [wineDuplications: WineDuplication, retailerWines: RetailerWine]

    static transients = [
            'name', 'fullName', 'rating', 'ratingHighNumber', 'ratingLowNumber',
            'priceHigh', 'priceLow', 'onSale', 'availableCountries', 'joinedAppellation'
    ]

    String id
    String vintage
    String maturity

    String ratingLow
    String ratingHigh
    String ratingQ

    String imageUrl

    WineDb wineDb
    Appellation appellation

    String mongoId
    String wineNId
    String wineNRating

    Source source

    Date dateCreated
    Date lastUpdated

    static constraints = {
        id length: 12
        maturity nullable: true, blank: true
        ratingLow nullable: true, blank: true, length: 10
        ratingHigh nullable: true, blank: true, length: 10
        ratingQ nullable: true, blank: true, length: 10
        imageUrl nullable: true, blank: true
        appellation nullable: true
        mongoId unique: true, nullable: true
        wineNId nullable: true
        wineNRating nullable: true, blank: true, length: 10

        wineDb unique: 'vintage'

        source nullable: true
    }

    static mapping = {
        datasource "wineDatabase"
        id column: 'id', generator: 'assigned'
        imageUrl sqlType: "text"
        wineNId column: 'wine_n_id'
        wineNRating column: 'wine_n_rating'
    }

    def getName() {
        return wineDb?.getDisplayName()
    }

    def getFullName() {
        return vintage + " " + wineDb?.getDisplayName()
    }

    def getRating() {

        if (StringUtils.isBlank(ratingLow) && StringUtils.isBlank(ratingHigh)) {
            return "NA"
        }

        Double low = getRatingLowNumber()
        Double high = getRatingHighNumber()
        if ( low > high ) {
            Double temp = low
            low = high
            high = temp
        }

        String template
        if (high == low || StringUtils.isBlank(ratingLow)) {
            template = '%2$s%3$s'
        } else if (StringUtils.isBlank(ratingHigh)) {
            template = '%1$s%3$s'
        } else {
            template = '%1$s - %2$s%3$s'
        }

        return sprintf(template, low, high, ratingQ?:"")
    }

    def getRatingHighNumber() {
        Double high = (double) 0
        try {
            high = ratingHigh?.find( /-?\d+\.\d*|-?\d*\.\d+|-?\d+/ )?.toDouble()
        } catch (Exception e) { /*do nothing*/ }

        return high
    }

    def getRatingLowNumber() {
        Double low = (double) 0
        try {
            low = ratingLow?.find( /-?\d+\.\d*|-?\d*\.\d+|-?\d+/ )?.toDouble()
        } catch (Exception e) { /*do nothing*/ }

        return low
    }

    def getPriceHigh() {
        BigDecimal high = RetailerWine.wineDatabase.createCriteria().get {
            projections {
                max('usdPrice')
            }
            and {
                isNotNull("usdPrice")
                eq("available", true)
                eq("wine", this)
            }
            maxResults(1)
        }

        return high
    }

    def getPriceLow() {
        BigDecimal low = RetailerWine.wineDatabase.createCriteria().get {
            projections {
                min('usdPrice')
            }
            and {
                isNotNull("usdPrice")
                eq("available", true)
                eq("wine", this)
            }
            maxResults(1)
        }

        return low
    }

    public Boolean isOnSale() {
        def avail = RetailerWine.wineDatabase.createCriteria().get {
            projections {
                count()
            }
            and {
                eq("available", true)
                eq("wine", this)
            }
            maxResults(1)
        }

        return avail > 0
    }

    /**
     * Get available countries (on sale)
     * @return List<Map<String, String>>
     *     [code: "", name: ""]
     */
    def getAvailableCountries() {

        def wines = RetailerWine.wineDatabase.createCriteria().list {
            projections {
                retailer {
                    country {
                        distinct("isoCode2")
                        property("name")
                    }
                }
            }
            and {
                eq("available", true)
                eq("wine", this)
            }
        }

        def countries = []
        wines?.each {
            countries.push([code: it?.getAt(0), name: it?.getAt(1)])
        }

        return countries
    }

    def getJoinedAppellation() {

        if (appellation) {
            def appList = []
            if (appellation.site) appList.push(appellation.site.name)
            if (appellation.locale) appList.push(appellation.locale.name)
            if (appellation.location) appList.push(appellation.location.name)
            if (appellation.region) appList.push(appellation.region.name)
            if (appellation.country) appList.push(appellation.country.name)

            if (appList?.size() > 0) {
                return appList?.join(", ")
            }
        }

        return ""
    }
}
