package wine.alert

import enums.AliasType
import groovy.time.TimeCategory
import groovyx.gpars.GParsPool
import org.apache.commons.collections.CollectionUtils
import org.apache.commons.io.FilenameUtils
import org.apache.commons.lang3.StringUtils
import org.springframework.beans.factory.annotation.Value
import utils.CharsetUtils
import wine.alert.utils.ServerUtils
import wine.alert.enums.EliminationType
import wine.alert.enums.KeywordType
import wine.alert.enums.ReservedType
import wine.alert.enums.SubstitutionType
import wine.db.ProducerKeyword

import java.text.SimpleDateFormat
import java.util.concurrent.atomic.AtomicInteger
import java.util.regex.Pattern

/**
 * Cron Job to watch and parse config files.<br>
 * Config files are files inside config directory ({application.config.location}).
 */
class ConfigFilesWatcherJob {

    private static final Pattern PATTERN_WINE_DB = Pattern.compile(".*/winedb.*\\.txt")
    private static final Pattern PATTERN_WINE_ALIASES = Pattern.compile(".*/wine-aliases\\.txt")
    private static final Pattern PATTERN_WINE_APPELLATIONS = Pattern.compile(".*/wine-appellations.*\\.txt")
    private static final Pattern PATTERN_REGION_MAP = Pattern.compile(".*/region-map.*\\.txt")

    private static final Pattern PATTERN_GLOBAL_NOISE = Pattern.compile(".*/global-noise.*\\.txt")
    private static final Pattern PATTERN_PRODUCER_NOISE = Pattern.compile(".*/producer-noise.*\\.txt")
    private static final Pattern PATTERN_WINE_NOISE = Pattern.compile(".*/wine-noise.*\\.txt")

    private static final Pattern PATTERN_MASS_CHANGES = Pattern.compile(".*/mass-changes\\.txt")
    private static final Pattern PATTERN_CUSTOM_CHANGES = Pattern.compile(".*/custom-changes\\.txt")
    private static final Pattern PATTERN_PRODUCER_ELIMINATIONS = Pattern.compile(".*/producer-elims\\.txt")
    private static final Pattern PATTERN_WINE_ELIMINATIONS = Pattern.compile(".*/wine-elims\\.txt")
    private static final Pattern PATTERN_COLOR_SUBSTITUTION = Pattern.compile(".*/color-substitutions\\.txt")
    private static final Pattern PATTERN_COLOR_NOT_SUBSTITUTION = Pattern.compile(".*/color-not-strings\\.txt")
    private static final Pattern PATTERN_SPREAD_TABLES = Pattern.compile(".*/[a-z]{2}[0-9]\\.txt")
    private static final Pattern PATTERN_RED_OPPOSITE = Pattern.compile(".*/red-opposites\\.txt")
    private static final Pattern PATTERN_RESERVE_WORDS = Pattern.compile(".*/reserve-words\\.txt")
    private static final Pattern PATTERN_ROSE_OPPOSITE = Pattern.compile(".*/rose-opposites\\.txt")
    private static final Pattern PATTERN_WHITE_OPPOSITE = Pattern.compile(".*/white-opposites\\.txt")

    def concurrent = false

    static triggers = {
        //cron name: 'ConfigFilesWatcher', cronExpression: "0 0 * * * ?"
        simple repeatInterval: 900000l, startDelay: 5000l // execute job once in 15 minutes
    }

    @Value('${application.config.location}')
    private String configLocation

    @Value('${application.config.backupLocation}')
    private String configBackupLocation

    def inputService

    def execute() {
        // execute job
        //log.info("Config file watcher is running...")

        if (!ServerUtils.isDefault()) {
            return
        }

        if (inputService.isImportConfigProcessing) {
            log.info("Import currently processing.")
            return
        }
        inputService.setIsImportConfigProcessing(true)

        String fileName
        new File(configLocation)?.eachFile() { file ->
            // Handle file modification process (eg: upload is in progress)
            def lastModified = 0
            while(lastModified != file.lastModified()) {
                sleep(20000)
                lastModified = file.lastModified()
            }

            fileName = file.getAbsolutePath()

            String patternString = ".*/wine-keywords-input.*\\.txt"
            if (fileName?.matches(Pattern.compile(patternString))) {
                log.info("New wine keyword file found: " + fileName)
                try {
                    inputService.importWineKeywords(fileName)
                    backupFile(fileName, true)
                } catch (Exception e){
                    log.error(e?.message, e)
                    backupFile(fileName, false)
                }
            }

            patternString = ".*/wine-bottle-sizes.*\\.txt"
            if (fileName?.matches(Pattern.compile(patternString))) {
                log.info("New wine bottle sizes file found: " + fileName)
                try {
                    inputService.importBottleSizes(fileName)
                    backupFile(fileName, true)
                } catch (Exception e){
                    log.error(e?.message, e)
                    backupFile(fileName, false)
                }
            }

            patternString = ".*/varieties.*\\.txt"
            if (fileName?.matches(Pattern.compile(patternString))) {
                log.info("New wine varieties file found: " + fileName)
                try {
                    inputService.importVarieties(fileName)
                    backupFile(fileName, true)
                } catch (FileNotFoundException e){
                    log.warn(e?.message)
                } catch (Exception e){
                    log.error(e?.message, e)
                    backupFile(fileName, false)
                }
            }

            patternString = ".*/config-file.*\\.txt"
            if (fileName?.matches(Pattern.compile(patternString))) {
                log.info("New Retailers Config file found: " + fileName)
                try {
                    inputService.importRetailers(fileName)
                    backupFile(fileName, true)
                } catch (Exception e){
                    log.error(e?.message, e)
                    backupFile(fileName, false)
                }
            }

            patternString = ".*/retailer-table.*\\.txt"
            if (fileName?.matches(Pattern.compile(patternString))) {
                log.info("New Retailers Table file found: " + fileName)
                try {
                    inputService.importRetailersTable(fileName)
                    backupFile(fileName, true)
                } catch (Exception e){
                    log.error(e?.message, e)
                    backupFile(fileName, false)
                }
            }

            if (fileName?.matches(PATTERN_WINE_DB)) {
                log.info("New Wine Alert DB file found: " + fileName)
                try {
                    inputService.importWineDb(fileName)
                    backupFile(fileName, true)
                } catch (Exception e){
                    log.error(e?.message, e)
                    backupFile(fileName, false)
                }
            }

            patternString = ".*/mongomap.*\\.txt"
            if (fileName?.matches(Pattern.compile(patternString))) {
                log.info("New Mongo Wines Map file found: " + fileName)
                try {
                    inputService.importMongoWinesMap(fileName)
                    backupFile(fileName, true)
                } catch (Exception e){
                    log.error(e?.message, e)
                    backupFile(fileName, false)
                }
            }

            patternString = ".*/producer-indexing.*\\.txt"
            if (fileName?.matches(Pattern.compile(patternString))) {
                log.info("New Producer Indexing Keywords found at: " + fileName)
                try {
                    importProducerIndexingKeywords(fileName)
                    backupFile(fileName, true)
                } catch (Exception e){
                    log.error(e?.message, e)
                    backupFile(fileName, false)
                }
            }

            if (fileName?.matches(PATTERN_WINE_APPELLATIONS)) {
                log.info("New Wine Appellations file found: " + fileName)
                try {
                    inputService.importWineAppellations(fileName)
                    backupFile(fileName, true)
                } catch (Exception e){
                    log.error(e?.message, e)
                    backupFile(fileName, false)
                }
            }

            if (fileName?.matches(PATTERN_REGION_MAP)) {
                log.info("New Region Maps file found: " + fileName)
                try {
                    inputService.importRegionMap(fileName)
                    backupFile(fileName, true)
                } catch (Exception e){
                    log.error(e?.message, e)
                    backupFile(fileName, false)
                }
            }

            if (fileName?.matches(PATTERN_GLOBAL_NOISE)) {
                log.info("New Global Noise file found: " + fileName)
                try {
                    inputService.importGlobalNoise(fileName)
                    backupFile(fileName, true)
                } catch (Exception e){
                    log.error(e?.message, e)
                    backupFile(fileName, false)
                }
            }

            if (fileName?.matches(PATTERN_PRODUCER_NOISE)) {
                log.info("New Producer Noise file found: " + fileName)
                try {
                    inputService.importProducerNoise(fileName)
                    backupFile(fileName, true)
                } catch (Exception e){
                    log.error(e?.message, e)
                    backupFile(fileName, false)
                }
            }

            if (fileName?.matches(PATTERN_WINE_NOISE)) {
                log.info("New Wine Noise file found: " + fileName)
                try {
                    inputService.importWineNoise(fileName)
                    backupFile(fileName, true)
                } catch (Exception e){
                    log.error(e?.message, e)
                    backupFile(fileName, false)
                }
            }

            if (fileName?.matches(PATTERN_WINE_ALIASES)) {
                log.info("New Wine Aliases found at: " + fileName)
                try {
                    inputService.importAliases(fileName, AliasType.WINE)
                    backupFile(fileName, true)
                } catch (Exception e){
                    log.error(e?.message, e)
                    backupFile(fileName, false)
                }
            }
        }

        new File(configLocation + "/keyword")?.eachFile() { file ->
            // Handle file modification process (eg: upload is in progress)
            def lastModified = 0
            while(lastModified != file.lastModified()) {
                sleep(20000)
                lastModified = file.lastModified()
            }

            fileName = file.getAbsolutePath()

            if (fileName?.matches(PATTERN_MASS_CHANGES)) {
                log.info("New Mass Changes found at: " + fileName)
                try {
                    inputService.importAliases(fileName, AliasType.MASS_CHANGES)
                    backupKeywordFile(fileName, true)
                } catch (Exception e){
                    log.error(e?.message, e)
                    backupKeywordFile(fileName, false)
                }
            }

            if (fileName?.matches(PATTERN_CUSTOM_CHANGES)) {
                log.info("New Custom Changes found at: " + fileName)
                try {
                    inputService.importAliases(fileName, AliasType.CUSTOM_CHANGES)
                    backupKeywordFile(fileName, true)
                } catch (Exception e){
                    log.error(e?.message, e)
                    backupKeywordFile(fileName, false)
                }
            }

            if (fileName?.matches(PATTERN_PRODUCER_ELIMINATIONS)) {
                log.info("New Producer eliminations found at: " + fileName)
                try {
                    inputService.importEliminations(fileName, EliminationType.PRODUCER)
                    backupKeywordFile(fileName, true)
                } catch (Exception e){
                    log.error(e?.message, e)
                    backupKeywordFile(fileName, false)
                }
            }

            if (fileName?.matches(PATTERN_WINE_ELIMINATIONS)) {
                log.info("New Wine eliminations found at: " + fileName)
                try {
                    inputService.importEliminations(fileName, EliminationType.WINE)
                    backupKeywordFile(fileName, true)
                } catch (Exception e){
                    log.error(e?.message, e)
                    backupKeywordFile(fileName, false)
                }
            }

            if (fileName?.matches(PATTERN_COLOR_SUBSTITUTION)) {
                log.info("New Color substitution found at: " + fileName)
                try {
                    inputService.importSubstitution(fileName, SubstitutionType.COLOR)
                    backupKeywordFile(fileName, true)
                } catch (Exception e){
                    log.error(e?.message, e)
                    backupKeywordFile(fileName, false)
                }
            }

            if (fileName?.matches(PATTERN_COLOR_NOT_SUBSTITUTION)) {
                log.info("New Color Not substitution found at: " + fileName)
                try {
                    inputService.importSubstitution(fileName, SubstitutionType.COLOR_NOT)
                    backupKeywordFile(fileName, true)
                } catch (Exception e){
                    log.error(e?.message, e)
                    backupKeywordFile(fileName, false)
                }
            }

            if (fileName?.matches(PATTERN_RED_OPPOSITE)) {
                log.info("New Red Opposite found at: " + fileName)
                try {
                    inputService.importReservedWords(fileName, ReservedType.RED_OPPOSITE)
                    backupKeywordFile(fileName, true)
                } catch (Exception e){
                    log.error(e?.message, e)
                    backupKeywordFile(fileName, false)
                }
            }

            if (fileName?.matches(PATTERN_RESERVE_WORDS)) {
                log.info("New Reserve Words found at: " + fileName)
                try {
                    inputService.importReservedWords(fileName, ReservedType.RESERVE_WORD)
                    backupKeywordFile(fileName, true)
                } catch (Exception e){
                    log.error(e?.message, e)
                    backupKeywordFile(fileName, false)
                }
            }

            if (fileName?.matches(PATTERN_ROSE_OPPOSITE)) {
                log.info("New Rose Opposite found at: " + fileName)
                try {
                    inputService.importReservedWords(fileName, ReservedType.ROSE_OPPOSITE)
                    backupKeywordFile(fileName, true)
                } catch (Exception e){
                    log.error(e?.message, e)
                    backupKeywordFile(fileName, false)
                }
            }

            if (fileName?.matches(PATTERN_WHITE_OPPOSITE)) {
                log.info("New White Opposite found at: " + fileName)
                try {
                    inputService.importReservedWords(fileName, ReservedType.WHITE_OPPOSITE)
                    backupKeywordFile(fileName, true)
                } catch (Exception e){
                    log.error(e?.message, e)
                    backupKeywordFile(fileName, false)
                }
            }
        }

        new File(configLocation + "/keyword/spread_tables")?.eachFile() { file ->
            // Handle file modification process (eg: upload is in progress)
            def lastModified = 0
            while(lastModified != file.lastModified()) {
                sleep(20000)
                lastModified = file.lastModified()
            }

            fileName = file.getAbsolutePath()

            if (fileName?.matches(PATTERN_SPREAD_TABLES)) {
                try {
                    String name = FilenameUtils.getBaseName(fileName)
                    String regionCode = name.take(2)
                    String eliminationNum = name.drop(2)
                    KeywordType keywordType = null
                    switch (eliminationNum) {
                        case "1":
                            keywordType = KeywordType.BASE
                            break
                        case "2":
                            keywordType = KeywordType.APPELLATION
                            break
                        case "3":
                            keywordType = KeywordType.COLOR
                            break
                        case "4":
                            keywordType = KeywordType.COMMON_MODIFIER
                            break
                    }
                    if (keywordType) {
                        log.info("New Spread tables ${regionCode}${eliminationNum} ${keywordType?.name()} found at: " + fileName)

                        inputService.importSpreadTable(fileName, regionCode, keywordType)
                        backupKeywordSpreadFile(fileName, true)
                    }
                } catch (Exception e){
                    log.error(e?.message, e)
                    backupKeywordSpreadFile(fileName, false)
                }
            }
        }

        inputService.setIsImportConfigProcessing(false)
    }

    /**
     * Backups file inside config directory ({application.config.location}) after parsed.
     * @param fileName the file name
     * @param success the parse status
     */
    void backupFile(String fileName, Boolean success) {
        String prefix = success ? "success" : "failed"

        try {
            Date date = new Date()
            SimpleDateFormat sdf = new SimpleDateFormat("yyyyMMddHHmmss")
            File dir = new File(configBackupLocation + "/" + prefix)
            dir.mkdirs()

            File file = new File(fileName)
            String newFileName = file.getName()?.replace(".txt", "-" + sdf.format(date) + ".txt")
            boolean fileMoved = file.renameTo(new File(dir, newFileName))
            if (!fileMoved) {
                log.warn("Failed to move file: " + fileName)
            }
        } catch (Exception e) {
            log.error(e?.message, e)
        }
    }

    /**
     * Backups file inside config/keyword directory ({application.config.location}/keyword) after parsed.
     * @param fileName the file name
     * @param success the parse status
     */
    void backupKeywordFile(String fileName, Boolean success) {
        String prefix = success ? "success" : "failed"

        try {
            Date date = new Date()
            SimpleDateFormat sdf = new SimpleDateFormat("yyyyMMddHHmmss")
            File dir = new File(configBackupLocation + "/" + prefix + "/keyword")
            dir.mkdirs()

            File file = new File(fileName)
            String newFileName = file.getName()?.replace(".txt", "-" + sdf.format(date) + ".txt")
            boolean fileMoved = file.renameTo(new File(dir, newFileName))
            if (!fileMoved) {
                log.warn("Failed to move file: " + fileName)
            }
        } catch (Exception e) {
            log.error(e?.message, e)
        }
    }

    /**
     * Backups file inside config/keyword/spread_tables directory ({application.config.location}/keyword/spread_tables) after parsed.
     * @param fileName the file name
     * @param success the parse status
     */
    void backupKeywordSpreadFile(String fileName, Boolean success) {
        String prefix = success ? "success" : "failed"

        try {
            Date date = new Date()
            SimpleDateFormat sdf = new SimpleDateFormat("yyyyMMddHHmmss")
            File dir = new File(configBackupLocation + "/" + prefix + "/keyword/spread_tables")
            dir.mkdirs()

            File file = new File(fileName)
            String newFileName = file.getName()?.replace(".txt", "-" + sdf.format(date) + ".txt")
            boolean fileMoved = file.renameTo(new File(dir, newFileName))
            if (!fileMoved) {
                log.warn("Failed to move file: " + fileName)
            }
        } catch (Exception e) {
            log.error(e?.message, e)
        }
    }

    /**
     * Parses a tsv file and update the {@link wine.db.ProducerKeyword}
     * @param pathname the file pathname to be parsed
     */
    void importProducerIndexingKeywords(String pathname) {
        log.info("Import Producer Keywords started.")
        File file = new File(pathname)
        if (!file || !file.isFile()) return

        Date date = new Date()
        String charset = CharsetUtils.detect(pathname) ?: "UTF-8"
        List<List<String>> records = new ArrayList<>()
        file.withReader(charset) { reader ->
            def line
            while ((line = reader.readLine()) != null) {
                def record = ((String) line)?.split("\t")?.toList()
                if (CollectionUtils.isNotEmpty(record)) {
                    String id = record.get(0)?.trim()
                    if (StringUtils.isNotBlank(id)) {
                        String keyword = record.size() > 1 ? record.get(1)?.trim() : ''
                        records.add([id, keyword])
                    }
                }
            }
        }

        log.info("Producer Keywords size: " + records.size())
        if (CollectionUtils.isNotEmpty(records)) {
            AtomicInteger counter = new AtomicInteger()
            records.collate(200).each { list ->
                ProducerKeyword.withNewTransaction {
                    GParsPool.withPool {
                        list.eachParallel { r ->
                            inputService.proceedProducerKeyword(r.get(0), r.get(1))
                            counter.getAndIncrement()
                        }
                    }
                }
                log.info("Producer Keywords processed: " + counter.intValue())
            }

            List<String> kwIds = new ArrayList<>()
            ProducerKeyword.withNewTransaction {
                kwIds = inputService.getProducerIdsOfProducerKeywords()
            }
            log.info("Producer Keywords in system: " + kwIds?.size())

            List<String> fileWineKeywords = records.collect { it.get(0) }
            List<String> tobeRemoved = kwIds?.findAll { !fileWineKeywords.contains(it) }
            if (CollectionUtils.isNotEmpty(tobeRemoved)) {
                tobeRemoved.unique()
                counter.set(0)
                tobeRemoved.collate(200).each { list ->
                    ProducerKeyword.withNewTransaction {
                        GParsPool.withPool {
                            list.eachParallel { id ->
                                inputService.removeProducerKeyword(id)
                                counter.getAndIncrement()
                            }
                        }
                    }
                    log.info("Producer Keywords removed: " + counter.intValue())
                }
            }
        }
        log.info("Import Producer Keywords completed. Execution times: " + TimeCategory.minus(new Date(), date))
    }
}
