package wine.alert

import enums.FileSource
import enums.PageType
import enums.Status
import groovy.time.TimeCategory
import groovyx.gpars.GParsPool
import wine.alert.utils.ServerUtils

/**
 * Cron Job for crawler.
 */
class CrawlAllJob {
    static triggers = {
        //cron name: "CrawlAll", cronExpression: "0 59 0 ? * *" // everyday at 0:59am
    }

    def concurrent = false

    def webCrawlerService
    def logsService
    def crawlerReportService

    def execute() {
        Date today = new Date()
        log.info("Crawl Job started.")

        // Single page WebCrawlers, use multiple thread
        WebCrawler.withTransaction {
            def webs = WebCrawler.createCriteria().list {
                projections {
                    property("code")
                    property("name")
                    property("url")
                }
                and {
                    ne("pageType", PageType.HTML)
                    ne("pageType", PageType.HTML_ASYNC)
                    or {
                        isNull("fileSource")
                        eq("fileSource", FileSource.INTERNAL)
                    }
                    or {
                        isNull("crawlSeparately")
                        eq("crawlSeparately", Boolean.FALSE)
                    }
                    eq("disabled", Boolean.FALSE)
                    ne("status", Status.DELETED)
                    if (ServerUtils.isDefault()) {
                        or {
                            eq("serverIdentifier", ServerUtils.DEFAULT_SERVER)
                            eq("serverIdentifier", "")
                            isNull("serverIdentifier")
                        }
                    } else {
                        eq("serverIdentifier", ServerUtils.getCurrent())
                    }
                }
                order("code", "asc")
            }

            if (webs?.size() > 0) {
                GParsPool.withPool {
                    webs?.eachParallel { def crawler ->
                        proceed(crawler)
                    }
                }
            }
        }

        // WebCrawlers with pagination, use single thread
        WebCrawler.withTransaction {
            def paginationWebs = WebCrawler.createCriteria().list {
                projections {
                    property("code")
                    property("name")
                    property("url")
                }
                and {
                    eq("pageType", PageType.HTML)
                    or {
                        isNull("fileSource")
                        eq("fileSource", FileSource.INTERNAL)
                    }
                    or {
                        isNull("crawlSeparately")
                        eq("crawlSeparately", Boolean.FALSE)
                    }
                    eq("disabled", Boolean.FALSE)
                    ne("status", Status.DELETED)
                    if (ServerUtils.isDefault()) {
                        or {
                            eq("serverIdentifier", ServerUtils.DEFAULT_SERVER)
                            eq("serverIdentifier", "")
                            isNull("serverIdentifier")
                        }
                    } else {
                        eq("serverIdentifier", ServerUtils.getCurrent())
                    }
                }
                order("code", "asc")
            }

            paginationWebs?.each { def crawler ->
                proceed(crawler)
            }
        }

        // HTML_ASYNC WebCrawlers, use single thread
        WebCrawler.withTransaction {
            def asyncPages = WebCrawler.createCriteria().list {
                projections {
                    property("code")
                    property("name")
                    property("url")
                }
                and {
                    eq("pageType", PageType.HTML_ASYNC)
                    or {
                        isNull("fileSource")
                        eq("fileSource", FileSource.INTERNAL)
                    }
                    or {
                        isNull("crawlSeparately")
                        eq("crawlSeparately", Boolean.FALSE)
                    }
                    eq("disabled", Boolean.FALSE)
                    ne("status", Status.DELETED)
                    if (ServerUtils.isDefault()) {
                        or {
                            eq("serverIdentifier", ServerUtils.DEFAULT_SERVER)
                            eq("serverIdentifier", "")
                            isNull("serverIdentifier")
                        }
                    } else {
                        eq("serverIdentifier", ServerUtils.getCurrent())
                    }
                }
                order("code", "asc")
            }

            asyncPages?.each { def crawler ->
                proceed(crawler)
            }
        }

        log.info("Crawl Job completed. Execution times: " + TimeCategory.minus(new Date(), today))
        crawlerReportService.sendToSlack()
    }

    def proceed(def crawler) {
        log.info("[" + crawler[0] + "] Crawling started.")
        try {
            JobLog.withNewTransaction {
                logsService.saveJobLogs(crawler[0] as String, crawler[1] as String, crawler[2] as String,
                        Status.RUNNING, "", 0, BigDecimal.ZERO, 0, 0, null, "", new Date(), null)
            }
        } catch (Exception e) {
            log.error(e?.message, e)
        }
        WebCrawler.withNewTransaction {
            webCrawlerService.parse(crawler[0] as String)
        }
    }
}
