package wine.alert

import enums.MatchedType
import enums.RollupType
import enums.Status
import groovy.time.TimeCategory
import groovyx.gpars.GParsPool
import wine.alert.utils.ServerUtils

/**
 * Cron Job for history match.
 */
class HistoryMatchJob {
    static triggers = {
        //cron name: "MatchAllHistory", cronExpression: "0 29 1 ? * *" // everyday at 1:29am
    }

    def concurrent = false

    def matcherService
    def logsService
    def outputService

    def execute() {
        if (!ServerUtils.isDefault()) {
            return
        }

        log.info( "History Match Job started." )

        Date today = new Date()
        List<String> crawlers = new ArrayList<>()
        WebCrawler.withTransaction {
            crawlers = WebCrawler.createCriteria().list {
                projections {
                    property("code")
                }
                and {
                    ne("disableHistory", Boolean.TRUE)
                    ne("status", Status.DELETED)
                    if (ServerUtils.isDefault()) {
                        or {
                            eq("serverIdentifier", ServerUtils.DEFAULT_SERVER)
                            eq("serverIdentifier", "")
                            isNull("serverIdentifier")
                        }
                    } else {
                        eq("serverIdentifier", ServerUtils.getCurrent())
                    }
                }
                order("code", "asc")
            }
        }

        if (crawlers?.size() > 0) {
            matcherService.clearRunningMatches(MatchedType.HISTORY)
            matcherService.addToRunningMatches(crawlers, MatchedType.HISTORY)

            GParsPool.withPool {
                crawlers?.eachParallel { String code ->
                    if (!matcherService.isTerminatedMatch(code, MatchedType.HISTORY)) {
                        Date crawlDate = outputService.getLatestCrawlDate(code)
                        if (crawlDate) {
                            MatchedLog.withNewTransaction {
                                logsService.saveMatchLog(code, crawlDate, Boolean.FALSE, Status.RUNNING, today)
                            }

                            WebCrawler.withNewTransaction {
                                matcherService.execute(code, Boolean.FALSE, crawlDate)
                            }
                        }
                        matcherService.removeFromRunningMatches(code, MatchedType.HISTORY)
                    } else {
                        log.warn("[" + code + "] was terminated from history match.")
                    }
                }
            }

            log.info( "History Match Job completed. Execution times: " + TimeCategory.minus(new Date(), today) )

            RollupJob.triggerNow([rollupType: RollupType.RP])
            RollupJob.triggerNow([rollupType: RollupType.WA])
        }
    }
}
