package wine.alert

import enums.FileSource
import enums.Status
import groovy.time.TimeCategory
import wine.alert.utils.ServerUtils

/**
 * Cron Job for crawler that run separately.<br>
 * This cron is needed because some retailer feeds aren't accessible at regular Crawler Cron time.
 */
class SeparateCrawlJob {
    static triggers = {
        //cron name: "SeparateCrawl", cronExpression: "0 59 1 ? * *" // everyday at 1:59am
    }

    def webCrawlerService
    def logsService

    def execute() {
        // execute job
        Date today = new Date()
        log.info( "Separate Crawl Job started." )

        // WebCrawlers as Separate job, use single thread
        WebCrawler.withTransaction {
            def separates = WebCrawler.createCriteria().list {
                projections {
                    property("code")
                    property("name")
                    property("url")
                }
                and {
                    or {
                        isNull("fileSource")
                        eq("fileSource", FileSource.INTERNAL)
                    }
                    eq("crawlSeparately", Boolean.TRUE)
                    eq("disabled", Boolean.FALSE)
                    ne("status", Status.DELETED)
                    if (ServerUtils.isDefault()) {
                        or {
                            eq("serverIdentifier", ServerUtils.DEFAULT_SERVER)
                            eq("serverIdentifier", "")
                            isNull("serverIdentifier")
                        }
                    } else {
                        eq("serverIdentifier", ServerUtils.getCurrent())
                    }
                }
                order("code", "asc")
            }

            separates?.each { def crawler ->
                proceed(crawler)
            }
        }

        log.info( "Separate Crawl Job completed. Execution times: " + TimeCategory.minus(new Date(), today) )
    }

    def proceed(def crawler) {
        log.info("[" + crawler[0] + "] Crawling started.")
        try {
            JobLog.withNewTransaction {
                logsService.saveJobLogs(crawler[0] as String, crawler[1] as String, crawler[2] as String,
                        Status.RUNNING, "", 0, BigDecimal.ZERO, 0, 0, null, "", new Date(), null)
            }
        } catch (Exception e) {
            log.error(e?.message, e)
        }
        WebCrawler.withNewTransaction {
            webCrawlerService.parse(crawler[0] as String)
        }
    }
}
