package wine.alert

import enums.MatchedType
import enums.Status
import groovy.time.TimeCategory
import groovyx.gpars.GParsPool

import java.text.SimpleDateFormat

/**
 * Cron Job for history match and keyword match.<br>
 * Triggered by user from dashboard.
 */
class WineMatchJob {
    static triggers = {
        //simple repeatInterval: 5000l // execute job once in 5 seconds
    }

    def matcherService
    def outputService
    def logsService

    def execute(context) {
        Date today = new Date()
        List<String> codes = context.mergedJobDataMap.get('codes')
        List<Boolean> runKeywords = context.mergedJobDataMap.get('runKeywords')
        List<Date> dates = context.mergedJobDataMap.get('dates')

        SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd")

        if (codes?.size() > 0) {

            Map<String, Date> keywordCodes = new HashMap<>()
            Map<String, Date> historyCodes = new HashMap<>()
            codes.eachWithIndex{ String entry, int i ->
                Date reqDate = dates.get(i)
                if (runKeywords.get(i)) {
                    keywordCodes.put(entry, reqDate)
                    matcherService.addToRunningMatches(entry, MatchedType.KEYWORD)
                } else {
                    historyCodes.put(entry, reqDate)
                    matcherService.addToRunningMatches(entry, MatchedType.HISTORY)
                }
            }

            GParsPool.withPool {
                historyCodes?.eachParallel { String k, Date v ->
                    if (!matcherService.isTerminatedMatch(k, MatchedType.HISTORY)) {
                        if (outputService.isValidCrawlOutput(k, v)) {
                            MatchedLog.withNewTransaction {
                                logsService.saveMatchLog(k, v, Boolean.FALSE, Status.RUNNING, today)
                            }

                            WebCrawler.withNewTransaction {
                                matcherService.execute(k, Boolean.FALSE, v)
                            }
                        } else {
                            log.warn("[" + k + "] match input not found at: " + (v ? sdf.format(v) : null))
                        }
                        matcherService.removeFromRunningMatches(k, MatchedType.HISTORY)
                    } else {
                        log.warn("[" + k + "] was terminated from history match.")
                    }
                }
            }

            keywordCodes?.each { String k, Date v ->
                if (!matcherService.isTerminatedMatch(k, MatchedType.KEYWORD)) {
                    if (outputService.isValidCrawlOutput(k, v)) {
                        MatchedLog.withNewTransaction {
                            logsService.saveMatchLog(k, v, Boolean.TRUE, Status.RUNNING, today)
                        }

                        WebCrawler.withNewTransaction {
                            matcherService.execute(k, Boolean.TRUE, v)
                        }
                    } else {
                        log.warn("[" + k + "] match input not found at: " + (v ? sdf.format(v) : null))
                    }
                    matcherService.removeFromRunningMatches(k, MatchedType.KEYWORD)
                } else {
                    log.warn("[" + k + "] was terminated from keyword match.")
                }
            }

            log.info( "Wine Match Job completed. Execution times: " + TimeCategory.minus(new Date(), today) )
        }
    }
}
