package wine.alert

import enums.MatchedType
import grails.transaction.Transactional
import org.springframework.beans.factory.annotation.Value

import java.text.SimpleDateFormat
import java.util.zip.ZipEntry
import java.util.zip.ZipOutputStream

/**
 * Service to archive unused files
 */
@Transactional
class ArchiveService {

    @Value('${application.archive.location}')
    private String archiveLocation

    @Value('${application.matched.backupLocation}')
    private String matchedBackupLocation

    @Value('${application.crawlOut.backupLocation}')
    private String crawlBackupLocation

    @Value('${application.history.backupLocation}')
    private String historyBackupLocation

    @Value('${application.download.location}')
    private String downloadLocation

    /**
     * Archives input (crawler output) backup files.
     * @param date the maximum file's date to be archived
     */
    void archiveCrawlerOutputBackup(Date date) {
        SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd")
        String formattedDate = sdf.format(date)

        File[] crawlBackupFiles = new File(crawlBackupLocation).listFiles(new FilenameFilter() {
            @Override
            boolean accept(File dir, String name) {
                if (name.length() < 14) return false
                String dn = name.reverse().substring(4, 14).reverse()
                Date fileDate = sdf.parse(dn)
                return fileDate.before(date)
            }
        })
        log.debug("Crawl output backup files: " + crawlBackupFiles.length)

        if (crawlBackupFiles.length > 0) {
            byte[] bytes = new byte[1024]

            String crawlZipFileName = archiveLocation + "/input_" + formattedDate + "_" + System.currentTimeMillis() + ".zip"
            FileOutputStream fosCrawlBackup = new FileOutputStream(crawlZipFileName)
            ZipOutputStream zipOutCrawlBackup = new ZipOutputStream(fosCrawlBackup)

            // Append new files
            crawlBackupFiles.each {
                FileInputStream fis = new FileInputStream(it)
                ZipEntry zipEntry = new ZipEntry(it.getName())
                zipOutCrawlBackup.putNextEntry(zipEntry)

                int length
                while((length = fis.read(bytes)) > 0) {
                    zipOutCrawlBackup.write(bytes, 0, length)
                }
                fis.close()
            }
            zipOutCrawlBackup.close()
            fosCrawlBackup.close()

            crawlBackupFiles.each {
                it.delete()
            }
        }
    }

    /**
     * Archives keyword match backup files.
     * @param date the maximum file's date to be archived
     */
    void archiveKeywordMatchBackup(Date date) {
        SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd")
        String formattedDate = sdf.format(date)

        File[] keywordMatchedBackupFiles = new File(matchedBackupLocation + File.separator + MatchedType.KEYWORD.value()).listFiles(new FilenameFilter() {
            @Override
            boolean accept(File dir, String name) {
                if (name.length() < 22) return false
                String dn = name.reverse().substring(12, 22).reverse()
                Date fileDate = sdf.parse(dn)
                return fileDate.before(date)
            }
        })
        log.debug("Keyword match backup files: " + keywordMatchedBackupFiles.length)

        if (keywordMatchedBackupFiles.length >0) {
            byte[] bytes = new byte[1024]

            String matchedKeywordZipFileName = archiveLocation + File.separator + "matched-keyword_" + formattedDate + "_" + System.currentTimeMillis() + ".zip"
            FileOutputStream fosMatchedKeywordBackup = new FileOutputStream(matchedKeywordZipFileName)
            ZipOutputStream zipOutMatchedKeywordBackup = new ZipOutputStream(fosMatchedKeywordBackup)

            // Append new files
            keywordMatchedBackupFiles.each {
                FileInputStream fis = new FileInputStream(it)
                ZipEntry zipEntry = new ZipEntry(it.getName())
                zipOutMatchedKeywordBackup.putNextEntry(zipEntry)

                int length
                while((length = fis.read(bytes)) > 0) {
                    zipOutMatchedKeywordBackup.write(bytes, 0, length)
                }
                fis.close()
            }
            zipOutMatchedKeywordBackup.close()
            fosMatchedKeywordBackup.close()

            keywordMatchedBackupFiles.each {
                it.delete()
            }
        }
    }

    /**
     * Archives history match backup files.
     * @param date the maximum file's date to be archived
     */
    void archiveHistoryMatchBackup(Date date) {
        SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd")
        String formattedDate = sdf.format(date)

        File[] historyMatchedBackupFiles = new File(matchedBackupLocation + File.separator + MatchedType.HISTORY.value()).listFiles(new FilenameFilter() {
            @Override
            boolean accept(File dir, String name) {
                if (name.length() < 22) return false
                String dn = name.reverse().substring(12, 22).reverse()
                Date fileDate = sdf.parse(dn)
                return fileDate.before(date)
            }
        })
        log.debug("History match backup files: " + historyMatchedBackupFiles.length)

        if (historyMatchedBackupFiles.length > 0) {
            byte[] bytes = new byte[1024]

            String matchedHistoryZipFileName = archiveLocation + File.separator + "matched-history_" + formattedDate + "_" + System.currentTimeMillis() + ".zip"
            FileOutputStream fosHistoryBackup = new FileOutputStream(matchedHistoryZipFileName)
            ZipOutputStream zipOutHistoryBackup = new ZipOutputStream(fosHistoryBackup)

            // Append new files
            historyMatchedBackupFiles.each {
                FileInputStream fis = new FileInputStream(it)
                ZipEntry zipEntry = new ZipEntry(it.getName())
                zipOutHistoryBackup.putNextEntry(zipEntry)

                int length
                while((length = fis.read(bytes)) > 0) {
                    zipOutHistoryBackup.write(bytes, 0, length)
                }
                fis.close()
            }
            zipOutHistoryBackup.close()
            fosHistoryBackup.close()

            historyMatchedBackupFiles.each {
                it.delete()
            }
        }
    }

    /**
     * Archives history backup files.
     * @param date the maximum file's date to be archived
     */
    void archiveHistoryBackup(Date date) {
        SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd")
        String formattedDate = sdf.format(date)

        File[] historyBackupFiles = new File(historyBackupLocation).listFiles(new FilenameFilter() {
            @Override
            boolean accept(File dir, String name) {
                if (name.length() < 14) return false
                if (name.reverse().substring(6, 7) == "-") { // old backup format
                    String dn = name.reverse().substring(4, 14).reverse()
                    Date fileDate = sdf.parse(dn)
                    return fileDate.before(date)
                }

                // new backup format
                String dn = name.reverse().substring(11, 21).reverse()
                Date fileDate = sdf.parse(dn)
                return fileDate.before(date)
            }
        })
        log.debug("History backup files: " + historyBackupFiles.length)

        if (historyBackupFiles.length > 0) {
            byte[] bytes = new byte[1024]

            String historyZipFileName = archiveLocation + File.separator + "history_" + formattedDate + "_" + System.currentTimeMillis() + ".zip"
            FileOutputStream fosHistoryBackup = new FileOutputStream(historyZipFileName)
            ZipOutputStream zipOutHistoryBackup = new ZipOutputStream(fosHistoryBackup)

            // Append new files
            historyBackupFiles.each {
                FileInputStream fis = new FileInputStream(it)
                ZipEntry zipEntry = new ZipEntry(it.getName())
                zipOutHistoryBackup.putNextEntry(zipEntry)

                int length
                while((length = fis.read(bytes)) > 0) {
                    zipOutHistoryBackup.write(bytes, 0, length)
                }
                fis.close()
            }
            zipOutHistoryBackup.close()
            fosHistoryBackup.close()

            historyBackupFiles.each {
                it.delete()
            }
        }
    }

    /**
     * Archives rollup files.
     * @param date the maximum file's date to be archived
     */
    void archiveRollup(Date date) {
        SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd")
        SimpleDateFormat sdf2 = new SimpleDateFormat("yy.MM.dd.HHmmss")
        String formattedDate = sdf.format(date)

        File[] rollupFiles = new File(downloadLocation).listFiles(new FilenameFilter() {
            @Override
            boolean accept(File dir, String name) {
                if (name.length() < 19 || (!name.startsWith("ForSale") && !name.startsWith("WAForSale"))) return false
                String dn = name.reverse().substring(4, 19).reverse()
                Date fileDate = sdf2.parse(dn)
                return fileDate.before(date)
            }
        })
        log.debug("Rollup files: " + rollupFiles.length)

        if (rollupFiles.length > 0) {
            byte[] bytes = new byte[1024]

            String historyZipFileName = archiveLocation + File.separator + "rollup_" + formattedDate + "_" + System.currentTimeMillis() + ".zip"
            FileOutputStream fosHistoryBackup = new FileOutputStream(historyZipFileName)
            ZipOutputStream zipOutHistoryBackup = new ZipOutputStream(fosHistoryBackup)

            // Append new files
            rollupFiles.each {
                FileInputStream fis = new FileInputStream(it)
                ZipEntry zipEntry = new ZipEntry(it.getName())
                zipOutHistoryBackup.putNextEntry(zipEntry)

                int length
                while((length = fis.read(bytes)) > 0) {
                    zipOutHistoryBackup.write(bytes, 0, length)
                }
                fis.close()
            }
            zipOutHistoryBackup.close()
            fosHistoryBackup.close()

            rollupFiles.each {
                it.delete()
            }
        }
    }
}
