package wine.alert

import enums.FileSource
import enums.Status
import grails.util.Environment
import org.springframework.web.client.RestTemplate

import java.time.LocalDateTime
import java.time.ZoneId

class CrawlerReportService {
    def activeCrawlerCount() {
        activeCrawlers().count()
    }

    def successfulCrawlerCount() {
        def oneDayAgo = Date.from(LocalDateTime.now().minusDays(1).atZone(ZoneId.systemDefault()).toInstant())

        JobLog.createCriteria().get {
            eq('status', Status.SUCCESS)
            ge('dateCreated', oneDayAgo)
            'in'('code', activeCrawlers().property('code'))
            projections {
                countDistinct('code')
            }
        }
    }

    def failedCrawlers() {
        def q = JobLog.withSession { session ->
            session.createSQLQuery("""select j.code
, (select date_created from job_log where code = j.code and status = 'SUCCESS' and id < j.id order by id desc limit 1) as last_success
, c.page_type, c.id
from job_log j
inner join web_crawler c on c.code = j.code
where j.status = 'FAILED' and j.date_created > date_sub(current_timestamp, interval 1 day) and j.code in (:active)
order by last_success desc, code""")
        }
        q.setParameterList('active', activeCrawlers().property('code').list())
        q.list()
    }

    def externalCrawlerCount() {
        WebCrawler.where {
            disabled == false && listing == true && status != Status.DELETED && fileSource == FileSource.EXTERNAL
        }.count()
    }

    def htmlCrawlerSql = """select id, code, total_records
    , case when total_records = 0 and total_records_yesterday = 0 then 0 when total_records_yesterday = 0 then 1 else (total_records - total_records_yesterday) * 1.0 / total_records_yesterday end as daily_change
    , case when total_records = 0 and avg = 0 then 0 when avg = 0 then 1 else (total_records - avg) * 1.0 / avg end as average_change
from (
    select id, code, total_records
    , (select total_records from job_log where code = j.code and id < j.id order by id desc limit 1) as total_records_yesterday
    , (select coalesce(avg(total_records), 0) from job_log where code = j.code and date_created > date_sub(current_timestamp, interval 30 day) and total_records > 0) as avg
from job_log j
where code in (select code from web_crawler c where c.page_type = :type) and date_created > date_sub(current_timestamp, interval 1 day) order by code) t"""

    def htmlCrawlers() {
        JobLog.withSession { session ->
            def q = session.createSQLQuery(htmlCrawlerSql)
            q.setParameter("type", "HTML").list()
        }
    }

    def htmlAsyncCrawlers() {
        JobLog.withSession { session ->
            def q = session.createSQLQuery(htmlCrawlerSql)
            q.setParameter("type", "HTML_ASYNC").list()
        }
    }

    def crawlersWithChange() {
        def successfulCrawlers = JobLog.withSession { session ->
            session.createSQLQuery("""select j.id, j.code, j.total_records, j.total_prices
    , (select total_records from job_log where code = j.code and id < j.id and status = 'SUCCESS' order by id desc limit 1) as last_total_records
    , (select total_prices from job_log where code = j.code and id < j.id and status = 'SUCCESS' order by id desc limit 1) as last_total_prices
    , c.currency_code 
from job_log j
inner join web_crawler c on c.code = j.code
inner join (select code, max(id) as id from job_log where status = 'SUCCESS' group by code) t on t.id = j.id""")
                .list()
        }
        successfulCrawlers.collect { r ->
            def recordChange = calcChange(r[2], r[4])
            def priceChange = calcChange(r[3], r[5])
            [code: r[1], totalRecords: r[2], totalPrices: r[3], lastTotalRecords: r[4], lastTotalPrices: r[5],
             recordChange: recordChange, priceChange: priceChange, currency: r[6]]
        }
    }

    // crawlers with no change for 5 successful runs
    // not a pure SQL solution, there should be a better one
    def crawlersWithNoChange() {
        // successful crawler codes with all total_records in order
        // note: group_concat() returns a limited sized result, should be fine as we'll archive job log older than 1 month
        def codeWithRecords = JobLog.withSession { session ->
            session.createSQLQuery("select j.code" +
                    ", group_concat(total_records order by id desc) as records " +
                    "from job_log j " +
                    "inner join (select code from job_log where status = 'SUCCESS' and date_created > date_sub(current_timestamp, interval 1 day)) t on j.code = t.code " +
                    "where j.status = 'SUCCESS' " +
                    "group by code").list()
        }

        // find crawlers with no change using total_records list
        def noChange = []
        codeWithRecords.each { r ->
            def records = r[1].split(',')
            def days = records.findIndexOf { it != records[0] }
            if (days == -1) days = records.size()
            if (days >= 5) {
                def lastSuccess = JobLog.createCriteria().get {
                    projections {
                        property('dateCreated')
                    }
                    ne('totalRecords', records[0].toInteger())
                    eq('code', r[0])
                    order('id', 'desc')
                    maxResults(1)
                }
                if (lastSuccess == null) {
                    // no change since day 1
                    lastSuccess = JobLog.createCriteria().get {
                        projections {
                            property('dateCreated')
                        }
                        eq('totalRecords', records[0].toInteger())
                        eq('code', r[0])
                        order('id', 'asc')
                        maxResults(1)
                    }
                    days = new Date() - lastSuccess
                }
                noChange << [
                        code: r[0],
                        days: days,
                        totalRecords: records[0],
                        since: lastSuccess
                ]
            }
        }
        noChange
    }

    def sendToSlack() {
        if (Environment.current != Environment.PRODUCTION) {
            return
        }

        def r = new RestTemplate()
        r.postForObject('https://hooks.slack.com/services/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', [
            attachments: [
                [
                    fallback  : 'Crawler report is ready',
                    title     : 'Click to view the detailed report on dashboard',
                    title_link: 'https://dashboard.winealert.com/jobLog/report',
                    fields    : [
                        [title: 'Active', value: activeCrawlerCount(), short: true],
                        [title: 'Successful', value: successfulCrawlerCount(), short: true],
                        [title: 'Failed', value: failedCrawlers().size(), short: true],
                        [title: 'External', value: externalCrawlerCount(), short: true]
                    ]
                ]
            ]
        ], String.class)
    }

    private def activeCrawlers() {
        WebCrawler.where { disabled == false && listing == true && status != Status.DELETED }
    }

    private def calcChange(a, b) {
        if (b == null || b == 0) {
            if (a > 0) 1
            else 0
        } else {
            (a - b) * 1.0 / b
        }
    }
}
