package wine.alert

import grails.plugins.rest.client.RestBuilder
import grails.transaction.Transactional
import org.grails.web.json.JSONObject
import org.springframework.beans.factory.annotation.Value

/**
 * Provides information of current currency rates in database.<br>
 * Provides method to update currency rates with third party currency rates.
 */
@Transactional
class CurrencyRateService {

    private static final String SYSTEM_BASE_CURRENCY = "USD"

    @Value('${application.fixer.url}')
    private String fixerUrl

    @Value('${application.fixer.apiKey}')
    private String fixerApiKey

    def restBuilder

    /**
     * Gets map of currency rates from currency_rate table ({@link wine.alert.CurrencyRate}).<br>
     * Base currency is {@link CurrencyRateService#SYSTEM_BASE_CURRENCY}
     * @return the map of currency code and rate
     */
    Map<String, BigDecimal> getRates() {
        Map<String, BigDecimal> currencyRates = new HashMap<>()
        def rates = CurrencyRate.findAllByBase(SYSTEM_BASE_CURRENCY)
        rates?.each {
            currencyRates.put(it.currency, it.rate)
            def currency = Currency.findByCode(it.currency)
            if (currency) {
                currency.aliases?.each { c ->
                    currencyRates.put(c, it.rate)
                }
            }
        }
        return currencyRates
    }

    /**
     * Updates currency_rate data ({@link wine.alert.CurrencyRate}) with third party currency rates.<br>
     * Base currency is {@link CurrencyRateService#SYSTEM_BASE_CURRENCY}
     */
    void updateRates() {
        def rates = getProviderRates()
        BigDecimal divider = rates?.getAt(SYSTEM_BASE_CURRENCY)
        rates?.each { k, v ->
            try {
                CurrencyRate currencyRate = CurrencyRate.findOrCreateByBaseAndCurrency(SYSTEM_BASE_CURRENCY, k)
                BigDecimal rate = new BigDecimal(v as String) / divider
                currencyRate.rate = rate?.setScale(10, BigDecimal.ROUND_HALF_UP)
                currencyRate.save()
            } catch (Exception e) {
                log.error(e?.message, e)
            }
        }
    }

    /**
     * Gets currency rates from Fixer API endpoints (https://fixer.io/documentation).
     * @return the map of currency rates
     */
    def getProviderRates() {
        try {
            restBuilder = new RestBuilder(connectTimeout: 6000, readTimeout: 20000)
            def resp = restBuilder.get(fixerUrl, [apiKey: fixerApiKey])
            resp.json instanceof JSONObject
            return resp.json?.rates
        } catch (Exception e) {
            log.error(e?.message, e)
        }
    }
}
