/**
 * Service for group crawler
 */

package wine.alert

import grails.async.Promise
import grails.async.Promises
import grails.transaction.Transactional
import org.apache.commons.collections.CollectionUtils
import wine.alert.crawler.DynamicCrawler
import wine.alert.crawler.FeedCrawler
import wine.alert.crawler.request.CrawlEndpoint
import wine.alert.crawler.request.CrawlRequest
import wine.alert.crawler.response.CrawlResponse
import wine.alert.crawler.response.FeedResponse
import wine.alert.utils.CrawlerUtils

import javax.validation.ValidationException
import java.util.concurrent.ConcurrentHashMap
import java.util.regex.Matcher
import java.util.regex.Pattern

@Transactional(readOnly = true)
class GroupCrawlerService {

    private static Map<String, Boolean> RUNNING_STATES = new ConcurrentHashMap<String, Boolean>()

    /**
     * Runs a group crawler
     * @param group the instance of {@link RetailerGroup}
     */
    void run(RetailerGroup group) {
        if (!group) return
        RUNNING_STATES.put(group.id.toString(), true)
        def auth = new PasswordAuthentication(group.feedUsername, group.feedPassword?.chars)
        CrawlEndpoint endpoint = new CrawlEndpoint(url: group.feedUrl, auth: auth, region: group.feedRegion, charset: group.charset)
        CrawlRequest request = new CrawlRequest(endpoint: endpoint, charset: group.charset)
        Promise p = Promises.task {
            try {
                executeRequest(request, group.id, group.clientIdPattern)
            } catch (Exception e) {
                log.error "An error occurred while running group [${group.name}]"
                log.error(e.message, e)
            }
        }
        p.onComplete { result ->
            RUNNING_STATES.remove(group.id.toString())
            log.info "Group [${group.name}] execution completed."
        }
    }

    /**
     * Assigns state. Eg: running status, etc
     * @param groups the list of {@link RetailerGroup}
     */
    void assignState(List<RetailerGroup> groups) {
        groups.each {
            it.running = RUNNING_STATES.get(it.id.toString()) == true
        }
    }

    /**
     * Executes feed and crawl requests
     * @param request the instance of {@link CrawlRequest}
     * @param groupId the group id
     * @param clientIdPattern the group client id pattern
     */
    private void executeRequest(CrawlRequest request, Long groupId, String clientIdPattern) {
        FeedCrawler feed = new FeedCrawler(request)
        FeedResponse response = feed.crawl()
        if (!response || CollectionUtils.isEmpty(response.urls)) {
            log.info("Empty response for Feed URL: ${request.endpoint?.url}")
            return
        }

        List<CrawlEndpoint> endpoints = getEndpoints(response, request.endpoint?.url, clientIdPattern)
        for (CrawlEndpoint endpoint in endpoints) {
            request.endpoint.url = endpoint.url
            String retailerCode = proceedRetailer(groupId, request, endpoint.id)

            // Run crawler for retailer
            DynamicCrawler crawler = new DynamicCrawler(request)
            CrawlerUtils.saveResult(crawler.crawl() as CrawlResponse, retailerCode)
        }
    }

    /**
     * Gets endpoints from retailer group feed response. Find the most recent resource for each unique client id.
     * @param response the instance of {@link FeedResponse}
     * @param feedUrl the feeds URL
     * @param clientIdPattern the group client id pattern
     * @return the list of {@link CrawlEndpoint}
     */
    private List<CrawlEndpoint> getEndpoints(FeedResponse response, String feedUrl, String clientIdPattern) {
        List<CrawlEndpoint> endpoints = new ArrayList<>()
        Set<String> retailers = new HashSet<>()
        for (String url in response.urls.sort().reverse()) {
            if (url == feedUrl) continue
            Pattern p = Pattern.compile(clientIdPattern)
            Matcher m = p.matcher(url.tokenize('/').last())
            if (!m.find()) continue
            String b = m.group()
            if (!retailers.contains(b)) {
                retailers.add(b)
                endpoints.add(new CrawlEndpoint(url: url, id: b))
                log.info("Found retailer [$b] with URL [$url]")
            }
        }
        return endpoints
    }

    /**
     * Proceed retailer
     * @param groupId the group id
     * @param request the request to adjust
     * @param clientId the client id
     * @return the retailer code
     */
    private String proceedRetailer(Long groupId, CrawlRequest request, String clientId) {
        String retailerCode = clientId
        WebCrawler.withNewTransaction {
            WebCrawler retailer = getOrCreateRetailer(groupId, clientId)
            if (retailer) {
                if (retailer.code != clientId) retailerCode = retailer.code
                if (retailer.charsetEncoding) {
                    request.charset = retailer.charsetEncoding
                    request.endpoint.charset = retailer.charsetEncoding
                }
            }
        }
        return retailerCode
    }

    /**
     * Gets or creates a retailer in a retailer group
     * @param groupId the group id
     * @param clientId the unique identifier of the retailer
     * @return the instance of {@link WebCrawler}
     */
    private WebCrawler getOrCreateRetailer(Long groupId, String clientId) {
        def query = WebCrawler.where {
            code == clientId || clientId == clientId
        }
        def exists = query.get()
        if (exists) return exists
        return createRetailer(groupId, clientId)
    }

    /**
     * Creates retailer in a retailer group
     * @param groupId the group id
     * @param clientId the unique identifier of the retailer
     * @return the instance of {@link WebCrawler}
     */
    private WebCrawler createRetailer(Long groupId, String clientId) {
        log.info("Saving new retailer [$clientId] for group [$groupId]")
        WebCrawler crawler = new WebCrawler()
        crawler.code = clientId
        crawler.name = clientId
        crawler.clientId = clientId
        crawler.group = RetailerGroup.get(groupId)
        crawler.validate()
        if (crawler.hasErrors()) {
            throw new ValidationException("Error while new retailer [$clientId] for group [$groupId]")
        }
        crawler.save(flush: true, failOnError: true)
        return crawler
    }
}
