package wine.alert

import enums.Status
import grails.transaction.Transactional
import wine.db.WineMerge

/**
 * Provides methods to interact with history files.<br>
 * History files are files that contains map between WineAlert's wine and retailer wine.
 */
@Transactional
class HistoryService {

    def outputService

    /**
     * Gets draft from matched table ({@link wine.alert.Matched}) and update related history files.<br>
     * matched is table to store validate keyword matches.
     */
    void processMatches() {
        List<Matched> drafts = Matched.createCriteria().list {
            eq("status", Status.DRAFT)
        } as List<Matched>

        Map<String, List<Matched>> grouped = drafts?.groupBy { it.webCrawler?.getHistoryFileName() }
        grouped?.each { historyName, matchedDrafts ->

            if (outputService.isHistoryFileWritten(historyName)) {
                List<String> historyList = outputService.loadRawHistories(historyName)

                int newRecords = 0
                matchedDrafts?.each { Matched matchedDraft ->
                    String value = matchedDraft.wineAlertId + "||" + matchedDraft.description
                    int index = historyList.indexOf(value)
                    if (index < 0) {
                        newRecords++
                        historyList.push(value)
                    } else {
                        matchedDraft.skipped = true
                    }
                }

                if (newRecords) {
                    def updated = outputService.updateHistoryFile(historyName, historyList)
                    if (updated) {
                        matchedDrafts?.each {
                            it.status = Status.PROCESSED
                            it.save()
                        }
                        log.info( "[${historyName}] history updated with " + newRecords + " record(s)." )
                    } else { // Update only skipped
                        matchedDrafts?.findAll {it.skipped}?.each {
                            it.status = Status.PROCESSED
                            it.save()
                        }
                    }
                } else {
                    matchedDrafts?.each {
                        it.status = Status.PROCESSED
                        it.save()
                    }
                    log.info( "[${historyName}] history update skipped, no new record." )
                }

            } else {
                log.info( "[${historyName}] history is writing record(s), update skipped to the next loop." )
            }
        }
    }

    /**
     * Gets merging wines from wine_merge table ({@link wine.db.WineMerge}) and update related history files.<br>
     * wine_merge is table to store wines merge history
     */
    void processMerges() {
        // Update History Files After Merge Wines
        List<WineMerge> winesToMerge = WineMerge.wineDatabase.createCriteria().list {
            eq("status", Status.QUEUED)
        } as List<WineMerge>

        List<String> allHistoryNames = getAllHistoryFileNames()
        winesToMerge.each { wm ->
            int skipped = 0
            allHistoryNames?.each { historyName ->
                if (outputService.isHistoryFileWritten(historyName)) {
                    List<String> histories = outputService.loadRawHistories(historyName)
                    if (histories && !histories.isEmpty()) {
                        int newRecords = 0
                        List<String> newHistories = []
                        histories?.each { text ->
                            if (text.startsWith(wm.sourceId)) {
                                text = text.replace(wm.sourceId, wm.destinationId)
                                newRecords++
                            }
                            newHistories.push(text)
                        }
                        if (newRecords > 0) {
                            outputService.updateHistoryFile(historyName, newHistories)
                            log.info( "[${historyName}] history updated with " + newRecords + " record(s)." )
                        } else {
                            log.info( "[${historyName}] history update skipped, no updated record." )
                        }
                    } else {
                        log.info( "[${historyName}] history update skipped, no record." )
                    }
                } else {
                    skipped++
                    log.info( "[${historyName}] history is writing record(s), update skipped to the next loop." )
                }
            }
            if (skipped == 0) {
                wm.status = Status.PROCESSED
                wm.save()
            }
        }
    }

    /**
     * Gets all history file names
     * @return the list of history file name
     */
    List<String> getAllHistoryFileNames() {
        def allWebCrawler = WebCrawler.list()
        List<String> allHistoryNames = new ArrayList<>()
        allWebCrawler.each {
            allHistoryNames.add(it?.getHistoryFileName())
        }
        return allHistoryNames.unique()
    }
}
