package wine.alert

import enums.AliasType
import enums.FileSource
import enums.Separator
import grails.transaction.Transactional
import groovy.time.TimeCategory
import groovyx.gpars.GParsPool
import org.apache.commons.collections.CollectionUtils
import org.apache.commons.lang3.StringUtils
import org.springframework.beans.factory.annotation.Value
import utils.CharsetUtils
import wine.alert.enums.EliminationType
import wine.alert.enums.KeywordType
import wine.alert.enums.ReservedType
import wine.alert.enums.SubstitutionType
import wine.alert.utils.IOUtils
import wine.alert.utils.IndexesUtils
import wine.db.*
import wine.db.common.BottleConf
import wine.db.utils.WineKeywordUtils

import javax.validation.ValidationException
import java.util.Locale
import java.util.concurrent.atomic.AtomicInteger

/**
 * Service to parse given input file.
 */
@Transactional
class InputService {

    @Value('${application.config.location}')
    private String configLocation

    Boolean isImportConfigProcessing = false

    def sessionFactory
    def messageSource

    def keywordsService
    def wineService
    def producerService
    def wineApiService

    /**
     * Parse given input file and update wine_keyword table ({@link wine.db.WineKeyword}).<br>
     * Input file format is text with Tab as separator.
     * @param newFile the file to be parsed
     */
    void importWineKeywords(String newFile) {
        log.info("Import wineKeywords started.")
        Date date = new Date()

        File file = new File(newFile)

        def keywordsMap = [:]

        String fileEncoding = CharsetUtils.detect(newFile) ?: "UTF-8"
        file.withReader(fileEncoding) { reader ->

            def line
            while ((line = reader.readLine())!=null) {

                def wineInfo = line?.split("\t")
                String wineId = wineInfo?.getAt(0)?.trim()
                //String wineName = wineInfo?.getAt(1)?.trim()
                //String fullKeyword = wineInfo?.getAt(2)?.trim()
                String producerKeyword = wineInfo?.size() > 3 ? wineInfo?.getAt(3)?.trim() : ""
                String baseKeyword = wineInfo?.size() > 4 ? wineInfo?.getAt(4)?.trim() : ""
                String appellationKeyword = wineInfo?.size() > 5 ? wineInfo?.getAt(5)?.trim() : ""
                String colorKeyword = wineInfo?.size() > 6 ? wineInfo?.getAt(6)?.trim() : ""
                String commonModifiersKeyword = wineInfo?.size() > 7 ? wineInfo?.getAt(7)?.trim() : ""
                String modifiersKeyword = wineInfo?.size() > 8 ? wineInfo?.getAt(8)?.trim() : ""
                String customNotKeyword = wineInfo?.size() > 9 ? wineInfo?.getAt(9)?.trim() : ""
                String globalNotKeyword =  wineInfo?.size() > 10 ? wineInfo?.getAt(10)?.trim() : ""

                if (wineId && !keywordsMap?.getAt(wineId)) {
                    keywordsMap.put(wineId, [
                            //wineName: wineName,
                            //fullKeyword: fullKeyword,
                            producerKeyword: producerKeyword,
                            baseKeyword: baseKeyword,
                            appellationKeyword: appellationKeyword,
                            colorKeyword: colorKeyword,
                            commonModifiersKeyword: commonModifiersKeyword,
                            modifiersKeyword: modifiersKeyword,
                            customNotKeyword: customNotKeyword,
                            globalNotKeyword: globalNotKeyword
                    ])
                }
            }
        }
        log.info("WineKeywords map size: " + keywordsMap?.size())

        AtomicInteger counter = new AtomicInteger()
        if (keywordsMap?.size() > 0) {
            int logInterval = getLogInterval(keywordsMap.size())

            counter.set(0)
            keywordsMap?.each { k, v ->
                WineKeyword.withNewTransaction {
                    proceedWineKeyword(k, v)
                }

                Integer cr = counter.getAndIncrement()
                if (cr % logInterval == 0) {
                    log.info("WineKeywords processed: " + cr)
                }
            }
            log.info("WineKeywords processed: " + counter.get())
            sessionFactory.currentSession.clear()
        }

        if (keywordsMap?.size() > 0) {
            List<String> collectionWineKeywords = new ArrayList<>()
            WineKeyword.withNewTransaction {
                collectionWineKeywords = WineKeyword.wineDatabase.createCriteria().list {
                    projections {
                        wineDb {
                            property("id")
                        }
                    }
                }
            }
            log.info("WineKeywords in system: " + collectionWineKeywords?.size())

            List<String> fileWineKeywords = keywordsMap.collect {(String) it.key}
            List<String> tobeRemoved = collectionWineKeywords.findAll {!fileWineKeywords.contains(it)}
            log.info("WineKeywords to be removed: " + tobeRemoved?.size())

            if (tobeRemoved?.size() > 0) {
                tobeRemoved.unique()
                int logInterval = getLogInterval(tobeRemoved.size())

                counter.set(0)
                tobeRemoved?.each { String id ->
                    WineKeyword.withNewTransaction {
                        removeKeyword(id)
                    }

                    Integer cr = counter.getAndIncrement()
                    if (cr % logInterval == 0) {
                        log.info("WineKeywords removed: " + cr)
                    }
                }
                log.info("WineKeywords removed: " + counter.get())
                sessionFactory.currentSession.flush()
            }
        }

        log.info("Import WineKeywords completed. Execution times: " + TimeCategory.minus(new Date(), date))

        RemapKeywordsJob.triggerNow()
    }

    /**
     * Saves wine keyword data from {@link InputService#importWineKeywords(String)}.<br>
     *     Deleted wineDb will be ignored.
     * @param wineDbId the id of the wineDb
     * @param keywordMap the keywords data mapping
     */
    void proceedWineKeyword(String wineDbId, def keywordMap) {
        WineDb wineDb = WineDb.wineDatabase.findByIdAndStatus(wineDbId, WineDb.Status.ACTIVE)
        if (wineDb) {
            try {
                WineKeyword wineKeyword = WineKeyword.wineDatabase.findByWineDb(wineDb)
                if (!wineKeyword) {
                    wineKeyword = new WineKeyword()
                    wineKeyword.wineDb = wineDb
                }
                //wineKeyword.pattern = keywordMap?.fullKeyword
                wineKeyword.producerKeyword = keywordMap?.producerKeyword
                wineKeyword.baseKeyword = keywordMap?.baseKeyword
                wineKeyword.appellationKeyword = keywordMap?.appellationKeyword
                wineKeyword.colorKeyword = keywordMap?.colorKeyword
                wineKeyword.commonModifiersKeyword = keywordMap?.commonModifiersKeyword
                wineKeyword.modifiersKeyword = keywordMap?.modifiersKeyword
                wineKeyword.customNotKeyword = keywordMap?.customNotKeyword
                wineKeyword.globalNotKeyword = keywordMap?.globalNotKeyword
                wineKeyword.pattern = WineKeywordUtils.joinPattern(wineKeyword)
                wineKeyword.save()
            } catch (Exception e) {
                log.error(e?.message, e)
            }
        }
    }

    /**
     * Delete record from WineKeyword based on given wine_db id
     * @param wineDbId the WineDb id
     */
    void removeKeyword(String wineDbId) {
        try {
            WineDb wineDb = WineDb.wineDatabase.findById(wineDbId)
            if (!wineDb) return

            WineKeyword wineKeyword = wineDb.wineKeyword
            if (!wineKeyword) return

            log.info("Removing wine keyword for ${wineDbId} with id: ${wineKeyword.id}")
            wineDb.wineKeyword = null
            wineKeyword.wineDb = null
            wineKeyword.delete()
        } catch (Exception e) {
            log.error(e?.message, e)
        }
    }

    /**
     * Parses given input file and update bottle_size table ({@link wine.db.BottleSize}).<br>
     * Input file format is text with Pipe ("|") as separator.
     * @param newFile the file to be parsed
     */
    void importBottleSizes(String newFile) {
        log.info("Import BottleSizes started.")
        Date date = new Date()
        File file = new File(newFile)

        List<BottleConf> bottles = []
        String fileEncoding = CharsetUtils.detect(newFile) ?: "UTF-8"
        file.withReader(fileEncoding) { reader ->
            def line
            while ((line = reader.readLine()) != null) {
                def dataList = line.tokenize("|")
                String bottleId = dataList?.getAt(0)?.trim()
                String bottleName = dataList?.getAt(1)?.trim()
                String bottleAliases = dataList?.drop(1)?.join("|")
                bottles.add(new BottleConf(
                        id: bottleId,
                        name: bottleName,
                        alias: bottleAliases
                ))
            }
        }
        log.info("BottleSizes to be saved: " + bottles.size())

        if (CollectionUtils.isNotEmpty(bottles)) {
            bottles.collate(50).each { list ->
                BottleSize.withNewTransaction {
                    list.each { bottle ->
                        proceedBottleSize(bottle)
                    }
                }
            }

            def ids = bottles.collect { it.id }
            BottleSize.withNewTransaction {
                def query = BottleSize.wineDatabase.where {
                    !(id in ids)
                }
                def affected = query.deleteAll()
                log.info("BottleSizes deleted: " + affected)
            }

            keywordsService.mapBottleAliases()
        }

        log.info("Import BottleSizes completed. Execution times: " + TimeCategory.minus(new Date(), date))
    }

    /**
     * Saves bottle size
     * @param bottle the instance of {@link BottleConf}
     */
    void proceedBottleSize(BottleConf bottle) {
        try {
            BottleSize bottleSize = BottleSize.wineDatabase.findOrCreateById(bottle.id)
            bottleSize.name = bottle.name
            bottleSize.alias = bottle.alias

            bottleSize.validate()
            if (bottleSize.hasErrors()) {
                def errMsgList = bottleSize.errors.allErrors.collect {
                    messageSource.getMessage(it, Locale.default)
                }
                throw new ValidationException(errMsgList?.join("\n"))
            }
            bottleSize.save()
        } catch (ValidationException e) {
            log.warn("Error processing Bottle Size: " + bottle.id + ", message: " + e?.message)
        } catch (Exception e) {
            log.error(e?.message, e)
        }
    }

    /**
     * Parse given input file and update variety table ({@link wine.db.Variety}).<br>
     * Input file format is text with Tab as separator.
     * @param newFile the file to be parsed
     */
    void importVarieties(String newFile) {
        log.info("Import Varieties started.")

        File file = new File(newFile)

        def varietiesMap = [:]
        String fileEncoding = CharsetUtils.detect(newFile) ?: "UTF-8"
        file.withReader(fileEncoding) { reader ->
            def line
            while ((line = reader.readLine())!=null) {
                List<String> dataList = line.tokenize("\t")
                String name = dataList?.getAt(0)?.trim()
                String color = dataList?.getAt(1)?.trim()
                if (!varietiesMap?.getAt(name)) {
                    varietiesMap.put(name, color)
                }
            }
        }
        log.info("Varieties map size: " + varietiesMap?.size())

        Variety.withNewTransaction {
            varietiesMap?.each { map ->
                Variety variety = Variety.wineDatabase.findOrCreateByName((String) map.key)
                variety.defaultColor = map.value
                variety.save()
            }

            // Cleanup
            def varieties = Variety.wineDatabase.findAll()
            varieties?.findAll()?.each { item ->
                if (!varietiesMap.containsKey(item.name)) {
                    item.delete()
                }
            }
        }
        log.info("Import Varieties map done.")
    }

    /**
     * Parse given input file and update web_crawler table ({@link wine.alert.WebCrawler}).<br>
     * Input file format is text with Tab as separator.
     * @param newFile the file to be parsed
     */
    void importRetailers(String newFile) {
        log.info("Import Retailers Config started.")
        Date date = new Date()

        File file = new File(newFile)

        def retailersMap = [:]

        String fileEncoding = CharsetUtils.detect(newFile) ?: "UTF-8"
        file.withReader(fileEncoding) { reader ->

            def line
            int i = 0
            while ((line = reader.readLine())!=null) {
                if (i >= 2) {
                    def dataList = line?.split("\t")?.toList()
                    String code = dataList?.getAt(1)?.trim()
                    String name = dataList?.getAt(2)?.trim()
                    String historyName = dataList?.getAt(3)?.trim()
                    if (historyName?.equalsIgnoreCase("-")) {
                        historyName = null
                    }

                    String delimiter = dataList?.getAt(4)?.trim()
                    Separator outputDelimiter = Separator.PIPE
                    try {
                        if ("\\t".equalsIgnoreCase(delimiter)) {
                            outputDelimiter = Separator.TAB
                        } else if ("\\s".equalsIgnoreCase(delimiter)) {
                            outputDelimiter = Separator.TAB
                        } else {
                            outputDelimiter = Separator.getByValue(delimiter)
                        }
                    } catch (Exception e) {
                        log.error(e?.message, e)
                    }

                    String currency = dataList?.getAt(5)?.trim()

                    String parsing = dataList?.getAt(6)?.trim()
                    Boolean vintageDefaultNv = Boolean.FALSE
                    Boolean bottleSizeDefaultBlank = Boolean.FALSE
                    if ("2"?.equalsIgnoreCase(parsing)) {
                        vintageDefaultNv = Boolean.TRUE
                    } else if ("4"?.equalsIgnoreCase(parsing)) {
                        bottleSizeDefaultBlank = Boolean.TRUE
                    }

                    String keywordIndexes = dataList?.getAt(7)?.trim()
                    if (StringUtils.isNotBlank(keywordIndexes)) {
                        keywordIndexes = IndexesUtils.decrease(keywordIndexes)
                    }

                    String historyIndexes = dataList?.getAt(8)?.trim()
                    if (StringUtils.isNotBlank(historyIndexes)) {
                        historyIndexes = IndexesUtils.decrease(historyIndexes)
                    }

                    String bottleSizeIndex = dataList?.getAt(9)?.trim()
                    if (StringUtils.isNotBlank(bottleSizeIndex)) {
                        bottleSizeIndex = IndexesUtils.decrease(bottleSizeIndex)
                    }

                    String vintageIndex = dataList?.getAt(10)?.trim()
                    if (StringUtils.isNotBlank(vintageIndex)) {
                        vintageIndex = IndexesUtils.decrease(vintageIndex)
                    }

                    String priceIndex = dataList?.getAt(11)?.trim()
                    if (StringUtils.isNotBlank(priceIndex)) {
                        priceIndex = IndexesUtils.decrease(priceIndex)
                    }

                    String taxStatusIndex = dataList?.getAt(12)?.trim()
                    if (StringUtils.isNotBlank(taxStatusIndex)) {
                        taxStatusIndex = IndexesUtils.decrease(taxStatusIndex)
                    }

                    String urlIndex = dataList?.getAt(13)?.trim()
                    if (StringUtils.isNotBlank(urlIndex)) {
                        urlIndex = IndexesUtils.decrease(urlIndex)
                    }

                    String descriptionIndexes = dataList?.getAt(14)?.trim()
                    if (StringUtils.isNotBlank(descriptionIndexes)) {
                        descriptionIndexes = IndexesUtils.decrease(descriptionIndexes)
                    }

                    String skuIndex = dataList?.getAt(15)?.trim()
                    if (StringUtils.isNotBlank(skuIndex)) {
                        skuIndex = IndexesUtils.decrease(skuIndex)
                    }

                    if (!retailersMap?.getAt(code)) {
                        retailersMap.put(code, [
                                name: name,
                                historyName: historyName,
                                outputDelimiter: outputDelimiter,
                                currency: currency,
                                vintageDefaultNv: vintageDefaultNv,
                                bottleSizeDefaultBlank: bottleSizeDefaultBlank,
                                keywordIndexes: keywordIndexes,
                                historyIndexes: historyIndexes,
                                bottleSizeIndex: bottleSizeIndex,
                                vintageIndex: vintageIndex,
                                priceIndex: priceIndex,
                                taxStatusIndex: taxStatusIndex,
                                urlIndex: urlIndex,
                                descriptionIndexes: descriptionIndexes,
                                skuIndex: skuIndex
                        ])
                    }
                }
                i++
            }
        }
        log.info("Retailers map size: " + retailersMap?.size())

        AtomicInteger counter = new AtomicInteger()
        if (retailersMap?.size() > 0) {
            int logInterval = getLogInterval(retailersMap.size())

            counter.set(0)
            retailersMap?.each { k, v ->
                WebCrawler.withNewTransaction {
                    proceedRetailer(k, v)
                }

                Integer cr = counter.getAndIncrement()
                if (cr % logInterval == 0) {
                    log.info("Retailers processed: " + cr)
                }
            }
            log.info("Retailers processed: " + counter.get())
            sessionFactory.currentSession.clear()
        }

        log.info("Import Retailers Config completed. Execution times: " + TimeCategory.minus(new Date(), date))
    }

    /**
     * Saves retailer data from {@link InputService#importRetailers(String)} to web_crawler table ({@link wine.alert.WebCrawler}).
     * @param code the retailer code
     * @param dataMap the retailer data mapping
     */
    void proceedRetailer(String code, def dataMap) {
        try {
            WebCrawler webCrawler = WebCrawler.findByCode(code)
            if (!webCrawler) {
                webCrawler = new WebCrawler()
                webCrawler.code = code
                webCrawler.fileSource = FileSource.EXTERNAL
                webCrawler.disabled = Boolean.TRUE
            }

            webCrawler.name = dataMap?.name
            webCrawler.currencyCode = dataMap?.currency
            webCrawler.historyName = dataMap?.historyName
            if (webCrawler.fileSource == FileSource.EXTERNAL) { // Do not update the output delimiter, it should be PIPE by default
                webCrawler.outputDelimiter = dataMap?.outputDelimiter
            }
            webCrawler.vintageDefaultNv = dataMap?.vintageDefaultNv
            webCrawler.bottleSizeDefaultBlank = dataMap?.bottleSizeDefaultBlank
            webCrawler.keywordIndexes = dataMap?.keywordIndexes
            webCrawler.historyIndexes = dataMap?.historyIndexes
            webCrawler.bottleSizeIndex = dataMap?.bottleSizeIndex ? Integer.valueOf(dataMap?.bottleSizeIndex) : null
            webCrawler.vintageIndex = dataMap?.vintageIndex ? Integer.valueOf(dataMap?.vintageIndex) : null
            webCrawler.priceIndex = dataMap?.priceIndex ? Integer.valueOf(dataMap?.priceIndex) : null
            webCrawler.taxStatusIndex = dataMap?.taxStatusIndex ? Integer.valueOf(dataMap?.taxStatusIndex) : null
            webCrawler.urlIndex = dataMap?.urlIndex ? Integer.valueOf(dataMap?.urlIndex) : null
            webCrawler.descriptionIndexes = dataMap?.descriptionIndexes
            webCrawler.skuIndex = dataMap?.skuIndex ? Integer.valueOf(dataMap?.skuIndex) : null

            webCrawler.validate()
            if (webCrawler.hasErrors()) {
                def errMsgList = webCrawler.errors.allErrors.collect{
                    messageSource.getMessage(it, Locale.default)
                }
                throw new ValidationException(errMsgList?.join("\n"))
            }
            webCrawler.save()
        } catch (ValidationException e) {
            log.warn("Error processing retailer: " + code + ", message: " + e?.message)
        } catch (Exception e) {
            log.error(e?.message, e)
        }
    }

    /**
     * Parse given input file and update retailer table ({@link wine.db.Retailer}) and web_crawler table ({@link wine.alert.WebCrawler}).<br>
     * Input file format is text with Tab as separator.
     * @param newFile the file to be parsed
     */
    void importRetailersTable(String newFile) {
        log.info("Import Retailers Table started.")
        Date date = new Date()

        File file = new File(newFile)

        List<Map<String, String>> retailerList = new ArrayList<>()

        String fileEncoding = CharsetUtils.detect(newFile) ?: "UTF-8"
        file.withReader(fileEncoding) { reader ->

            int i = 0
            String line
            while ((line = reader.readLine())!=null) {
                if (i >= 2) {
                    Collection<String> row = line?.split("\t")
                    String code = row?.getAt(0)?.trim()
                    String name = row?.getAt(1)?.trim()
                    String address = row?.getAt(2)?.trim()
                    String city = row?.getAt(3)?.trim()
                    String state = row?.getAt(4)?.trim()
                    String postalCode = row?.getAt(5)?.trim()
                    String country = row?.getAt(6)?.trim()
                    String shipTo = row?.getAt(7)?.trim()
                    String phone = row?.getAt(8)?.trim()
                    String fax = row?.getAt(9)?.trim()
                    String email = row?.getAt(10)?.trim()
                    String webUrl = row?.size() > 11 ? row?.getAt(11)?.trim() : ""
                    String listing = row?.size() > 12 ? row?.getAt(12)?.trim() : ""
                    String download = row?.size() > 13 ? row?.getAt(13)?.trim() : ""
                    String currency = row?.size() > 14 ? row?.getAt(14)?.trim() : null

                    retailerList.add([
                            code: code,
                            name: name,
                            address: address,
                            city: city,
                            state: state,
                            postalCode: postalCode,
                            country: country,
                            shipTo: shipTo,
                            phone: phone,
                            fax: fax,
                            email: email,
                            webUrl: webUrl,
                            listing: listing,
                            download: download,
                            currency: currency
                    ])
                }
                i++
            }
        }
        log.info("Retailers map size: " + retailerList.size())

        AtomicInteger counter = new AtomicInteger()
        if (retailerList.size() > 0) {
            int logInterval = getLogInterval(retailerList.size())

            counter.set(0)
            retailerList.each { item ->
                Retailer.withNewTransaction {
                    proceedRetailerTable(item)
                }
                WebCrawler.withNewTransaction {
                    proceedRetailerConfigFromTable(item)
                }

                Integer cr = counter.getAndIncrement()
                if (cr % logInterval == 0) {
                    log.info("Retailers Table processed: " + cr)
                }
            }
            log.info("Retailers Table processed: " + counter.get())
            sessionFactory.currentSession.clear()
        }

        log.info("Import Retailers Table completed. Execution times: " + TimeCategory.minus(new Date(), date))
    }

    /**
     * Proceed retailer data from {@link InputService#importRetailersTable(String)} and update retailer table ({@link wine.db.Retailer})
     * @param dataMap the retailer data mapping
     */
    private void proceedRetailerTable(Map<String, String> dataMap) {
        String code = dataMap?.code
        if (!code) return

        Retailer retailer = Retailer.wineDatabase.findByCode(code)
        if (!retailer) {
            retailer = new Retailer()
            retailer.code = code
        }

        retailer.name = dataMap?.name
        retailer.email = dataMap?.email
        retailer.webUrl = dataMap?.webUrl
        retailer.address = dataMap?.address
        retailer.postalCode = dataMap?.postalCode
        retailer.city = dataMap?.city
        retailer.state = dataMap?.state
        retailer.phone = dataMap?.phone
        retailer.shipTo = dataMap?.shipTo
        retailer.country = Country.wineDatabase.findByName(dataMap?.country as String)

        retailer.validate()
        if (retailer.hasErrors()) {
            def errMsgList = retailer.errors.allErrors.collect{
                messageSource.getMessage(it, Locale.default)
            }
            throw new RuntimeException(errMsgList?.join("\n"))
        }

        retailer.save(flush: true)
    }

    /**
     * Proceed retailer data from {@link InputService#importRetailersTable(String)} and update web_crawler table ({@link wine.alert.WebCrawler}).
     * @param dataMap the retailer data mapping
     */
    private void proceedRetailerConfigFromTable(Map<String, String> dataMap) {
        String code = dataMap?.code
        if (!code) return

        WebCrawler webCrawler = WebCrawler.findByCode(code)
        if (!webCrawler) {
            webCrawler = new WebCrawler()
            webCrawler.code = code
            webCrawler.fileSource = FileSource.EXTERNAL
            webCrawler.disabled = Boolean.TRUE
        }

        webCrawler.name = dataMap?.name
        webCrawler.address = dataMap?.address
        webCrawler.city = dataMap?.city
        webCrawler.state = dataMap?.state
        webCrawler.zip = dataMap?.zip
        webCrawler.country = dataMap?.country
        webCrawler.shipToCountry = dataMap?.shipTo
        webCrawler.phone = dataMap?.phone
        webCrawler.fax = dataMap?.fax
        webCrawler.email = dataMap?.email

        try {
            webCrawler.listing = dataMap?.listing?.toBoolean()
            webCrawler.download = dataMap?.download?.toBoolean()
        } catch (Exception e) {}

        webCrawler.currencyCode = dataMap?.currency

        webCrawler.validate()
        if (webCrawler.hasErrors()) {
            def errMsgList = webCrawler.errors.allErrors.collect{
                messageSource.getMessage(it, Locale.default)
            }
            throw new RuntimeException(errMsgList?.join("\n"))
        }

        webCrawler.save(flush: true)
    }

    /**
     * Parse given input file and update mongo map at wine table ({@link wine.db.Wine}).<br>
     * Mongo map is mapping between WineAlert wine ({@link wine.db.Wine}) and RobertParker's wine<br>
     * Input file format is text with Tab as separator.
     * @param newFile the file to be parsed
     */
    void importMongoWinesMap(String newFile) {
        log.info("Import Mongo Wines Map started.")
        Date date = new Date()

        File file = new File(newFile)

        Map<String, String> vintageMap = new HashMap<>()
        Vintage.withNewTransaction {
            Vintage.wineDatabase.findAll()?.each {
                vintageMap.put(it.id, it.name)
            }
        }

        Map<String, String> mongoMap = new HashMap<>()
        Map<String, String> duplicateMap = new HashMap<>()

        String fileEncoding = CharsetUtils.detect(newFile) ?: "UTF-8"
        file.withReader(fileEncoding) { reader ->
            def line
            while ((line = reader.readLine())!=null) {
                List<String> dataList = line?.split("\t")?.toList()
                String wineId = dataList?.getAt(0)?.trim()
                String mongoId = dataList?.getAt(1)?.trim()
                if (wineId && wineId != "z" && mongoId) {
                    if (!mongoMap.containsKey(wineId)) {
                        mongoMap[wineId] = mongoId
                    } else {
                        duplicateMap[mongoId] = wineId
                    }
                }
            }
        }
        log.info("Mongo Wines Map size: " + mongoMap?.size())
        log.info("Duplicate Mongo Wines Map size: " + duplicateMap?.size())

        AtomicInteger counter = new AtomicInteger()
        if (mongoMap && !mongoMap.isEmpty()) {
            int logInterval = getLogInterval(mongoMap.size())

            mongoMap.each { wineId, mongoId ->
                Wine.withNewTransaction {
                    saveMongoMap(wineId, mongoId, vintageMap)
                }

                Integer cr = counter.getAndIncrement()
                if (cr % logInterval == 0) {
                    log.info("Mongo Wines Map processed: " + cr)
                }
            }
            log.info("Mongo Wines Map processed: " + counter.get())
            sessionFactory.currentSession.flush()

            // Cleanup invalid map
            def wines = new ArrayList()
            Wine.withNewTransaction {
                wines = Wine.wineDatabase.createCriteria().list {
                    projections {
                        property("id")
                    }
                    ne("mongoId", "")
                    isNotNull("mongoId")
                }
            }
            wines?.findAll {!mongoMap.containsKey(it)}?.each { item ->
                Wine.withNewTransaction {
                    removeMongoMap(item)
                }
            }
            sessionFactory.currentSession.clear()
        }

        if (duplicateMap && !duplicateMap.isEmpty()) {
            int logInterval = getLogInterval(duplicateMap.size())

            counter.set(0)
            duplicateMap.each { k, v ->
                WineDuplication.withNewTransaction {
                    proceedDuplicateWineMap(k, v)
                }

                Integer cr = counter.getAndIncrement()
                if (cr % logInterval == 0) {
                    log.info("Duplicate Mongo Wines Map processed: " + cr)
                }
            }
            log.info("Duplicate Mongo Wines Map processed: " + counter.get())
            sessionFactory.currentSession.clear()

            // Remove invalid duplications
            def wds = new ArrayList()
            WineDuplication.withNewTransaction {
                wds = WineDuplication.wineDatabase.createCriteria().list {
                    projections {
                        property("mongoId")
                    }
                }
            }
            def invalidList = wds?.findAll { !duplicateMap.containsKey(it) }
            if (invalidList && !invalidList.isEmpty()) {
                log.info("Total of Invalid Duplicate Mongo Wines Map: " + invalidList?.size())
                invalidList?.each { mongoId ->
                    WineDuplication.withNewTransaction {
                        removeWineDuplication(mongoId)
                    }
                }
            }
            sessionFactory.currentSession.clear()
        }

        log.info("Import Mongo Wines Map completed. Execution times: " + TimeCategory.minus(new Date(), date))
    }

    /**
     * Saves wine mongo map from {@link InputService#importMongoWinesMap(String)}.
     * @param wineId the wine id
     * @param mongoId the mongo id (sku/uuid of wine at RobertParker.com)
     * @param vintageMap the map of vintage, key is the id and value is vintage
     */
    void saveMongoMap(String wineId, String mongoId, Map<String, String> vintageMap) {
        try {
            // Search old map
            Wine old = Wine.wineDatabase.findByMongoId(mongoId)
            if (old && old.id != wineId) {
                old.mongoId = null
                old.save()
            }

            // Save new map
            Wine wine = Wine.wineDatabase.findById(wineId)
            if (wine == null) {
                String wineDbId = wineId.take(9)
                WineDb wineDb = WineDb.wineDatabase.findById(wineDbId)
                if (!wineDb) {
                    throw new ValidationException("WineDb not found.")
                }

                String vintageId = wineId.length() > 9 ? wineId.substring(9) : null
                if (!vintageMap?.containsKey(vintageId)) {
                    throw new ValidationException("Vintage not found.")
                }

                // Search bad given vintage (Resolve duplicate wineDb-vintage)
                Wine tobeRepaired = Wine.wineDatabase.findByWineDbAndVintage(wineDb, vintageMap?.get(vintageId))
                if (tobeRepaired) {
                    Vintage correctVintage = Vintage.wineDatabase.findById(tobeRepaired.id.substring(9))
                    if (!correctVintage) {
                        throw new ValidationException("Vintage correction not found for: " + tobeRepaired.id)
                    }
                    tobeRepaired.vintage = correctVintage.name
                    tobeRepaired.save()
                }

                wine = new Wine()
                wine.id = wineId
                wine.vintage = vintageMap?.get(vintageId)
                wine.wineDb = wineDb
            }
            wine.mongoId = mongoId
            wine.save()

        } catch (ValidationException e) {
            log.warn("Error processing wineId: " + wineId + ", mongoId: " + mongoId + ", message: " + e?.message)
        } catch (Exception e) {
            log.error("Error processing wineId: " + wineId + ", mongoId: " + mongoId)
            log.error(e?.message, e)
        }
    }

    /**
     * Removes mapping between wine ({@link wine.db.Wine}) and RobertParker's wine based on given wine id
     * @param wineId the id of wine
     */
    void removeMongoMap(String wineId) {
        try {
            Wine wine = Wine.wineDatabase.findById(wineId)
            wine.mongoId = null
            wine.save()
        } catch (Exception e) {
            log.error(e?.message, e)
        }
    }

    /**
     * Saves duplicate mongo map from {@link InputService#importMongoWinesMap(String)}.
     * @param mongoId the mongo id (sku/uuid of wine at RobertParker.com)
     * @param wineId the id of wine
     */
    void proceedDuplicateWineMap(String mongoId, String wineId) {
        try {
            WineDuplication duplication = WineDuplication.wineDatabase.findByMongoId(mongoId)
            if (!duplication) {
                Wine wine = Wine.wineDatabase.findById(wineId)
                if (!wine) {
                    log.info("Wine not found: " + wineId)
                    return
                }
                duplication = new WineDuplication()
                duplication.mongoId = mongoId
                duplication.wine = wine
                duplication.save()
            } else {
                Wine wine = Wine.wineDatabase.findById(wineId)
                if (!wine) {
                    log.info("Wine not found: " + wineId)
                    return
                }
                duplication.wine = wine
                duplication.save()
            }
        } catch (ValidationException e) {
            log.warn("Error processing duplication map mongoId: " + mongoId + ", wineId: " + wineId + ", message: " + e?.message)
        } catch (Exception e) {
            log.error("Error processing duplication map mongoId: " + mongoId + ", wineId: " + wineId)
            log.error(e?.message, e)
        }
    }

    /**
     * Removes duplicate mongo map from wine_duplication table ({@link wine.db.WineDuplication}) based on given mongo id
     * @param mongoId the mongo id (sku/uuid of wine at RobertParker.com)
     */
    void removeWineDuplication(String mongoId) {
        try {
            WineDuplication wd = WineDuplication.wineDatabase.findByMongoId(mongoId)
            if (wd) {
                wd.delete()
            }
        } catch (Exception e) {
            log.error(e?.message, e)
        }
    }

    /**
     * Parse given input file update wine_db table ({@link wine.db.WineDb}).<br>
     * Input file format is text with Tab as separator.
     * @param newFile the file to be parsed
     */
    void importWineDb(String newFile) {
        log.info("Import WineDb started.")
        Date date = new Date()

        File file = new File(newFile)

        List<Map<String, String>> wineDbList = new ArrayList<>()
        Map<String, List<String>> producers = new HashMap<>()

        String fileEncoding = CharsetUtils.detect(newFile) ?: "UTF-8"
        file.withReader(fileEncoding) { reader ->

            int c = 0
            def line
            while ((line = reader.readLine())!=null) {
                c += 1
                if (c == 1) continue
                Collection<String> row = line?.split("\t")
                String id = row?.getAt(0)?.trim()
                String wineNId = row?.getAt(3)?.trim()
                String producerName = row?.getAt(4)?.trim()
                String producerNameShow = row?.getAt(5)?.trim()
                String name = row?.getAt(6)?.trim()
                String variety = row?.getAt(7)?.trim()
                String color = row?.getAt(8)?.trim()
                String sweetness = row?.getAt(9)?.trim()
                String type = row?.getAt(10)?.trim()
                String wineCountry = row?.size() > 11 ? row?.getAt(11)?.trim() : ''
                String wineRegion = row?.size() > 12 ? row?.getAt(12)?.trim() : ''
                String wineLocation = row?.size() > 13 ? row?.getAt(13)?.trim() : ''
                String wineLocale = row?.size() > 14 ? row?.getAt(14)?.trim() : ''
                String wineSite = row?.size() > 15 ? row?.getAt(15)?.trim() : ''

                producerName = StringUtils.replace(producerName, "“", "\"") // left double quotation mark
                producerName = StringUtils.replace(producerName, "”", "\"") // right double quotation mark
                if (StringUtils.startsWith(producerName, "\"") && StringUtils.endsWith(producerName, "\"")) { // Only if start and end with quote. Handler for excel quote escape
                    producerName = StringUtils.substring(producerName, 1, producerName.length() - 1)
                }
                producerName = StringUtils.replace(producerName, "\"\"", "\"")

                producerNameShow = StringUtils.replace(producerNameShow, "“", "\"") // left double quotation mark
                producerNameShow = StringUtils.replace(producerNameShow, "”", "\"") // right double quotation mark
                if (StringUtils.startsWith(producerNameShow, "\"") && StringUtils.endsWith(producerNameShow, "\"")) { // Only if start and end with quote. Handler for excel quote escape
                    producerNameShow = StringUtils.substring(producerNameShow, 1, producerNameShow.length() - 1)
                }
                producerNameShow = StringUtils.replace(producerNameShow, "\"\"", "\"")

                name = StringUtils.replace(name, "“", "\"") // left double quotation mark
                name = StringUtils.replace(name, "”", "\"") // right double quotation mark
                if (StringUtils.startsWith(name, "\"") && StringUtils.endsWith(name, "\"")) { // Only if start and end with quote. Handler for excel quote escape
                    name = StringUtils.substring(name, 1, name.length() - 1)
                }
                name = StringUtils.replace(name, "\"\"", "\"")

                wineDbList.add([
                        id: id,
                        wineNId: wineNId,
                        name: name,
                        variety: variety,
                        color: color,
                        sweetness: sweetness,
                        type: type,
                        wineCountry: wineCountry,
                        wineRegion: wineRegion,
                        wineLocation: wineLocation,
                        wineLocale: wineLocale,
                        wineSite: wineSite
                ])
                String producerId = id?.take(6)
                if (!producers.containsKey(producerId)) {
                    producers.put(producerId, [producerName, producerNameShow])
                }
            }
        }

        log.info("Producers size: " + producers?.size())
        log.info("WineDb data size: " + wineDbList?.size())

        AtomicInteger counter = new AtomicInteger()
        if (producers?.size() > 0) {
            int logInterval = getLogInterval(producers.size())

            counter.set(0)
            GParsPool.withPool {
                producers?.eachParallel { k, v ->
                    Producer.withNewTransaction {
                        proceedProducer(k, v)
                    }

                    Integer cr = counter.getAndIncrement()
                    if (cr % logInterval == 0) {
                        log.info("Producers processed: " + cr)
                    }
                }
            }
            log.info("Producers processed: " + counter.get())
            sessionFactory.currentSession.clear()
        }

        if (wineDbList?.size() > 0) {
            int logInterval = getLogInterval(wineDbList.size())

            counter.set(0)
            GParsPool.withPool {
                wineDbList?.eachParallel { wineDb ->
                    WineDb.withNewTransaction {
                        proceedWineDb(wineDb)
                    }

                    Integer cr = counter.getAndIncrement()
                    if (cr % logInterval == 0) {
                        log.info("WineDb data processed: " + cr)
                    }
                }
            }
            log.info("WineDb data processed: " + counter.get())
            sessionFactory.currentSession.clear()
        }

        if (wineDbList?.size() > 0) {
            List<String> collectionWineDbs = new ArrayList<>()
            WineDb.withNewTransaction {
                collectionWineDbs = WineDb.wineDatabase.createCriteria().list {
                    projections {
                        property("id")
                    }
                    eq("status", WineDb.Status.ACTIVE)
                }
            }
            log.info("Active WineDbs in system: " + collectionWineDbs?.size())

            List<String> fileWineDbs = wineDbList.collect {(String) it.id}
            List<String> tobeRemoved = collectionWineDbs.findAll {!fileWineDbs.contains(it)}
            log.info("WineDbs to be removed: " + tobeRemoved?.size())

            if (tobeRemoved?.size() > 0) {
                tobeRemoved.unique()
                int logInterval = getLogInterval(tobeRemoved.size())

                List<String> toBeDeletedIndex = new ArrayList<>()
                counter.set(0)
                tobeRemoved?.each { String id ->
                    WineDb.withNewTransaction {
                        if (removeWineDb(id)) {
                            toBeDeletedIndex.add(id)
                        }
                    }

                    Integer cr = counter.getAndIncrement()
                    if (cr % logInterval == 0) {
                        log.info("WineDbs removed: " + cr)
                    }
                }
                //wineApiService.indexingWineDb(toBeDeletedIndex)

                log.info("WineDbs removed: " + counter.get())
                sessionFactory.currentSession.clear()
            }
        }

        log.info("Import WineDb data completed. Execution times: " + TimeCategory.minus(new Date(), date))
    }

    /**
     * Saves producer data from {@link InputService#importWineDb(String)}.
     * @param id the producer id
     * @param data the producer data mapping
     */
    private void proceedProducer(String id, List<String> data) {
        try {
            Producer producer = Producer.wineDatabase.findById(id)
            if (!producer) {
                producer = new Producer()
                producer.id = id
            }
            producer.name = data?.size() > 0 ? data.getAt(0) : ''
            producer.nameShow = data?.size() > 1 ? data.getAt(1) : ''

            producer.validate()
            if (producer.hasErrors()) {
                def errMsgList = producer.errors.allErrors.collect{
                    messageSource.getMessage(it, Locale.default)
                }
                throw new ValidationException(errMsgList?.join("\n"))
            }

            Boolean wineApiReindexRequired = producerService.isWineApiReindexRequired(producer)

            producer.save()

            if (wineApiReindexRequired) {
                //wineApiService.indexingWineDb(producer.wineDbs?.id)
            }
        } catch (ValidationException e) {
            log.warn("Error processing producer: " + id + ", message: " + e?.message)
        } catch (Exception e) {
            log.error("Error processing producer: " + id)
            log.error(e?.message, e)
        }
    }

    /**
     * Saves wine_db data from {@link InputService#importWineDb(String)}.<br>
     *     Deleted wineDb will be ignored.
     * @param data the wine_db data mapping
     */
    private void proceedWineDb(Map<String, String> data) {
        String id = data.id

        Boolean wineApiReindexRequired
        try {
            String producerId = id?.take(6)
            WineDb wineDb = WineDb.wineDatabase.findById(id)
            if (wineDb && wineDb.status!=WineDb.Status.ACTIVE) return

            if (!wineDb) {
                wineDb = new WineDb()
                wineDb.id = id
                wineDb.status = WineDb.Status.ACTIVE
                wineApiReindexRequired = true
            }
            wineDb.producer = Producer.wineDatabase.findById(producerId)
            wineDb.name = data.name
            wineDb.wineNId = data.wineNId
            wineDb.variety = data.variety
            wineDb.color = data.color
            wineDb.sweetness = data.sweetness
            wineDb.type = data.type
            wineDb.wineCountry = data.wineCountry
            wineDb.wineRegion = data.wineRegion
            wineDb.wineLocation = data.wineLocation
            wineDb.wineLocale = data.wineLocale
            wineDb.wineSite = data.wineSite

            wineDb.validate()
            if (wineDb.hasErrors()) {
                def errMsgList = wineDb.errors.allErrors.collect{
                    messageSource.getMessage(it, Locale.default)
                }
                throw new ValidationException(errMsgList?.join("\n"))
            }

            if(!wineApiReindexRequired) {
                 wineApiReindexRequired = wineService.isWineApiReindexRequired(wineDb)
            }

            wineDb.save()

        } catch (ValidationException e) {
            log.warn("Error processing wine db: " + id + ", message: " + e?.message)
        } catch (Exception e) {
            log.error("Error processing wine db: " + id)
            log.error(e?.message, e)
        }

        if (wineApiReindexRequired) {
            //wineApiService.indexingWineDb([id])
        }
    }

    /**
     * Removes wine db from wine_db table ({@link wine.db.WineDb}) based on given wine db id
     * @param wineDbId the wine db id
     */
    boolean removeWineDb(String wineDbId) {
        try {
            WineDb wineDb = WineDb.wineDatabase.findById(wineDbId)
            if (wineDb) {
                wineService.deleteWineDb(wineDb)
                return true
            }
        } catch (Exception e) {
            log.error(e?.message, e)
        }
        return false
    }

    /**
     * Parse given input file and update wine_appellation table ({@link wine.db.WineAppellation})<br>
     * Input file format is text with Tab as separator.<br>
     * 0: country<br>
     * 1: region<br>
     * 2: location<br>
     * 3: locale<br>
     * 4: site
     * @param fileName the location of file to be parsed
     */
    void importWineAppellations(String fileName) {
        log.info("Import Wine Appellations started.")
        List<List<String>> appellationList = IOUtils.getContents(fileName, "\t", true)
        log.info("Wine Appellations size: " + appellationList?.size())

        if (appellationList?.size() > 0) {
            AtomicInteger counter = new AtomicInteger()
            int logInterval = getLogInterval(appellationList.size())

            counter.set(0)
            String importTag = new Date().getTime().toString()
            GParsPool.withPool {
                appellationList?.eachParallel { k ->
                    WineAppellation.withNewTransaction {
                        proceedWineAppellation(k, importTag)
                    }

                    Integer cr = counter.getAndIncrement()
                    if (cr % logInterval == 0) {
                        log.info("Producers processed: " + cr)
                    }
                }
            }
            log.info("Producers processed: " + counter.get())

            Integer deleted = WineAppellation.wineDatabase.executeUpdate("DELETE FROM WineAppellation WHERE importTag IS NULL OR importTag!=:importTag", [importTag: importTag])
            log.debug("Deleted Wine Appellations: " + deleted)

            sessionFactory.currentSession.clear()
        }
        log.info("Import Wine Appellations completed.")
    }

    /**
     * Saves wine_appellation data from {@link InputService#importWineAppellations(String)}.<br>
     * 0: country<br>
     * 1: region<br>
     * 2: location<br>
     * 3: locale<br>
     * 4: site
     * @param data the data to be saved
     * @param importTag the unique tag
     */
    private void proceedWineAppellation(List<String> data, String importTag) {
        if (!data || !data?.getAt(0)) return
        try {
            WineAppellation wineAppellation = WineAppellation.wineDatabase.createCriteria().get {
                and {
                    eq("country", data?.getAt(0))
                    if (data?.getAt(1)) {
                        eq("region", data?.getAt(1))
                    } else {
                        or {
                            isNull("region")
                            eq("region", "")
                        }
                    }
                    if (data?.getAt(2)) {
                        eq("location", data?.getAt(2))
                    } else {
                        or {
                            isNull("location")
                            eq("location", "")
                        }
                    }
                    if (data?.getAt(3)) {
                        eq("locale", data?.getAt(3))
                    } else {
                        or {
                            isNull("locale")
                            eq("locale", "")
                        }
                    }
                    if (data?.getAt(4)) {
                        eq("site", data?.getAt(4))
                    } else {
                        or {
                            isNull("site")
                            eq("site", "")
                        }
                    }
                }
                maxResults(1)
            }

            if (!wineAppellation) {
                wineAppellation = new WineAppellation()
            }
            wineAppellation.country = data?.getAt(0)
            wineAppellation.region = data?.getAt(1)
            wineAppellation.location = data?.getAt(2)
            wineAppellation.locale = data?.getAt(3)
            wineAppellation.site = data?.getAt(4)
            wineAppellation.importTag = importTag

            // Get Region Code from the existing
            String countryRegion = data?.getAt(0) ?: ""
            countryRegion = countryRegion + (data?.getAt(1) ? " " + data?.getAt(1) : "")
            String query = "FROM WineAppellation WHERE CONCAT_WS(' ', country, region) = :countryRegion"
            def results = WineAppellation.wineDatabase.executeQuery(query, [countryRegion: countryRegion], [max:1])
            if (results) {
                wineAppellation.regionCode = results?.getAt(0)?.regionCode
            }

            wineAppellation.save(flush: true)
        } catch (Exception e) {
            log.error("Error when processing wine appellation: " + data)
            log.error(e?.message, e)
        }
    }

    /**
     * Saves producer_keyword data from {@link ConfigFilesWatcherJob#importProducerIndexingKeywords(String)}.
     * @param prodId the producer id
     * @param prodKeyword the producer keyword
     */
    void proceedProducerKeyword(String prodId, String prodKeyword) {
        Producer producer = Producer.wineDatabase.findById(prodId)
        if (producer) {
            try {
                ProducerKeyword producerKeyword = ProducerKeyword.wineDatabase.findByProducer(producer)
                if (!producerKeyword) {
                    producerKeyword = new ProducerKeyword()
                    producerKeyword.producer = producer
                }
                producerKeyword.keyword = prodKeyword
                producerKeyword.save()
            } catch (Exception e) {
                log.error(e?.message, e)
            }
        }
    }

    /**
     * Gets producer ids of producer keyword
     * @return the list of producer id
     */
    List<String> getProducerIdsOfProducerKeywords() {
        def ids = ProducerKeyword.wineDatabase.createCriteria().list {
            projections {
                producer {
                    property("id")
                }
            }
        }
        return (List<String>) ids
    }

    /**
     * Deletes record from ProducerKeyword based on given producer id
     * @param producerId the producer id
     */
    void removeProducerKeyword(String producerId) {
        try {
            Producer producer = Producer.wineDatabase.findById(producerId)
            if (!producer) return

            ProducerKeyword producerKeyword = producer.producerKeyword
            if (!producerKeyword) return

            log.debug("Removing producer keyword for ${producerId} with id: ${producerKeyword.id}")
            producer.producerKeyword = null
            producerKeyword.producer = null
            producerKeyword.delete()
        } catch (Exception e) {
            log.error(e?.message, e)
        }
    }

    /**
     * Parse given input file and update region_code_id at wine_appellation table ({@link wine.db.WineAppellation#regionCode})<br>
     * Input file format is text with Tab as separator.
     * @param newFile the file to be parsed
     */
    void importRegionMap(String newFile) {
        log.info("Import Region Maps started.")
        Date date = new Date()

        File file = new File(newFile)

        List<Map<String, String>> regions = new ArrayList<>()
        String fileEncoding = CharsetUtils.detect(newFile) ?: "UTF-8"
        file.withReader(fileEncoding) { reader ->
            def line
            while ((line = reader.readLine())!=null) {
                def regionInfo = line?.split("\t")
                String countryRegion = regionInfo?.getAt(0)?.trim()
                if (StringUtils.startsWith(countryRegion, "\"") && StringUtils.endsWith(countryRegion, "\"")) { // Only if start and end with quote. Handler for excel quote escape
                    countryRegion = StringUtils.substring(countryRegion, 1, countryRegion.length() - 1)
                }
                String code = regionInfo?.size() > 1 ? regionInfo?.getAt(1)?.trim() : null
                if (code) {
                    regions.add([code: code, countryRegion: countryRegion])
                }
            }
        }
        log.info("Region Maps size: " + regions?.size())

        AtomicInteger counter = new AtomicInteger()
        if (regions?.size() > 0) {
            int logInterval = getLogInterval(regions.size())

            counter.set(0)
            regions?.each { item ->
                WineAppellation.withNewTransaction {
                    proceedRegionMap(item)
                }

                Integer cr = counter.getAndIncrement()
                if (cr % logInterval == 0) {
                    log.info("Region Maps processed: " + cr)
                }
            }
            log.info("Region Maps processed: " + counter.get())
            sessionFactory.currentSession.clear()
        }

        log.info("Import Region Maps completed. Execution times: " + TimeCategory.minus(new Date(), date))
    }

    /**
     * Saves region map from {@link InputService#importRegionMap(String)}.
     * @param region the region data to be saved
     */
    void proceedRegionMap(Map<String, String> region) {
        RegionCode regionCode = RegionCode.wineDatabase.findById(region.code)
        if (regionCode) {
            try {
                String query = "FROM WineAppellation WHERE CONCAT_WS(' ', country, region) = :countryRegion"
                def results = WineAppellation.wineDatabase.executeQuery(query, [countryRegion: region.countryRegion])
                if (results) {
                    results?.each { WineAppellation appellation ->
                        appellation.regionCode = regionCode
                        appellation.save()
                    }
                }
            } catch (Exception e) {
                log.error(e?.message, e)
            }
        }
    }

    /**
     * Parse given input file and update global_noise table ({@link wine.alert.GlobalNoise})<br>
     * Input file format is text with Tab as separator.<br>
     * Input file doesn't contains header.
     * @param newFile the file to be parsed
     */
    void importGlobalNoise(String newFile) {
        Date date = new Date()

        File file = new File(newFile)

        List<String> noises = new ArrayList<>()
        String fileEncoding = CharsetUtils.detect(newFile) ?: "UTF-8"
        file.withReader(fileEncoding) { reader ->
            def line
            while ((line = reader.readLine())!=null) {
                if (line) noises.add((String) line)
            }
        }
        log.info("Global noise size: " + noises?.size())

        AtomicInteger counter = new AtomicInteger()
        if (noises?.size() > 0) {
            int logInterval = getLogInterval(noises.size())

            counter.set(0)
            noises?.each { item ->
                GlobalNoise.withNewTransaction {
                    try {
                        if (GlobalNoise.countByNoise(item) == 0) {
                            new GlobalNoise(noise: item).save()
                        }
                    } catch (Exception e) {
                        log.error(e?.message, e)
                    }
                }

                Integer cr = counter.getAndIncrement()
                if (cr % logInterval == 0) {
                    log.info("Global noise processed: " + cr)
                }
            }
            log.info("Global noise processed: " + counter.get())
            sessionFactory.currentSession.clear()
        }

        log.info("Import Global noise completed. Execution times: " + TimeCategory.minus(new Date(), date))
    }

    /**
     * Parse given input file and update producer_noise table ({@link wine.db.ProducerNoise})<br>
     * Input file format is text with Tab as separator.<br>
     * Input file contains header: Producer ID, Noise.
     * @param newFile the file to be parsed
     */
    void importProducerNoise(String newFile) {
        Date date = new Date()

        File file = new File(newFile)

        int i = 0
        List<List<String>> noises = new ArrayList<>()
        String fileEncoding = CharsetUtils.detect(newFile) ?: "UTF-8"
        file.withReader(fileEncoding) { reader ->
            def line
            while ((line = reader.readLine())!=null) {
                i += 1
                if (i == 1) continue

                List<String> noiseInfo = ((String) line)?.tokenize("\t")
                if (noiseInfo && noiseInfo.size() > 1) {
                    noises.add(noiseInfo)
                }
            }
        }
        log.info("Producer noise size: " + noises?.size())

        AtomicInteger counter = new AtomicInteger()
        if (noises?.size() > 0) {
            int logInterval = getLogInterval(noises.size())

            counter.set(0)
            noises?.each { item ->
                ProducerNoise.withNewTransaction {
                    try {
                        def c = ProducerNoise.wineDatabase.createCriteria().count {
                            eq("noise", item.get(1))
                            producer {
                                eq("id", item.get(0))
                            }
                        }
                        if (c == 0) {
                            Producer producer = Producer.wineDatabase.findById(item.get(0))
                            if (producer) {
                                new ProducerNoise(noise: item.get(1), producer: producer).save()
                            }
                        }
                    } catch (Exception e) {
                        log.error(e?.message, e)
                    }
                }

                Integer cr = counter.getAndIncrement()
                if (cr % logInterval == 0) {
                    log.info("Producer noise processed: " + cr)
                }
            }
            log.info("Producer noise processed: " + counter.get())
            sessionFactory.currentSession.clear()
        }

        log.info("Import Producer noise completed. Execution times: " + TimeCategory.minus(new Date(), date))
    }

    /**
     * Parse given input file and update wine_noise table ({@link wine.db.WineNoise})<br>
     *     Deleted wineDb will be ignored.<br>
     *     Input file format is text with Tab as separator.<br>
     *     Input file contains header: Wine Alert ID, Noise.
     * @param newFile the file to be parsed
     */
    void importWineNoise(String newFile) {
        Date date = new Date()

        File file = new File(newFile)

        int i = 0
        List<List<String>> noises = new ArrayList<>()
        String fileEncoding = CharsetUtils.detect(newFile) ?: "UTF-8"
        file.withReader(fileEncoding) { reader ->
            def line
            while ((line = reader.readLine())!=null) {
                i += 1
                if (i == 1) continue

                List<String> noiseInfo = ((String) line)?.tokenize("\t")
                if (noiseInfo && noiseInfo.size() > 1) {
                    noises.add(noiseInfo)
                }
            }
        }
        log.info("Wine noise size: " + noises?.size())

        AtomicInteger counter = new AtomicInteger()
        if (noises?.size() > 0) {
            int logInterval = getLogInterval(noises.size())

            counter.set(0)
            noises?.each { item ->
                WineNoise.withNewTransaction {
                    try {
                        def c = WineNoise.wineDatabase.createCriteria().count {
                            eq("noise", item.get(1))
                            wineDb {
                                eq("id", item.get(0))
                            }
                        }
                        if (c == 0) {
                            WineDb wineDb = WineDb.wineDatabase.findByIdAndStatus(item.get(0), WineDb.Status.ACTIVE)
                            if (wineDb) {
                                new WineNoise(noise: item.get(1), wineDb: wineDb).save()
                            }
                        }
                    } catch (Exception e) {
                        log.error(e?.message, e)
                    }
                }

                Integer cr = counter.getAndIncrement()
                if (cr % logInterval == 0) {
                    log.info("Wine noise processed: " + cr)
                }
            }
            log.info("Wine noise processed: " + counter.get())
            sessionFactory.currentSession.clear()
        }

        log.info("Import Wine noise completed. Execution times: " + TimeCategory.minus(new Date(), date))
    }

    /**
     * Parse given input file and update word_alias table ({@link wine.alert.WordAlias})<br>
     * Input file format is text with Pipe ("|") as separator.
     * @param newFile the file to be parsed
     * @param aliasType the alias type from enum {@link enums.AliasType}
     */
    void importAliases(String newFile, AliasType aliasType) {
        File file
        try {
            file = new File(newFile)
            if (!file) {
                log.error("File not found.")
                return
            }
        } catch (Exception e) {
            log.warn(e?.message, e)
            return
        }
        log.info("Import ${aliasType?.name()} aliases started.")

        List<List<String>> aliases = new ArrayList<>()
        String fileEncoding = CharsetUtils.detect(newFile) ?: "UTF-8"
        file.withReader(fileEncoding) { reader ->
            def line
            while ((line = reader.readLine()) != null) {
                def words = line?.tokenize("|")
                if (words && words?.size() > 0) aliases.add(words)
            }
        }
        log.info("${aliasType?.name()} Aliases size: ${aliases.size()}")

        if (CollectionUtils.isNotEmpty(aliases)) {
            int c = 0
            aliases.collate(50).each { list ->
                WordAlias.withNewTransaction {
                    list.each { k ->
                        proceedAlias(k, aliasType)
                        c++
                        if (c % 5000 == 0) {
                            log.info("${aliasType?.name()} Aliases processed: $c")
                        }
                    }
                }
                sessionFactory.currentSession.clear()
            }
            log.info("${aliasType?.name()} Aliases processed: $c")

            List<String> words = aliases.collect { it.getAt(0)?.replace(';', '|') }
            words.removeAll { !it }
            WordAlias.withNewTransaction {
                Integer affected = deleteUnusedAliases(words, aliasType)
                log.info("Deleted ${aliasType.name()} aliases: ${affected}")
            }
            sessionFactory.currentSession.clear()

            if (aliasType == AliasType.WINE) RemapKeywordsJob.triggerNow()
        }
        log.info("Import ${aliasType?.name()} aliases completed.")
    }

    /**
     * Saves word_alias data from {@link InputService#importAliases(String, AliasType)}.
     * @param words the list of word, first word is the original word and the rest is word alias.
     * @param aliasType the alias type
     */
    void proceedAlias(List<String> words, AliasType aliasType) {
        try {
            String word = words?.getAt(0)
            String alias = words?.drop(1)?.join("|")
            if (aliasType == AliasType.CUSTOM_CHANGES) {
                word = StringUtils.replace(word, ';', '|')
                alias = StringUtils.replace(alias, ';', '|')
            }

            def wordAlias = WordAlias.findByWordAndAliasType(word, aliasType)
            if (wordAlias) {
                wordAlias.alias = alias
                wordAlias.save()
            } else {
                new WordAlias(
                        word: word,
                        alias: alias,
                        aliasType: aliasType
                ).save()
            }
        } catch (Exception e) {
            log.error(e?.message, e)
        }
    }

    /**
     * Deletes unused aliases
     * @param validWords the list of valid words
     * @param aliasType the alias type
     * @return the total affected rows
     */
    Integer deleteUnusedAliases(List<String> validWords, AliasType aliasType) {
        if (CollectionUtils.isEmpty(validWords)) return 0
        def criteria = WordAlias.where {
            aliasType == aliasType && !(word in validWords)
        }
        Integer c = criteria.deleteAll()?.intValue()
        return c
    }

    /**
     * Parses given input file and update word_elimination table ({@link wine.alert.WordElimination})<br>
     * Input file format is text with Pipe ("|") as separator.
     * @param newFile the file to be parsed
     * @param eliminationType the elimination type from enum {@link wine.alert.enums.EliminationType}
     */
    void importEliminations(String newFile, EliminationType eliminationType) {
        File file
        try {
            file = new File(newFile)
            if (!file) {
                log.error("File not found.")
                return
            }
        } catch (Exception e) {
            log.warn(e?.message, e)
            return
        }
        log.info("Import ${eliminationType?.name()} Eliminations started.")

        List<String> eliminations = new ArrayList<>()
        String fileEncoding = CharsetUtils.detect(newFile) ?: "UTF-8"
        file.withReader(fileEncoding) { reader ->
            def line
            while ((line = reader.readLine()) != null) {
                def words = line?.tokenize("|")
                if (words && words?.size() > 0) eliminations.add(words.get(0))
            }
        }
        log.info("${eliminationType?.name()} Eliminations size: ${eliminations.size()}")

        if (CollectionUtils.isNotEmpty(eliminations)) {
            int c = 0
            eliminations.collate(50).each { list ->
                WordElimination.withNewTransaction {
                    list.each { k ->
                        proceedElimination(k, eliminationType)
                        c++
                        if (c % 5000 == 0) {
                            log.info("${eliminationType?.name()} Eliminations processed: $c")
                        }
                    }
                }
                sessionFactory.currentSession.clear()
            }
            log.info("${eliminationType?.name()} Eliminations processed: $c")

            WordElimination.withNewTransaction {
                Integer affected = deleteUnusedEliminations(eliminations, eliminationType)
                log.info("Deleted ${eliminationType.name()} eliminations: ${affected}")
            }
            sessionFactory.currentSession.clear()
        }
        log.info("${eliminationType?.name()} Eliminations completed.")
    }

    /**
     * Saves word_elimination data from {@link InputService#importEliminations(String, EliminationType)}.
     * @param word the word to save in elimination table
     * @param eliminationType the elimination type
     */
    void proceedElimination(String word, EliminationType eliminationType) {
        try {
            def elimination = WordElimination.findByWordAndEliminationType(word, eliminationType)
            if (!elimination) {
                new WordElimination(
                        word: word,
                        eliminationType: eliminationType
                ).save()
            }
        } catch (Exception e) {
            log.error(e?.message, e)
        }
    }

    /**
     * Deletes unused eliminations
     * @param validWords the list of valid words
     * @param eliminationType the elimination type
     * @return the total affected rows
     */
    Integer deleteUnusedEliminations(List<String> validWords, EliminationType eliminationType) {
        if (CollectionUtils.isEmpty(validWords)) return 0
        def criteria = WordElimination.where {
            eliminationType == eliminationType && !(word in validWords)
        }
        Integer c = criteria.deleteAll()?.intValue()
        return c
    }

    /**
     * Parses given input file and update keyword_substitution table ({@link wine.alert.KeywordSubstitution})<br>
     * Input file format is text with Tab as separator.
     * @param newFile the file to be parsed
     * @param substituteType the substitution type from enum {@link wine.alert.enums.SubstitutionType}
     */
    void importSubstitution(String newFile, SubstitutionType substituteType) {
        File file
        try {
            file = new File(newFile)
            if (!file) {
                log.error("File not found.")
                return
            }
        } catch (Exception e) {
            log.warn(e?.message, e)
            return
        }
        log.info("Import ${substituteType?.name()} substitutions started.")

        List<List<String>> substitutions = []
        String fileEncoding = CharsetUtils.detect(newFile) ?: "UTF-8"
        file.withReader(fileEncoding) { reader ->
            def line
            while ((line = reader.readLine()) != null) {
                def words = line?.tokenize("\t")
                if (words && words?.size() > 1) {
                    substitutions.add([words.get(0), words.get(1)])
                }
            }
        }
        log.info("${substituteType?.name()} substitutions size: ${substitutions.size()}")

        if (CollectionUtils.isNotEmpty(substitutions)) {
            int c = 0
            substitutions.collate(50).each { list ->
                KeywordSubstitution.withNewTransaction {
                    list.each { s ->
                        proceedSubstitution(s[0], s[1], substituteType)
                        c++
                        if (c % 5000 == 0) {
                            log.info("${substituteType?.name()} substitutions processed: $c")
                        }
                    }
                }
                sessionFactory.currentSession.clear()
            }
            log.info("${substituteType?.name()} substitutions processed: $c")

            def keywords = substitutions.collect { it[0] }
            KeywordSubstitution.withNewTransaction {
                Integer affected = deleteUnusedSubstitutions(keywords, substituteType)
                log.info("Deleted ${substituteType.name()} substitutions: ${affected}")
            }
            sessionFactory.currentSession.clear()
        }
        log.info("${substituteType?.name()} substitutions completed.")
    }

    /**
     * Saves keyword_substitution data from {@link InputService#importSubstitution(String, SubstitutionType)}.
     * @param keyword the keyword to save in keyword_substitution table
     * @param substituteType the substitution type
     */
    void proceedSubstitution(String keyword, String substitution, SubstitutionType substituteType) {
        try {
            def keywordSubstitution = KeywordSubstitution.findByKeywordAndSubstitutionType(keyword, substituteType)
            if (keywordSubstitution) {
                keywordSubstitution.substitution = substitution
                keywordSubstitution.save()
            } else {
                new KeywordSubstitution(
                        keyword: keyword,
                        substitution: substitution,
                        substitutionType: substituteType
                ).save()
            }
        } catch (Exception e) {
            log.error(e?.message, e)
        }
    }

    /**
     * Deletes unused substitutions
     * @param validWords the list of valid words
     * @param substituteType the substitution type
     * @return the total affected rows
     */
    Integer deleteUnusedSubstitutions(List<String> validWords, SubstitutionType substituteType) {
        if (CollectionUtils.isEmpty(validWords)) return 0
        def criteria = KeywordSubstitution.where {
            substitutionType == substituteType && !(keyword in validWords)
        }
        Integer c = criteria.deleteAll()?.intValue()
        return c
    }

    /**
     * Parses given input file and update keyword_spread table ({@link wine.alert.KeywordSpread})<br>
     * Input file format is text file.
     * @param newFile the file to be parsed
     * @param regionCode the region code from region_code table ({@link wine.db.RegionCode})
     * @param keywordType the keyword type from enum {@link wine.alert.enums.KeywordType}
     */
    void importSpreadTable(String newFile, String regionCode, KeywordType keywordType) {
        File file
        try {
            file = new File(newFile)
            if (!file) {
                log.error("File not found.")
                return
            }
        } catch (Exception e) {
            log.warn(e?.message, e)
            return
        }
        log.info("Import ${keywordType?.name()} Spread tables [$regionCode] started.")

        List<String> spreads = new ArrayList<>()
        String fileEncoding = CharsetUtils.detect(newFile) ?: "UTF-8"
        file.withReader(fileEncoding) { reader ->
            def line
            while ((line = reader.readLine()) != null) {
                if (line) spreads.add(line)
            }
        }
        log.info("${keywordType?.name()} Spread tables [$regionCode] size: ${spreads?.size()}")

        if (CollectionUtils.isNotEmpty(spreads)) {
            int c = 0
            spreads.collate(50).each { list ->
                KeywordSpread.withNewTransaction {
                    list.each { k ->
                        proceedSpreadTable(k, regionCode, keywordType)
                        c++
                        if (c % 5000 == 0) {
                            log.info("${keywordType?.name()} Spread tables [$regionCode] processed: $c")
                        }
                    }
                }
                sessionFactory.currentSession.clear()
            }
            log.info("${keywordType?.name()} Spread tables [$regionCode] processed: $c")

            KeywordSpread.withNewTransaction {
                Integer affected = deleteUnusedSpreadTables(spreads, regionCode, keywordType)
                log.info("Deleted ${keywordType.name()} spread tables [$regionCode]: ${affected}")
            }
            sessionFactory.currentSession.clear()
        }
        log.info("${keywordType?.name()} Spread tables [$regionCode] completed.")
    }

    /**
     * Saves keyword_spread data from {@link InputService#importSpreadTable(String, String, KeywordType)}.
     * @param keyword the keyword to save in keyword_substitution table
     * @param regionCode the region code from region_code table ({@link wine.db.RegionCode})
     * @param keywordType the keyword type from enum {@link wine.alert.enums.KeywordType}
     */
    void proceedSpreadTable(String keyword, String regionCode, KeywordType keywordType) {
        try {
            def keywordSpread = KeywordSpread.findByKeywordAndRegionCodeAndKeywordType(keyword, regionCode, keywordType)
            if (!keywordSpread) {
                new KeywordSpread(
                        keyword: keyword,
                        regionCode: regionCode,
                        keywordType: keywordType
                ).save()
            }
        } catch (Exception e) {
            log.error(e?.message, e)
        }
    }

    /**
     * Deletes unused spread tables
     * @param validWords the list of valid words
     * @param regionCode the region code from region_code table ({@link wine.db.RegionCode})
     * @param keywordType the keyword type from enum {@link wine.alert.enums.KeywordType}
     * @return the total affected rows
     */
    Integer deleteUnusedSpreadTables(List<String> validWords, String regionCode, KeywordType keywordType) {
        if (CollectionUtils.isEmpty(validWords)) return 0
        def criteria = KeywordSpread.where {
            keywordType == keywordType && regionCode == regionCode && !(keyword in validWords)
        }
        Integer c = criteria.deleteAll()?.intValue()
        return c
    }

    /**
     * Parses given input file and update word_reserved table ({@link wine.alert.WordReserved})<br>
     * Input file format is text file.
     * @param newFile the file to be parsed
     * @param reservedType the reserved type from enum {@link wine.alert.enums.ReservedType}
     */
    void importReservedWords(String newFile, ReservedType reservedType) {
        File file
        try {
            file = new File(newFile)
            if (!file) {
                log.error("File not found.")
                return
            }
        } catch (Exception e) {
            log.warn(e?.message, e)
            return
        }
        log.info("Import ${reservedType?.name()} reserved words started.")

        List<String> words = new ArrayList<>()
        String fileEncoding = CharsetUtils.detect(newFile) ?: "UTF-8"
        file.withReader(fileEncoding) { reader ->
            def line
            while ((line = reader.readLine()) != null) {
                if (line) words.add(line)
            }
        }
        log.info("${reservedType?.name()} reserved words size: ${words.size()}")

        if (CollectionUtils.isNotEmpty(words)) {
            int c = 0
            words.collate(50).each { list ->
                WordReserved.withNewTransaction {
                    list.each { k ->
                        proceedReservedWord(k, reservedType)
                        c++
                        if (c % 5000 == 0) {
                            log.info("${reservedType?.name()} reserved words processed: $c")
                        }
                    }
                }
                sessionFactory.currentSession.clear()
            }
            log.info("${reservedType?.name()} reserved words processed: $c")

            WordReserved.withNewTransaction {
                Integer affected = deleteUnusedReservedWords(words, reservedType)
                log.info("Deleted ${reservedType.name()} reserved words: ${affected}")
            }
            sessionFactory.currentSession.clear()
        }
        log.info("${reservedType?.name()} reserved words completed.")
    }

    /**
     * Saves word_reserved data from {@link InputService#importReservedWords(String, ReservedType)}.
     * @param word the word to save in word_reserved table
     * @param reservedType the reserved type from enum {@link wine.alert.enums.ReservedType}
     */
    void proceedReservedWord(String word, ReservedType reservedType) {
        try {
            def elimination = WordReserved.findByWordAndReservedType(word, reservedType)
            if (!elimination) {
                new WordReserved(
                        word: word,
                        reservedType: reservedType
                ).save()
            }
        } catch (Exception e) {
            log.error(e?.message, e)
        }
    }

    /**
     * Deletes unused reserved words
     * @param validWords the list of valid words
     * @param reservedType the reserved type from enum {@link wine.alert.enums.ReservedType}
     * @return the total affected rows
     */
    Integer deleteUnusedReservedWords(List<String> validWords, ReservedType reservedType) {
        if (CollectionUtils.isEmpty(validWords)) return 0
        def criteria = WordReserved.where {
            reservedType == reservedType && !(word in validWords)
        }
        Integer c = criteria.deleteAll()?.intValue()
        return c
    }

    /**
     * Gets log interval
     * @param dataSize the data size
     * @return the interval size
     */
    private static int getLogInterval(Integer dataSize) {
        if (!dataSize) return 1
        int logInterval = Math.floor(dataSize / 4).intValue()
        if (logInterval == 0) logInterval = 1
        return logInterval
    }
}
