package wine.alert

import enums.AliasType
import grails.compiler.GrailsCompileStatic
import grails.transaction.Transactional
import groovy.transform.TypeCheckingMode
import org.apache.commons.lang3.StringUtils
import wine.db.BottleSize
import wine.db.Vintage
import wine.db.WineDb
import wine.db.WineKeyword

import java.util.concurrent.ConcurrentHashMap
import java.util.concurrent.ConcurrentLinkedQueue
import java.util.regex.Pattern

/**
 * In memory data processing
 */
@GrailsCompileStatic
@Transactional(readOnly = true)
class KeywordsService {

    private static final String REGEX_PARSE_PREFIX = "(?i)(^|\\b|\\p{Punct}|\\s)"
    private static final String REGEX_PARSE_SUFFIX = "(\\s|\\p{Punct}|\\b|\$)"

    private static final String STR_SPACE = " "
    private static final String STR_BLANK = ""

    private static final Comparator<String> reverseLengthComparator = new Comparator<String>() {
        public int compare(String a, String b) {
            int r = Integer.compare(b.length(), a.length())
            if (r != 0) {
                return r
            }
            return b.compareTo(a)
        }
    }

    private ConcurrentHashMap<String, List<String>> candidatesMap // [$keyword: [$wineDbId, $wineDbId, ...]]
    private ConcurrentHashMap<String, String> patternMap // [$wineDbId: $pattern]
    private ConcurrentHashMap<String, String> normalizedPatternMap // [$wineDbId: $normalizedPattern]
    private ConcurrentHashMap<String, String> normalizedSortedPatternMap // [$wineDbId: $normalizedSortedPattern]
    private ConcurrentHashMap<String, List<String>> requiredWords // [$wineDbId: [$word, $word, ...]]
    private TreeMap<String, String> aliasesMap // [$keyword: $replacement]
    private TreeMap<String, String> bottleAliasesMap // [$name: $id]
    private ConcurrentHashMap<String, String> vintagesMap // [$id, $year]
    private Queue<String> bottleSizes = new ConcurrentLinkedQueue<String>()
    private Queue<String> bottleExclusions = new ConcurrentLinkedQueue<String>()
    private ConcurrentHashMap<String, String> bottlesMap // [$id: $name]

    def outputService

    /**
     * Prepares all required data for history match and keyword match.
     */
    void prepare() {
        mapCandidates()
        mapAliases()
        mapBottleAliases()
        mapVintages()
    }

    /**
     * Gets list of wine's candidate based on provided word<br>
     * Refer to {@link KeywordsService#mapCandidates} for more details how to map the wine's candidates.
     * @param word the word from wine name
     * @return the list of wine db id
     */
    public List<String> getWineCandidates(String word) {
        if (word == null) return []
        return candidatesMap?.get(word) as List
    }

    /**
     * Gets wine_db's keyword/pattern
     * @param wineDbId
     * @return the wine db pattern/keyword
     */
    public String getPattern(String wineDbId) {
        if (StringUtils.isBlank(wineDbId)) {
            return STR_BLANK
        }
        return patternMap?.get(wineDbId) ?: STR_BLANK
    }

    /**
     * Gets wine_db's normalized and sorted keyword/pattern.<br>
     * Refer to {@link KeywordsService#mapCandidates} for more details how to normalize and sort the keyword/pattern.
     * @param wineDbId
     * @return the normalized and sorted keyword/pattern
     */
    public String getNormalizedSortedPattern(String wineDbId) {
        if (StringUtils.isBlank(wineDbId)) {
            return STR_BLANK
        }
        return normalizedSortedPatternMap?.get(wineDbId) ?: STR_BLANK
    }

    /**
     * Gets wine_db's normalized keyword/pattern.<br>
     * Refer to {@link KeywordsService#mapCandidates} for more details how to normalize the keyword/pattern.
     * @param wineDbId
     * @return normalized keyword/pattern
     */
    public String getNormalizedPattern(String wineDbId) {
        if (StringUtils.isBlank(wineDbId)) {
            return STR_BLANK
        }
        return normalizedPatternMap?.get(wineDbId) ?: STR_BLANK
    }

    /**
     * Gets wine_db's required words.<br>
     * Refer to {@link KeywordsService#mapCandidates} for more details how required words are mapped.
     * @param wineDbId
     * @return the list of required words
     */
    public List<String> getRequiredWords(String wineDbId) {
        return requiredWords?.get(wineDbId)
    }

    /**
     * Gets bottle id based on given size.
     * @param size the bottle size. Size must be in bottle table ({@link wine.db.BottleSize}) name or alias.
     * @return the bottle size id
     */
    public String getBottleId(String size) {
        if (StringUtils.isBlank(size)) {
            return null
        }
        return bottleAliasesMap?.get(size?.toLowerCase())
    }

    /**
     * Gets bottle size based on given bottle id.
     * @param id the bottle size id
     * @return the bottle size name
     */
    String getBottleSize(String id) {
        if (StringUtils.isBlank(id)) return null
        return bottlesMap?.get(id)
    }

    /**
     * Gets vintage id based on given vintage.
     * @param year the vintage name. Vintage must be in vintage table ({@link wine.db.Vintage}) name.
     * @return the vintage id
     */
    public String getVintageId(String year) {
        if (StringUtils.isBlank(year)) return null
        String id = ""
        for (def v : vintagesMap) {
            if (year?.equalsIgnoreCase(v.value)) {
                id = v.key
                break
            }
        }
        return id
    }

    /**
     * Gets vintage name based on given vintage id.
     * @param the vintage id
     * @return the vintage name
     */
    public String getVintage(String id) {
        if (StringUtils.isBlank(id)) {
            return null
        }
        return vintagesMap?.get(id)
    }

    /**
     * Search for bottle size in text.<br>
     * Refer to {@link KeywordsService#mapCandidates} for more details how bottle sizes are mapped.
     * @param text the text to look into.
     * @return the bottle size name
     */
    public String parseBottleSize(String text) {
        if (StringUtils.isBlank(text)) return null
        for (def b : bottleAliasesMap) {
            if (text.find(REGEX_PARSE_PREFIX + Pattern.quote(b.key) + REGEX_PARSE_SUFFIX)) {
                return b.key
            }
        }
        return null
    }

    /**
     * Gets list of wine's candidate based on provided wine name.
     * @param name the wine name form retailer's wine
     * @return list of wine db id
     */
    public List<String> getCandidateList(String name) {
        List<String> candidateList = new ArrayList<>()
        name?.split("\\s+")?.toUnique()?.each { String word ->

            // first check the token itself
            List<String> wordCandidates = this.getWineCandidates(word)
            if (wordCandidates && !wordCandidates.isEmpty()) {
                candidateList.addAll(wordCandidates)
            }

            // then check all possible substrings with wildcard suffixes '_'
            // ie: apple
            // apple, apple_, appl_, app_, ap_, a_
            for (int i = word.length(); i > 0; i--) {
                String subToken = word.substring(0, i) + Calculator.WILDCARD_SYMBOL
                List<String> subCandidates = this.getWineCandidates(subToken)
                if (subCandidates && !subCandidates.isEmpty()) {
                    candidateList.addAll(subCandidates)
                }
            }
        }
        return candidateList?.unique()
    }

    /**
     * Replace given wine name with the standard wine db standard naming convention.
     * @param inputText the wine name/retailer's wine name
     * @return the standardized wine name
     */
    public String replaceAliases(String inputText) {
        if (StringUtils.isBlank(inputText)) {
            return null
        }
        inputText = inputText?.toLowerCase()
        aliasesMap?.each { String k, String v ->
            inputText = inputText?.replaceAll(REGEX_PARSE_PREFIX + Pattern.quote(k) + REGEX_PARSE_SUFFIX, STR_SPACE + v + STR_SPACE)
            inputText = inputText?.replaceAll("\\s+", STR_SPACE)
        }
        return inputText?.trim()
    }

    /**
     * Removes bottle sizes from given wine name.
     * @param inputText the wine name/retailer's wine name
     * @return the wine name without bottle sizes
     */
    public String removeSizeInfo(String inputText) {
        if (StringUtils.isBlank(inputText)) {
            return null
        }
        inputText = inputText?.toLowerCase()
        bottleSizes?.each { String bottle ->
            if (!bottleExclusions.contains(bottle)) {
                inputText = inputText.replaceAll(REGEX_PARSE_PREFIX + Pattern.quote(bottle) + REGEX_PARSE_SUFFIX, STR_SPACE)
            }
        }
        return inputText?.trim()
    }

    /**
     * Prepares required data (related to wine db keyword) for keyword match.<br>
     * Maps candidates, map between word and the list of wine db id<br>
     * Maps pattern, map between wine db id and wine db pattern/keyword<br>
     * Maps normalized & sorted pattern, map between wine db id and normalized & sorted pattern/keyword<br>
     * Maps normalized pattern, map between wine db id and normalized pattern/keyword<br>
     * Maps required words, map between wine db id and required words
     */
    @GrailsCompileStatic(TypeCheckingMode.SKIP)
    public void mapCandidates() {
        log.info("Mapping keywords started")

        candidatesMap = new ConcurrentHashMap<String, List<String>>()
        patternMap = new ConcurrentHashMap<String, String>()
        normalizedSortedPatternMap = new ConcurrentHashMap<String, String>()
        normalizedPatternMap = new ConcurrentHashMap<String, String>()
        requiredWords = new ConcurrentHashMap<String, List<String>>()

        def keywords = WineKeyword.wineDatabase.createCriteria().list {
            projections {
                wineDb {
                    property("id")
                }
                property("pattern")
                property("producerKeyword")
            }
            wineDb {
                eq("status", WineDb.Status.ACTIVE)
            }
            //wineDb { eq("id", "al1170113") }
            //wineDb { like("id", "al1100%") }
        }

        keywords.each { wk ->
            String wineId = wk.getAt(0)
            String pattern = wk.getAt(1)
            String producerKeyword = wk.getAt(2)

            patternMap.put(wineId, pattern)
            normalizedPatternMap.put(wineId, Calculator.normalizePattern(pattern))
            normalizedSortedPatternMap.put(wineId, Calculator.normalizeAndSortPattern(pattern))

            List<String> rws = Calculator.requiredWords(pattern, true)
            if (rws && !rws.isEmpty()) {
                requiredWords.put(wineId, rws)
            }

            List<String> words = new ArrayList<>()
            List<String> requiredWords = Calculator.requiredWords(pattern, false)
            if (requiredWords.isEmpty()) {
                // If required words for non "or logic" is empty
                // use all words include the "or logic"
                words = producerKeyword ? Calculator.truePositiveWords(producerKeyword) : Calculator.truePositiveWords(pattern)
            } else {
                // Get the most lower usage of word
                int score = 0
                String candidateWord = null
                requiredWords.eachWithIndex { word, index ->
                    if (index == 0 || !candidatesMap.containsKey(word) || !candidatesMap.get(word) || candidatesMap.get(word).size() == 0 || candidatesMap.get(word).size() < score) {
                        candidateWord = word
                        score = candidatesMap.get(word) ? candidatesMap.get(word).size() : 0
                    }
                }
                if (candidateWord) words.add(candidateWord)
            }

            if (words && !words.isEmpty()) {
                words?.each { candidate ->
                    if (candidate.contains(" ")) {
                        candidate = candidate.split("\\s+").toList().get(0)
                    }
                    if (!candidatesMap.containsKey(candidate)) candidatesMap.put(candidate, new ArrayList<String>())
                    if (!candidatesMap.get(candidate).contains(wineId)) {
                        candidatesMap.get(candidate).add(wineId)
                    }
                }
            } else {
                if (pattern) log.warn("No candidate words for: " + wineId)
            }
        }

        int counter = 0
        candidatesMap?.each { k, v ->
            counter += v.size()
        }
        log.info("Total of candidates: " + counter)

        log.info("Mapping keywords done.")
    }

    /**
     * Prepares required data (related to wine alias name) for history and keyword match.<br>
     * Maps wine name aliases, map between alias word and standard word of wine name
     */
    @GrailsCompileStatic(TypeCheckingMode.SKIP)
    public void mapAliases() {
        log.info("Mapping aliases started.")

        HashMap<String, String> alias = new HashMap<String, String>()
        def wordAliases = WordAlias.findAllByAliasType(AliasType.WINE)
        wordAliases.each {
            String replacement = it.word?.toLowerCase()
            it.aliases?.each { String a ->
                if (!alias?.containsKey(a)) {
                    alias.put(a?.toLowerCase(), replacement)
                }
            }
        }

        aliasesMap = new TreeMap<String, String>(reverseLengthComparator)
        aliasesMap.putAll(alias)

        log.info("Mapping aliases done.")
    }

    /**
     * Prepares required data (related to bottle size) for history and keyword match.<br>
     * Maps bottle aliases, map between bottle alias name and standard name<br>
     * List all bottle sizes in correct order. Bottle size with the longer name first<br>
     * List bottle size exclusions. The bottle sizes that skipped when do wine name clean up at {@link KeywordsService#removeSizeInfo(String)}.
     */
    @GrailsCompileStatic(TypeCheckingMode.SKIP)
    public void mapBottleAliases() {
        def bottles = BottleSize.wineDatabase.findAll()
        mapBottleAliases((Set<BottleSize>) bottles)
    }

    /**
     * Prepares required data (related to bottle size) for history and keyword match.<br>
     * Maps bottle aliases, map between bottle alias name and standard name<br>
     * List all bottle sizes in correct order. Bottle size with the longer name first<br>
     * List bottle size exclusions. The bottle sizes that skipped when do wine name clean up at {@link KeywordsService#removeSizeInfo(String)}.
     * @param bottles the bottle sizes
     */
    @GrailsCompileStatic(TypeCheckingMode.SKIP)
    public void mapBottleAliases(Set<BottleSize> bottles) {
        log.info("Mapping bottles started.")
        bottlesMap = new ConcurrentHashMap<String, String>()

        HashMap<String, String> alias = new HashMap<String, String>()
        bottleExclusions.removeAll()
        ArrayList<String> bottle = new ArrayList<String>()

        bottles.each {
            it.aliases?.each { String al ->
                alias.put(al?.toLowerCase(), it.id)
                if (!bottle.contains(al)) {
                    bottle.add(al)
                }
            }
            it.exclusions?.each { String es ->
                if (!bottleExclusions.contains(es)) {
                    bottleExclusions.add(es)
                }
            }
            bottlesMap.put(it.id, it.name)
        }

        bottle = bottle.sort { -it.size() }
        bottleSizes.removeAll()
        bottle.each { bottleSizes.add(it) }

        bottleAliasesMap = new TreeMap<String, String>(reverseLengthComparator)
        bottleAliasesMap.putAll(alias)

        log.info("Mapping bottles completed.")
    }

    /**
     * Prepares required data (related to vintage) for history and keyword match.<br>
     * Maps vintages, map between vintage id and vintage name
     */
    @GrailsCompileStatic(TypeCheckingMode.SKIP)
    public void mapVintages() {
        log.info("Mapping vintages started.")
        vintagesMap = new ConcurrentHashMap<String, String>()
        Vintage.wineDatabase.findAll().each { Vintage vintage ->
            vintagesMap.put(vintage.id, vintage.name)
        }
        log.info("Mapping vintages done.")
    }
}
