package wine.alert

import enums.Status
import grails.transaction.Transactional

/**
 * Service for logs related.
 */
@Transactional
class LogsService {

    /**
     * Saves each crawler logs to crawl_log table {@link CrawlLog}.
     * @param code the retailer code
     * @param url the feed url
     * @param status the crawl status {@link Status}
     * @param attempt the number of attempts
     * @param message the status message
     * @param causedBy the error cause
     */
    void saveLogs(String code, String url, Status status, Integer attempt, String message, String causedBy) {
        try {
            CrawlLog crawlLog = new CrawlLog()
            crawlLog.code = code
            crawlLog.url = url
            crawlLog.status = status
            crawlLog.attempt = attempt
            crawlLog.message = message
            crawlLog.causedBy = causedBy

            crawlLog.validate()
            if (!crawlLog.hasErrors()) {
                crawlLog.save(flush: true)
            }
        } catch (Exception e) {
            log.error(e?.message, e)
        }
    }

    /**
     * Saves crawler job status to job_log table {@link JobLog}.
     * @param crawler the instance of {@link WebCrawler}
     * @param status the crawl status {@link Status}
     * @param message the status message
     * @param totalRecords the total of found rows
     * @param totalPrices the price summary
     * @param totalValidation the total of ran validations
     * @param validationPassed the total of passed validations
     * @param validationStatus the validation status {@link Status}
     * @param validationMessage the validation message
     * @param dateStart the date of the job start
     * @param dateEnd the date of the job end
     * @return the saved instance id
     */
    def saveCrawlerLog(WebCrawler crawler, Status status, String message, Integer totalRecords, BigDecimal totalPrices, Integer totalValidation, Integer validationPassed, Status validationStatus, String validationMessage, Date dateStart, Date dateEnd) {
        saveJobLogs(crawler.code, crawler.name, crawler.url, status, message, totalRecords, totalPrices, totalValidation, validationPassed, validationStatus, validationMessage, dateStart, dateEnd)
    }

    /**
     * Saves crawler job status to job_log table {@link JobLog}.
     * @param code the retailer code
     * @param name the retailer name
     * @param url the feed url
     * @param status the crawl status {@link Status}
     * @param message the status message
     * @param totalRecords the total of found rows
     * @param totalPrices the price summary
     * @param totalValidation the total of ran validations
     * @param validationPassed the total of passed validations
     * @param validationStatus the validation status {@link Status}
     * @param validationMessage the validation message
     * @param dateStart the date of the job start
     * @param dateEnd the date of the job end
     * @return the saved instance id
     */
    def saveJobLogs(String code, String name, String url, Status status, String message, Integer totalRecords, BigDecimal totalPrices, Integer totalValidation,
                    Integer validationPassed, Status validationStatus, String validationMessage, Date dateStart, Date dateEnd) {
        try {
            Date today = new Date().clearTime()
            JobLog jobLog = JobLog.createCriteria().get {
                and {
                    eq("code", code)
                    ge("dateCreated", today)
                }
                maxResults(1)
            }

            if (!jobLog) {
                jobLog = new JobLog()
                jobLog.code = code
                jobLog.name = name
                jobLog.url = url
            }

            jobLog.dateStart = dateStart
            jobLog.status = status
            jobLog.message = message
            jobLog.totalRecords = totalRecords
            jobLog.totalPrices = totalPrices?.setScale(2, BigDecimal.ROUND_HALF_UP)
            jobLog.totalValidation = totalValidation
            jobLog.validationPassed = validationPassed
            jobLog.validationStatus = validationStatus
            jobLog.validationMessage = validationMessage
            jobLog.dateEnd = dateEnd
            jobLog.message = message

            jobLog.validate()
            if (!jobLog.hasErrors()) {
                jobLog.save(flush: true)
                return jobLog.id
            }

        } catch (Exception e) {
            log.error(e?.message, e)
        }

        return null
    }

    /**
     * Saves the wine match log to matched_log {@link MatchedLog}.
     * @param code the retailer code
     * @param crawlDate the input date
     * @param processKeyword the keyword match indicator. Keyword match if true, else history match.
     * @param status the wine match status
     * @param dateStart the date of the job start
     * @return the saved instance id
     */
    public saveMatchLog(String code, Date crawlDate, Boolean processKeyword, Status status, Date dateStart) {
        return saveMatchLog(code, crawlDate, processKeyword, status, "", dateStart, null, null, null, null, null)
    }

    /**
     * Saves the wine match log to matched_log {@link MatchedLog}.
     * @param code the retailer code
     * @param crawlDate the input date
     * @param processKeyword the keyword match indicator. Keyword match if true, else history match.
     * @param status the wine match status
     * @param message the status message
     * @param dateStart the date of the job start
     * @param totalRecords the total rows of the data
     * @param historyMatched the total of history matched
     * @param keywordMatched the total of keyword matched
     * @param ambiguousMatched the total of ambiguous matched
     * @param unknownWines the total of unknown
     * @return the saved instance id
     */
    public saveMatchLog(String code, Date crawlDate, Boolean processKeyword, Status status, String message, Date dateStart,
                        Integer totalRecords, Integer historyMatched, Integer keywordMatched, Integer ambiguousMatched, Integer unknownWines) {

        try {

            Date today = new Date().clearTime()
            MatchedLog matchedLog = MatchedLog.createCriteria().get {
                and {
                    eq("code", code)
                    eq("processKeyword", processKeyword)
                    ge("lastUpdated", today)
                }
                order("lastUpdated", "desc")
                maxResults(1)
            }

            if (!matchedLog) {
                matchedLog = new MatchedLog()
                matchedLog.code = code
            }
            matchedLog.dateStart = dateStart
            matchedLog.crawlDate = crawlDate
            matchedLog.processKeyword = processKeyword
            matchedLog.status = status
            matchedLog.message = message
            matchedLog.dateEnd = new Date()
            matchedLog.totalRecords = totalRecords ?: 0
            matchedLog.historyMatched = historyMatched ?: 0
            matchedLog.keywordMatched = keywordMatched ?: 0
            matchedLog.ambiguousMatched = ambiguousMatched ?: 0
            matchedLog.unknownWines = unknownWines ?: 0

            matchedLog.validate()
            if (!matchedLog.hasErrors()) {
                matchedLog.save(flush: true)
                return matchedLog.id
            }

        } catch (Exception e) {
            log.error(e?.message, e)
        }

        return null
    }
}
