package wine.alert

import enums.*
import grails.compiler.GrailsCompileStatic
import grails.transaction.Transactional
import groovy.time.TimeCategory
import groovy.transform.TypeCheckingMode
import groovyx.gpars.GParsPool
import org.apache.commons.collections.CollectionUtils
import org.apache.commons.lang.StringUtils
import org.springframework.beans.factory.annotation.Value
import wine.alert.model.MatchInput
import wine.alert.model.MatchOutput
import wine.db.WineService
import wine.db.common.SimpleModel

import java.text.SimpleDateFormat
import java.util.concurrent.ConcurrentHashMap
import java.util.concurrent.CopyOnWriteArrayList

/**
 * Used to do matching process between WineAlert's wine and retailer's wine
 */
@Transactional
@GrailsCompileStatic
class MatcherService {

    @Value('${application.output.lineTerminator}')
    private LineTerminator outputLineTerminator

    @Value('${application.output.encoding}')
    private String outputEncoding

    @Value('${application.bottleSize}')
    private String defaultBottleSize

    @Value('${application.matcher.maxThreads}')
    private Integer maxThreads

    private static final String VINTAGE_MISSING = "vintage missing"
    private static final String SIZE_MISSING = "size missing"
    private static final String NOT_FOUND = "notfound"
    private static
    final String DEFAULT_MATCH_TITLE = "Wine Id|Wine Name|Match|Errors|Bottle Size|Vintage|Price|Tax Status|URL|Retailer Description|Keyword String|History String|SKU"

    private List<String> RUNNING_KEYWORD_MATCHES = new CopyOnWriteArrayList<String>()
    private List<String> RUNNING_HISTORY_MATCHES = new CopyOnWriteArrayList<String>()

    KeywordsService keywordsService
    OutputService outputService
    LogsService logsService
    WineService wineService

    /**
     * Adds list of retailer code to the running matches
     * @param retailerCodeList the list of retailer code
     * @param matchedType the enum match type {@link MatchedType}
     */
    public void addToRunningMatches(List<String> retailerCodeList, MatchedType matchedType) {
        retailerCodeList?.each {
            addToRunningMatches(it, matchedType)
        }
    }

    /**
     * Adds retailer code to the current running matches
     * @param retailerCode the retailer code
     * @param matchedType the enum match type {@link MatchedType}
     */
    public void addToRunningMatches(String retailerCode, MatchedType matchedType) {
        if (StringUtils.isNotBlank(retailerCode)) {
            if (matchedType == MatchedType.KEYWORD && !RUNNING_KEYWORD_MATCHES.contains(retailerCode)) {
                RUNNING_KEYWORD_MATCHES.add(retailerCode)
            } else if (matchedType == MatchedType.HISTORY && !RUNNING_HISTORY_MATCHES.contains(retailerCode)) {
                RUNNING_HISTORY_MATCHES.add(retailerCode)
            }
        }
    }

    /**
     * Removes retailer code from the running matches
     * @param retailerCode the retailer code
     * @param matchedType the enum match type {@link MatchedType}
     */
    public void removeFromRunningMatches(String retailerCode, MatchedType matchedType) {
        if (StringUtils.isNotBlank(retailerCode)) {
            if (matchedType == MatchedType.KEYWORD) {
                RUNNING_KEYWORD_MATCHES.remove(retailerCode)
            } else if (matchedType == MatchedType.HISTORY) {
                RUNNING_HISTORY_MATCHES.remove(retailerCode)
            }
        }
    }

    /**
     * Gets list of running matches
     * @param matchedType the enum match type {@link MatchedType}
     * @return the list of retailer code
     */
    List<String> getRunningMatches(MatchedType matchedType) {
        if (matchedType == MatchedType.KEYWORD) {
            return RUNNING_KEYWORD_MATCHES.findAll { it != null }
        } else {
            return RUNNING_HISTORY_MATCHES.findAll { it != null }
        }
    }

    /**
     * Checks existence of retailer code in running matches
     * @param retailerCode the retailer code
     * @param matchedType the enum match type {@link MatchedType}
     * @return true if terminated, or else false
     */
    Boolean isTerminatedMatch(String retailerCode, MatchedType matchedType) {
        Boolean terminated = Boolean.FALSE
        if (StringUtils.isNotBlank(retailerCode)) {
            if (matchedType == MatchedType.KEYWORD) {
                terminated = !RUNNING_KEYWORD_MATCHES.contains(retailerCode)
            } else if (matchedType == MatchedType.HISTORY) {
                terminated = !RUNNING_HISTORY_MATCHES.contains(retailerCode)
            }
        }
        return terminated
    }

    /**
     * Removes all retailer codes from running matches
     * @param matchedType the enum match type {@link MatchedType}
     */
    public void clearRunningMatches(MatchedType matchedType) {
        if (matchedType == MatchedType.KEYWORD) {
            RUNNING_KEYWORD_MATCHES.clear()
        } else {
            RUNNING_HISTORY_MATCHES.clear()
        }
    }

    /**
     * Executes match processing for given retailer code
     * @param retailerCode the retailer code
     * @param runKeyword the option to run keyword match processing
     * @param date the input date (crawler output date)
     */
    public void execute(String retailerCode, Boolean runKeyword, Date date) {
        execute(WebCrawler.findByCode(retailerCode), runKeyword, date)
    }

    /**
     * Executes match processing for given WebCrawler instance
     * @param webCrawler the WebCrawler {@link WebCrawler} instance
     * @param runKeyword the option to run keyword match processing
     * @param date the input date (crawler output date)
     */
    @GrailsCompileStatic(TypeCheckingMode.SKIP)
    public void execute(WebCrawler webCrawler, Boolean runKeyword, Date date) {
        if (!webCrawler) {
            throw new RuntimeException("Retailer not found")
        }

        // Refresh hibernate session
        webCrawler.refresh()

        if (!date) date = new Date()
        Date dateStart = new Date()
        SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd")
        Integer hMatch = 0, kMatch = 0, aMatch = 0, uMatch = 0

        String code = webCrawler.code
        String fileName = code + "-" + sdf.format(date)
        log.info("[" + fileName + "] Matching process started")

        String inputDelimiter = webCrawler?.outputDelimiter?.delimiter() ?: "\\|"   // Crawl output delimiter
        String fileEncoding = this.outputEncoding ?: "UTF-8"
        if (StringUtils.isNotBlank(webCrawler.outputEncoding)) {
            fileEncoding = webCrawler.outputEncoding
        }

        List<String> outputTitles = webCrawler.outputTitle?.tokenize(inputDelimiter) ?: new ArrayList<>()
        if (outputTitles?.size() == 0) {
            def cps = CrawlProperty.createCriteria().list {
                projections {
                    property("wineProperty")
                    property("customLabel")
                }
                and {
                    eq("webCrawler", webCrawler)
                    eq("dataOutput", Boolean.TRUE)
                }
                order("propertyIndex", "asc")
            }
            cps?.each {
                if (it[1]) {
                    outputTitles.add(it[1])
                } else {
                    WineProperty wp = WineProperty.valueOf(it[0] as String)
                    outputTitles.add(wp.value())
                }
            }
        }

        int rawWinesSize = 0
        Status status = Status.RUNNING
        String message = ""
        try {
            List<String> outputList = outputService.loadRetailerOutput(webCrawler, date) as List ?: new ArrayList<String>()
            if (CollectionUtils.isNotEmpty(outputList)) {
                rawWinesSize = outputList.size()
                Map<String, String> historyMatches = outputService.loadHistories(webCrawler)
                ConcurrentHashMap<String, Integer> map = new ConcurrentHashMap<>()
                Set<String> ids = map.newKeySet()

                Integer counter = 0
                int logInterval = Math.floor(rawWinesSize / 2).intValue()
                if (logInterval == 0) logInterval = 1
                def results = outputList.collect { String output ->
                    MatchInput matchInput = parseInput(webCrawler, output)
                    MatchOutput matchOutput = new MatchOutput(matchInput, output, inputDelimiter)
                    String bottleId = matchInput.getBottleSizeId()

                    // HISTORY MATCH
                    String historyId = matchOutput.historyString ? historyMatches.get(matchOutput.historyString) : ""
                    if (StringUtils.isNotBlank(historyId)) {
                        matchOutput.setWineId(historyId)
                        matchOutput.setMatch(MatchedType.HISTORY.value())
                        matchOutput.setWineName("")

                        // Get vintage from id
                        if (StringUtils.isBlank(matchOutput.vintage)) {
                            String vintage = getVintageFromWineId(historyId)
                            matchOutput.setVintage(vintage)
                        }

                        // Get bottle size from id
                        String historyBottleId = getBottleIdFromWineId(historyId) ?: ""
                        if (StringUtils.isNotBlank(historyBottleId)) {
                            bottleId = historyBottleId
                        }
                    } else if (runKeyword) { // KEYWORD MATCH
                        if (StringUtils.isNotBlank(matchOutput.vintage)) {
                            AtomicKwMatch matchResult = calculateMatch(matchOutput.keywordString)
                            if (matchResult.getMatchesSize() == 1) { // single match
                                matchOutput.setMatch(MatchedType.KEYWORD.value())

                                if (StringUtils.isBlank(bottleId)) { // size missing
                                    matchOutput.setErrors(SIZE_MISSING)
                                } else {
                                    String wineDbId = matchResult.getFirstMatch() ?: ""
                                    Boolean inSet = ids.contains(wineDbId)
                                    Boolean exists = inSet || wineService.isWineDbExists(wineDbId)
                                    if (exists) {
                                        if (!inSet) ids.add(wineDbId)
                                        String vintageId = keywordsService.getVintageId(matchOutput.vintage) ?: ""
                                        matchOutput.setWineDbId(wineDbId)
                                        matchOutput.setWineId(wineDbId + vintageId + bottleId)
                                    } else { // wine not found in wineDb
                                        matchOutput.setErrors(NOT_FOUND)
                                        log.warn("WineDB not found: ${wineDbId}")
                                    }
                                }
                            } else if (matchResult.getMatchesSize() > 1) { // ambiguous
                                matchOutput.setMatch(MatchedType.KEYWORD.value())
                                matchOutput.setErrors(MatchedType.AMBIGUOUS.value())
                                matchOutput.setWineId(matchResult.getMatches()?.join(","))
                            } else { // unknown
                                matchOutput.setMatch(MatchedType.UNKNOWN.value())
                                matchOutput.setErrors(MatchedType.UNKNOWN.value())
                            }
                        } else {
                            // vintage missing
                            matchOutput.setMatch(MatchedType.UNKNOWN.value())
                            matchOutput.setErrors(VINTAGE_MISSING)
                        }
                    }

                    if (StringUtils.isNotBlank(bottleId)) {
                        String bottle = keywordsService.getBottleSize(bottleId)
                        if (bottle) matchOutput.setBottleSize(bottle)
                    }

                    counter++
                    if (counter % logInterval == 0) {
                        log.info("[${fileName}] Processed wines: ${counter} of ${rawWinesSize}")
                    }

                    return matchOutput
                }
                log.info("[${fileName}] Processed wines: ${counter} of ${rawWinesSize}")

                String outputDelimiter = Separator.PIPE.value() // Rollup will be always use PIPE as INPUT
                LineTerminator lineTerminator = outputLineTerminator ?: LineTerminator.CRLF
                String newLine = lineTerminator.value()
                StringBuilder stringBuilder = new StringBuilder()
                stringBuilder.append(DEFAULT_MATCH_TITLE.split("\\|").join(outputDelimiter)).append(outputDelimiter)
                stringBuilder.append(outputTitles?.join(outputDelimiter) ?: "").append(newLine)

                setWineNames(results)
                results?.each {
                    if (it.match == MatchedType.HISTORY.value()) {
                        hMatch++
                    } else if (it.match == MatchedType.KEYWORD.value() && !it.errors) {
                        kMatch++
                    } else if (it.errors == MatchedType.AMBIGUOUS.value()) {
                        aMatch++
                    } else {
                        uMatch++
                    }

                    stringBuilder.append(it.getOutput()).append(newLine)
                }
                outputService.saveMatchedToFile(code, fileEncoding, stringBuilder.toString(), date, runKeyword)
            }

            status = Status.SUCCESS
        } catch (Exception e) {
            log.error(e?.message, e)
            status = Status.FAILED
            message = e?.message ?: ""
        }

        MatchedLog.withNewTransaction {
            logsService.saveMatchLog(code, date, runKeyword, status, message, dateStart, rawWinesSize, hMatch, kMatch, aMatch, uMatch)
        }

        String matchName = runKeyword ? "Keyword" : "History"
        log.info("[" + fileName + "] " + matchName + " match done. " +
                "Total data: " + rawWinesSize + ", History: " + hMatch + ", Keywords: " + kMatch + ", Ambiguous: " + aMatch + ", Unknowns: " + uMatch +
                ". Execution times: " + TimeCategory.minus(new Date(), dateStart))
    }

    /**
     * Parse crawler output as wine match input.
     * @param webCrawler the WebCrawler {@link WebCrawler} instance
     * @param outputLine the crawler output (1 row)
     * @return the wine match input {@link MatchInput}
     */
    private MatchInput parseInput(WebCrawler webCrawler, String outputLine) {
        String separator = webCrawler?.outputDelimiter?.delimiter() ?: "\\|"
        List<String> wineInfo = outputLine?.split(separator)?.toList()

        String vintage = webCrawler.vintageIndex != null ? wineInfo?.getAt(webCrawler.vintageIndex)?.trim() : ""
        if (vintage?.equalsIgnoreCase("N.V.")) vintage = "NV"
        vintage = vintage?.length() == 4 || vintage?.equalsIgnoreCase("NV") ? vintage : "" // validate
        String originalSizeInfo = webCrawler.bottleSizeIndex != null ? wineInfo?.getAt(webCrawler.bottleSizeIndex) : ""

        // Set size only if it's valid
        String size = keywordsService.getBottleId(originalSizeInfo) ? originalSizeInfo : ""

        MatchInput matchInput = new MatchInput()
        try {
            String priceString = webCrawler.priceIndex != null ? wineInfo?.getAt(webCrawler.priceIndex) : ""
            BigDecimal price = Calculator.parseMoney(priceString)
            matchInput.setPrice(price)

            String sku = webCrawler.skuIndex != null ? wineInfo?.getAt(webCrawler.skuIndex) : ""
            matchInput.setSku(sku)

            String taxStatus = webCrawler.taxStatusIndex != null ? wineInfo?.getAt(webCrawler.taxStatusIndex) : ""
            matchInput.setTaxStatus(taxStatus)

            String url = webCrawler.urlIndex != null ? wineInfo?.getAt(webCrawler.urlIndex) : ""
            matchInput.setUrl(url)

            // process description
            // add space between prop values
            StringBuilder descBuild = new StringBuilder()
            webCrawler.listDescriptionIndexes?.each { Integer it ->
                String text = it != null && wineInfo?.size() > it ? wineInfo?.get(it) : ""
                if (text?.trim()?.length() > 0) {
                    descBuild.append(text).append(" ")
                }
            }
            matchInput.setDescription(descBuild?.toString()?.trim())

            // history input :: do not run aliases replacement for history match
            // no space between prop values
            StringBuilder hInput = new StringBuilder()
            webCrawler.listHistoryIndexes?.each { Integer it ->
                String text = it != null && wineInfo?.size() > it ? wineInfo?.get(it) : ""
                if (text?.trim()?.length() > 0) {
                    hInput.append(text)
                }
            }
            String historyInput = Calculator.cleanup(hInput?.toString())
            matchInput.setHistoryText(historyInput)
            matchInput.setOriginalHistoryText(hInput?.toString()?.trim())

            // keyword input
            // add space between prop values
            StringBuilder kInput = new StringBuilder()
            webCrawler.listKeywordIndexes?.each { Integer it ->
                String text = it != null && wineInfo?.size() > it ? wineInfo?.get(it) : ""
                if (text?.trim()?.length() > 0) {
                    kInput.append(text).append(" ")
                }
            }
            matchInput.setOriginalKeywordText(kInput?.toString()?.trim())

            String matchesInput = kInput?.toString()
            matchesInput = Calculator.cleanupNumericHtml(matchesInput)
            matchesInput = keywordsService.replaceAliases(matchesInput) // replace with aliases before cleanup

            // Remove bottle size from input keywords (after searching the bottle size and vintage)
            matchesInput = keywordsService.removeSizeInfo(matchesInput)
            //matchesInput = Calculator.removeVintages(matchesInput) // Remove the vintages for now
            matchesInput = Calculator.cleanup(matchesInput)
            matchesInput = keywordsService.replaceAliases(matchesInput) // replace with aliases after cleanup
            matchInput.setKeywordText(matchesInput?.trim())

            // Vintage Parser
            if (StringUtils.isBlank(vintage)) { // search from name/desc/etc
                vintage = Calculator.parseVintage(matchInput.originalKeywordText)
            }

            // Use NV if the config is default to NV
            // This vintage search should be after history match
            if (StringUtils.isBlank(vintage) && webCrawler.vintageDefaultNv) {
                vintage = "NV"
            }

            matchInput.setVintage(vintage ?: "")

            // Bottle Size Parser
            if (StringUtils.isBlank(size)) { // search from name/desc/etc
                size = keywordsService.parseBottleSize(originalSizeInfo)
                if (StringUtils.isBlank(size)) {
                    size = keywordsService.parseBottleSize(matchInput.description)
                }
            }

            // Use default bottle size if bottleSizeDefaultBlank == false
            // This size search should be after history match
            if (StringUtils.isBlank(size) && !webCrawler.bottleSizeDefaultBlank) {
                size = defaultBottleSize
            }

            matchInput.setSize(size ?: "")

            String bottleId = keywordsService.getBottleId(size)
            if (StringUtils.isNotBlank(bottleId)) { // if valid bottle
                matchInput.setBottleSizeId(bottleId)
            }

        } catch (Exception e) {
            log.error(e?.message, e)
        }

        return matchInput
    }

    /**
     * Search the best match of given retailer's wine name
     * @param searchText the wine name
     * @return the keyword match result {@link AtomicKwMatch}
     */
    @GrailsCompileStatic(TypeCheckingMode.SKIP)
    private AtomicKwMatch calculateMatch(String searchText) {
        //Date date = new Date()

        AtomicKwMatch kwMatch = new AtomicKwMatch()
        int threads = maxThreads

        List<String> candidates = keywordsService.getCandidateList(searchText)
        GParsPool.withPool(threads) {
            candidates?.eachParallel { String id ->
                try {
                    List<String> requiredWords = keywordsService.getRequiredWords(id)
                    String pattern = keywordsService.getPattern(id)
                    String normalizedSortedPattern = keywordsService.getNormalizedSortedPattern(id)
                    String normalizedPattern = keywordsService.getNormalizedPattern(id)
                    int score = Calculator.getScore(searchText, pattern, normalizedSortedPattern, normalizedPattern, kwMatch.getScore(), requiredWords)
                    kwMatch.set(id, score)
                } catch (Exception e) {
                    log.error(e?.message, e)
                }
            }
        }
        //long milliseconds = (new Date()).getTime()-date.getTime()

        return kwMatch
    }

    /**
     * Gets vintage information of the given wine id
     * @param wineId the wine id
     * @return the vintage information
     */
    private String getVintageFromWineId(String wineId) {
        if (StringUtils.isBlank(wineId) || wineId?.length() < 14) return null

        String vId = wineId?.substring(9, 12)
        return keywordsService.getVintage(vId)
    }

    /**
     * Gets bottle size id of the given wine id
     * @param wineId the wine id
     * @return the bottle size id
     */
    private static String getBottleIdFromWineId(String wineId) {
        if (StringUtils.isBlank(wineId) || wineId?.length() < 14) return null

        return wineId?.substring(12)
    }

    /**
     * Sets wine wines
     * @param outputs the list of {@link MatchOutput}
     */
    private void setWineNames(List<MatchOutput> outputs) {
        if (CollectionUtils.isEmpty(outputs)) return
        List<MatchOutput> keywords = outputs.findAll { it.match == MatchedType.KEYWORD.value() && !it.errors }
        if (CollectionUtils.isEmpty(keywords)) return
        List<String> ids = keywords.collect { it.wineDbId }
        ids.removeAll { !it }
        if (CollectionUtils.isEmpty(ids)) return
        ids.unique()
        List<SimpleModel> wines = new ArrayList<>()
        ids.collate(50).each { inIds ->
            def imWines = wineService.getWinesWithDescription(inIds)
            if (CollectionUtils.isNotEmpty(imWines)) wines.addAll(imWines)
        }
        if (CollectionUtils.isNotEmpty(wines)) {
            keywords.each { k ->
                SimpleModel wine = wines.find { it.id == k.wineDbId }
                if (wine) k.setWineName(wine.label)
            }
        }
    }
}
