package wine.alert

import enums.*
import grails.transaction.Transactional
import groovy.io.FileType
import groovyx.gpars.GParsPool
import org.apache.commons.lang.StringUtils
import org.springframework.beans.factory.annotation.Value
import utils.CharsetUtils
import wine.alert.model.RollupOutput
import wine.alert.utils.CrawlerUtils
import wine.db.utils.WineUtils

import java.nio.file.Files
import java.nio.file.Paths
import java.nio.file.StandardCopyOption
import java.nio.file.attribute.PosixFilePermission
import java.text.SimpleDateFormat
import java.util.regex.Pattern
import com.amazonaws.services.s3.model.CannedAccessControlList
import com.amazonaws.services.s3.transfer.Upload
import java.util.zip.ZipEntry
import java.util.zip.ZipOutputStream

/**
 * Service for files related operation.<br>
 * Crawler output, history file, history match output, keyword match output, rollup<br>
 * Wine match statistic dump file
 */
@Transactional
class OutputService {

    @Value('${application.crawlOut.location}')
    private String crawlOutLocation

    @Value('${application.crawlOut.backupLocation}')
    private String crawlBackupLocation

    @Value('${application.history.location}')
    private String historyLocation

    @Value('${application.history.backupLocation}')
    private String historyBackupLocation

    @Value('${application.matched.location}')
    private String matchedLocation

    @Value('${application.matched.backupLocation}')
    private String matchedBackupLocation

    @Value('${application.download.location}')
    private String downloadLocation

    @Value('${application.output.lineTerminator}')
    private String outputLineTerminator

    @Value('${application.output.encoding}')
    private String outputEncoding

    @Value('${application.s3bucket.path.rollup}')
    private String rollupPath

    private static String ROLLUP_CHARSET = "UTF-16LE"
    private static int MAX_RECORDS_PER_LOAD = 10000

    Boolean matchDownloadProcessing = false

    private static final Set<PosixFilePermission> FILE_PERMISSIONS = new HashSet<PosixFilePermission>()
    static {
        FILE_PERMISSIONS.add(PosixFilePermission.OWNER_READ)
        FILE_PERMISSIONS.add(PosixFilePermission.OWNER_WRITE)
        FILE_PERMISSIONS.add(PosixFilePermission.OWNER_EXECUTE)
        FILE_PERMISSIONS.add(PosixFilePermission.GROUP_READ)
        FILE_PERMISSIONS.add(PosixFilePermission.GROUP_WRITE)
        FILE_PERMISSIONS.add(PosixFilePermission.GROUP_EXECUTE)
        FILE_PERMISSIONS.add(PosixFilePermission.OTHERS_READ)
    }

    def keywordsService
    def matcherService
    def wineService
    //def amazonS3Service

    /**
     * Saves crawler output as a text file
     * @param webCrawler the WebCrawler instance {@link WebCrawler}
     * @param data the crawler data
     */
    void saveToFile(WebCrawler webCrawler, def data) {

        try {
            def date = new Date()
            def sdf = new SimpleDateFormat("yyyy-MM-dd")
            String fileName = this.crawlOutLocation + File.separator + webCrawler.code + "-" + sdf.format(date) + ".txt"
            String lineTerminator = getOutputLineTerminator()
            String outputDelimiter = webCrawler.outputDelimiter?.value() ?: "|"
            String fileEncoding = outputEncoding
            if (StringUtils.isNotBlank(webCrawler.outputEncoding)) {
                fileEncoding = webCrawler.outputEncoding
            }

            def contentProperties = CrawlProperty.createCriteria().list(sort: "propertyIndex", order: "asc") {
                and {
                    eq("webCrawler", webCrawler)
                    eq("dataOutput", true)
                }
            }

            StringBuilder stringBuilder = new StringBuilder()
            if (webCrawler.outputTitle) {
                stringBuilder.append(webCrawler.outputTitle).append(lineTerminator)
            } else {
                contentProperties.eachWithIndex { CrawlProperty prop, int i ->
                    if (i > 0) stringBuilder.append(outputDelimiter)
                    if (prop.customLabel) {
                        stringBuilder.append(prop.customLabel)
                    } else {
                        stringBuilder.append(prop.wineProperty?.value())
                    }
                }
                stringBuilder.append(lineTerminator)
            }

            data?.each { def dataMap ->

                StringBuilder innerBuilder = new StringBuilder()
                contentProperties.eachWithIndex { CrawlProperty prop, int i ->
                    if (i > 0) innerBuilder.append(outputDelimiter)

                    String indexName = prop.indexName
                    if (dataMap && dataMap?.containsKey(indexName)) {
                        innerBuilder.append(dataMap?.getAt(indexName) ?: "")
                    }
                }

                stringBuilder.append(innerBuilder).append(lineTerminator)
            }

            CrawlerUtils.backupPriorResults(webCrawler?.code)

            File file = new File(fileName)
            file.getParentFile().mkdirs()
            file.createNewFile()
            file.write(stringBuilder.toString(), fileEncoding)

            Files.setPosixFilePermissions(file.toPath(), FILE_PERMISSIONS)

        } catch (Exception e) {
            log.error(e?.message, e)
        }
    }

    /**
     * Checks if crawler output is exists for the given retailer code and given date
     * @param code the retailer code
     * @param date the file's date
     * @return true if file exists, else false
     */
    def isValidCrawlOutput(String code, Date date) {
        if (code && date) {
            try {
                def sdf = new SimpleDateFormat("yyyy-MM-dd")
                String filePath = crawlOutLocation + File.separator + code + "-" + sdf.format(date) + ".txt"
                File file = new File(filePath)
                return file.isFile()
            } catch (Exception e) {
                log.error(e.toString())
            }
        }
        return Boolean.FALSE
    }

    /**
     * Gets crawler output of the given WebCrawler and date
     * @param webCrawler the WebCrawler instance {@link WebCrawler}
     * @param date the file's date
     * @return the list of row of the crawler results
     */
    List<String> loadRetailerOutput(WebCrawler webCrawler, Date date) {
        List<String> outputList = new ArrayList()
        try {
            SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd")
            String fileName = crawlOutLocation + File.separator + webCrawler.code + "-" + sdf.format(date) + ".txt"
            File file = new File(fileName)
            if (file.exists()) {
                log.info("[${webCrawler.code}-${sdf.format(date)}] Loading crawler output...")

                String fileEncoding = CharsetUtils.detect(fileName) ?: "UTF-8"
                if (StringUtils.isNotBlank(webCrawler.outputEncoding)) {
                    fileEncoding = webCrawler.outputEncoding
                }
                String line
                file.withReader(fileEncoding) { reader ->
                    int c = 0
                    while ((line = reader.readLine()) != null) {
                        if (c > 0) { // skip first row
                            if (StringUtils.isNotBlank(line)) {
                                outputList.add(line?.trim())
                            }
                        }
                        c++
                    }
                }
            } else {
                log.warn("[${webCrawler.code}-${sdf.format(date)}] File not found.")
            }
        } catch (Exception e) {
            log.error(e?.message, e)
        }
        return outputList
    }

    /**
     * Gets history data of the given WebCrawler instance.
     * Charset is "Windows-1252" to support Windows encoding
     * @param webCrawler the WebCrawler instance {@link WebCrawler}
     * @return the map between wine id and the validated match (history file)
     */
    Map<String, String> loadHistories(WebCrawler webCrawler) {
        Map<String, String> histories = new HashMap()
        try {
            String historyName = webCrawler.historyFileName
            log.info("[" + webCrawler.code + "] Loading history file [${historyName}] ...")

            String fileName = historyLocation + File.separator + historyName + "-history.txt"
            File file = new File(fileName)
            if (!file.exists()) {
                return histories
            }

            def line
            file.withReader(outputEncoding) { reader ->
                while ((line = reader.readLine()) != null) {
                    String[] h = ((String) line)?.split("\\|\\|")
                    if (h.length > 1) {
                        String id = h[0]
                        String name = Calculator.cleanup(h[1])
                        histories.put(name, id)
                    }
                }
            }
            log.info("[" + webCrawler.code + "] History record(s) loaded for [${historyName}]: " + histories?.size())
        } catch (Exception e) {
            log.error(e?.message, e)
        }
        return histories
    }

    /**
     * Gets all available crawler outputs.
     * @return the list of the file name
     */
    public List<String> getAvailableCrawlerOutput() {
        List<String> files = []
        try {
            new File(crawlOutLocation)?.eachFile() { file ->
                if (file.isFile() && !file.isHidden()) {
                    String fileName = file.getName()
                    if (!fileName.startsWith(".")) {
                        files.add(fileName)
                    }
                }
            }
        } catch (FileNotFoundException e) {
            log.warn(e?.message)
        } catch (Exception e) {
            log.error(e?.message, e)
        }
        return files?.sort()
    }

    /**
     * Gets all available matches.
     * @param matchedType the match type {@link MatchedType}
     * @return the list of the file name
     */
    public List<String> getAvailableMatchedOutput(MatchedType matchedType) {
        List<String> files = []
        String folderName = matchedLocation
        try {
            if (MatchedType.KEYWORD == matchedType) {
                folderName += File.separator + MatchedType.KEYWORD.value() + File.separator
            } else {
                folderName += File.separator + MatchedType.HISTORY.value() + File.separator
            }
            new File(folderName)?.eachFile() { file ->
                if (file.isFile() && !file.isHidden()) {
                    String fileName = file.getName()
                    if (!fileName.startsWith(".")) {
                        files.add(fileName)
                    }
                }
            }
        } catch (FileNotFoundException e) {
            log.warn(e?.message)
        } catch (Exception e) {
            log.error(e?.message, e)
        }
        return files?.sort()
    }

    /**
     * Gets all available matches.
     * @param matchedType the match type {@link MatchedType}
     * @return the list of retailer code
     */
    public List<String> getAvailableMatched(MatchedType matchedType) {
        List<String> codes = []
        getAvailableMatchedOutput(matchedType)?.each {
            String code = it?.substring(0, it.length() - 23)
            if (StringUtils.isNotBlank(code) && !codes.contains(code)) {
                codes.add(code)
            }
        }
        return codes?.sort()
    }

    /**
     * Gets all available matches as map.
     * @param matchedType the match type {@link MatchedType}
     * @return the map of retailer code and the matched date
     */
    public Map<String, Date> getAvailableMatchedMap(MatchedType matchedType) {
        Map<String, Date> matchedMap = new HashMap<>()
        SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd")
        getAvailableMatchedOutput(matchedType)?.each {
            try {
                String code = it?.substring(0, it.length() - 23)
                if (StringUtils.isNotBlank(code) && !matchedMap.containsKey(code)) {
                    String d = it?.replace(code + "-", "")?.replace("-matched.txt", "")
                    if (StringUtils.isNotBlank(d)) {
                        if (d.length() > 10) {
                            d = d?.reverse()?.take(10)?.reverse()
                        }
                        Date date = sdf.parse(d)

                        matchedMap.put(code, date)
                    }
                }
            } catch (Exception e) {
                log.error(e?.message, e)
            }
        }
        return matchedMap
    }

    /**
     * Gets crawler result file's absolute path by the given retailer code.
     * @param retailerCode the retailer code
     * @return The absolute path of latest crawl output
     */
    String getCrawlFileName(String retailerCode) {
        if (StringUtils.isBlank(retailerCode)) return null
        String fileName = ""
        try {
            new File(crawlOutLocation)?.eachFile() { file ->
                if (file.isFile()) {
                    String absolutePath = file.getAbsolutePath()
                    String patternString = ".*/" + retailerCode + "-.*\\.txt"
                    if (absolutePath?.matches(Pattern.compile(patternString))) {
                        if (StringUtils.isBlank(fileName) || absolutePath.compareTo(fileName) > 0) {
                            fileName = absolutePath
                        }
                    }
                }
            }
        } catch (FileNotFoundException e) {
            log.warn(e?.message)
        } catch (Exception e) {
            log.error(e?.message, e)
        }
        return fileName
    }

    /**
     * Gets wine match result file's absolute path by the given retailer code.
     * @param retailerCode the retailer code
     * @param matchedType the match type {@link MatchedType}
     * @return The absolute path of latest match
     */
    String getMatchedFileName(String retailerCode, MatchedType matchedType) {
        if (StringUtils.isBlank(retailerCode)) return null
        String fileName = ""
        String folderName = matchedLocation
        try {
            if (MatchedType.KEYWORD == matchedType) {
                folderName += File.separator + MatchedType.KEYWORD.value() + File.separator
            } else {
                folderName += File.separator + MatchedType.HISTORY.value() + File.separator
            }
            new File(folderName)?.eachFile() { file ->
                if (file.isFile()) {
                    String absolutePath = file.getAbsolutePath()
                    String patternString = ".*/" + retailerCode + "-.*-matched\\.txt"
                    if (absolutePath?.matches(Pattern.compile(patternString))) {
                        if (StringUtils.isBlank(fileName) || absolutePath.compareTo(fileName) > 0) {
                            fileName = absolutePath
                        }
                    }
                }
            }
        } catch (FileNotFoundException e) {
            log.warn(e?.message)
        } catch (Exception e) {
            log.error(e?.message, e)
        }
        return fileName
    }

    /**
     * Gets matches output of the given retailer code and match type.
     * @param retailerCode the retailer code
     * @param matchedType the match type {@link MatchedType}
     * @return the list of row of the match results
     */
    def loadMatchedResults(String retailerCode, MatchedType matchedType) {
        def crawler = WebCrawler.findByCode(retailerCode)
        if (!crawler) return null
        return loadMatchedResults(crawler, matchedType)
    }

    /**
     * Gets matches output of the given WebCrawler instance and match type.
     * @param retailerCode the retailer code
     * @param matchedType the match type {@link MatchedType}
     * @return the list of row of the match results
     */
    def loadMatchedResults(WebCrawler webCrawler, MatchedType matchedType) {
        def outputList = []
        try {
            String fileName = getMatchedFileName(webCrawler.code, matchedType)
            File file = new File(fileName)
            if (file.exists()) {
                if (!file.canRead()) {
                    throw new Exception("No permission to Read the file: " + fileName)
                }

                String fileEncoding = CharsetUtils.detect(fileName) ?: "UTF-8"
                if (StringUtils.isNotBlank(webCrawler.outputEncoding)) {
                    fileEncoding = webCrawler.outputEncoding
                }

                def line
                file.withReader(fileEncoding) { reader ->
                    int c = 0
                    while ((line = reader.readLine()) != null) {
                        if (c > 0) { // skip first row
                            outputList.add(line)
                        }
                        c++
                    }
                }
            } else {
                throw new Exception("File not found: " + fileName)
            }
        } catch (Exception e) {
            log.warn(e.toString())
        }
        return outputList
    }

    /**
     * Gets all keyword matches.
     * @param retailerCode the retailer code
     * @param matchedFilter the match filter, "rp wines" or "non rp wines"
     * @param isNarrow : If true, keyword matches and unknown matched will be loaded. Else, only keyword matches will be loaded
     * @return the offset list and keyword match list
     */
    def getKeywordMatches(String retailerCode, String matchedFilter, Boolean isNarrow, Integer offset) {
        def kwMatched = []
        if (!retailerCode) return kwMatched
        List<Integer> offsets = new ArrayList<>()
        def output = loadMatchedResults(retailerCode, MatchedType.KEYWORD)
        if (output && !output.isEmpty()) {
            if (offset) {
                output = output.drop(offset).take(MAX_RECORDS_PER_LOAD)
            } else {
                int totalRows = output.size()
                if (totalRows > MAX_RECORDS_PER_LOAD) {
                    output = output.take(MAX_RECORDS_PER_LOAD)
                    Integer totalPages = Math.ceil(totalRows / MAX_RECORDS_PER_LOAD).intValue()
                    (1..(totalPages - 1)).each {
                        offsets.push(it * MAX_RECORDS_PER_LOAD)
                    }
                }
            }

            GParsPool.withPool {
                kwMatched = output.collectParallel { String line ->
                    String delimiter = Separator.PIPE.delimiter()
                    List<String> wineInfo = line?.split(delimiter)?.toList()
                    return parseKeywordMatch(retailerCode, matchedFilter, isNarrow, wineInfo)
                }
            }
        }
        if (kwMatched) kwMatched.removeIf { !it }

        return [offsets: offsets, matches: kwMatched]
    }

    /**
     * Parse keyword match.
     * @param retailerCode the retailer code
     * @param matchedFilter the match filter, "rp wines" or "non rp wines"
     * @param isNarrow : If true, keyword matches and unknown matched will be loaded. Else, only keyword matches will be loaded
     * @param wineInfo the wine information
     * @return the map of keyword match properties
     */
    def parseKeywordMatch(String retailerCode, String matchedFilter, Boolean isNarrow, List<String> wineInfo) {
        String matchType = wineInfo?.getAt(2)?.trim()
        String errorString = wineInfo?.getAt(3)?.trim()
        Boolean isValidOutput = isNarrow ? !"history"?.equalsIgnoreCase(matchType) : "keyword"?.equalsIgnoreCase(matchType) && StringUtils.isBlank(errorString)
        if (isValidOutput) {
            def map = [:]
            map.retailerCode = retailerCode

            // Process id
            String wineAlertId = wineInfo?.getAt(0)?.trim()
            map.wineAlertId = wineAlertId
            map.id = wineAlertId != "?" ? wineAlertId?.take(12) : ""
            map.wineDbId = wineAlertId != "?" ? wineAlertId?.take(9) : ""

            //map.wineName = wineInfo?.getAt(1)?.trim()
            map.bottleSize = wineInfo?.getAt(4)?.trim()

            String vintage = wineInfo?.getAt(5)?.trim()
            if (vintage?.equalsIgnoreCase("N.V.")) vintage = "NV"
            map.vintage = vintage

            map.price = wineInfo?.getAt(6)?.trim()
            map.taxStatus = wineInfo?.getAt(7)?.trim()
            map.url = wineInfo?.getAt(8)?.trim()
            map.description = wineInfo?.getAt(9)?.trim()

            map.historyString = wineInfo?.getAt(11)?.trim()
            map.sku = wineInfo?.getAt(12)?.trim()

            def wine = wineService.getWinePropertiesForMatchValidation(map.id, map.wineDbId)

            if ("rp wines".equalsIgnoreCase(matchedFilter) && StringUtils.isBlank((String) wine?.getAt(14))) {
                return null
            }

            if ("non rp wines".equalsIgnoreCase(matchedFilter) && StringUtils.isNotBlank((String) wine?.getAt(14))) {
                return null
            }

            String keywordString = wineInfo?.getAt(10)
            map.originalKeywordString = keywordString

            keywordString = Calculator.cleanupNumericHtml(keywordString)
            keywordString = keywordsService.replaceAliases(keywordString) // replace with aliases before cleanup

            keywordString = keywordsService.removeSizeInfo(keywordString)
            keywordString = Calculator.removeVintages(keywordString, (String) wine?.getAt(12))
            keywordString = Calculator.cleanup(keywordString)
            keywordString = keywordsService.replaceAliases(keywordString) // replace with aliases after cleanup
            map.keywordString = keywordString

            map.maturity = wine?.getAt(13)
            map.mongoId = wine?.getAt(14)

            map.color = wine?.getAt(0)
            map.wineType = wine?.getAt(1)
            map.sweetness = wine?.getAt(2)
            map.variety = wine?.getAt(3)
            map.waNote = wine?.getAt(4)
            map.wineName = WineUtils.getDisplayName((String) wine?.getAt(11), (String) wine?.getAt(5))
            map.country = wine?.getAt(6)
            map.region = wine?.getAt(7)
            map.location = wine?.getAt(8)
            map.locale = wine?.getAt(9)
            map.site = wine?.getAt(10)

            map.pattern = wine?.getAt(12)

            map.wineNoises = []

            return map
        }
    }

    /**
     * Saves wine match results to text file
     * @param code the retailer code
     * @param fileEncoding the output file encoding
     * @param textData the wine match results
     * @param date the match date
     * @param runKeyword : If true, mean keyword match. Else, history match.
     */
    void saveMatchedToFile(String code, String fileEncoding, String textData, Date date, Boolean runKeyword) {
        def sdf = new SimpleDateFormat("yyyy-MM-dd")
        String prefix = MatchedType.HISTORY.value()
        if (runKeyword) prefix = MatchedType.KEYWORD.value()
        try {
            // Backup old matched results
            backupMatchFile(code, runKeyword)

            log.info("[" + code + "-" + sdf.format(date) + "] Saving matched results...")
            String fileName = this.matchedLocation + File.separator + prefix + File.separator + code + "-" + sdf.format(date) + "-matched.txt"
            File file = new File(fileName)
            file.getParentFile().mkdirs()
            file.createNewFile()
            file.write(textData, fileEncoding)

            Files.setPosixFilePermissions(file.toPath(), FILE_PERMISSIONS)
        } catch (Exception e) {
            log.error(e?.message, e)
        }
    }

    /**
     * Backups old wine match result file.
     * @param retailerCode the retailer code
     * @param runKeyword : If true, mean keyword match. Else, history match.
     */
    void backupMatchFile(String retailerCode, Boolean runKeyword) {
        if (StringUtils.isBlank(retailerCode)) return
        try {
            String prefix = MatchedType.HISTORY.value()
            if (runKeyword) {
                prefix = MatchedType.KEYWORD.value()
            }
            String patternString = ".*/" + retailerCode + "-.*-matched\\.txt"
            Pattern p = Pattern.compile(patternString)
            File dir = new File(matchedBackupLocation + File.separator + prefix)
            dir.mkdirs()
            new File(matchedLocation + File.separator + prefix).eachFile() { file ->
                if (file.isFile()) {
                    String fileName = file.getAbsolutePath()
                    if (fileName.find(p)) {
                        String destination = dir.absolutePath + File.separator + file.getName()
                        boolean fileMoved = file.renameTo(new File(destination))
                        if (fileMoved) {
                            Files.setPosixFilePermissions(Paths.get(destination), FILE_PERMISSIONS)
                        } else {
                            log.warn("Failed to move file: " + fileName)
                        }
                    }
                }
            }
        } catch (FileNotFoundException e) {
            log.warn(e?.message)
        } catch (Exception e) {
            log.error(e?.message, e)
        }
    }

    /**
     * Backups old history file.
     * @param historyName the history file name
     */
    void backupHistoryFile(String historyName) {
        if (StringUtils.isBlank(historyName)) return
        File dir = new File(historyBackupLocation)
        dir.mkdirs()

        String fileName = historyLocation + File.separator + historyName + "-history.txt"
        File original = new File(fileName)
        if (!original.exists()) {
            return
        }

        String backupSuffix = "-" + new Date().format('yyyy-MM-dd-HHmmss') + ".txt"
        String newName = original.getName()?.replace(".txt", backupSuffix)
        def target = Paths.get(dir.absolutePath, newName)

        def copy = Files.copy(original.toPath(), target, StandardCopyOption.REPLACE_EXISTING)
        try {
            Files.setPosixFilePermissions(copy, FILE_PERMISSIONS)
        } catch (Exception e) {
            log.warn(e.toString())
        }
    }

    /**
     * Gets the latest crawler output date of given retailer code
     * @param retailerCode the retailer code
     * @return the date of the latest crawler output
     */
    Date getLatestCrawlDate(String retailerCode) {
        if (StringUtils.isBlank(retailerCode)) return null
        SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd")
        try {
            retailerCode = retailerCode?.trim()
            String absolutePath = getCrawlFileName(retailerCode)
            if (StringUtils.isNotBlank(absolutePath)) {
                String fileName = absolutePath.split("/")?.last()
                String d = fileName?.replace(retailerCode + "-", "")?.replace(".txt", "")
                if (StringUtils.isNotBlank(d)) {
                    if (d.length() > 10) {
                        d = d?.reverse()?.take(10)?.reverse()
                    }
                    Date date = sdf.parse(d)
                    return date
                }
            }
        } catch (Exception e) {
            log.error(e?.message, e)
        }
        return null
    }

    /**
     * Gets the latest wine match output file name of the given retailer code.
     * @param retailerCode the retailer code
     * @param matchedType the match type {@link MatchedType}
     * @return the latest wine match output file name
     */
    String findLatestMatchedOutput(String retailerCode, MatchedType matchedType) {
        if (StringUtils.isBlank(retailerCode)) return null

        List<String> filenames = []
        try {
            String patternString = ".*/" + retailerCode + "-.*-matched\\.txt"
            Pattern p = Pattern.compile(patternString)
            new File(matchedLocation + File.separator + matchedType?.value()).eachFile() { file ->
                String fileName = file.getAbsolutePath()
                if (fileName.find(p)) {
                    filenames.add(fileName)
                }
            }
        } catch (Exception e) {
            log.error(e?.message, e)
        }

        if (filenames.size() > 0) {
            // Sort Descending
            filenames.sort { a, b -> b <=> a }
            return filenames.first()
        }
        return null
    }

    /**
     * Cleans old files. Crawler output, history match, keyword match.
     * @param codes the list of retailer code
     * @param matchedType the match type {@link MatchedType}
     */
    void cleanupCrawlOutputAndMatchOutput(List<String> codes, MatchedType matchedType) {
        if (codes && !codes.isEmpty()) {
            def allCrawlerOutFiles = []
            new File(crawlOutLocation).eachFile() { file ->
                if (!file.isHidden() && !file.isDirectory()) {
                    String fileName = file.getName()
                    if (!fileName.startsWith(".")) {
                        allCrawlerOutFiles.add(file)
                    }
                }
            }
            def allMatchedOutFiles = []
            new File(matchedLocation + File.separator + matchedType?.value()).eachFile() { file ->
                if (!file.isHidden() && !file.isDirectory()) {
                    String fileName = file.getName()
                    if (!fileName.startsWith(".")) {
                        allMatchedOutFiles.add(file)
                    }
                }
            }
            GParsPool.withPool {
                codes.eachParallel { String code ->
                    cleanupCrawlOutput(code, allCrawlerOutFiles)
                    cleanupMatchedOutput(code, matchedType, allMatchedOutFiles)
                }
            }
        }
    }

    /**
     * Cleans old crawler output files.
     * @param retailerCode the retailer code
     * @param files the list of files to backup
     */
    public void cleanupCrawlOutput(String retailerCode, List<File> files) {
        if (StringUtils.isBlank(retailerCode)) return

        try {
            List<String> filenames = []
            String patternString = ".*/" + retailerCode + "-.*\\.txt"
            Pattern p = Pattern.compile(patternString)
            files?.each { file ->
                String fileName = file.getAbsolutePath()
                if (fileName.find(p)) {
                    filenames.add(fileName)
                }
            }

            // Move all except the most updated one
            if (filenames.size() > 1) {
                // Sort Descending
                filenames.sort { a, b -> b <=> a }
                filenames.remove(0)

                File dir = new File(crawlBackupLocation)
                filenames?.each {
                    try {
                        File file = new File(it)
                        String destination = dir.absolutePath + File.separator + file.getName()
                        boolean fileMoved = file.renameTo(new File(destination))
                        if (fileMoved) {
                            Files.setPosixFilePermissions(Paths.get(destination), FILE_PERMISSIONS)
                        } else {
                            log.warn("Failed to move file: " + it)
                        }
                    } catch (Exception ex) {
                        log.error(ex?.message, ex)
                    }
                }
            }
        } catch (Exception e) {
            log.error(e?.message, e)
        }
    }

    /**
     * Cleans old wine match output files.
     * @param retailerCode the retailer code
     * @param matchedType the match type {@link MatchedType}
     * @param files the list of files to backup
     */
    public void cleanupMatchedOutput(String retailerCode, MatchedType matchedType, List<File> files) {
        if (StringUtils.isBlank(retailerCode)) return

        try {
            List<String> filenames = []
            String patternString = ".*/" + retailerCode + "-.*-matched\\.txt"
            Pattern p = Pattern.compile(patternString)
            files?.each { file ->
                String fileName = file.getAbsolutePath()
                if (fileName.find(p)) {
                    filenames.add(fileName)
                }
            }

            // Move all except the most updated one
            if (filenames.size() > 1) {
                // Sort Descending
                filenames.sort { a, b -> b <=> a }
                filenames.remove(0)

                File dir = new File(matchedBackupLocation + File.separator + matchedType?.value())
                filenames?.each {
                    try {
                        File file = new File(it)
                        String destination = dir.absolutePath + File.separator + file.getName()
                        boolean fileMoved = file.renameTo(new File(destination))
                        if (fileMoved) {
                            Files.setPosixFilePermissions(Paths.get(destination), FILE_PERMISSIONS)
                        } else {
                            log.warn("Failed to move file: " + it)
                        }
                    } catch (Exception ex) {
                        log.error(ex?.message, ex)
                    }
                }
            }
        } catch (Exception e) {
            log.error(e?.message, e)
        }
    }

    /**
     * Checks if history file writing done.
     * @param historyFileName the history file name
     * @return true if wring done, else false
     */
    Boolean isHistoryFileWritten(String historyFileName) {
        try {
            String fileName = historyLocation + File.separator + historyFileName + "-history.txt"
            File file = new File(fileName)
            if (!file.exists()) {
                return true
            }

            long prevLength = 0
            sleep(300)
            long newLength = file.length()
            int loop = 0
            while (prevLength != newLength || loop < 3) {
                prevLength = newLength
                sleep(250)
                newLength = file.length()
                loop++
            }
            return prevLength == newLength
        } catch (Exception e) {
            log.error(e?.message, e)
        }
        return false
    }

    /**
     * Gets history data as raw text.<br>
     * Charset is "Windows-1252" to support Windows encoding
     * @param historyFileName the history file name
     * @return the list of row of the history file
     */
    def loadRawHistories(String historyFileName) {
        List<String> histories = new ArrayList<>()
        try {
            log.info("Loading history [${historyFileName}] ...")

            String fileName = historyLocation + File.separator + historyFileName + "-history.txt"
            File file = new File(fileName)
            if (!file.exists()) {
                return histories
            }

            String line
            file.withReader(outputEncoding) { reader ->
                while ((line = reader.readLine()) != null) {
                    histories.push(line)
                }
            }
            log.info("[${historyFileName}] histories size: " + histories?.size())
        } catch (Exception e) {
            log.error(e?.message, e)
        }
        return histories
    }

    /**
     * Updated history file with the given data.<br>
     * Output Charset is "Windows-1252" to support Windows encoding
     * @param historyFileName the history file name
     * @param data the data to be written into history file
     * @return the status of operation
     */
    def updateHistoryFile(String historyFileName, List<String> data) {
        if (!data || data.isEmpty()) {
            log.info("[${historyFileName}] New history data is empty.")
            return false
        }
        try {
            backupHistoryFile(historyFileName)

            log.info("Updating history file ${data?.size()} of [${historyFileName}] ...")
            String lineTerminator = getOutputLineTerminator()
            String tmpFileName = this.historyLocation + File.separator + historyFileName + "-history.tmp"
            File tmpFile = new File(tmpFileName)
            tmpFile.getParentFile().mkdirs()
            tmpFile.createNewFile()
            tmpFile.write(data?.join(lineTerminator) + lineTerminator, outputEncoding)

            String fileName = this.historyLocation + File.separator + historyFileName + "-history.txt"
            boolean fileMoved = tmpFile.renameTo(new File(fileName))
            if (!fileMoved) {
                log.warn("Failed to move file: " + fileName)
                return false
            }
            try {
                Files.setPosixFilePermissions(Paths.get(fileName), FILE_PERMISSIONS)
            } catch (Exception ex) {
                log.error(ex?.message, ex)
            }
            log.info("History file [${historyFileName}] updated.")
        } catch (Exception e) {
            log.error(e?.message, e)
            return false
        }
        return true
    }

    @Deprecated
    def getLatestDownloadFile(String fileNamePattern, boolean returnFile) {
        File dumpDir = new File(downloadLocation)
        def files = []
        dumpDir?.eachFile(FileType.FILES) {
            if (it.name?.matches(Pattern.compile(fileNamePattern))) {
                files << it
            }
        }
        if (files.isEmpty()) {
            return returnFile ? null : false
        } else {
            // sort descending by name
            return returnFile ? files?.toSorted { a, b -> b.name <=> a.name }?.first() : true
        }
    }

    /**
     * Save RobertParker Rollup output to text file.<br>
     * Output Charset is "UTF-16LE" for legacy app support
     * @param headerRow the header of the file
     * @param rollupOutputList the data
     * @param separator the data separator
     * @param fileName the file name
     */
    void createRPRollupFile(List<String> headerRow, List<RollupOutput> rollupOutputList, String separator, String fileName) {
        String lineTerminator = getOutputLineTerminator()
        StringBuilder sb = new StringBuilder()
        sb.append(headerRow.join(separator)).append(lineTerminator)
        rollupOutputList.each { RollupOutput output ->
            sb.append(output.toString(separator, RollupType.RP)).append(lineTerminator)
        }

        File file = new File(downloadLocation + File.separator + fileName)
        file.getParentFile().mkdirs()
        file.createNewFile()
        file.write(sb.toString(), ROLLUP_CHARSET)

        Files.setPosixFilePermissions(file.toPath(), FILE_PERMISSIONS)
    }

    /**
     * Save WineAlert Rollup output to text file.<br>
     * Output Charset is "UTF-16LE" for legacy app support
     * @param headerRow the header of the file
     * @param rollupOutputList the data
     * @param separator the data separator
     * @param fileName the file name
     */
    void createWineAlertRollupFile(List<String> headerRow, List<RollupOutput> rollupOutputList, String separator, String fileName) {
        String lineTerminator = getOutputLineTerminator()
        StringBuilder sb = new StringBuilder()
        sb.append(headerRow.join(separator)).append(lineTerminator)
        rollupOutputList.each { RollupOutput output ->
            sb.append(output.toString(separator, RollupType.WA)).append(lineTerminator)
        }

        File file = new File(downloadLocation + File.separator + fileName)
        file.getParentFile().mkdirs()
        file.createNewFile()
        file.write(sb.toString(), ROLLUP_CHARSET)

        Upload upload = amazonS3Service.transferFile(rollupPath + "/" + file.getName(), file, CannedAccessControlList.Private)
        while (!upload.done) {
            sleep(300)
            log.info("${upload.progress.percentTransferred}%")
        }

        String zipFileName = downloadLocation + File.separator + fileName.substring(0, fileName.lastIndexOf(".")) + ".zip"
        FileOutputStream fos = new FileOutputStream(zipFileName);
        ZipOutputStream zipWARollupOut = new ZipOutputStream(fos);
        FileInputStream fis = new FileInputStream(file);
        ZipEntry zipEntry = new ZipEntry(file.getName());
        zipWARollupOut.putNextEntry(zipEntry);

        byte[] bytes = new byte[1024];
        int length;
        while ((length = fis.read(bytes)) >= 0) {
            zipWARollupOut.write(bytes, 0, length);
        }
        zipWARollupOut.close();
        fis.close();
        fos.close();

        File zipedFile = new File(zipFileName)
        Upload uploadZip = amazonS3Service.transferFile(rollupPath + "/" + zipedFile.getName(), zipedFile, CannedAccessControlList.Private)
        while (!uploadZip.done) {
            sleep(300)
            log.info("${uploadZip.progress.percentTransferred}%")
        }
        zipedFile.delete()

        Files.setPosixFilePermissions(file.toPath(), FILE_PERMISSIONS)
    }

    /**
     * Gets rollup file.
     * @param filename the file name
     * @return the rollup File
     */
    def getRollupResultFile(String filename) {
        return new File("${downloadLocation}/${filename}")
    }

    /**
     * Populates wine match related dates of the given webCrawlerList
     * @param IN webCrawlerList
     * @param OUT inputDates
     * @param OUT historyMatchedDates
     * @param OUT keywordMatchedDates
     */
    void populateWineMatchDates(List<WebCrawler> webCrawlerList, Map<Long, Date> inputDates, Map<Long, Date> historyMatchedDates, Map<Long, Date> keywordMatchedDates) {
        SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd")
        sdf.setTimeZone(TimeZone.getTimeZone("GMT"))

        // Check crawler output
        def availableCrawlOutput = this.getAvailableCrawlerOutput()
        def crawlCodes = []
        def crawlDates = [:]
        availableCrawlOutput?.each {
            try {
                String code = it?.substring(0, it.length() - 15)
                if (StringUtils.isNotBlank(code) && !crawlCodes.contains(code)) {
                    crawlCodes.add(code)
                }

                String stringDate = it?.replace(code + "-", "")?.replace(".txt", "")
                Date date = sdf.parse(stringDate)
                crawlDates.put(code, date)
            } catch (Exception e) {
                log.warn(e?.message)
            }
        }

        // Check history output
        def availableHistoryOutput = getAvailableMatchedOutput(MatchedType.HISTORY)
        def historyCodes = []
        def historyDates = [:]
        availableHistoryOutput?.each {
            try {
                String code = it?.substring(0, it.length() - 23)
                if (StringUtils.isNotBlank(code) && !historyCodes.contains(code)) {
                    historyCodes.add(code)
                }

                String stringDate = it?.replace(code + "-", "")?.replace("-matched.txt", "")
                Date date = sdf.parse(stringDate)
                historyDates.put(code, date)
            } catch (Exception e) {
                log.warn(e?.message)
            }
        }

        // Check keyword output
        def availableKeywordOutput = getAvailableMatchedOutput(MatchedType.KEYWORD)
        def keywordCodes = []
        def keywordDates = [:]
        availableKeywordOutput?.each {
            try {
                String code = it?.substring(0, it.length() - 23)
                if (StringUtils.isNotBlank(code) && !keywordCodes.contains(code)) {
                    keywordCodes.add(code)
                }

                String stringDate = it?.replace(code + "-", "")?.replace("-matched.txt", "")
                Date date = sdf.parse(stringDate)
                keywordDates.put(code, date)
            } catch (Exception e) {
                log.warn(e?.message)
            }
        }

        webCrawlerList?.each { WebCrawler w ->
            if (crawlCodes.contains(w.code)) inputDates.put(w.id, (Date) crawlDates.get(w.code))
            if (historyCodes.contains(w.code)) historyMatchedDates.put(w.id, (Date) historyDates.get(w.code))
            if (keywordCodes.contains(w.code)) keywordMatchedDates.put(w.id, (Date) keywordDates.get(w.code))
        }
    }

    /**
     * Exports wine match statistic.<br>
     * Output Charset is "Windows-1252" to support Windows encoding
     * @param matchedType the match type {@link MatchedType}
     * @return the statistic File
     */
    def dumpWineMatch(String matchedType) {
        String separator = Separator.TAB.value()
        Date now = new Date()
        if (matchDownloadProcessing) {
            log.warn("Wine Match is currently processing.")
            return
        }

        try {
            matchDownloadProcessing = true
            log.info('Loading Wine Match from database')
            List<WebCrawler> webCrawlerList = WebCrawler.createCriteria().list {
                ne("status", Status.DELETED)
            }
            int total = webCrawlerList.size()
            log.info("Total record(s): " + total)

            Boolean isKeyword = MatchedType.KEYWORD.name() == matchedType

            // Populate additional data
            Map<Long, Date> inputDates = [:]
            Map<Long, Date> historyMatchedDates = [:]
            Map<Long, Date> keywordMatchedDates = [:]
            populateWineMatchDates(webCrawlerList, inputDates, historyMatchedDates, keywordMatchedDates)

            def runningKeywordMatches = matcherService.getRunningMatches(MatchedType.KEYWORD)
            def runningHistoryMatches = matcherService.getRunningMatches(MatchedType.HISTORY)

            def headerRow = [
                    "Retailer Code",
                    "Input Date",
                    "Matched Date",
                    "Latest Run",
                    "Records",
                    "History",
                    "Keyword",
                    "Ambiguous",
                    "Unknown",
                    "Status"
            ]
            StringBuilder sb = new StringBuilder()
            sb.append(headerRow.join(separator)).append("\n")

            String dateFormat = "MM/dd/yyyy"
            String longDateFormat = "MM/dd/yyyy HH:mm:ss"
            for (int i = 0; i < webCrawlerList.size(); i++) {
                WebCrawler it = webCrawlerList.get(i)
                Date inputDate = inputDates.get(it.id)
                Date keywordMatchedDate = keywordMatchedDates.get(it.id)
                Date historyMatchedDate = historyMatchedDates.get(it.id)

                MatchedLog latestKeywordMatched = it.latestKeywordMatched
                Date latestKeywordDate = latestKeywordMatched?.dateStart

                MatchedLog latestHistoryMatched = it.latestHistoryMatched
                Date latestHistoryDate = latestHistoryMatched?.dateStart

                sb.append(it.code ?: "").append(separator)
                sb.append(inputDate ? inputDate.format(dateFormat) : "").append(separator)

                if (isKeyword) {
                    sb.append(inputDate && keywordMatchedDate ? keywordMatchedDate?.format(dateFormat) : "").append(separator)
                } else {
                    sb.append(inputDate && historyMatchedDate ? historyMatchedDate?.format(dateFormat) : "").append(separator)
                }

                if (isKeyword) {
                    sb.append(inputDate && latestKeywordDate ? latestKeywordDate?.format(longDateFormat) : "").append(separator)
                } else {
                    sb.append(inputDate && latestHistoryDate ? latestHistoryDate?.format(longDateFormat) : "").append(separator)
                }

                sb.append(isKeyword ? (latestKeywordMatched?.totalRecords ?: "") : (latestHistoryMatched?.totalRecords ?: "")).append(separator)

                if (isKeyword) {
                    sb.append(latestKeywordMatched?.historyMatched != null && inputDate ? latestKeywordMatched?.historyMatched : "").append(separator)
                } else {
                    sb.append(latestHistoryMatched?.historyMatched != null && inputDate ? latestHistoryMatched?.historyMatched : "").append(separator)
                }

                sb.append(isKeyword && inputDate ? (latestKeywordMatched?.keywordMatched ?: "") : "").append(separator)
                sb.append(isKeyword && inputDate ? (latestKeywordMatched?.ambiguousMatched ?: "") : "").append(separator)
                sb.append(isKeyword && inputDate ? (latestKeywordMatched?.unknownWines ?: "") : "").append(separator)

                if (isKeyword) {
                    sb.append(runningKeywordMatches?.contains(it.code) ? Status.RUNNING.name() : (inputDate && latestKeywordMatched?.status ? latestKeywordMatched?.status?.name() : "")).append("\n")
                } else {
                    sb.append(runningHistoryMatches?.contains(it.code) ? Status.RUNNING.name() : (inputDate && latestHistoryMatched?.status ? latestHistoryMatched?.status?.name() : "")).append("\n")
                }
            }

            // write file
            String fileName = "WineMatch-" + now.format('yyyy-MM-dd') + ".txt"
            log.info("Creating file " + fileName)
            File file = new File(downloadLocation + File.separator + fileName)
            file.getParentFile().mkdirs()
            file.createNewFile()
            file.write(sb.toString(), outputEncoding)

            Files.setPosixFilePermissions(file.toPath(), FILE_PERMISSIONS)

            log.info("WineMatch is ready to be downloaded at " + file.path)
            matchDownloadProcessing = false

            return file
        } catch (Exception e) {
            log.error(e?.message, e)
            matchDownloadProcessing = false
        }
        return null
    }

    /**
     * Gets file output line terminator.
     * @return the line terminator {@link LineTerminator}
     */
    private String getOutputLineTerminator() {
        LineTerminator terminator = LineTerminator.CRLF
        try {
            terminator = outputLineTerminator as LineTerminator
        } catch (Exception ignored) {
        }
        return terminator?.value()
    }
}
