package wine.alert

import enums.MatchedType
import enums.RollupType
import enums.Separator
import enums.Status
import grails.transaction.Transactional
import groovyx.gpars.GParsPool
import org.apache.commons.lang.StringUtils
import wine.alert.model.RollupOutput
import wine.db.BottleSize
import wine.db.Retailer
import wine.db.Vintage
import wine.db.Wine

import java.math.RoundingMode

/**
 * Service to create Rollup.
 */
@Transactional
class RollupService {

    private static final List<String> RP_HEADER_LIST = [
            "wine_alert_id",
            "wine_id",
            "vintage",
            "bottle_size",
            "retailer",
            "price",
            "currency",
            "tax_notes",
            "wine_url",
            "actual_retail_description",
            "override_price_exception",
            "auction",
            "retailer_name",
            "address",
            "city",
            "state",
            "zip",
            "country",
            "ship_to_country",
            "phone",
            "url",
            "Listing",
            "Download",
            "your_code"
    ]

    private static final List<String> WA_HEADER_LIST = [
            "wine_alert_id",
            "wine_id",
            "vintage",
            "bottle_size",
            "retailer",
            "price",
            "currency",
            "tax_notes",
            "wine_url",
            "actual_retail_description",
            "override_price_exception",
            "auction",
            "retailer_name",
            "address",
            "city",
            "state",
            "zip",
            "country",
            "ship_to_country",
            "phone",
            "url",
            "Listing",
            "Download",
            "your_code",
            "equalized_price"
    ]

    def outputService
    def currencyRateService

    /**
     * Creates RobertParker Rollup file.
     * @throws Exception
     */
    public void executeForRP() throws Exception {
        log.info("RP Rollup started.")

        RollupLog rollupLog = new RollupLog(type: RollupType.RP, status: Status.RUNNING, dateStart: new Date())
        RollupLog.withNewTransaction {
            rollupLog.save(flush: true)
        }

        try {
            def webCrawlerList = outputService.getAvailableMatched(MatchedType.HISTORY)

            // cache vintage and bottle size
            Map<String, Vintage> vintageMap = [:]
            Vintage.wineDatabase.findAll().each { Vintage v ->
                vintageMap.put(v.id, v)
            }
            Map<String, BottleSize> bottleSizeMap = [:]
            BottleSize.wineDatabase.findAll().each { BottleSize b ->
                bottleSizeMap.put(b.id, b)
            }

            Map<String, Retailer> retailersMap = [:]
            Retailer.wineDatabase.findAll().each { Retailer r ->
                r.country?.id
                retailersMap.put(r.code?.toLowerCase(), r)
            }

            Map<String, BigDecimal> currencyRates = currencyRateService.getRates()

            // rollup each retailers
            List<RollupOutput> rollupOutputList = []
            int totalRecords = webCrawlerList.size()
            int processingRecords = 1
            for (String code : webCrawlerList) {

                RollupLog.withNewTransaction {
                    rollupLog.message = "Processing ${processingRecords}/${totalRecords}"
                    rollupLog.save(flush: true)
                }

                WebCrawler crawler = WebCrawler.findByCode(code)
                if (crawler) {
                    // Refresh hibernate session
                    crawler.refresh()

                    String inputDelimiter = Separator.PIPE.delimiter() // Always use PIPE, Matched Output Always use PIPE as Separator
                    Retailer retailer = retailersMap.get(crawler.code?.toLowerCase())

                    log.info("Processing history match RP rollup output of [${crawler.code}]")
                    List<String> historyMatches = outputService.loadMatchedResults(crawler, MatchedType.HISTORY)

                    GParsPool.withPool {
                        List<RollupOutput> results = historyMatches?.collectParallel { String line ->
                            try {
                                line = line?.replaceAll("\\u0000", "")
                                String[] row = line?.split(inputDelimiter)
                                String mongoId = getMongoIdByWineId(row?.getAt(0)?.take(12) as String)
                                return RollupOutput.create(crawler, retailer, row, vintageMap, bottleSizeMap, mongoId)
                            } catch (Exception e) {
                                log.error(e?.message, e)
                            }
                        }

                        // Remove null and duplicates
                        def newList = results?.findAll { it != null }
                        newList?.unique {[it.actualRetailerDescription, it.price]}
                        rollupOutputList.addAll(newList)
                    }
                }
                processingRecords++
            }

            // Equalize all prices
            equalizePrice(rollupOutputList, currencyRates)

            // Filter by median price per wineId
            List<RollupOutput> filteredList = filterByMedianPricePerWineId(rollupOutputList)

            log.info("Saving RP Rollup result to file...")

            Date now = new Date()
            String fileName = "ForSale${now.format('yy.MM.dd.HHmmss')}.txt"

            outputService.createRPRollupFile(RP_HEADER_LIST, filteredList?.sort {it.wineAlertId}, Separator.TAB.value(), fileName)

            rollupLog.filename = fileName
            rollupLog.dateEnd = new Date()
            rollupLog.status = Status.SUCCESS
            rollupLog.message = null

            log.info("RP Rollup completed. File name: ${fileName}")
        } catch (Exception e) {
            log.error(e?.message, e)

            rollupLog.dateEnd = new Date()
            rollupLog.status = Status.FAILED
            rollupLog.message = e?.message
        } finally {
            RollupLog.withNewTransaction {
                rollupLog.save(flush: true)
            }
        }
    }

    /**
     * Creates WineAlert Rollup file.
     * @throws Exception
     */
    public void executeForWineAlert() throws Exception {
        log.info("WineAlert Rollup started.")

        RollupLog rollupLog = new RollupLog(type: RollupType.WA, status: Status.RUNNING, dateStart: new Date())
        RollupLog.withNewTransaction {
            rollupLog.save(flush: true)
        }

        try {
            def webCrawlerList = outputService.getAvailableMatched(MatchedType.HISTORY)

            // cache vintage and bottle size
            Map<String, Vintage> vintageMap = [:]
            Vintage.wineDatabase.findAll().each { Vintage v ->
                vintageMap.put(v.id, v)
            }
            Map<String, BottleSize> bottleSizeMap = [:]
            BottleSize.wineDatabase.findAll().each { BottleSize b ->
                bottleSizeMap.put(b.id, b)
            }

            Map<String, Retailer> retailersMap = [:]
            Retailer.wineDatabase.findAll().each { Retailer r ->
                retailersMap.put(r.code, r)
            }

            Map<String, BigDecimal> currencyRates = currencyRateService.getRates()

            // rollup each retailers
            List<RollupOutput> rollupOutputList = []
            int totalRecords = webCrawlerList.size()
            int processingRecords = 1
            for (String code : webCrawlerList) {

                RollupLog.withNewTransaction {
                    rollupLog.message = "Processing ${processingRecords}/${totalRecords}"
                    rollupLog.save(flush: true)
                }

                WebCrawler crawler = WebCrawler.findByCodeAndListing(code,true)
                if (crawler) {
                    // Refresh hibernate session
                    crawler.refresh()

                    String inputDelimiter = Separator.PIPE.delimiter() // Always use PIPE, Matched Output Always use PIPE as Separator
                    Retailer retailer = retailersMap.get(crawler.code)

                    log.info("Processing history match WineAlert rollup output of [${crawler.code}]")
                    List<String> historyMatches = outputService.loadMatchedResults(crawler, MatchedType.HISTORY)

                    GParsPool.withPool {
                        List<RollupOutput> results = historyMatches?.collectParallel { String line ->
                            try {
                                line = line?.replaceAll("\\u0000", "")
                                String[] row = line?.split(inputDelimiter)
                                String mongoId = getMongoIdByWineId(row?.getAt(0)?.take(12) as String)
                                return RollupOutput.createForWineAlert(crawler, retailer, row, vintageMap, bottleSizeMap, mongoId)
                            } catch (Exception e) {
                                log.error(e?.message, e)
                            }
                        }

                        // Remove null and duplicates
                        def newList = results?.findAll { it != null }
                        newList?.unique {[it.actualRetailerDescription, it.price]}
                        rollupOutputList.addAll(newList)
                    }
                }
                processingRecords++
            }

            // Equalize all prices
            equalizePrice(rollupOutputList, currencyRates)

            // Filter by median price per wineId
            List<RollupOutput> filteredList = filterByMedianPricePerWineId(rollupOutputList)

            log.info("Saving WineAlert Rollup result to file...")

            Date now = new Date()
            String fileName = "WAForSale${now.format('yy.MM.dd.HHmmss')}.txt"

            outputService.createWineAlertRollupFile(WA_HEADER_LIST, filteredList?.sort {it.wineAlertId}, Separator.TAB.value(), fileName)

            rollupLog.filename = fileName
            rollupLog.dateEnd = new Date()
            rollupLog.status = Status.SUCCESS
            rollupLog.message = null

            log.info("WineAlert Rollup completed. File name: ${fileName}")
        } catch (Exception e) {
            log.error(e?.message, e)

            rollupLog.dateEnd = new Date()
            rollupLog.status = Status.FAILED
            rollupLog.message = e?.message
        } finally {
            RollupLog.withNewTransaction {
                rollupLog.save(flush: true)
            }
        }
    }

    /**
     * Filters rollup data by the median of the price.
     * @param rollupOutputList the rollup data to be filtered
     * @return the filtered rollup data
     */
    private List<RollupOutput> filterByMedianPrice(List<RollupOutput> rollupOutputList) {
        List<RollupOutput> filteredList = new ArrayList<>()

        // Ignore records with price less than 5.5 $
        BigDecimal lowFilter = new BigDecimal(5.5)
        List<BigDecimal> priceList = []
        rollupOutputList.each {
            if (it.equalizedPrice != null && it.equalizedPrice >= lowFilter && it.listing) {
                priceList << new BigDecimal(it.equalizedPrice)
            }
        }

        if (priceList.size() > 0) {
            try {
                // Calculate median
                priceList.sort()
                int midNumber = (int)(priceList.size() / 2)
                BigDecimal median = priceList.size() % 2 != 0 ? priceList[midNumber] : (priceList[midNumber] + priceList[midNumber - 1]) / 2

                // Ignore records with price > median * 4
                // Ignore records with price < median / 4
                // IMPORTANT: If the listing is FALSE, include every records (Ignore the price validation)
                BigDecimal maxPrice = median * 4
                BigDecimal minPrice = median / 4
                filteredList = rollupOutputList.findAll {
                    !it.listing || (it.equalizedPrice != null && it.equalizedPrice <= maxPrice && it.equalizedPrice >= minPrice)
                }
            } catch (Exception e) {
                log.warn(e?.message)
            }
        } else {
            filteredList = rollupOutputList.findAll {
                !it.listing || it.equalizedPrice >= lowFilter
            }
        }
        return filteredList
    }

    /**
     * Filters rollup data by the median of the same wine prices.
     * @param rollupOutputList the rollup data to be filtered
     * @return the filtered rollup data
     */
    private List<RollupOutput> filterByMedianPricePerWineId(List<RollupOutput> rollupOutputList) {
        List<RollupOutput> result = new ArrayList<>()

        // Group by wineId
        Map<String, List<RollupOutput>> grouped = rollupOutputList.groupBy { it?.wineId }

        // Filter each group of wineId
        BigDecimal lowFilter = new BigDecimal(5.5)
        grouped.each { wineId, list ->
            List<RollupOutput> filteredList
            if (list.size() == 1) {
                // if only 1 wine in a group, no need to calculate & filter by median
                // Ignore records with price less than 5.5 $
                // IMPORTANT: If the listing is FALSE, include every records (Ignore the price validation)
                filteredList = list.findAll {
                    !it.listing || it.equalizedPrice >= lowFilter
                }
            } else {
                // Filter by median price
                filteredList = filterByMedianPrice(list)
            }

            if (filteredList.size() > 0) {
                result.addAll(filteredList)
            }
        }
        return result
    }

    /**
     * Equalizes prices.
     * @param rollupOutputList the rollup data to be processed
     * @param currencyRates the map of currency rates
     */
    private void equalizePrice(List<RollupOutput> rollupOutputList, Map<String, BigDecimal> currencyRates) {
        // Convert price to USD
        // Convert price to standard 750 ml
        for (RollupOutput output : rollupOutputList) {
            if (output.price != null) {
                try {
                    BigDecimal price = output.price.divide(currencyRates.get(output.currency ?: "USD"), 2, RoundingMode.CEILING)
                    output.equalizedPrice = standardizePrice(output.bottleSize, price)
                } catch (Exception e) {
                    log.warn(e?.message)
                }
            }
        }
    }

    /**
     * Standardizes prices.
     * @param bottleSize the bottle size
     * @param price the price of the wine in the current bottle size
     * @return the standardized price
     */
    private BigDecimal standardizePrice(String bottleSize, BigDecimal price) {
        if (StringUtils.isBlank(bottleSize) || price == null) {
            return null
        }

        try {
            BigDecimal newPrice = new BigDecimal(price)
            def standardSize = 750
            // using reference from http://www.wrathofgrapes.com/bottles.html
            switch (bottleSize.toLowerCase()) {
                case ~/1 l/:
                    newPrice = standardSize / 1000 * price
                    break
                case ~/1.5 l - magnum/:
                    newPrice = standardSize / 1500 * price
                    break
                case ~/10 standard bottles/:
                    newPrice = price / 10
                    break
                case ~/100 ml/:
                    newPrice = standardSize / 100 * price
                    break
                case ~/11 half bottles/:
                    newPrice = standardSize / 375 * price / 11
                    break
                case ~/12 bottle case - 187 ml - splits/:
                    newPrice = standardSize / 187 * price / 12
                    break
                case ~/12 bottle case - 1 l bottles/:
                    newPrice = standardSize / 1000 * price / 12
                    break
                case ~/12 bottle case - 500 ml bottles/:
                    newPrice = standardSize / 500 * price / 12
                    break
                case ~/12 bottle case - 620 ml clavelins \(jura\)/:
                    newPrice = standardSize / 620 * price / 12
                    break
                case ~/12 bottle case - 700 ml bottles/:
                    newPrice = standardSize / 700 * price / 12
                    break
                case ~/12 bottle case - double magnums/:
                    newPrice = standardSize / 3000 * price / 12
                    break
                case ~/12 bottle case - half bottles/:
                    newPrice = standardSize / 375 * price / 12
                    break
                case ~/12 bottle case - magnums/:
                    newPrice = standardSize / 1500 * price / 12
                    break
                case ~/12 bottle case - standard bottles/:
                    newPrice = price / 12
                    break
                case ~/12 bottle case - standard bottles original wood/:
                    newPrice = price / 12
                    break
                case ~/12 l - balthazar/:
                    newPrice = standardSize / 12000 * price
                    break
                case ~/13 bottle case - standard bottles/:
                    newPrice = price / 13
                    break
                case ~/15 bottle case - standard bottles/:
                    newPrice = price / 15
                    break
                case ~/15 l - nebuchadnezzar/:
                    newPrice = standardSize / 15000 * price
                    break
                case ~/16 standard bottles/:
                    newPrice = price / 16
                    break
                case ~/17 l/:
                    newPrice = standardSize / 17000 * price
                    break
                case ~/18 l - melchior/:
                    newPrice = standardSize / 18000 * price
                    break
                case ~/18 standard bottles/:
                    newPrice = price / 18
                    break
                case ~/18.5 l/:
                    newPrice = standardSize / 18500 * price
                    break
                case ~/187 ml - split/:
                    newPrice = standardSize / 187 * price
                    break
                case ~/2 bottle case - half bottles/:
                    newPrice = standardSize / 375 * price / 2
                    break
                case ~/2 bottle case - magnums/:
                    newPrice = standardSize / 1500 * price / 2
                    break
                case ~/2 standard bottles/:
                    newPrice = price / 2
                    break
                case ~/2.1 l - tappit hen/:
                    newPrice = standardSize / 2100 * price
                    break
                case ~/2.25 l - marie-jeanne/:
                    newPrice = standardSize / 2250 * price
                    break
                case ~/2.5 l - marie-jeanne/:
                    newPrice = standardSize / 2500 * price
                    break
                case ~/20 l - solomon/:
                    newPrice = standardSize / 20000 * price
                    break
                case ~/200 ml/:
                    newPrice = standardSize / 200 * price
                    break
                case ~/24 bottle case - half bottles/:
                    newPrice = standardSize / 375 * price / 24
                    break
                case ~/24 bottle case - half bottles original wood/:
                    newPrice = standardSize / 375 * price / 24
                    break
                case ~/24 bottle case - splits/:
                    newPrice = standardSize / 187 * price / 24
                    break
                case ~/24 standard bottles/:
                    newPrice = price / 24
                    break
                case ~/250 ml - one third bottle/:
                    newPrice = standardSize / 250 * price
                    break
                case ~/27 l/:
                    newPrice = standardSize / 27000 * price
                    break
                case ~/3 bottle case - double magnums/:
                    newPrice = standardSize / 3000 * price / 3
                    break
                case ~/3 bottle case - jeroboams/:
                    newPrice = standardSize / 3000 * price / 3
                    break
                case ~/3 bottle case - standard bottles/:
                    newPrice = price / 3
                    break
                case ~/3 bottle case - standard bottles original wood/:
                    newPrice = price / 3
                    break
                case ~/3 half bottles/:
                    newPrice = standardSize / 375 * price / 3
                    break
                case ~/3 l - double magnum/:
                    newPrice = standardSize / 3000 * price
                    break
                case ~/3 l - double magnum - etched/:
                    newPrice = standardSize / 3000 * price
                    break
                case ~/3 l - jeroboam \(champagne\/burgundy\)/:
                    newPrice = standardSize / 3000 * price
                    break
                case ~/3 magnums/:
                    newPrice = standardSize / 1500 * price / 3
                    break
                case ~/3 pack - half bottles/:
                    newPrice = standardSize / 375 * price / 3
                    break
                case ~/3 pack 187 ml - splits/:
                    newPrice = standardSize / 187 * price / 3
                    break
                case ~/30 l/:
                    newPrice = standardSize / 30000 * price
                    break
                case ~/30 standard bottles/:
                    newPrice = price / 30
                    break
                case ~/36 standard bottles/:
                    newPrice = price / 36
                    break
                case ~/375 ml - half bottle/:
                    newPrice = standardSize / 375 * price
                    break
                case ~/4 magnums/:
                    newPrice = standardSize / 1500 * price / 3
                    break
                case ~/4 pack - half bottles/:
                    newPrice = standardSize / 375 * price / 4
                    break
                case ~/4 pack 187 ml - splits/:
                    newPrice = standardSize / 187 * price / 4
                    break
                case ~/4 standard bottles/:
                    newPrice = price / 4
                    break
                case ~/4.5 l - jeroboam/:
                    newPrice = standardSize / 4500 * price
                    break
                case ~/5 l - jeroboam/:
                    newPrice = standardSize / 5000 * price
                    break
                case ~/5 standard bottles/:
                    newPrice = price / 5
                    break
                case ~/5.5 l/:
                    newPrice = standardSize / 5500 * price
                    break
                case ~/50 ml/:
                    newPrice = standardSize / 50 * price
                    break
                case ~/500 ml - two thirds bottle/:
                    newPrice = standardSize / 500 * price
                    break
                case ~/6 bottle case - 500 ml bottles/:
                    newPrice = standardSize / 500 * price / 6
                    break
                case ~/6 bottle case - half bottles/:
                    newPrice = standardSize / 375 * price / 6
                    break
                case ~/6 bottle case - magnums/:
                    newPrice = standardSize / 1500 * price / 6
                    break
                case ~/6 bottle case - magnums original wood/:
                    newPrice = standardSize / 1500 * price / 6
                    break
                case ~/6 bottle case - standard bottles/:
                    newPrice = price / 6
                    break
                case ~/6 bottle case - standard bottles original wood/:
                    newPrice = price / 6
                    break
                case ~/6 l - imperial/:
                    newPrice = standardSize / 6000 * price
                    break
                case ~/6 l - methuselah/:
                    newPrice = standardSize / 6000 * price
                    break
                case ~/620 ml - clavelin \(jura\)/:
                    newPrice = standardSize / 620 * price
                    break
                case ~/7 standard bottles/:
                    newPrice = price / 7
                    break
                case ~/700 ml/:
                    newPrice = standardSize / 700 * price
                    break
                case ~/750 ml - standard bottle/:
                    // do nothing
                    break
                case ~/750 ml - standard bottle w\/screwcap/:
                    // do nothing
                    break
                case ~/8 standard bottles/:
                    newPrice = price / 8
                    break
                case ~/9 l - salmanazar/:
                    newPrice = standardSize / 9000 * price
                    break
                case ~/9 standard bottles/:
                    newPrice = price / 9
                    break
                default:
                    // do nothing
                    break
            }
            /*if (newPrice.compareTo(price) != 0) {
                log.info("${price} ${bottleSize} === ${newPrice} ${standardSize}ml")
            }*/
            return newPrice
        } catch (Exception e) {
            log.warn(e?.message)
        }
        return null
    }

    /**
     * Gets mongo id of the given wine id.
     * @param wineId the wine id
     * @return the mongo id
     */
    private String getMongoIdByWineId(String wineId) {
        if (StringUtils.isBlank(wineId) || wineId?.length() != 12) return null
        String mongoId = Wine.wineDatabase.createCriteria().get {
            projections {
                property("mongoId")
            }
            and {
                eq("id", wineId)
                ne("mongoId", "")
                isNotNull("mongoId")
            }
            maxResults(1)
        }
        return mongoId
    }
}
