package wine.alert

import enums.Status
import grails.transaction.Transactional
import org.springframework.beans.factory.annotation.Value
import wine.alert.model.ValidatorResult
import wine.alert.utils.CrawlDataValidator

/**
 * Service to validate crawler results.<br>
 * Currently used only for indicator.
 */
@Transactional
class ValidatorService {

    @Value('${application.validation.price.active}')
    private Boolean validateTotalPrices

    @Value('${application.validation.price.allowedMargin}')
    private BigDecimal allowedPriceMargin

    /**
     * Validates crawler results.
     * @param webCrawler the WebCrawler instance {@link WebCrawler}
     * @param data the crawler results
     * @param totalPrices the total prices of the crawler results
     * @return the validation result {@link wine.alert.model.ValidatorResult}
     */
    public ValidatorResult validateAll(WebCrawler webCrawler, def data, BigDecimal totalPrices) {
        ValidatorResult validatorResult = new ValidatorResult()
        try {
            // Global Validation :: Total Price Margin
            if (this.validateTotalPrices) {
                def processName = "isValidTotalPrices"
                def globalArgs = [totalPrices?.toString(), webCrawler.getAveragePrices()?.toString(), webCrawler.allowedMargin?.toString() ?: this.allowedPriceMargin?.toString()]
                def g = this.validate(processName, data, globalArgs)
                if (g == Status.SUCCESS) {
                    validatorResult.addPass()
                } else if (g == Status.WARN) {
                    validatorResult.addWarn(processName)
                } else {
                    validatorResult.addFailed(processName)
                }
            }

            // PerRetailer Validation
            def validators = CrawlValidator.createCriteria().list(sort: "processOrder", order: "asc") {
                and {
                    eq("webCrawler", webCrawler)
                    eq("disabled", false)
                }
            }

            validators?.each { CrawlValidator crawlValidator ->
                def args = crawlValidator?.processArgs?.tokenize("\n")
                def v = this.validate(crawlValidator.processName, data, args)
                if (v == Status.SUCCESS) {
                    validatorResult.addPass()
                } else if (v == Status.WARN) {
                    validatorResult.addWarn(crawlValidator.processName)
                } else {
                    validatorResult.addFailed(crawlValidator.processName)
                }
            }
        } catch (Exception e) {
            log.error(e?.message, e)
        }
        return validatorResult
    }

    /**
     * Validates crawler results using the given function.
     * @param functionName the function name from {@link wine.alert.utils.CrawlDataValidator}
     * @param data the crawler results
     * @param args the validation arguments
     * @return the validation status {@link Status}
     */
    public Status validate(String functionName, def data, def args) {
        try {
            def cDV = new CrawlDataValidator()
            return cDV."${functionName}"(data, *args)
        } catch (Exception e) {
            log.error(e?.message, e)
        }
        return Status.FAILED
    }
}
