package wine.alert

import com.opencsv.CSVParser
import com.opencsv.CSVParserBuilder
import enums.*
import grails.converters.JSON
import grails.io.IOUtils
import grails.transaction.Transactional
import groovy.json.JsonOutput
import groovyx.gpars.GParsPool
import io.github.bonigarcia.wdm.ChromeDriverManager
import org.apache.commons.lang.StringUtils
import org.apache.commons.net.ftp.FTPClient
import org.apache.commons.net.ftp.FTPReply
import org.apache.commons.net.util.Base64
import org.apache.http.client.utils.URIBuilder
import org.apache.poi.hssf.usermodel.HSSFWorkbook
import org.apache.poi.ss.usermodel.Cell
import org.apache.poi.ss.usermodel.CellType
import org.apache.poi.ss.usermodel.Hyperlink
import org.apache.poi.ss.usermodel.Row
import org.apache.poi.xssf.usermodel.XSSFWorkbook
import org.grails.web.json.JSONElement
import org.jsoup.Connection
import org.jsoup.Jsoup
import org.jsoup.nodes.Document
import org.jsoup.nodes.Element
import org.jsoup.parser.Parser
import org.jsoup.select.Elements
import org.openqa.selenium.*
import org.openqa.selenium.chrome.ChromeDriver
import org.openqa.selenium.chrome.ChromeOptions
import org.openqa.selenium.support.ui.ExpectedConditions
import org.openqa.selenium.support.ui.WebDriverWait
import wine.alert.model.ValidatorResult
import wine.alert.parser.CsvReader

/**
 * Service to for crawler
 */
@Transactional
class WebCrawlerService {

    private static final String USER_AGENT = "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/74.0.3729.169 Safari/537.36"
    private static final int CONNECTION_TIMEOUT = 10000
    private static final int READ_TIMEOUT = 1200000
    private static final int RETRY_ATTEMPTS = 5
    private static final int MAX_RETAILER_THREADS = 10
    private static final int DOCUMENT_READY_WAIT = 5000 // Wait for append/document ready. Tested using "petersham"
    private static final int DETAILS_DOCUMENT_READY_WAIT = 100

    def outputService
    def validatorService
    def logsService
    def postProcessingService

    /**
     * Crawls the retailer feed.
     * @param code the web_crawler code
     */
    void parse(String code) {
        parse(WebCrawler.findByCodeAndDisabledAndStatusNotEqual(code, Boolean.FALSE, Status.DELETED))
    }

    /**
     * Crawls the retailer feed.
     * @param webCrawler the WebCrawler instance {@link WebCrawler}
     */
    void parse(WebCrawler webCrawler) {
        if (!webCrawler) {
            log.warn("Retailer not found.")
            return
        }

        List<Map<String, Object>> data = new ArrayList<>()
        Date dateStart = new Date()
        Status status = Status.RUNNING
        String jobMessage = ""
        BigDecimal totalPrices = BigDecimal.ZERO
        Status validationStatus = Status.SUCCESS
        String failedProcess = ""
        int numOfGlobalValidators = 1
        Integer passedCount = 0
        Integer validatorSize = CrawlValidator.where {
            webCrawler == webCrawler
        }.count().intValue()

        try {
            if (webCrawler.saveAsIs) {
                data = this.saveAsIs(webCrawler)
            } else {
                switch (webCrawler.pageType) {
                    case PageType.HTML:
                        data = this.parseHtml(webCrawler)
                        break
                    case PageType.HTML_ASYNC:
                        data = this.parseHtmlAsync(webCrawler)
                        break
                    case PageType.SEPARATED_VALUES:
                        data = this.parseSeparatedValues(webCrawler)
                        break
                    case PageType.XML:
                        data = this.parseXml(webCrawler)
                        break
                    case PageType.CSV:
                        data = this.parseCsv(webCrawler)
                        break
                    case PageType.EXCEL: case PageType.EXCEL_2007:
                        data = this.parseExcel(webCrawler)
                        break
                    case PageType.JSON:
                        data = this.parseJson(webCrawler)
                        break
                    case PageType.ENTHUSIAST:
                        data = this.parseEnthusiast(webCrawler)
                        break
                    default:
                        break
                }
            }

            if (data?.isEmpty()) {
                throw new Exception("No record found.")
            }

            // Process post processing
            postProcessingService.process(webCrawler, data)

            // Remove duplicates
            if (webCrawler.removeDuplicate) {
                if (webCrawler.duplicateIndex != null || webCrawler.urlIndex != null || webCrawler.skuIndex != null) {
                    Integer didx = webCrawler.duplicateIndex != null ? webCrawler.duplicateIndex : (webCrawler.urlIndex != null ? webCrawler.urlIndex : webCrawler.skuIndex)
                    CrawlProperty crawlProperty = CrawlProperty.findByWebCrawlerAndPropertyIndex(webCrawler, didx)
                    if (crawlProperty) {
                        data.unique {[it[(crawlProperty.getIndexName())]]}
                    }
                } else {
                    data.unique()
                }
            }

            // Clean up and calculate the total prices
            totalPrices = this.getTotalPrices(webCrawler, data)

            // Process validators
            ValidatorResult v = validatorService.validateAll(webCrawler, data, totalPrices)
            passedCount = v.passedCount ?: 0
            if (v.failedCount > 0) {
                validationStatus = Status.FAILED
                failedProcess = "Failed validations: " + v.failedProcessList?.join(", ")
            }

            // Process output
            outputService.saveToFile(webCrawler, data)

            status = Status.SUCCESS
        } catch (Exception e) {
            if (e?.message == "No record found.") {
                log.warn(e?.message)
            } else {
                log.error(e?.message, e)
            }

            status = Status.FAILED
            jobMessage = e?.message
        }

        try {
            Date dateEnd = new Date()
            logsService.saveJobLogs(webCrawler.code, webCrawler.name, webCrawler.url, status, jobMessage, data?.size(), totalPrices,
                    validatorSize + numOfGlobalValidators, passedCount, validationStatus, failedProcess, dateStart, dateEnd)
        } catch (Exception e) {
            log.error(e?.message, e)
        }
        log.info("[" + webCrawler.code + "] Crawling completed. Found ${data?.size()} record(s)")
    }

    /**
     * Crawls the html page
     * @param webCrawler the WebCrawler instance {@link WebCrawler}
     * @return the crawler results
     */
    List<Map<String, Object>> parseHtml(WebCrawler webCrawler) {
        List<String> urlList = null
        try {
            urlList = getUrls(webCrawler)
        } catch (Exception e) {
            log.error(e?.message, e)
        }
        if (!urlList || urlList?.isEmpty()) return Collections.EMPTY_LIST

        def crawlProperties = CrawlProperty.createCriteria().list {
            and {
                eq("dataOutput", true)
                eq("webCrawler", webCrawler)
            }
            order("propertyIndex", "asc")
        }

        List<Map<String, Object>> resultList = new ArrayList<>()
        Integer numberOfThreads = webCrawler.maxThreads ?: MAX_RETAILER_THREADS
        GParsPool.withPool( numberOfThreads ) {

            resultList = urlList?.collectManyParallel { String uriString ->

                def localResult = []
                if (webCrawler.paginationType == PaginationType.FOLLOW) {
                    String nextUrl = uriString
                    while (nextUrl!=null) {
                        int attempt = 0
                        String prevUrl = nextUrl
                        Status status = null
                        Exception exception = null
                        while (attempt < RETRY_ATTEMPTS) {
                            try {
                                status = Status.SUCCESS
                                Document doc = getDocument(nextUrl, webCrawler)
                                def items = parseDoc(doc, webCrawler, crawlProperties)
                                if (items && !items.isEmpty()) localResult.addAll(items)

                                // Pagination
                                if (webCrawler.paginationSelector) {
                                    Element element = doc?.select(webCrawler.paginationSelector)?.first()
                                    if (element) {
                                        // base url is set by getDocument()
                                        nextUrl = element.absUrl("href")
                                    }
                                }
                            } catch (FileNotFoundException e) {
                                log.error(e?.message)
                                status = Status.FAILED
                                exception = e
                            } catch (Exception e) {
                                log.error(e?.message, e)
                                status = Status.FAILED
                                exception = e
                            }

                            attempt += 1
                            if (status == Status.SUCCESS) break
                            if (webCrawler.threadIntervals != null) Thread.sleep(webCrawler.threadIntervals) // Delay when failed
                        }

                        CrawlLog.withNewTransaction {
                            logsService.saveLogs(webCrawler.code, prevUrl, status, attempt, exception?.message ?: "", exception?.cause?.message ?: "")
                        }

                        if (nextUrl==prevUrl) nextUrl = null
                        if (status == Status.FAILED) break
                        if (webCrawler.threadIntervals != null) Thread.sleep(webCrawler.threadIntervals)
                    }
                } else {
                    int attempt = 0
                    Status status = null
                    Exception exception = null
                    while (attempt < RETRY_ATTEMPTS) {
                        try {
                            status = Status.SUCCESS
                            Document doc = getDocument(uriString, webCrawler)
                            localResult = parseDoc(doc, webCrawler, crawlProperties)
                        } catch (FileNotFoundException e) {
                            log.error(e?.message)
                            status = Status.FAILED
                            exception = e
                        } catch (Exception e) {
                            log.error(e?.message, e)
                            status = Status.FAILED
                            exception = e
                        }

                        attempt += 1
                        if (status == Status.SUCCESS) break
                    }

                    CrawlLog.withNewTransaction {
                        logsService.saveLogs(webCrawler.code, uriString, status, attempt, exception?.message ?: "", exception?.cause?.message ?: "")
                    }

                    if (webCrawler.threadIntervals != null) Thread.sleep(webCrawler.threadIntervals)
                }
                return localResult
            }
        }
        return resultList
    }

    /**
     * Gets HTML element from website
     * @param uri the url of the web page
     * @param webCrawler the web crawler
     * @return the HTML element as Document {@link Document}
     */
    Document getDocument(String uri, WebCrawler webCrawler) {
        String charsetEncoding = webCrawler.charsetEncoding ?: "UTF-8"
        URL url = new URL(uri)
        if (webCrawler.requestType != null && webCrawler.requestType == RequestType.POST) { // POST METHOD
            Map<String, String> formData = splitQueryString(url)
            return Jsoup.connect(uri).data(formData).postDataCharset(charsetEncoding).post()
        } else { // GET METHOD
            return Jsoup.connect(uri).timeout(READ_TIMEOUT).userAgent(USER_AGENT)
                    .ignoreContentType(true).maxBodySize(0) // handle special case
                    .execute()
                    .bufferUp() // https://github.com/jhy/jsoup/issues/1218
                    .charset(charsetEncoding).parse()
        }
    }

    /**
     * Parse document for html type page.
     * @param doc the document to be parsed. Instance of {@link Document}
     * @param webCrawler the WebCrawler instance {@link WebCrawler}
     * @param contentProperties the web crawler properties. List of {@link CrawlProperty}
     * @return List of a Map of Retailer Wine's properties
     */
    def parseDoc(Document doc, WebCrawler webCrawler, def contentProperties) {

        def localResult = []
        if (!doc) return localResult
        String charsetEncoding = webCrawler.charsetEncoding ?: "UTF-8"

        LinkedList<Element> elementList = doc?.select(webCrawler.selector)
        elementList?.each { Element el ->

            def resultMap = [:]

            // Wine Details Page processing
            Element detailsEl = null
            if (webCrawler.digFurther) {
                String detailsUrl
                def detailsUrlElement = el?.select(webCrawler.detailsLinkSelector?.trim())?.first()
                if (webCrawler.detailsLinkPlacement == DataPlacement.PROPERTY) {
                    detailsUrl = detailsUrlElement?.absUrl(webCrawler.detailsLinkPropertyName?.trim())
                } else {
                    detailsUrl = detailsUrlElement?.absUrl("href")
                }

                if (detailsUrl) {
                    detailsEl = this.getDetailsElement(detailsUrl, charsetEncoding, webCrawler.detailsSelectors)
                }
            }

            boolean ignore = false
            for (CrawlProperty prop: contentProperties) {

                Element newEl = prop.useDetails ? detailsEl?.clone() : el?.clone()
                Element contentElement

                // manual separate selectors by comma
                def propList = prop.useDetails ? prop.detailsSelectors?.split(",") : prop.selectors?.split(",")
                for (String newProp: propList) {
                    newProp = newProp?.trim()
                    if (StringUtils.isNotBlank(newProp)) {
                        if (newProp.startsWith("nextSibling")) {
                            // handle pages where properties can't be selected into one element
                            // e.g. https://www.italianwinemerchants.com/Wine-Price-s/58.htm?searching=Y&sort=1&cat=58&show=100&page=1
                            contentElement = el.nextElementSibling().selectFirst(newProp.replaceFirst("nextSibling", ""))
                        } else {
                            contentElement = newEl?.select(newProp)?.first()
                        }

                        if (contentElement) {
                            Element newContentElement = contentElement
                            if (StringUtils.isNotBlank(prop.noises)) {
                                def noiseList = prop.noises?.split(",")
                                noiseList.each { String noise ->
                                    if (newContentElement?.select(noise?.trim()) != null) {
                                        newContentElement?.select(noise?.trim())?.remove()
                                    }
                                }
                            }

                            String contentText
                            if (prop.dataPlacement == DataPlacement.PROPERTY) {
                                contentText = newContentElement?.attr(prop.propertyName)
                            } else if (prop.wineProperty == WineProperty.URL) {
                                contentText = newContentElement?.absUrl("href")
                            } else {
                                contentText = newContentElement?.text()
                            }

                            if (StringUtils.isNotBlank(prop.textToRemove)) {
                                contentText = contentText?.replace(prop.textToRemove, "")
                            }

                            contentText = Calculator.cleanupSpaces(contentText)

                            if (shouldIgnore(contentText, prop.ignoreIf, prop.ignoreValues)) {
                                ignore = true
                                break
                            }
                            resultMap[prop.indexName] = contentText
                            break
                        } else {
                            log.debug("${newProp} not found in html:\n${newEl}")
                        }
                    }
                }

                // Need this for post processing index
                if (prop.indexName && !resultMap.containsKey(prop.indexName)) {
                    resultMap[prop.indexName] = null
                }
            }

            if (!ignore && !resultMap.isEmpty() && (webCrawler.minValuesExist == null || resultMap?.size() >= webCrawler.minValuesExist)) {
                localResult.add(resultMap)
            }
        }
        return localResult
    }

    /**
     * Gets element of the wine detail page based on the given selector.
     * @param detailsUrl the url of the detail page
     * @param charsetEncoding the html page encoding
     * @param selectors the detail page selector
     * @return Element the element in wine page
     */
    private Element getDetailsElement(String detailsUrl, String charsetEncoding, String selectors) {
        if (StringUtils.isBlank(detailsUrl)) return  null

        Element element = null
        try {
            Document doc = Jsoup.connect(detailsUrl).timeout(READ_TIMEOUT).userAgent(USER_AGENT)
                .ignoreContentType(true).maxBodySize(0) // handle special case
                .execute()
                .bufferUp() // https://github.com/jhy/jsoup/issues/1218
                .charset(charsetEncoding).parse()

            def elementList = doc?.select(selectors) as Elements
            element = elementList?.first()
        } catch (Exception e) {
            log.error(e?.message, e)
        }
        return element
    }

    /**
     * Crawls async website using selenium
     * @param webCrawler the WebCrawler instance {@link WebCrawler}
     * @return List of a Map of Retailer Wine's properties
     */
    List<Map<String, Object>> parseHtmlAsync(WebCrawler webCrawler) {
        List<Map<String, Object>> resultList = new ArrayList<>()
        Integer asyncWaitTimeOut = webCrawler.asyncWaitTimeOut ?: 30
        List<String> urlList = getUrls(webCrawler)

        ChromeDriverManager.chromedriver().setup()
        ChromeOptions options = new ChromeOptions()
        options.addArguments("--headless")
        // make window higher, so that element is not blocked and clickable after scroll
        // e.g. abvintners
        options.addArguments("window-size=800,1080")
        WebDriver webDriver = new ChromeDriver(options)
        WebDriver detailsWebDriver = webCrawler.detailsIsAsync ? new ChromeDriver(options) : null // Used for details page
        JavascriptExecutor je = webDriver as JavascriptExecutor

        urlList.each {
            try {
                webDriver.get(it)

                WebDriverWait wait = new WebDriverWait(webDriver, asyncWaitTimeOut)
                JSWaiter.setDriver(webDriver, wait)
                JSWaiter.waitJQueryAngular(true)
                wait.until(ExpectedConditions.presenceOfElementLocated(By.cssSelector(webCrawler.asyncWaitSelector)))
                JSWaiter.sleep(DOCUMENT_READY_WAIT)

                if (webCrawler.usePagination) {
                    WebElement button = webDriver.findElement(By.cssSelector(webCrawler.paginationSelector))
                    if (webCrawler.paginationType == PaginationType.STREAM) {
                        int i = 0
                        while (button != null && i < 2000) { // maximum page 2000
                            try {
                                if (button != null) {
                                    je.executeScript("arguments[0].scrollIntoView();", button)
                                    button.click()
                                    wait.until(ExpectedConditions.presenceOfElementLocated(By.cssSelector(webCrawler.asyncWaitSelector)))
                                    JSWaiter.sleep(DOCUMENT_READY_WAIT)
                                    button = webDriver.findElement(By.cssSelector(webCrawler.paginationSelector))
                                }
                            } catch (Exception ex) {
                                log.info("Reach the end of the page.")
                                break
                            }
                            i++
                        }
                        // Populate elements
                        List<WebElement> webElements = webDriver.findElements(By.cssSelector(webCrawler.selector))
                        resultList.addAll(populateAsyncResults(webElements, webCrawler, detailsWebDriver))
                    } else {
                        List<WebElement> webElements = webDriver.findElements(By.cssSelector(webCrawler.selector))
                        List<Map<String, Object>> result = populateAsyncResults(webElements, webCrawler, detailsWebDriver)
                        if (result && !result.isEmpty()) {
                            resultList.addAll(result)
                        }

                        String prevPage = null
                        String nextPage
                        int i = 0
                        while (button != null && i < 2000) { // maximum page 2000
                            try {
                                if (button != null) {
                                    nextPage = button.getAttribute("href")
                                    if (nextPage) { // Visit the next page
                                        if (nextPage == prevPage) break

                                        prevPage = nextPage
                                        String nextUrl = nextPage
                                        if (StringUtils.isNotBlank(nextUrl) && !nextUrl?.startsWith("http") && StringUtils.isNotBlank(webCrawler.baseUrl)) {
                                            nextUrl = webCrawler.baseUrl?.endsWith("/") ? webCrawler.baseUrl?.reverse()?.drop(1)?.reverse() + nextUrl : webCrawler.baseUrl + nextUrl
                                        }

                                        webDriver.get(nextUrl)
                                        JSWaiter.waitJQueryAngular(false)
                                        wait.until(ExpectedConditions.presenceOfElementLocated(By.cssSelector(webCrawler.asyncWaitSelector)))
                                    } else { // Leave action to the button
                                        je.executeScript("arguments[0].scrollIntoView();", button)
                                        button.click()
                                        wait.until(ExpectedConditions.presenceOfElementLocated(By.cssSelector(webCrawler.asyncWaitSelector)))
                                    }
                                    JSWaiter.sleep(DOCUMENT_READY_WAIT)

                                    List<WebElement> nextWebElements = webDriver.findElements(By.cssSelector(webCrawler.selector))
                                    List<Map<String, Object>> nextResult = populateAsyncResults(nextWebElements, webCrawler, detailsWebDriver)
                                    if (nextResult && !nextResult.isEmpty()) {
                                        resultList.addAll(nextResult)
                                    }

                                    // Refresh button element with the new one in the current page
                                    button = webDriver.findElement(By.cssSelector(webCrawler.paginationSelector))
                                }
                            } catch (Exception ex) {
                                ex.printStackTrace()
                                log.info("Reach the end of the page.")
                                break
                            }
                            i++
                        }
                    }
                } else {
                    List<WebElement> webElements = webDriver.findElements(By.cssSelector(webCrawler.selector))
                    List<Map<String, Object>> result = populateAsyncResults(webElements, webCrawler, detailsWebDriver)
                    if (result && !result.isEmpty()) {
                        resultList.addAll(result)
                    }
                }
            } catch (TimeoutException e) {
                log.error(e?.message)
            } catch (Exception e) {
                log.error(e?.message, e)
            }
        }
        webDriver.quit()
        if (detailsWebDriver) {
            detailsWebDriver.quit()
        }
        return resultList
    }

    /**
     * Populates Retailer's Wines from the given html elements.
     * @param webElements the list of WebElement {@link WebElement} to be parsed
     * @param webCrawler the WebCrawler instance {@link WebCrawler}
     * @param detailsWebDriver the Web Driver for async details page
     * @return List of a Map of Retailer Wine's properties
     */
    List<Map<String, Object>> populateAsyncResults(
            List<WebElement> webElements, WebCrawler webCrawler, WebDriver detailsWebDriver) {
        List<Map<String, Object>> resultList = new ArrayList<>()
        List<CrawlProperty> crawlProperties = CrawlProperty.createCriteria().list {
            and {
                eq("dataOutput", true)
                eq("webCrawler", webCrawler)
            }
            order("propertyIndex", "asc")
        } as List<CrawlProperty>

        webElements?.each { WebElement webElement ->
            try {
                String content = webElement.getAttribute("outerHTML")
                if (content.startsWith('<tr')) {
                    content = '<table>' + content + '</table>'
                }
                if (StringUtils.isNotBlank(content)) {
                    Map<String, Object> result = parseResultFromString(
                            content, webCrawler, crawlProperties, detailsWebDriver)
                    if (result && !result.isEmpty()) {
                        resultList.add(result)
                    }
                }
            } catch (Exception ex) {
                log.error(ex?.message, ex)
            }
        }
        return resultList
    }

    /**
     * Converts HTML string to Wine information.
     * @param content the HTML content
     * @param webCrawler the WebCrawler instance {@link WebCrawler}
     * @param contentProperties the web crawler properties. List of {@link CrawlProperty}
     * @param detailsWebDriver the Web Driver for async details page
     * @return the map of wine's properties
     */
    Map<String, Object> parseResultFromString(
            String content, WebCrawler webCrawler, List<CrawlProperty> contentProperties,
            WebDriver detailsWebDriver) {
        if (StringUtils.isBlank(content)) return null

        Document doc = Jsoup.parse(content, webCrawler.baseUrl ?: "")
        Map<String, Object> resultMap = new HashMap<>()
        String charsetEncoding = webCrawler.charsetEncoding ?: "UTF-8"

        // Wine Details Page processing
        Element detailsEl = null
        if (webCrawler.digFurther) {
            String detailsUrl
            Element detailsUrlElement = doc?.select(webCrawler.detailsLinkSelector?.trim())?.first()
            if (webCrawler.detailsLinkPlacement == DataPlacement.PROPERTY) {
                detailsUrl = detailsUrlElement?.absUrl(webCrawler.detailsLinkPropertyName?.trim())
            } else {
                detailsUrl = detailsUrlElement?.absUrl("href")
            }

            if (detailsUrl) {
                if (webCrawler.detailsIsAsync) {
                    detailsEl = getDetailsElementAsync(
                            detailsUrl, charsetEncoding, webCrawler.detailsSelectors,
                            detailsWebDriver, webCrawler.detailsSelectorToWait, webCrawler.asyncWaitTimeOut)
                } else {
                    detailsEl = getDetailsElement(detailsUrl, charsetEncoding, webCrawler.detailsSelectors)
                }
            }
        }

        boolean ignore = false
        for (CrawlProperty prop: contentProperties) {

            Element newEl = prop.useDetails ? detailsEl?.clone() : doc?.clone()
            Element contentElement

            // manual separate selectors by comma
            def propList = prop.useDetails ? prop.detailsSelectors?.split(",") : prop.selectors?.split(",")
            for (String newProp: propList) {
                newProp = newProp?.trim()
                if (StringUtils.isNotBlank(newProp)) {
                    contentElement = newEl?.select(newProp)?.first()
                    if (contentElement) {
                        Element newContentElement = contentElement
                        if (StringUtils.isNotBlank(prop.noises)) {
                            def noiseList = prop.noises?.split(",")
                            noiseList.each { String noise ->
                                if (newContentElement?.select(noise?.trim()) != null) {
                                    newContentElement?.select(noise?.trim())?.remove()
                                }
                            }
                        }

                        String contentText
                        if (prop.dataPlacement == DataPlacement.PROPERTY) {
                            contentText = newContentElement?.attr(prop.propertyName)
                        } else if (prop.wineProperty == WineProperty.URL) {
                            contentText = newContentElement?.absUrl("href")
                        } else {
                            contentText = newContentElement?.text()
                        }

                        if (StringUtils.isNotBlank(prop.textToRemove)) {
                            contentText = contentText?.replace(prop.textToRemove, "")
                        }

                        contentText = Calculator.cleanupSpaces(contentText)

                        if (shouldIgnore(contentText, prop.ignoreIf, prop.ignoreValues)) {
                            ignore = true
                            break
                        }
                        resultMap[prop.indexName] = contentText
                        break
                    }
                }
            }

            // Need this for post processing index
            if (prop.indexName && !resultMap.containsKey(prop.indexName)) {
                resultMap[prop.indexName] = null
            }
        }

        if (!ignore && !resultMap.isEmpty() && (webCrawler.minValuesExist == null || resultMap?.size() >= webCrawler.minValuesExist)) {
            return resultMap
        }
        return null
    }

    /**
     * Crawls text separated values feed from html page.
     * @param webCrawler the WebCrawler instance {@link WebCrawler}
     * @return List of a Map of Retailer Wine's properties
     */
    List<Map<String, Object>> parseSeparatedValues(WebCrawler webCrawler) {
        List<Map<String, Object>> resultList = new ArrayList<>()
        if (!webCrawler.selector) {
            log.info("Missing selector for [${webCrawler.code}]")
            return resultList
        }

        String charsetEncoding = webCrawler.charsetEncoding ?: "UTF-8"

        List<CrawlProperty> contentProperties = CrawlProperty.createCriteria().list(sort: "propertyIndex", order: "asc") {
            and {
                eq("webCrawler", webCrawler)
                eq("dataOutput", true)
            }
        }

        int attempt = 0
        while( attempt < RETRY_ATTEMPTS) {

            resultList = new ArrayList<>()
            Status status = Status.SUCCESS
            String crawlMessage = ""
            String crawlCauseMessage = ""

            try {
                // Get input stream from the URL
                URL url = new URL(webCrawler.url)
                URLConnection urlConnection = url.openConnection()
                urlConnection.setReadTimeout(READ_TIMEOUT)
                urlConnection.setConnectTimeout(CONNECTION_TIMEOUT)
                urlConnection.setRequestProperty("User-Agent", USER_AGENT)
                InputStream inputStream = urlConnection.inputStream

                // Parse document using input stream and specify the charset
                Document doc = Jsoup.parse(inputStream, charsetEncoding, webCrawler.url)
                Elements elementList = doc?.select(webCrawler.selector)

                // Need this because some documents use html break
                // Replace <br> with \n
                elementList?.select("br")?.append("br2l")
                String content = elementList?.text()?.replaceAll("br2l", "\n")

                Integer count = 0
                Integer rowStar = webCrawler.rowStart != null ? webCrawler.rowStart : 0
                Separator separator = webCrawler.delimiter
                String delimiter = (webCrawler.dataBorder ?: "") + separator?.delimiter()

                content?.splitEachLine(delimiter) { items ->

                    if (count >= rowStar) {
                        // Remove string borders
                        if (webCrawler.dataBorder?.length() > 0) {
                            items?.eachWithIndex { String item, int i ->
                                items[i] = item?.replace(webCrawler.dataBorder, "")
                            }
                        }

                        def lineData = [:]
                        boolean ignore = false
                        for (int i = 0; i < contentProperties.size(); i++) {
                            CrawlProperty prop = contentProperties[i]

                            String indexName = prop.indexName

                            String mapVal = items[prop.propertyIndex]
                            if (mapVal && prop.parseAsHtml) {
                                mapVal = Jsoup.parse("<div>" + mapVal + "</div>", "", Parser.xmlParser()).text()
                            }

                            // Remove last string border
                            if (contentProperties?.size() - 1 == i && contentProperties?.size() > 1 && webCrawler.dataBorder?.length() > 0) {
                                mapVal = mapVal?.reverse()?.replace(webCrawler.dataBorder, "")?.reverse()
                            }

                            mapVal = mapVal?.trim()
                            if (shouldIgnore(mapVal, prop.ignoreIf, prop.ignoreValues)) {
                                ignore = true
                                break
                            }
                            lineData[indexName] = mapVal
                        }

                        if (!ignore && (webCrawler.minValuesExist == null || lineData?.size() >= webCrawler.minValuesExist)) {
                            resultList.add(lineData)
                        }
                    }
                    count += 1
                }

                inputStream.close()
            } catch (Exception e) {
                log.error(e?.message, e)

                status = Status.FAILED
                crawlMessage = e?.message
                crawlCauseMessage = e?.cause?.message
            }

            attempt += 1
            CrawlLog.withNewTransaction {
                logsService.saveLogs(webCrawler.code, webCrawler.url, status, attempt, crawlMessage?:"", crawlCauseMessage?:"")
            }

            if (status == Status.SUCCESS) break
        }
        return resultList
    }

    /**
     * Crawls xml feed.
     * @param webCrawler the WebCrawler instance {@link WebCrawler}
     * @return List of a Map of Retailer Wine's properties
     */
    def parseXml(WebCrawler webCrawler) {
        List<Map<String, Object>> resultList = new ArrayList<>()
        if (!webCrawler.selector) {
            log.info("Missing selector for [${webCrawler.code}]")
            return resultList
        }

        String charsetEncoding = webCrawler.charsetEncoding ?: "UTF-8"

        List<CrawlProperty> contentProperties = CrawlProperty.createCriteria().list(sort: "propertyIndex", order: "asc") {
            and {
                eq("webCrawler", webCrawler)
                eq("dataOutput", true)
            }
        }

        int attempt = 0
        while( attempt < RETRY_ATTEMPTS) {

            resultList = new ArrayList<>()
            Status status = Status.SUCCESS
            String crawlMessage = ""
            String crawlCauseMessage = ""

            try {
                Document doc = Jsoup.connect(webCrawler.url).userAgent(USER_AGENT).timeout(READ_TIMEOUT)
                    .ignoreContentType(true).maxBodySize(0).parser(Parser.xmlParser())
                    .execute().charset(charsetEncoding).parse()

                Elements items = doc?.select(webCrawler.selector)
                if (items != null) {
                    for (Element e : items) {
                        def resultMap = [:]
                        boolean ignore = false
                        for (int i = 0; i < contentProperties.size(); i++) {
                            CrawlProperty prop = contentProperties[i]

                            String indexName = prop.indexName
                            Element propEl = null
                            if (prop.selectors) {
                                propEl = e.getElementsByTag(prop.selectors)?.first()
                            }

                            if (propEl != null) {
                                String contentText
                                if (prop.dataPlacement == DataPlacement.PROPERTY) {
                                    contentText = propEl?.attr(prop.propertyName)
                                } else if (prop.wineProperty == WineProperty.URL) {
                                    contentText = propEl?.absUrl("href")
                                } else {
                                    contentText = propEl?.text()
                                }
                                contentText = contentText?.trim()?.replace("\u000b", "") // Remove Vertical Tab

                                if (StringUtils.isNotBlank(prop.textToRemove)) {
                                    contentText = contentText?.replace(prop.textToRemove, "")
                                }

                                if (shouldIgnore(contentText, prop.ignoreIf, prop.ignoreValues)) {
                                    ignore = true
                                    break
                                }
                                resultMap[indexName] = contentText
                            } else {
                                resultMap[indexName] = null
                            }
                        }

                        if (!ignore && !resultMap?.isEmpty() && (webCrawler.minValuesExist == null || resultMap?.size() >= webCrawler.minValuesExist)) {
                            resultList.add(resultMap)
                        }
                    }
                }
            } catch (Exception e) {
                log.error(e?.message, e)

                status = Status.FAILED
                crawlMessage = e?.message
                crawlCauseMessage = e?.cause?.message
            }

            attempt += 1
            CrawlLog.withNewTransaction {
                logsService.saveLogs(webCrawler.code, webCrawler.url, status, attempt, crawlMessage?:"", crawlCauseMessage?:"")
            }

            if (status == Status.SUCCESS) break
        }
        return resultList
    }

    /**
     * Crawls CSV feed.
     * @param webCrawler the WebCrawler instance {@link WebCrawler}
     * @return List of a Map of Retailer Wine's properties
     */
    def parseCsv(WebCrawler webCrawler) {
        List<Map<String, Object>> resultList = new ArrayList<>()
        String url = webCrawler.url
        int rowStart = webCrawler.rowStart != null ? webCrawler.rowStart : 0

        List<CrawlProperty> crawlProperties = CrawlProperty.createCriteria().list(sort: "propertyIndex", order: "asc") {
            and {
                eq("webCrawler", webCrawler)
                eq("dataOutput", true)
            }
        }

        int attempt = 0
        String content = ""
        while( attempt < RETRY_ATTEMPTS) {

            resultList = new ArrayList<>()
            Status status = Status.SUCCESS
            String crawlMessage = ""
            String crawlCauseMessage = ""

            try {
                if (StringUtils.isBlank(content)) { // Skip download if already success before
                    if (webCrawler.requestType == RequestType.FTP) {
                        content = getFTPContent(url)

                        if (StringUtils.isBlank(content)) {
                            throw new RuntimeException("Empty result from ftp for: " + url)
                        }
                    } else if (webCrawler.asyncDownload) {
                        // Async Type
                        // Get content using selenium
                        content = getAsyncContent(webCrawler)
                        if (StringUtils.isBlank(content)) {
                            throw new RuntimeException("Empty result from: " + url)
                        }
                    } else {
                        def bytes
                        if (url.startsWith('http')) {
                            bytes = Jsoup.connect(url).maxBodySize(0).ignoreContentType(true).execute().bodyAsBytes()
                        } else {
                            def is = new File(url).newInputStream()
                            bytes = is.bytes
                        }
                        if (webCrawler.charsetEncoding) {
                            content = new String(bytes, webCrawler.charsetEncoding)
                        } else {
                            content = new String(bytes)
                        }

                        if (StringUtils.isBlank(content)) {
                            throw new RuntimeException("Empty result from: " + url)
                        }
                    }
                }

                String separator = StringUtils.isNotBlank(webCrawler.delimiter?.value()) ? webCrawler.delimiter?.value() : ','
                CSVParser parser = new CSVParserBuilder()
                        .withSeparator(separator.toCharacter())
                        .withIgnoreQuotations(webCrawler.csvQuoteMode == null)
                        .build()

                CsvReader csvReader = new CsvReader()
                def lines = csvReader.parseText(content, parser)
                int count = 0
                for (String[] record in lines) {
                    if (count >= rowStart) {
                        def resultMap = [:]
                        boolean ignore = false

                        for (CrawlProperty prop: crawlProperties) {
                            String indexName = prop.indexName

                            String contentText = ""
                            if (record?.size() > prop.propertyIndex) {
                                contentText = record?.toList()?.getAt(prop.propertyIndex)
                                if (contentText && prop.parseAsHtml) {
                                    contentText = Jsoup.parse("<div>" + contentText + "</div>", "", Parser.xmlParser()).text()
                                }
                                contentText = contentText?.replace("\"", "") // Remove quote
                            }
                            contentText = contentText?.trim()

                            if (StringUtils.isNotBlank(prop.textToRemove)) {
                                contentText = contentText?.replace(prop.textToRemove, "")
                            }

                            if (shouldIgnore(contentText, prop.ignoreIf, prop.ignoreValues)) {
                                ignore = true
                                break
                            }
                            resultMap[indexName] = contentText
                        }

                        if (!ignore && !resultMap?.isEmpty() && (webCrawler.minValuesExist == null || resultMap?.size() >= webCrawler.minValuesExist)) {
                            resultList.add(resultMap)
                        }
                    }
                    count += 1
                }

            } catch (Exception e) {
                log.error(e?.message, e)

                status = Status.FAILED
                crawlMessage = e?.message
                crawlCauseMessage = e?.cause?.message
            }

            attempt += 1
            CrawlLog.withNewTransaction {
                logsService.saveLogs(webCrawler.code, webCrawler.url, status, attempt, crawlMessage?:"", crawlCauseMessage?:"")
            }

            if (status == Status.SUCCESS) break
        }
        return resultList
    }

    /**
     * Crawls text separated values feed with saveAsIs {@link WebCrawler#saveAsIs} enabled.
     * @param webCrawler the WebCrawler instance {@link WebCrawler}
     * @return List of a Map of Retailer Wine's properties
     */
    def saveAsIs(WebCrawler webCrawler) {
        List<Map<String, Object>> resultList = new ArrayList<>()
        String charsetEncoding = webCrawler.charsetEncoding ?: "UTF-8"
        int rowStart = webCrawler.rowStart ?: 0
        Separator separator = webCrawler.delimiter
        String delimiter = (webCrawler.dataBorder ?: "") + separator?.delimiter()

        List<CrawlProperty> contentProperties = CrawlProperty.createCriteria().list(sort: "propertyIndex", order: "asc") {
            and {
                eq("webCrawler", webCrawler)
                eq("dataOutput", true)
            }
        }

        int attempt = 0
        while( attempt < RETRY_ATTEMPTS) {

            resultList = new ArrayList<>()
            Status status = Status.SUCCESS
            String crawlMessage = ""
            String crawlCauseMessage = ""

            try {
                def response = Jsoup.connect(webCrawler.url).timeout(READ_TIMEOUT).userAgent(USER_AGENT).maxBodySize(0).ignoreContentType(true).execute()
                if (response.statusCode() != 200) {
                    throw new Exception("Connection failed: " + response.statusCode())
                }
                String content = new String(response.bodyAsBytes(), charsetEncoding)

                Integer count = 0
                content?.splitEachLine(delimiter) { items ->

                    if (count >= rowStart) {

                        // Remove string borders
                        if (webCrawler.dataBorder?.length() > 0) {
                            items?.eachWithIndex { String item, int i ->
                                items[i] = item?.replace(webCrawler.dataBorder, "")
                            }
                        }

                        def lineData = [:]
                        boolean ignore = false
                        for (int i = 0; i < contentProperties.size(); i++) {
                            CrawlProperty prop = contentProperties[i]
                            String indexName = prop.indexName

                            String mapVal = items[prop.propertyIndex]
                            if (mapVal && prop.parseAsHtml) {
                                mapVal = Jsoup.parse("<div>" + mapVal + "</div>", "", Parser.xmlParser()).text()
                            }

                            // Remove last string border
                            if (contentProperties?.size() - 1 == i && contentProperties?.size() > 1 && webCrawler.dataBorder?.length() > 0) {
                                mapVal = mapVal?.reverse()?.replace(webCrawler.dataBorder, "")?.reverse()
                            }

                            if (StringUtils.isNotBlank(prop.textToRemove)) {
                                mapVal = mapVal?.replace(prop.textToRemove, "")
                            }

                            mapVal = mapVal?.trim()
                            if (shouldIgnore(mapVal, prop.ignoreIf, prop.ignoreValues)) {
                                ignore = true
                                break
                            }
                            lineData[indexName] = mapVal
                        }

                        if (!ignore && (webCrawler.minValuesExist == null || lineData?.size() >= webCrawler.minValuesExist)) {
                            resultList.add(lineData)
                        }
                    }
                    count += 1
                }

            } catch (Exception e) {
                log.error(e?.message, e)

                status = Status.FAILED
                crawlMessage = e?.message
                crawlCauseMessage = e?.cause?.message
            }

            attempt += 1
            CrawlLog.withNewTransaction {
                logsService.saveLogs(webCrawler.code, webCrawler.url, status, attempt, crawlMessage?:"", crawlCauseMessage?:"")
            }

            if (status == Status.SUCCESS) break
        }
        return resultList
    }

    /**
     * Crawls Spreadsheet feed.
     * @param webCrawler the WebCrawler instance {@link WebCrawler}
     * @return List of a Map of Retailer Wine's properties
     */
    def parseExcel(WebCrawler webCrawler) {
        List<Map<String, Object>> resultList = new ArrayList<>()
        String charsetEncoding = webCrawler.charsetEncoding ?: "UTF-8"
        int rowStart = webCrawler.rowStart ?: 0

        List<CrawlProperty> contentProperties = CrawlProperty.createCriteria().list(sort: "propertyIndex", order: "asc") {
            and {
                eq("webCrawler", webCrawler)
                eq("dataOutput", true)
            }
        }

        int attempt = 0
        while( attempt < RETRY_ATTEMPTS) {

            resultList = new ArrayList<>()
            Status status = Status.SUCCESS
            String crawlMessage = ""
            String crawlCauseMessage = ""

            try {
                // Get input stream from the URL
                URL url = new URL(webCrawler.url)
                URLConnection urlConnection = url.openConnection()
                urlConnection.setReadTimeout(READ_TIMEOUT)
                urlConnection.setConnectTimeout(CONNECTION_TIMEOUT)
                urlConnection.setRequestProperty("User-Agent", USER_AGENT)
                InputStream inputStream = urlConnection.inputStream

                // Get the workbook instance for EXCEL file
                def workbook
                if (webCrawler.pageType == PageType.EXCEL) {
                    workbook = new HSSFWorkbook(inputStream)
                } else {
                    workbook = new XSSFWorkbook(inputStream)
                }

                // Get first sheet from the workbook
                def sheet = workbook.getSheetAt(0)

                // Iterate through each rows from first sheet
                Iterator<Row> rowIterator = sheet.iterator()

                Integer count = 0
                while(rowIterator.hasNext()) {

                    Row row = rowIterator.next()
                    if (count >= rowStart) {
                        def resultMap = [:]
                        boolean ignore = false
                        for (int i = 0; i < contentProperties.size(); i++) {
                            CrawlProperty prop = contentProperties[i]

                            String indexName = prop.indexName
                            resultMap[indexName] = null

                            try {
                                Cell cell = row.getCell(prop.propertyIndex, Row.MissingCellPolicy.CREATE_NULL_AS_BLANK)
                                if (cell) {
                                    switch(cell.getCellTypeEnum()) { // TODO: Change this when updated to version 4.0 of POI
                                        case CellType.BOOLEAN:
                                            resultMap[indexName] = cell?.getBooleanCellValue()
                                            break
                                        case CellType.NUMERIC:
                                            if (prop.dataType == DataType.STRING) {
                                                // WARN: DataFormatter will not work in some cases
                                                cell.setCellType(CellType.STRING)
                                                resultMap[indexName] = cell?.getStringCellValue()
                                            } else if (prop.dataType == DataType.INTEGER) {
                                                Double val = cell?.getNumericCellValue()
                                                resultMap[indexName] = val?.toInteger()
                                            } else {
                                                resultMap[indexName] = cell?.getNumericCellValue()
                                            }
                                            break
                                        case CellType.STRING:
                                            if (prop.wineProperty == WineProperty.URL && prop.dataPlacement == DataPlacement.HYPERLINK) {
                                                Hyperlink hyperlink = cell.getHyperlink()
                                                resultMap[indexName] = hyperlink?.getAddress() ?: hyperlink?.getLabel()
                                            } else {
                                                resultMap[indexName] = cell?.getStringCellValue()
                                            }
                                            break
                                    }
                                }
                            } catch (Exception ex) {
                                log.error(ex?.message, ex)
                            }

                            String mapVal = resultMap[indexName]?.toString()
                            if (shouldIgnore(mapVal, prop.ignoreIf, prop.ignoreValues)) {
                                ignore = true
                                break
                            }
                        }

                        if (!ignore && !resultMap?.isEmpty() && (webCrawler.minValuesExist == null || resultMap?.size() >= webCrawler.minValuesExist)) {
                            resultList.add(resultMap)
                        }
                    }
                    count += 1
                }

                inputStream.close()
            } catch (Exception e) {
                log.error(e?.message, e)

                status = Status.FAILED
                crawlMessage = e?.message
                crawlCauseMessage = e?.cause?.message
            }

            attempt += 1
            CrawlLog.withNewTransaction {
                logsService.saveLogs(webCrawler.code, webCrawler.url, status, attempt, crawlMessage?:"", crawlCauseMessage?:"")
            }

            if (status == Status.SUCCESS) break
        }
        return resultList
    }

    /**
     * Crawls JSON feed.
     * @param webCrawler the WebCrawler instance {@link WebCrawler}
     * @return List of a Map of Retailer Wine's properties
     */
    def parseJson(WebCrawler webCrawler) {
        List<Map<String, Object>> resultList = new ArrayList<>()
        //String charsetEncoding = webCrawler.charsetEncoding ?: "UTF-8"

        List<CrawlProperty> contentProperties = CrawlProperty.createCriteria().list(sort: "propertyIndex", order: "asc") {
            and {
                eq("webCrawler", webCrawler)
                eq("dataOutput", true)
            }
        }

        List<String> urls = getUrls(webCrawler)
        urls?.each { url ->
            try {
                JSONElement json
                if (webCrawler.authType == AuthType.NTLM) {
                    String content = getContentWithNTLM(webCrawler, url)
                    if (StringUtils.isNotBlank(content)) {
                        try {
                            json = JSON.parse(content)
                        } catch(Exception e) {
                            log.error(e?.message, e)
                        }
                    }
                } else {
                    json = getJSONContent(webCrawler, url)
                }

                if (json) {
                    def list = json
                    if (StringUtils.isNotBlank(webCrawler.selector)) {
                        if (webCrawler.selector?.contains(".")) {
                            webCrawler.selector?.split("\\.")?.each {
                                list = list."${it}"
                            }
                        } else {
                            list =  json."$webCrawler.selector"
                        }
                    }
                    if (list instanceof List) {
                        list.each { item ->
                            if (item) {
                                def lineData = [:]
                                boolean ignore = false
                                for (int i = 0; i < contentProperties.size(); i++) {
                                    CrawlProperty prop = contentProperties[i]
                                    String indexName = prop.indexName

                                    String mapVal
                                    if (prop.selectors?.contains(".")) {
                                        def innerItem = item
                                        prop.selectors?.split("\\.")?.each {
                                            if (innerItem != null) {
                                                innerItem = innerItem."${it}"
                                            }
                                            if (innerItem && innerItem instanceof List) {
                                                innerItem = innerItem.getAt(0)
                                            }
                                        }
                                        mapVal = innerItem ? innerItem.toString() : ''
                                    } else {
                                        mapVal = item."$prop.selectors" != null ? item."$prop.selectors" : ""
                                    }

                                    if (StringUtils.isNotBlank(prop.textToRemove)) {
                                        mapVal = mapVal?.replace(prop.textToRemove, "")
                                    }

                                    mapVal = mapVal?.trim()
                                    if (shouldIgnore(mapVal, prop.ignoreIf, prop.ignoreValues)) {
                                        ignore = true
                                        break
                                    }
                                    lineData[indexName] = mapVal
                                }

                                if (!ignore && (webCrawler.minValuesExist == null || lineData?.size() >= webCrawler.minValuesExist)) {
                                    resultList.add(lineData)
                                }
                            }
                        }
                    }
                }
            } catch (Exception e) {
                log.error(e?.message, e)
            }
        }
        return resultList
    }

    def parseEnthusiast(WebCrawler webCrawler) {
        List<Map<String, Object>> resultList = new ArrayList<>()
        String charset = webCrawler.charsetEncoding ?: 'UTF-8'
        List<CrawlProperty> contentProperties = getCrawlProperties(webCrawler)

        URL url = new URL(webCrawler.url)
        String[] parts = url.query?.split("&")
        Map payload = new HashMap()
        for (String part : parts) {
            String[] keyVal = part.split("=")
            payload.put(keyVal[0], URLDecoder.decode(keyVal[1], 'UTF-8'))
        }

        URIBuilder builder = new URIBuilder(webCrawler.url)
        builder.removeQuery()
        String idSource = payload.get('IDSource')
        payload.remove('IDSource')
        try {
            List<Long> ids = new ArrayList<>()
            if (idSource) {
                File file = new File(idSource)
                String fileContent = file.text
                if (StringUtils.isNotBlank(fileContent)) {
                    ids.addAll(fileContent.tokenize(',').collect { it.toLong() })
                }
            } else {
                Connection conn = Jsoup.connect(builder.toString())
                        .header("Content-Type", "application/json")
                        .header("Accept", "application/json")
                        .followRedirects(true)
                        .ignoreHttpErrors(true)
                        .ignoreContentType(true)
                        .userAgent(USER_AGENT)
                        .method(Connection.Method.POST)
                        .requestBody(JsonOutput.toJson(payload).toString())
                        .timeout(CONNECTION_TIMEOUT)
                        .maxBodySize(0)
                JSONElement json = JSON.parse(conn.execute().charset(charset).body())
                if (json instanceof List) {
                    ids.addAll(json.collect { it as Long })
                }
            }

            if (ids) {
                payload.remove('DTG')
                payload.remove('Func')
                payload.put('Func', 'GetWineData')
                ids.collate(50)?.each { list ->
                    payload.remove('IDs')
                    payload.put('IDs', list)
                    Connection conn2 = Jsoup.connect(builder.toString())
                            .header("Content-Type", "application/json")
                            .header("Accept", "application/json")
                            .followRedirects(true)
                            .ignoreHttpErrors(true)
                            .ignoreContentType(true)
                            .userAgent(USER_AGENT)
                            .method(Connection.Method.POST)
                            .requestBody(JsonOutput.toJson(payload).toString())
                            .timeout(CONNECTION_TIMEOUT)
                            .maxBodySize(0)
                    JSONElement json2 = JSON.parse(conn2.execute().charset(charset).body())
                    if (json2 instanceof List) {
                        json2.each { item ->
                            if (item) {
                                def lineData = [:]
                                boolean ignore = false
                                for (int i = 0; i < contentProperties.size(); i++) {
                                    CrawlProperty prop = contentProperties[i]
                                    String indexName = prop.indexName

                                    String mapVal
                                    if (prop.selectors?.contains(".")) {
                                        def innerItem = item
                                        prop.selectors?.split("\\.")?.each {
                                            if (innerItem != null) {
                                                innerItem = innerItem."${it}"
                                            }
                                            if (innerItem && innerItem instanceof List) {
                                                innerItem = innerItem.getAt(0)
                                            }
                                        }
                                        mapVal = innerItem ? innerItem.toString() : ''
                                    } else {
                                        mapVal = item."$prop.selectors" != null ? item."$prop.selectors" : ""
                                    }

                                    mapVal = StringUtils.trim(mapVal)
                                    if (StringUtils.isNotBlank(mapVal) && prop.parseAsHtml) {
                                        mapVal = Jsoup.parse("<div>" + mapVal + "</div>", "", Parser.xmlParser()).text()
                                    }
                                    if (StringUtils.isNotBlank(mapVal) && StringUtils.isNotBlank(prop.textToRemove)) {
                                        mapVal = mapVal.replace(prop.textToRemove, "")
                                    }

                                    if (shouldIgnore(mapVal, prop.ignoreIf, prop.ignoreValues)) {
                                        ignore = true
                                        break
                                    }
                                    lineData[indexName] = mapVal
                                }

                                if (!ignore && (webCrawler.minValuesExist == null || lineData?.size() >= webCrawler.minValuesExist)) {
                                    resultList.add(lineData)
                                }
                            }
                        }
                    }
                }
            }
        } catch (Exception ex) {
            log.error(ex?.message, ex)
        }
        return resultList
    }

    List<CrawlProperty> getCrawlProperties(WebCrawler webCrawler) {
        def result = CrawlProperty.where {
            webCrawler == webCrawler && dataOutput == true
        }.sort('propertyIndex', 'asc').list()
        return result
    }

    /**
     * Gets JSONElement {@link JSONElement} from feed.
     * @param webCrawler the WebCrawler instance {@link WebCrawler}
     * @param crawlUrl the feed url
     * @return the JSONElement
     */
    private JSONElement getJSONContent(WebCrawler webCrawler, String crawlUrl) {
        JSONElement jsonElement = null
        String auth = webCrawler.authType == AuthType.BASIC ? getBasicAuthHeader(webCrawler) : ""
        String charsetEncoding = webCrawler.charsetEncoding ?: "UTF-8"

        int attempt = 0
        while( attempt < RETRY_ATTEMPTS) {
            Status status = Status.SUCCESS
            String crawlMessage = ""
            String crawlCauseMessage = ""

            try {
                if (crawlUrl.startsWith('http')) {
            Connection conn = Jsoup.connect(crawlUrl)
                .userAgent(USER_AGENT)
                .timeout(CONNECTION_TIMEOUT)
                .ignoreContentType(true).maxBodySize(0)
            if (webCrawler.requestType == RequestType.GET) {
                if (webCrawler.authType == AuthType.BASIC) {
                    conn.header('Authorization', 'Basic ' + auth)
                }
            } else {
                conn = conn.method(Connection.Method.POST)
                // work around for form data
                String params = crawlUrl.replace(webCrawler.url + "?", "")
                params.split("&").each {
                    String[] kv = it.split("=")
                    conn.data(kv[0], kv[1])
                }
            }
                jsonElement = JSON.parse(conn.execute().charset(charsetEncoding).body())
                } else {
                    def is = new File(crawlUrl).newInputStream()
                    jsonElement = JSON.parse(is, charsetEncoding)
                }
            } catch (Exception e) {
                log.error(e?.message, e)

                status = Status.FAILED
                crawlMessage = e?.message
                crawlCauseMessage = e?.cause?.message
            }

            attempt += 1
            CrawlLog.withNewTransaction {
                logsService.saveLogs(webCrawler.code, crawlUrl, status, attempt, crawlMessage?:"", crawlCauseMessage?:"")
            }

            if (status == Status.SUCCESS) break
        }
        return jsonElement
    }

    /**
     * Gets content from FTP feed.
     * @param url the ftp url
     * @return the feed's content
     */
    private String getFTPContent(String url) {
        FTPClient client = new FTPClient()
        try {
            def fs = url.tokenize("/")
            String serverAddress = fs?.get(0)
            client.connect(serverAddress)

            int reply = client.getReplyCode()
            if (!FTPReply.isPositiveCompletion(reply)) {
                client.disconnect()
                throw new Exception("FTP server refused connection: " + serverAddress)
            }

            client.enterLocalPassiveMode()

            if ( !client.login("anonymous", "") ) {
                throw new Exception("Could not login to FTP Server: " + serverAddress)
            }

            String filename = "/" + fs?.drop(1)?.join("/")
            InputStream inputStream = client.retrieveFileStream(filename)

            def line
            StringBuilder stringBuilder = new StringBuilder()
            inputStream?.withReader("UTF-8") { reader ->
                while ((line = reader.readLine())!=null) {
                    stringBuilder.append(line).append("\n")
                }
            }

            return stringBuilder.toString()

        } catch (IOException e) {
            log.error(e?.message, e)
        } catch (Exception e) {
            log.error(e?.message, e)
        } finally {
            try {
                client.disconnect()
            } catch (IOException ex) {
                log.warn(ex?.message)
            }
        }
        return null
    }

    /**
     * Gets content from async html page. Eg: download url is generated.<br>
     * Download url will be generated using selenium, and file will be downloaded using URLConnection
     * @param webCrawler the WebCrawler instance {@link WebCrawler}
     * @return the feed's content
     */
    private String getAsyncContent(WebCrawler webCrawler) {
        String content = null
        try {
            ChromeDriverManager.chromedriver().setup()
            ChromeOptions options = new ChromeOptions()
            options.addArguments("--headless")

            WebDriver webDriver = new ChromeDriver(options)
            webDriver.get(webCrawler.url)

            Integer asyncWaitTimeOut = webCrawler.asyncWaitTimeOut ?: 30
            WebDriverWait wait = new WebDriverWait(webDriver, asyncWaitTimeOut)
            WebElement element = wait.until(ExpectedConditions.presenceOfElementLocated(By.cssSelector(webCrawler.asyncWaitSelector)))
            String downloadUrl = element.getAttribute("href")

            Set<Cookie> cookies = webDriver.manage().getCookies()
            String cookieString = ""
            for (Cookie cookie : cookies) {
                cookieString += cookie.getName() + "=" + cookie.getValue() + ";";
            }

            URL url = new URL(downloadUrl)
            URLConnection urlConnection = url.openConnection()
            urlConnection.setReadTimeout(READ_TIMEOUT)
            urlConnection.setConnectTimeout(CONNECTION_TIMEOUT)
            urlConnection.setRequestProperty("User-Agent", USER_AGENT)
            urlConnection.setRequestProperty("Cookie", cookieString)
            InputStream inputStream = urlConnection.inputStream

            content = IOUtils.toString(new InputStreamReader(inputStream))

            webDriver.quit()
        } catch (Exception e) {
            log.error(e?.message, e)
        }
        return content
    }

    /**
     * Generates basic auth header.
     * @param webCrawler the WebCrawler instance {@link WebCrawler}
     * @return the basic auth header
     */
    private static String getBasicAuthHeader(WebCrawler webCrawler) {
        String authString = webCrawler.authUsername + ":" + webCrawler.authPassword
        return new String( Base64.encodeBase64(authString.getBytes()) )
    }

    /**
     * Gets content from feed with NTLM authentication.
     * @param webCrawler the WebCrawler instance {@link WebCrawler}
     * @param crawlUrl the feed url
     * @return the feed's content
     */
    private String getContentWithNTLM(WebCrawler webCrawler, String crawlUrl) {
        String content = null
        Authenticator.setDefault(new Authenticator() {
            @Override
            public PasswordAuthentication getPasswordAuthentication() {
                return new PasswordAuthentication(webCrawler.authUsername, webCrawler.authPassword?.toCharArray())
            }
        })

        int attempt = 0
        while( attempt < RETRY_ATTEMPTS) {
            Status status = Status.SUCCESS
            String crawlMessage = ""
            String crawlCauseMessage = ""

            try {
                URL url = new URL(crawlUrl)
                URLConnection urlConnection = url.openConnection()
                urlConnection.setReadTimeout(READ_TIMEOUT)
                urlConnection.setConnectTimeout(CONNECTION_TIMEOUT)
                urlConnection.setRequestProperty("User-Agent", USER_AGENT)
                InputStream inputStream = urlConnection.inputStream

                content = IOUtils.toString(new InputStreamReader(inputStream))
            } catch (Exception e) {
                log.error(e?.message, e)

                status = Status.FAILED
                crawlMessage = e?.message
                crawlCauseMessage = e?.cause?.message
            }

            attempt += 1
            CrawlLog.withNewTransaction {
                logsService.saveLogs(webCrawler.code, crawlUrl, status, attempt, crawlMessage?:"", crawlCauseMessage?:"")
            }

            if (status == Status.SUCCESS) break
        }
        return content
    }

    /**
     * Converts URI String to Map for POST request purpose.
     * @param url the feed url
     * @return the map of query string
     * @throws UnsupportedEncodingException
     */
    public static Map<String, String> splitQueryString(URL url) throws UnsupportedEncodingException {
        Map<String, String> query_pairs = new HashMap<>()
        if (!url) return query_pairs

        String query = url.getQuery()
        String[] pairs = query?.split("&")
        for (String pair : pairs) {
            int idx = pair?.indexOf("=")
            query_pairs.put(URLDecoder.decode(pair?.substring(0, idx), "UTF-8"), URLDecoder.decode(pair?.substring(idx + 1), "UTF-8"))
        }
        return query_pairs
    }

    /**
     * Gets feeds urls.
     * @param webCrawler the WebCrawler instance {@link WebCrawler}
     * @return the list of feed url
     */
    private static List<String> getUrls(WebCrawler webCrawler) {
        List<String> urlList = new ArrayList<>()
        if (webCrawler.usePagination && webCrawler.paginationType != PaginationType.STREAM) {
            int itemsPerPage = webCrawler.itemsPerPage ?: 10
            String limitParams = webCrawler.pageLimitName ? "&"+webCrawler.pageLimitName+"="+itemsPerPage : ""
            int firstPage = 1
            if (webCrawler.firstPage != null) {
                firstPage = webCrawler.firstPage
            }

            if (webCrawler.paginationType == PaginationType.PAGE_NUMBER) {
                if (webCrawler.multipleUrl) {
                    webCrawler.crawlUrls?.each { CrawlUrl crawlUrl ->
                        Integer fPage = crawlUrl.firstPage ?: firstPage
                        (fPage..crawlUrl.lastPage)?.each {
                            String newUrl = crawlUrl.url + "?" + (crawlUrl.paginationParams ?: "") + limitParams + "&" + webCrawler.pageNumberName + "=" + it
                            urlList.add(newUrl)
                        }
                    }
                } else {
                    (firstPage..webCrawler.lastPage)?.each {
                        String newUrl = webCrawler.url + "?" + (webCrawler.paginationParams ?: "") + limitParams + "&" + webCrawler.pageNumberName + "=" + it
                        urlList.add(newUrl)
                    }
                }
            } else if(webCrawler.paginationType == PaginationType.LIMIT_OFFSET) {
                if (webCrawler.multipleUrl) {
                    webCrawler.crawlUrls?.each { CrawlUrl crawlUrl ->
                        int lastOffset = crawlUrl.lastOffset ?: 0
                        if (crawlUrl.itemsPerPage) itemsPerPage = crawlUrl.itemsPerPage
                        0.step lastOffset, itemsPerPage, {
                            String newUrl = crawlUrl.url + "?" + (crawlUrl.paginationParams ?: "") + limitParams + "&" + webCrawler.pageOffsetName + "=" + it
                            urlList.add(newUrl)
                        }
                    }
                } else {
                    int lastOffset = webCrawler.lastOffset ?: 0
                    0.step lastOffset, itemsPerPage, {
                        String newUrl = webCrawler.url + "?" + (webCrawler.paginationParams ?: "") + limitParams + "&" + webCrawler.pageOffsetName + "=" + it
                        urlList.add(newUrl)
                    }
                }
            } else if (webCrawler.paginationType == PaginationType.PAGE_PATTERN) {
                if (webCrawler.multipleUrl) {
                    webCrawler.crawlUrls?.each { CrawlUrl crawlUrl ->
                        if (crawlUrl.firstPage != null) {
                            firstPage = crawlUrl.firstPage
                        }
                        (firstPage..crawlUrl.lastPage)?.each {
                            String newUrl = crawlUrl.url?.replace("{{page}}", it?.toString())
                            urlList.add(newUrl)
                        }
                    }
                } else {
                    (firstPage..webCrawler.lastPage)?.each {
                        String newUrl = webCrawler.url?.replace("{{page}}", it?.toString())
                        urlList.add(newUrl)
                    }
                }
            } else if (webCrawler.paginationType == PaginationType.FOLLOW) {
                if (webCrawler.multipleUrl) {
                    webCrawler.crawlUrls?.each { CrawlUrl crawlUrl ->
                        urlList.add(crawlUrl.url)
                    }
                } else {
                    urlList.add(webCrawler.url)
                }
            }
        } else {
            if (webCrawler.multipleUrl) {
                webCrawler.crawlUrls?.each { CrawlUrl crawlUrl ->
                    urlList.add(crawlUrl.url)
                }
            } else {
                urlList.add(webCrawler.url)
            }
        }
        return urlList
    }

    /**
     * Calculates total prices of crawler results.
     * @param webCrawler the WebCrawler instance {@link WebCrawler}
     * @param data the crawler results
     * @return the price summary
     */
    public BigDecimal getTotalPrices(WebCrawler webCrawler, def data) {
        BigDecimal totalPrices = BigDecimal.ZERO
        String priceIndexName = ""

        try {
            if (webCrawler.priceIndex != null) {
                def priceIndex = webCrawler.priceIndex
                def cProp = CrawlProperty.createCriteria().list(sort: "propertyIndex", order: "asc") {
                    and {
                        eq("webCrawler", webCrawler)
                        eq("dataOutput", true)
                    }
                }
                cProp?.eachWithIndex { CrawlProperty prop, int i ->
                    if(i == priceIndex) {
                        priceIndexName = prop.indexName
                    }
                }
            }

            data?.each {
                if (StringUtils.isNotBlank(priceIndexName) && it && it?.containsKey(priceIndexName)) {
                    def v = it.getAt(priceIndexName)
                    if (v) {
                        if (v instanceof BigDecimal) {
                            totalPrices = totalPrices.add(v)
                        } else {
                            def p = Calculator.parseMoney(v instanceof String ? (String) v : v?.toString())
                            if (p) {
                                totalPrices = totalPrices.add(p)
                            }
                        }
                    }
                }
            }
        } catch (Exception e) {
            log.error(e?.message, e)
        }
        return totalPrices
    }

    /**
     * Gets element of the wine detail page from async type page based on the given selector.
     * @param detailsUrl the url of the detail page
     * @param charsetEncoding the html page encoding
     * @param selectors the detail page selector
     * @param asyncWaitSelector the selector element to wait
     * @param asyncWaitTimeOut the waiting time out
     * @param webDriver the Web Driver
     * @return Element the element in wine page
     */
    private Element getDetailsElementAsync(
            String detailsUrl, String charsetEncoding, String selectors,
            WebDriver webDriver, String asyncWaitSelector, Integer asyncWaitTimeOut) {
        if (StringUtils.isBlank(detailsUrl)) return  null
        if (!asyncWaitTimeOut) asyncWaitTimeOut = 30

        Element element = null
        try {
            webDriver.get(detailsUrl)

            WebDriverWait wait = new WebDriverWait(webDriver, asyncWaitTimeOut)
            JSWaiter.setDriver(webDriver, wait)
            JSWaiter.waitJQueryAngular(true)
            wait.until(ExpectedConditions.presenceOfElementLocated(By.cssSelector(asyncWaitSelector ?: selectors)))
            if (asyncWaitSelector && !asyncWaitSelector.equalsIgnoreCase(selectors)) {
                JSWaiter.sleep(DETAILS_DOCUMENT_READY_WAIT)
            }

            List<WebElement> webElements = webDriver.findElements(By.cssSelector(selectors))
            if (webElements && !webElements.isEmpty()) {
                String content = webElements.get(0)?.getAttribute("outerHTML")
                if (content) {
                    Document doc = Jsoup.parse(content, "")
                    def elementList = doc?.select(selectors) as Elements
                    element = elementList?.first()
                }
            }
        } catch (TimeoutException e) {
            log.error(e?.message)
        } catch (Exception e) {
            log.error(e?.message, e)
        } finally {
            // reset url so that hashtag (#) doesn't affect wait condition
            // e.g. petersham wine detail page
            webDriver.get('about:blank')
        }
        return element
    }

    /**
     * Check if the current row of crawler result should be ignored.
     * @param text the property value to be checked
     * @param ignoreIf the condition {@link IgnoreIf}
     * @param ignoreValues the condition values
     * @return the test result
     */
    private static Boolean shouldIgnore(String text, IgnoreIf ignoreIf, String ignoreValues) {
        text = text?.trim()
        Boolean ignore = false

        if (ignoreIf != IgnoreIf.NOT_SET) {
            if (ignoreIf == IgnoreIf.BLANK) {
                text = Calculator.cleanupSpaces(text)
                if (StringUtils.isBlank(text)) {
                    ignore = true
                }
            } else if (ignoreIf == IgnoreIf.EQUALS) {
                List<String> iValues = ignoreValues?.tokenize("\n")
                for (int j=0; j < iValues.size(); j++) {
                    String iv = iValues[j]
                    if (text?.equalsIgnoreCase(iv)) {
                        ignore = true
                        break
                    }
                }
            } else if (ignoreIf == IgnoreIf.CONTAINS) {
                List<String> iValues = ignoreValues?.tokenize("\n")
                for (int j=0; j < iValues.size(); j++) {
                    String iv = iValues[j]
                    if (text?.contains(iv)) {
                        ignore = true
                        break
                    }
                }
            }
        }

        return ignore
    }
}
