package wine.db

import enums.LineTerminator
import grails.transaction.Transactional
import groovy.io.FileType
import org.springframework.beans.factory.annotation.Value
import wine.alert.GlobalNoise
import wine.alert.enums.DumpType
import wine.db.utils.WineKeywordUtils
import wine.db.utils.WineUtils

import java.util.regex.Pattern

/**
 * Service to dump and download certain data from database with particular format.<br>
 * Output Charset is "Windows-1252" to support Windows encoding
 */
@Transactional
class DownloadService {

    private static final String SEPARATOR = "\t"

    private static final String WINE_DB_NAME_PREFIX = "winedb-"
    private static final String WINE_DB_HEADER = "WineAlert ID\tWineAl\tProdShow LabelName  ColorClass\tVinN\tProd\tProdShow\tLabelName\tVariety\tColorClass\tDryness\tWineType\tCountry\tRegion\tSub-region\tAppellation\tSub-appellation\tStatus\tLast Modified\n"

    private static final String MONGO_MAP_NAME_PREFIX = "mongomap-"
    private static final String WINE_KEYWORD_NAME_PREFIX = "wine-keywords-input-"
    private static final String PRODUCER_KEYWORD_NAME_PREFIX = "producer-indexing-"
    private static final String GLOBAL_NOISE_PREFIX = "global-noise-"
    private static final String PRODUCER_NOISE_PREFIX = "producer-noise-"
    private static final String WINE_NOISE_PREFIX = "wine-noise-"

    @Value('${application.download.location}')
    private String downloadLocation

    @Value('${application.output.lineTerminator}')
    private String outputLineTerminator

    @Value('${application.output.encoding}')
    private String outputEncoding

    DumpType runningDump = null

    /**
     * Dump WineDb database.<br>
     * Output format is tab separated value, similar to winedb.txt that used as input {@link wine.alert.InputService#importWineDb(String)} with additional field "Last Modified"
     */
    void dumpWineDb() {
        try {
            runningDump = DumpType.WINE_DB
            Integer total = WineDb.wineDatabase.count()
            Integer max = 5000
            Integer offset = 0
            Integer pageCount = Math.ceil(total / max).intValue()

            String lineTerminator = getOutputLineTerminator()
            StringBuilder sb = new StringBuilder(WINE_DB_HEADER)
            for (int i = 1; i <= pageCount; i++) {
                log.info("Processing wine " + (i * max) + " of " + total)

                def wineDbList = WineDb.wineDatabase.createCriteria().list(max: max, offset: offset, sort: 'id') {
                    createAlias("producer", "p")
                    projections {
                        property("id")
                        property("name")
                        property("color")
                        property("type")
                        property("sweetness")
                        property("variety")
                        property("wineCountry")
                        property("wineRegion")
                        property("wineLocation")
                        property("wineLocale")
                        property("wineSite")
                        property("wineNId")
                        property("p.name")
                        property("p.nameShow")
                        property("status")
                        property("lastUpdated")
                    }
                    not { eq("id", "z") }
                    not { like("id", "%xxx") }
                }
                wineDbList?.each { wineDb ->
                    String description = WineUtils.getDescription(wineDb[13], wineDb[1], wineDb[2])

                    sb.append(wineDb[0] ?: "").append(SEPARATOR)
                    sb.append(wineDb[0]?.take(6) ?: "").append(SEPARATOR)
                    sb.append(description ?: "").append(SEPARATOR) // Desc
                    sb.append(wineDb[11] ?: "").append(SEPARATOR) // VinN
                    sb.append(wineDb[12] ?: "").append(SEPARATOR) // Prod
                    sb.append(wineDb[13] ?: "").append(SEPARATOR) // Prod Show
                    sb.append(wineDb[1] ?: "").append(SEPARATOR) // Name
                    sb.append(wineDb[5] ?: "").append(SEPARATOR) // Variety
                    sb.append(wineDb[2] ?: "").append(SEPARATOR) // Color
                    sb.append(wineDb[4] ?: "").append(SEPARATOR) // Sweetness
                    sb.append(wineDb[3] ?: "").append(SEPARATOR) // Type
                    sb.append(wineDb[6] ?: "").append(SEPARATOR) // Country
                    sb.append(wineDb[7] ?: "").append(SEPARATOR) // Region
                    sb.append(wineDb[8] ?: "").append(SEPARATOR) // Sub Region
                    sb.append(wineDb[9] ?: "").append(SEPARATOR) // Appellation
                    sb.append(wineDb[10] ?: "").append(SEPARATOR) // Sub Appellation
                    sb.append(wineDb[14] ?: "").append(SEPARATOR) // Status
                    Date lastModified = wineDb[15]
                    sb.append(lastModified ? lastModified.format("MM/dd/yyyy HH:mm:ss") : "") // Last modified
                    sb.append(lineTerminator)
                }
                offset += max
            }

            Date date = new Date()
            String fileName = WINE_DB_NAME_PREFIX + "${date.format('yyyy-MM-dd-HHmmss')}.txt"
            File file = new File(downloadLocation + File.separator + fileName)
            file.getParentFile().mkdirs()
            file.createNewFile()
            file.append(sb.toString(), outputEncoding)

            log.info("WineDb dump is ready for download at: " + file.path)
        } catch(Exception e) {
            log.error(e?.message, e)
        } finally {
            runningDump = null
        }
    }

    /**
     * Dump MongoMap database.<br>
     * Output format is tab separated value, similar to mongomap.txt that used as input {@link wine.alert.InputService#importMongoWinesMap(String)} with additional field "Last Modified"
     */
    void dumpMongoMap() {
        try {
            runningDump = DumpType.MONGO_MAP
            Integer total = Wine.wineDatabase.count()
            Integer max = 5000
            Integer offset = 0
            Integer pageCount = Math.ceil(total / max).intValue()

            String lineTerminator = getOutputLineTerminator()
            StringBuilder sb = new StringBuilder()
            for (int i = 1; i <= pageCount; i++) {
                log.info("Processing wine " + (i * max) + " of " + total)

                def wineList = Wine.wineDatabase.createCriteria().list(max: max, offset: offset, sort: 'id') {
                    projections {
                        property("id")
                        property("mongoId")
                        property("lastUpdated")
                    }
                    and {
                        isNotNull("mongoId")
                        ne("mongoId", "")
                    }
                }
                wineList?.each { wine ->
                    sb.append(wine?.getAt(0) ?: "").append(SEPARATOR)
                    sb.append(wine?.getAt(1) ?: "").append(SEPARATOR)
                    Date lastModified = wine?.getAt(2)
                    sb.append(lastModified ? lastModified.format("MM/dd/yyyy HH:mm:ss") : "")
                    sb.append(lineTerminator)

                    // Duplications
                    def duplications = WineDuplication.wineDatabase.createCriteria().list {
                        createAlias("wine", "w")
                        projections {
                            property("w.id")
                            property("mongoId")
                            property("lastUpdated")
                        }
                        eq("w.id", wine?.getAt(0))
                    }
                    duplications?.each { dup ->
                        sb.append(dup?.getAt(0) ?: "").append(SEPARATOR)
                        sb.append(dup?.getAt(1) ?: "").append(SEPARATOR)
                        Date dupLastModified = dup?.getAt(2)
                        sb.append(dupLastModified ? dupLastModified.format("MM/dd/yyyy HH:mm:ss") : "")
                        sb.append(lineTerminator)
                    }
                }
                offset += max
            }

            Date date = new Date()
            String fileName = MONGO_MAP_NAME_PREFIX + "${date.format('yyyy-MM-dd-HHmmss')}.txt"
            File file = new File(downloadLocation + File.separator + fileName)
            file.getParentFile().mkdirs()
            file.createNewFile()
            file.append(sb.toString(), outputEncoding)

            log.info("Mongo Map dump is ready for download at: " + file.path)
        } catch(Exception e) {
            log.error(e?.message, e)
        } finally {
            runningDump = null
        }
    }

    /**
     * Dump WineKeyword database.<br>
     * Output format is tab separated value, similar to wine-keywords-input.txt that used as input {@link wine.alert.InputService#importWineKeywords(String)} with additional field "Last Modified"
     */
    void dumpWineKeyword() {
        try {
            runningDump = DumpType.WINE_KEYWORD
            Integer total = WineKeyword.wineDatabase.count()
            Integer max = 5000
            Integer offset = 0
            Integer pageCount = Math.ceil(total / max).intValue()

            String lineTerminator = getOutputLineTerminator()
            Date date = new Date()
            String fileName = WINE_KEYWORD_NAME_PREFIX + "${date.format('yyyy-MM-dd-HHmmss')}.txt"
            String filePath = downloadLocation + File.separator + fileName
            File file = new File(filePath)
            file.getParentFile().mkdirs()
            file.createNewFile()

            for (int i = 1; i <= pageCount; i++) {
                log.info("Processing WineKeywords " + (i * max) + " of " + total)

                def wineKeywordList = WineKeyword.wineDatabase.createCriteria().list(max: max, offset: offset, sort: 'id') {
                    createAlias("wineDb", "wd")
                    createAlias("wd.producer", "p")
                    projections {
                        property("wd.id")
                        property("wd.name")
                        property("wd.color")
                        property("p.nameShow")
                        property("pattern")
                        property("producerKeyword")
                        property("baseKeyword")
                        property("appellationKeyword")
                        property("colorKeyword")
                        property("commonModifiersKeyword")
                        property("modifiersKeyword")
                        property("customNotKeyword")
                        property("globalNotKeyword")
                        property("wd.status")
                        property("lastUpdated")
                    }
                    not { eq("wd.id", "z") }
                    not { like("wd.id", "%xxx") }
                }
                StringBuilder sb = new StringBuilder()
                wineKeywordList?.each { wineDb ->
                    String description = WineUtils.getDescription(wineDb[3], wineDb[1], wineDb[2])
                    String producerKeyword = wineDb[5]
                    String baseKeyword = wineDb[6]
                    String appellationKeyword = wineDb[7]
                    String colorKeyword = wineDb[8]
                    String commonModifiersKeyword = wineDb[9]
                    String modifiersKeyword = wineDb[10]
                    String customNotKeyword = wineDb[11]
                    String globalNotKeyword = wineDb[12]
                    String pattern = WineKeywordUtils.joinPatternFromStrings(producerKeyword, baseKeyword, appellationKeyword,
                            colorKeyword, commonModifiersKeyword, modifiersKeyword, customNotKeyword, globalNotKeyword)
                    Date lastModified = (Date) wineDb[14]

                    sb.append(wineDb[0] ?: "").append(SEPARATOR) // WineDb ID
                    sb.append(description).append(SEPARATOR) // Description
                    sb.append(pattern ?: "").append(SEPARATOR) // Full Keyword
                    sb.append(producerKeyword ?: "").append(SEPARATOR) // Producer Keyword
                    sb.append(baseKeyword ?: "").append(SEPARATOR) // Base Keyword
                    sb.append(appellationKeyword ?: "").append(SEPARATOR) // Appellation Keyword
                    sb.append(colorKeyword ?: "").append(SEPARATOR) // Color Keyword
                    sb.append(commonModifiersKeyword ?: "").append(SEPARATOR) // Common Modifiers Keyword
                    sb.append(modifiersKeyword ?: "").append(SEPARATOR) // Modifiers Keyword
                    sb.append(customNotKeyword ?: "").append(SEPARATOR) // Custom Not Keyword
                    sb.append(globalNotKeyword ?: "").append(SEPARATOR) // Global Not Keyword
                    sb.append(wineDb[13] ?: "").append(SEPARATOR) // Status
                    sb.append(lastModified ? lastModified.format("MM/dd/yyyy HH:mm:ss") : "") // Last modified
                    sb.append(lineTerminator)
                }
                File file2 = new File(filePath)
                file2.append(sb.toString(), outputEncoding)
                offset += max
            }

            log.info("WineKeyword dump is ready for download at: " + file.path)
        } catch (Exception e) {
            log.error(e?.message, e)
        } finally {
            runningDump = null
        }
    }

    /**
     * Dump ProducerKeyword database.<br>
     * Output format is tab separated value, similar to producer-indexing.txt that used as input {@link wine.alert.ConfigFilesWatcherJob#importProducerIndexingKeywords(String)} with additional field "Last Modified"
     */
    void dumpProducerKeyword() {
        try {
            runningDump = DumpType.PRODUCER_KEYWORD
            Integer total = ProducerKeyword.wineDatabase.count()
            Integer max = 5000
            Integer offset = 0
            Integer pageCount = Math.ceil(total / max).intValue()

            String lineTerminator = getOutputLineTerminator()
            StringBuilder sb = new StringBuilder()
            for (int i = 1; i <= pageCount; i++) {
                log.info("Processing ProducerKeyword " + (i * max) + " of " + total)

                def keywordList = ProducerKeyword.wineDatabase.createCriteria().list(max: max, offset: offset, sort: 'id') {
                    projections {
                        property("keyword")
                        property("lastUpdated")
                        producer {
                            property("id")
                        }
                    }
                }
                keywordList?.each { pk ->
                    sb.append(pk[2] ?: "").append(SEPARATOR) // Producer ID
                    sb.append(pk[0] ?: "").append(SEPARATOR) // Keyword

                    Date lastModified = pk[1]
                    sb.append(lastModified ? lastModified.format("MM/dd/yyyy HH:mm:ss") : "") // Last modified
                    sb.append(lineTerminator)
                }
                offset += max
            }

            Date date = new Date()
            String fileName = PRODUCER_KEYWORD_NAME_PREFIX + "${date.format('yyyy-MM-dd-HHmmss')}.txt"
            File file = new File(downloadLocation + File.separator + fileName)
            file.getParentFile().mkdirs()
            file.createNewFile()
            file.append(sb.toString(), outputEncoding)

            log.info("ProducerKeyword dump is ready for download at: " + file.path)
        } catch(Exception e) {
            log.error(e?.message, e)
        } finally {
            runningDump = null
        }
    }

    /**
     * Dump GlobalNoise database.<br>
     * Output format is tab separated value, similar to global-noise.txt that used as input {@link wine.alert.InputService#importGlobalNoise(String)} with additional field "Last Modified"
     */
    void dumpGlobalNoises() {
        try {
            runningDump = DumpType.GLOBAL_NOISE
            Integer total = GlobalNoise.count()
            Integer max = 5000
            Integer offset = 0
            Integer pageCount = Math.ceil(total / max).intValue()

            String lineTerminator = getOutputLineTerminator()
            StringBuilder sb = new StringBuilder()
            for (int i = 1; i <= pageCount; i++) {
                log.info("Processing GlobalNoise " + (i * max) + " of " + total)

                def noiseList = GlobalNoise.createCriteria().list(max: max, offset: offset, sort: 'id') {
                    projections {
                        property("noise")
                        property("lastUpdated")
                    }
                }
                noiseList?.each { pk ->
                    sb.append(pk[0] ?: "").append(SEPARATOR) // Noise

                    Date lastModified = pk[1]
                    sb.append(lastModified ? lastModified.format("MM/dd/yyyy HH:mm:ss") : "") // Last modified
                    sb.append(lineTerminator)
                }
                offset += max
            }

            Date date = new Date()
            String fileName = GLOBAL_NOISE_PREFIX + "${date.format('yyyy-MM-dd-HHmmss')}.txt"
            File file = new File(downloadLocation + File.separator + fileName)
            file.getParentFile().mkdirs()
            file.createNewFile()
            file.append(sb.toString(), outputEncoding)

            log.info("GlobalNoise dump is ready for download at: " + file.path)
        } catch(Exception e) {
            log.error(e?.message, e)
        } finally {
            runningDump = null
        }
    }

    /**
     * Dump ProducerNoise database.<br>
     * Output format is tab separated value, similar to producer-noise.txt that used as input {@link wine.alert.InputService#importProducerNoise(String)} with additional field "Last Modified"
     */
    void dumpProducerNoises() {
        try {
            runningDump = DumpType.PRODUCER_NOISE
            Integer total = ProducerNoise.wineDatabase.count()
            Integer max = 5000
            Integer offset = 0
            Integer pageCount = Math.ceil(total / max).intValue()

            String lineTerminator = getOutputLineTerminator()
            StringBuilder sb = new StringBuilder("Producer Id\tNoise\tLast Modified\n")
            for (int i = 1; i <= pageCount; i++) {
                log.info("Processing ProducerNoise " + (i * max) + " of " + total)

                def noiseList = ProducerNoise.wineDatabase.createCriteria().list(max: max, offset: offset, sort: 'id') {
                    projections {
                        property("noise")
                        property("lastUpdated")
                        producer {
                            property("id")
                        }
                    }
                }
                noiseList?.each { pk ->
                    sb.append(pk[0] ?: "").append(SEPARATOR) // Noise
                    sb.append(pk[2] ?: "").append(SEPARATOR) // Producer ID

                    Date lastModified = pk[1]
                    sb.append(lastModified ? lastModified.format("MM/dd/yyyy HH:mm:ss") : "") // Last modified
                    sb.append(lineTerminator)
                }
                offset += max
            }

            Date date = new Date()
            String fileName = PRODUCER_NOISE_PREFIX + "${date.format('yyyy-MM-dd-HHmmss')}.txt"
            File file = new File(downloadLocation + File.separator + fileName)
            file.getParentFile().mkdirs()
            file.createNewFile()
            file.append(sb.toString(), outputEncoding)

            log.info("ProducerNoise dump is ready for download at: " + file.path)
        } catch(Exception e) {
            log.error(e?.message, e)
        } finally {
            runningDump = null
        }
    }

    /**
     * Dump WineNoise database.<br>
     * Output format is tab separated value, similar to producer-noise.txt that used as input {@link wine.alert.InputService#importWineNoise(String)} with additional field "Last Modified"
     */
    void dumpWineNoises() {
        try {
            runningDump = DumpType.WINE_NOISE
            Integer total = WineNoise.wineDatabase.count()
            Integer max = 5000
            Integer offset = 0
            Integer pageCount = Math.ceil(total / max).intValue()

            String lineTerminator = getOutputLineTerminator()
            StringBuilder sb = new StringBuilder("Wine Alert Id\tNoise\tLast Modified\n")
            for (int i = 1; i <= pageCount; i++) {
                log.info("Processing WineNoise " + (i * max) + " of " + total)

                def noiseList = WineNoise.wineDatabase.createCriteria().list(max: max, offset: offset, sort: 'id') {
                    projections {
                        property("noise")
                        property("lastUpdated")
                        wineDb {
                            property("id")
                        }
                    }
                }
                noiseList?.each { pk ->
                    sb.append(pk[0] ?: "").append(SEPARATOR) // Noise
                    sb.append(pk[2] ?: "").append(SEPARATOR) // Wine Db ID

                    Date lastModified = pk[1]
                    sb.append(lastModified ? lastModified.format("MM/dd/yyyy HH:mm:ss") : "") // Last modified
                    sb.append(lineTerminator)
                }
                offset += max
            }

            Date date = new Date()
            String fileName = WINE_NOISE_PREFIX + "${date.format('yyyy-MM-dd-HHmmss')}.txt"
            File file = new File(downloadLocation + File.separator + fileName)
            file.getParentFile().mkdirs()
            file.createNewFile()
            file.append(sb.toString(), outputEncoding)

            log.info("WineNoise dump is ready for download at: " + file.path)
        } catch(Exception e) {
            log.error(e?.message, e)
        } finally {
            runningDump = null
        }
    }

    /**
     * Gets the latest dump file name based on dump type
     * @param dumpType One of enum {@link DumpType}
     * @return the latest dump file name
     */
    String getLatestDumpFileName(DumpType dumpType) {
        if (dumpType == DumpType.WINE_DB) {
            return getLatestDownloadFileName(WINE_DB_NAME_PREFIX + ".+\\.txt")
        } else if (dumpType == DumpType.MONGO_MAP) {
            return getLatestDownloadFileName(MONGO_MAP_NAME_PREFIX + ".+\\.txt")
        } else if (dumpType == DumpType.WINE_KEYWORD) {
            return getLatestDownloadFileName(WINE_KEYWORD_NAME_PREFIX + ".+\\.txt")
        } else if (dumpType == DumpType.PRODUCER_KEYWORD) {
            return getLatestDownloadFileName(PRODUCER_KEYWORD_NAME_PREFIX + ".+\\.txt")
        } else if (dumpType == DumpType.GLOBAL_NOISE) {
            return getLatestDownloadFileName(GLOBAL_NOISE_PREFIX + ".+\\.txt")
        } else if (dumpType == DumpType.PRODUCER_NOISE) {
            return getLatestDownloadFileName(PRODUCER_NOISE_PREFIX + ".+\\.txt")
        } else if (dumpType == DumpType.WINE_NOISE) {
            return getLatestDownloadFileName(WINE_NOISE_PREFIX + ".+\\.txt")
        }
        return null
    }

    /**
     * Gets the latest dump file based on dump type
     * @param dumpType One of enum {@link DumpType}
     * @return the latest dump file
     */
    def getLatestDumpFile(DumpType dumpType) {
        if (dumpType == DumpType.WINE_DB) {
            return getLatestDownloadFile(WINE_DB_NAME_PREFIX + ".+\\.txt")
        } else if (dumpType == DumpType.MONGO_MAP) {
            return getLatestDownloadFile(MONGO_MAP_NAME_PREFIX + ".+\\.txt")
        } else if (dumpType == DumpType.WINE_KEYWORD) {
            return getLatestDownloadFile(WINE_KEYWORD_NAME_PREFIX + ".+\\.txt")
        } else if (dumpType == DumpType.PRODUCER_KEYWORD) {
            return getLatestDownloadFile(PRODUCER_KEYWORD_NAME_PREFIX + ".+\\.txt")
        } else if (dumpType == DumpType.GLOBAL_NOISE) {
            return getLatestDownloadFile(GLOBAL_NOISE_PREFIX + ".+\\.txt")
        } else if (dumpType == DumpType.PRODUCER_NOISE) {
            return getLatestDownloadFile(PRODUCER_NOISE_PREFIX + ".+\\.txt")
        } else if (dumpType == DumpType.WINE_NOISE) {
            return getLatestDownloadFile(WINE_NOISE_PREFIX + ".+\\.txt")
        }
        return null
    }

    /**
     * Gets the latest dump file name based on given pattern
     * @param fileNamePattern Text regex pattern of file path/location
     * @return the latest download file name
     */
    private String getLatestDownloadFileName(String fileNamePattern) {
        File dumpDir = new File(downloadLocation)
        def files = []
        dumpDir?.eachFile(FileType.FILES) {
            if (it.name?.matches(Pattern.compile(fileNamePattern))) {
                files << it.name
            }
        }
        if (files.isEmpty()) {
            return null
        } else {
            return files?.toSorted { a, b -> b <=> a }?.first()
        }
    }

    /**
     * Gets the latest dump file name based on given pattern
     * @param fileNamePattern Text regex pattern of file path/location
     * @return the latest download file
     */
    def getLatestDownloadFile(String fileNamePattern) {
        File dumpDir = new File(downloadLocation)
        def files = []
        dumpDir?.eachFile(FileType.FILES) {
            if (it.name?.matches(Pattern.compile(fileNamePattern))) {
                files << it
            }
        }
        if (files.isEmpty()) {
            return null
        } else {
            return files?.toSorted { a, b -> b.name <=> a.name }?.first()
        }
    }

    /**
     * Gets the correct line terminator for the file generated
     * @return the line terminator
     */
    private String getOutputLineTerminator() {
        LineTerminator terminator = LineTerminator.CRLF
        try {
            terminator = outputLineTerminator as LineTerminator
        } finally {}
        return terminator?.value()
    }
}
