package wine.db

import grails.transaction.Transactional
import grails.web.servlet.mvc.GrailsParameterMap
import org.apache.commons.collections.CollectionUtils
import org.hibernate.sql.JoinType
import wine.db.common.SimpleModel
import wine.db.enums.Source
import wine.db.utils.WineKeywordUtils
import wine.db.utils.WineUtils

/**
 * Service to interact with wine related tables.
 */
@Transactional
class WineService {

    /**
     * Merge two or more wines
     * @param wineDb the wine_db as merge destination
     * @param params the grails parameter map that contains list of wineMerge as wine id
     */
    void mergeWineDb(WineDb wineDb, GrailsParameterMap params) {

        List<String> wineMerges = params?.list("wineMerge")?.findAll { it != wineDb.id }
        wineMerges?.each { String mergeId ->

            WineDb merging = WineDb.wineDatabase.findById(mergeId)
            if (merging) {

                merging.wines?.each { Wine wine ->
                    String mongoId = wine.mongoId

                    wine.mongoId = null
                    wine.validate()
                    if (wine.hasErrors()) {
                        log.error(wine.errors)
                        return
                    }
                    wine.save()

                    Boolean isNewWine = Boolean.FALSE
                    Wine sourceWine = Wine.wineDatabase.findByWineDbAndVintage(wineDb, wine.vintage)
                    if (sourceWine == null) {
                        isNewWine = Boolean.TRUE
                        sourceWine = new Wine(
                                vintage: wine.vintage,
                                maturity: wine.maturity,
                                ratingLow: wine.ratingLow,
                                ratingHigh: wine.ratingHigh,
                                ratingQ: wine.ratingQ,
                                imageUrl: wine.imageUrl,
                                appellation: wine.appellation,
                                mongoId: mongoId,
                                wineNId: wine.wineNId,
                                wineNRating: wine.wineNRating,
                                wineDb: wineDb
                        )
                        sourceWine.id = wineDb.id + wine.id?.reverse()?.take(3)?.reverse()

                        sourceWine.validate()
                        if (sourceWine.hasErrors()) {
                            log.error(sourceWine.errors)
                            return
                        }
                        sourceWine.save()
                    }

                    if (!sourceWine.mongoId) {
                        sourceWine.mongoId = mongoId
                        sourceWine.validate()
                        if (sourceWine.hasErrors()) {
                            log.error(sourceWine.errors)
                            return
                        }
                        sourceWine.save()
                    } else {
                        new WineDuplication(mongoId: mongoId, wine: sourceWine).save()
                    }

                    wine.wineDuplications?.findAll()?.each { WineDuplication wineDuplication ->
                        wine.removeFromWineDuplications(wineDuplication)
                        sourceWine.addToWineDuplications(wineDuplication)
                    }

                    wine.retailerWines?.findAll()?.each { RetailerWine retailerWine ->
                        if (isNewWine) {
                            RetailerWine sourceRetailerWine = new RetailerWine(
                                    name: retailerWine.name,
                                    quantity: retailerWine.quantity,
                                    taxStatus: retailerWine.taxStatus,
                                    description: retailerWine.description,
                                    sku: retailerWine.sku,
                                    url: retailerWine.url,
                                    currency: retailerWine.currency,
                                    price: retailerWine.price,
                                    usdPrice: retailerWine.usdPrice,
                                    bottleSize: retailerWine.bottleSize,
                                    retailer: retailerWine.retailer,
                                    available: retailerWine.available,
                                    lastAvailable: retailerWine.lastAvailable
                            )
                            sourceRetailerWine.validate()
                            if (sourceRetailerWine.hasErrors()) {
                                log.error(sourceRetailerWine.errors)
                                return
                            }
                            sourceWine.addToRetailerWines(sourceRetailerWine)
                        } else {
                            RetailerWine rw = RetailerWine.wineDatabase.findByWineAndRetailerAndBottleSizeAndSku(sourceWine, retailerWine.retailer, retailerWine.bottleSize, retailerWine.sku)
                            if (!rw) {
                                RetailerWine sourceRetailerWine = new RetailerWine(
                                        name: retailerWine.name,
                                        quantity: retailerWine.quantity,
                                        taxStatus: retailerWine.taxStatus,
                                        description: retailerWine.description,
                                        sku: retailerWine.sku,
                                        url: retailerWine.url,
                                        currency: retailerWine.currency,
                                        price: retailerWine.price,
                                        usdPrice: retailerWine.usdPrice,
                                        bottleSize: retailerWine.bottleSize,
                                        retailer: retailerWine.retailer,
                                        available: retailerWine.available,
                                        lastAvailable: retailerWine.lastAvailable,
                                )
                                sourceRetailerWine.validate()
                                if (sourceRetailerWine.hasErrors()) {
                                    log.error(sourceRetailerWine.errors)
                                    return
                                }
                                sourceWine.addToRetailerWines(sourceRetailerWine)
                            }
                        }
                    }
                    sourceWine.validate()
                    if (sourceWine.hasErrors()) {
                        return
                    }
                    sourceWine.save()
                }

                new WineMerge(sourceId: mergeId, destinationId: wineDb.id).save()
            }
        }
    }

    /**
     * Removes wine_db
     * @param wineDb the WineDb instance to be removed
     */
    void deleteWineDb(WineDb wineDb) {
        List<Wine> wines = Wine.wineDatabase.createCriteria().list {
            eq("wineDb", wineDb)
        }
        List<String> wineIdList = new ArrayList<>()
        wines?.each {
            wineIdList.add((String) it.id)
            DeletedWine deletedWine = new DeletedWine(
                    objectId: it.id,
                    vintage: it.vintage,
                    maturity: it.maturity,
                    wineDbId: it.wineDbId,
                    appellationId: it.appellationId?.toString(),
                    imageUrl: it.imageUrl,
                    ratingLow: it.ratingLow,
                    ratingHigh: it.ratingHigh,
                    ratingQ: it.ratingQ,
                    mongoId: it.mongoId,
                    wineNId: it.wineNId,
                    wineNRating: it.wineNRating
            )

            deletedWine.validate()
            if (deletedWine.hasErrors()) {
                log.error(deletedWine.errors)
                return
            }
            deletedWine.save()
        }

        if (wineIdList && !wineIdList.isEmpty()) {
            Integer affectedRw = RetailerWine.executeUpdate("DELETE FROM RetailerWine WHERE wine.id IN(:wineIdList)", [wineIdList: wineIdList])
            log.debug("Deleted RetailerWines: " + affectedRw)
            Integer affectedWd = WineDuplication.executeUpdate("DELETE FROM WineDuplication WHERE wine.id IN(:wineIdList) ", [wineIdList: wineIdList])
            log.debug("Deleted WineDuplications: " + affectedWd)
            Integer affectedWines = Wine.executeUpdate("DELETE FROM Wine WHERE id IN(:wineIdList)", [wineIdList: wineIdList])
            log.debug("Deleted Wines: " + affectedWines)
        }
        WineNoise.executeUpdate("DELETE FROM WineNoise WHERE wineDb=:wineDb", [wineDb: wineDb])

        wineDb.status = WineDb.Status.DELETED

        wineDb.validate()
        if (wineDb.hasErrors()) {
            log.error(wineDb.errors)
            return
        }

        wineDb.save()
    }

    /**
     * Get map of wine or wined_db properties
     * @param wineId the of the wine {@link Wine#id}
     * @param wineDbId the of the wine_db {@link WineDb#id}
     * @return the list of wine/wine_db properties
     */
    @Transactional(readOnly = true)
    def getWinePropertiesForMatchValidation(String wineId, String wineDbId) {
        def wine = Wine.wineDatabase.createCriteria().get {
            createAlias("wineDb", "wd")
            createAlias("wd.producer", "pd")
            createAlias("wd.wineKeyword", "wk", JoinType.LEFT_OUTER_JOIN)
            projections {
                property('wd.color')
                property('wd.type')
                property('wd.sweetness')
                property('wd.variety')
                property('wd.note')
                property('wd.name')
                property('wd.wineCountry')
                property('wd.wineRegion')
                property('wd.wineLocation')
                property('wd.wineLocale')
                property('wd.wineSite')
                property('pd.nameShow')
                property('wk.pattern')
                property('maturity')
                property('mongoId')
            } as List<String>
            eq("id", wineId)
            maxResults(1)
        }
        if (!wine) {
            wine = WineDb.wineDatabase.createCriteria().get {
                createAlias("producer", "pd")
                createAlias("wineKeyword", "wk", JoinType.LEFT_OUTER_JOIN)
                projections {
                    property('color')
                    property('type')
                    property('sweetness')
                    property('variety')
                    property('note')
                    property('name')
                    property('wineCountry')
                    property('wineRegion')
                    property('wineLocation')
                    property('wineLocale')
                    property('wineSite')
                    property('pd.nameShow')
                    property('wk.pattern')
                }
                eq("id", wineDbId)
                maxResults(1)
            } as List<String>
            if (wine) {
                wine.push("")
                wine.push("")
            }
        }
        return wine
    }

    /**
     * Checks whether reindex at wine-data API endpoint required.<br>
     *     Have to be called before save the modified instance.
     * @param wineDb the wineDb instance
     * @return true if reindex required
     */
    Boolean isWineApiReindexRequired(WineDb wineDb) {
        return wineDb.isDirty("name") || wineDb.isDirty("color") || wineDb.isDirty("producer") || wineDb.isDirty("status")
    }

    /**
     * Checks if WineDb is exists
     * @param wineDbId the WineDb ID
     * @return true if exists
     */
    @Transactional(readOnly = true)
    Boolean isWineDbExists(String wineDbId) {
        Integer c = WineDb.where {
            id == wineDbId && status == WineDb.Status.ACTIVE
        }.count()?.intValue()
        return c && c > 0
    }

    /**
     * Gets wines with description
     * @param ids the list of {@link WineDb#id}
     * @return the list of {@link SimpleModel}
     */
    @Transactional(readOnly = true)
    List<SimpleModel> getWinesWithDescription(List<String> ids) {
        if (CollectionUtils.isEmpty(ids)) return Collections.emptyList()
        def wines = WineDb.createCriteria().list {
            projections {
                property('id')      //0
                property('name')    //1
                property('color')   //2
                producer {
                    property('nameShow') //3
                }
            }
            'in'('id', ids)
        }
        List<SimpleModel> list = wines?.collect {
            String label = WineUtils.getDescription(it[3], it[1], it[2])
            return new SimpleModel(id: it[0], label: label)
        }
        return list
    }

    /**
     * Saves keywords
     * @param keywords the json list of keywords
     * @param source the source data
     * @return the list of {@link WineKeyword}
     */
    List<WineKeyword> saveKeywords(keywords, Source source) {
        def list = []
        keywords?.each { data ->
            String wineDbId = data.wineDbId
            Integer c = WineKeyword.where {
                wineDb { id == wineDbId }
            }.count()?.intValue()
            if (!c) {
                WineDb wineDb = WineDb.wineDatabase.findById(wineDbId)
                if (wineDb) {
                    String producerKeyword = data.producerKeyword
                    String appellationKeyword = data.appellationKeyword
                    String baseKeyword = data.baseKeyword
                    String colorKeyword = data.colorKeyword
                    String commonModifiersKeyword = data.commonModifiersKeyword
                    String globalNotKeyword = data.globalNotKeyword
                    String modifiersKeyword = data.modifiersKeyword
                    WineKeyword wineKeyword = saveKeyword(wineDb, producerKeyword, baseKeyword, appellationKeyword, colorKeyword, commonModifiersKeyword, modifiersKeyword, null, globalNotKeyword, source)
                    list.push(wineKeyword)
                }
            }
        }
        return list
    }

    /**
     * Saves keyword
     * @param wineDb the wine db
     * @param producerKeyword the producer keyword
     * @param baseKeyword the base keyword
     * @param appellationKeyword the appellation keyword
     * @param colorKeyword the color keyword
     * @param commonModifiersKeyword the common modifier keyword
     * @param modifiersKeyword the modifier keyword
     * @param customNotKeyword the custom not keyword
     * @param globalNotKeyword the global not keyword
     * @param source the source data
     * @return the instance of {@link WineKeyword}
     */
    WineKeyword saveKeyword(WineDb wineDb, String producerKeyword, String baseKeyword, String appellationKeyword, String colorKeyword, String commonModifiersKeyword, String modifiersKeyword, String customNotKeyword, String globalNotKeyword, Source source) {
        WineKeyword wineKeyword = new WineKeyword()
        wineKeyword.wineDb = wineDb
        wineKeyword.producerKeyword = producerKeyword
        wineKeyword.baseKeyword = baseKeyword
        wineKeyword.appellationKeyword = appellationKeyword
        wineKeyword.colorKeyword = colorKeyword
        wineKeyword.commonModifiersKeyword = commonModifiersKeyword
        wineKeyword.modifiersKeyword = modifiersKeyword
        wineKeyword.customNotKeyword = customNotKeyword
        wineKeyword.globalNotKeyword = globalNotKeyword
        wineKeyword.pattern = WineKeywordUtils.joinPattern(wineKeyword)
        wineKeyword.source = source

        wineKeyword.validate()
        if (wineKeyword.hasErrors()) {
            return wineKeyword
        }
        wineKeyword.save()
        wineKeyword
    }
}
