package wine.alert.crawler

import org.apache.commons.collections.CollectionUtils
import org.apache.commons.lang3.StringUtils
import org.jsoup.Jsoup
import org.jsoup.parser.Parser
import wine.alert.crawler.request.CrawlElement
import wine.alert.crawler.request.CrawlEndpoint
import wine.alert.crawler.request.CrawlRequest
import wine.alert.crawler.response.IResponse
import wine.alert.crawler.response.Property
import wine.alert.crawler.rest.*

/**
 * An abstract implementation of crawler service.
 */
abstract class AbstractCrawler implements ICrawler {

    CrawlRequest request
    IRest rest

    IResponse crawl() {
        if (request.endpoint.isFile()) return getLocalFile(request.endpoint)
        if (request.endpoint.isWeb()) return getWeb(request.endpoint)
        if (request.endpoint.isFtp()) return getFtp(request.endpoint)
        if (request.endpoint.isSftp()) return getSftp(request.endpoint)
        if (request.endpoint.isS3()) return getS3(request.endpoint)
        return null
    }

    IResponse parse(IResponse response) {
        IResponse crawlResponse
        if (CollectionUtils.isNotEmpty(request.elements)) {
            crawlResponse = getFixedProps(response)
        } else {
            crawlResponse = getAllProps(response)
        }
        done()
        return crawlResponse
    }

    /**
     * Gets content form web
     * @param endpoint the instance of {@link CrawlEndpoint}
     * @return the content
     */
    protected IResponse getWeb(CrawlEndpoint endpoint) {
        rest = new WebRest(endpoint)
        return parse(rest.execute())
    }

    /**
     * Gets content form FTP server
     * @param endpoint the instance of {@link CrawlEndpoint}
     * @return the content
     */
    protected IResponse getFtp(CrawlEndpoint endpoint) {
        rest = new FtpRest(endpoint)
        return parse(rest.execute())
    }

    /**
     * Gets content form SFTP server
     * @param endpoint the instance of {@link CrawlEndpoint}
     * @return the content
     */
    protected IResponse getSftp(CrawlEndpoint endpoint) {
        rest = new SftpRest(endpoint)
        return parse(rest.execute())
    }

    /**
     * Gets content form s3 bucket
     * @param endpoint the instance of {@link CrawlEndpoint}
     * @return the content
     */
    protected IResponse getS3(CrawlEndpoint endpoint) {
        rest = new S3Rest(endpoint)
        return parse(rest.execute())
    }

    /**
     * Gets content form file
     * @param endpoint the instance of {@link CrawlEndpoint}
     * @return the content
     */
    protected IResponse getLocalFile(CrawlEndpoint endpoint) {
        rest = new FileRest(endpoint)
        return parse(rest.execute())
    }

    /**
     * Does necessary clean up
     */
    protected void done() {
        if (!rest) return
        rest.done()
    }

    /**
     * Checks whether the record should be ignored
     * @param properties the list of {@link Property}
     * @return Boolean
     */
    protected Boolean shouldIgnoreRecord(List<Property> properties) {
        if (CollectionUtils.isEmpty(properties)) return true
        Boolean hasIgnore = properties.count { it.ignore }.intValue() > 0
        if (hasIgnore) return true
        Integer blank = properties.count { StringUtils.isBlank(it.value) }.intValue()
        if (blank == properties.size()) return true
        Integer nonBlank = properties.count { StringUtils.isNotBlank(it.value) }.intValue()
        return request.minValueExists && request.minValueExists > nonBlank
    }

    /**
     * Prepares value
     * @param value the value
     * @param element the instance of {@link CrawlElement}
     * @return the instance of {@link Property}
     */
    protected static Property prepare(Object value, CrawlElement element) {
        String val = null
        if (value) {
            val = value instanceof String ? value : value.toString()
            if (element.isHtml) {
                val = Jsoup.parse("<div>" + val + "</div>", "", Parser.xmlParser()).text()
            }
            if (element.noises) {
                val = StringUtils.remove(val, element.noises)
            }
        }
        Property property = new Property(
                order: element.order,
                key: element.key,
                value: val,
                ignore: shouldIgnoreValue(val, element)
        )
        return property
    }

    /**
     * Checks whether the value should be ignored
     * @param value the value to be checked
     * @param element the instance of {@link CrawlElement}
     * @return the test result
     */
    private static Boolean shouldIgnoreValue(String value, CrawlElement element) {
        value = StringUtils.trim(value)
        Boolean ignore = false
        if (element.ignoreType == CrawlElement.IgnoreType.NOT_SET) return ignore

        if (element.ignoreType == CrawlElement.IgnoreType.BLANK) {
            if (StringUtils.isBlank(value)) {
                ignore = true
            }
        } else if (element.ignoreType == CrawlElement.IgnoreType.EQUALS && StringUtils.isNotBlank(element.ignoreValues)) {
            List<String> iValues = element.ignoreValues.tokenize("\n")
            for (int j = 0; j < iValues.size(); j++) {
                String iv = iValues[j]
                if (value.equalsIgnoreCase(iv)) {
                    ignore = true
                    break
                }
            }
        } else if (element.ignoreType == CrawlElement.IgnoreType.CONTAINS && StringUtils.isNotBlank(element.ignoreValues)) {
            List<String> iValues = element.ignoreValues.tokenize("\n")
            for (int j = 0; j < iValues.size(); j++) {
                String iv = iValues[j]
                if (value.contains(iv)) {
                    ignore = true
                    break
                }
            }
        }

        return ignore
    }
}
