package wine.alert.crawler

import org.apache.commons.io.IOUtils
import org.apache.http.MethodNotSupportedException
import wine.alert.crawler.request.CrawlRequest
import wine.alert.crawler.response.CommonResponse
import wine.alert.crawler.response.CrawlResponse
import wine.alert.crawler.response.IResponse

/**
 * Crawler that support dynamic parser.<br>
 * Support for json, xls, xlsx, xml and csv format.
 */
class DynamicCrawler extends AbstractCrawler {

    DynamicCrawler(CrawlRequest request) {
        this.request = request
    }

    IResponse parse(IResponse response) {
        if (!response || !response.inputStream) return null
        ByteArrayOutputStream baos = new ByteArrayOutputStream()
        IOUtils.copy(response.inputStream, baos)
        byte[] bytes = baos.toByteArray()
        ByteArrayInputStream bais = new ByteArrayInputStream(bytes)
        CommonResponse commonResponse = new CommonResponse(
                inputStream: bais,
                charset: response.charset
        )

        CrawlResponse crawlResponse = parseJson(commonResponse)
        if (!validate(crawlResponse)) {
            bais.reset()
            crawlResponse = parseExcel(commonResponse)
        }
        if (!validate(crawlResponse)) {
            // Try with the opposite of ooxml original value
            bais.reset()
            request.ooxml = !request.ooxml
            crawlResponse = parseExcel(commonResponse)
        }

        // XML parser should be after xlsx since xlsx is xml
        if (!validate(crawlResponse)) {
            bais.reset()
            crawlResponse = parseXml(commonResponse)
        }

        // CSV parser at the end, usually always return rows
        if (!validate(crawlResponse)) {
            bais.reset()
            crawlResponse = parseCsv(commonResponse)
        }
        bais.close()
        done()
        return crawlResponse
    }

    /**
     * Parses json format
     * @param response the instance of {@link CommonResponse}
     * @return the instance of {@link CrawlResponse}
     */
    CrawlResponse parseJson(CommonResponse response) {
        CrawlResponse crawlResponse = null
        try {
            crawlResponse = new JsonCrawler(request).parse(response) as CrawlResponse
        } catch (Exception ignored) {
        }
        return crawlResponse
    }

    /**
     * Parses Excel format
     * @param response the instance of {@link CommonResponse}
     * @return the instance of {@link CrawlResponse}
     */
    CrawlResponse parseExcel(CommonResponse response) {
        CrawlResponse crawlResponse = null
        try {
            crawlResponse = new ExcelCrawler(request).parse(response) as CrawlResponse
        } catch (Exception ignored) {
        }
        return crawlResponse
    }

    /**
     * Parses XML format
     * @param response the instance of {@link CommonResponse}
     * @return the instance of {@link CrawlResponse}
     */
    CrawlResponse parseXml(CommonResponse response) {
        CrawlResponse crawlResponse = null
        try {
            crawlResponse = new XmlCrawler(request).parse(response) as CrawlResponse
        } catch (Exception ignored) {
        }
        return crawlResponse
    }

    /**
     * Parses CSV format
     * @param response the instance of {@link CommonResponse}
     * @return the instance of {@link CrawlResponse}
     */
    CrawlResponse parseCsv(CommonResponse response) {
        CrawlResponse crawlResponse = null
        try {
            crawlResponse = new CsvCrawler(request).parse(response) as CrawlResponse
        } catch (Exception ignored) {
        }
        return crawlResponse
    }

    IResponse getFixedProps(IResponse response) {
        throw new MethodNotSupportedException("Method getFixedProps is not supported by DynamicCrawler")
    }

    IResponse getAllProps(IResponse response) {
        throw new MethodNotSupportedException("Method getAllProps is not supported by DynamicCrawler")
    }

    /**
     * Validates response
     * @param response the instance of {@link CrawlResponse}
     * @return Boolean
     */
    private static Boolean validate(CrawlResponse response) {
        return response && response.total >= 2
    }
}
