package wine.alert.crawler

import org.apache.poi.hssf.usermodel.HSSFWorkbook
import org.apache.poi.ss.usermodel.*
import org.apache.poi.xssf.usermodel.XSSFWorkbook
import wine.alert.crawler.request.CrawlElement
import wine.alert.crawler.request.CrawlRequest
import wine.alert.crawler.response.CrawlResponse
import wine.alert.crawler.response.IResponse
import wine.alert.crawler.response.Property

/**
 * Crawler for xls and xlsx file.
 */
class ExcelCrawler extends AbstractCrawler {

    ExcelCrawler(CrawlRequest request) {
        this.request = request
    }

    CrawlResponse getFixedProps(IResponse response) {
        if (!response || !response.inputStream) null
        CrawlResponse crawlResponse = new CrawlResponse()
        Integer i = 0
        Iterator<Row> iterator = getRows(response)
        while (iterator.hasNext()) {
            i++
            Row row = iterator.next()
            if (request.start && i < request.start) continue
            List<Property> props = new ArrayList<>()
            for (CrawlElement element : request.elements) {
                Cell cell = row.getCell(element.index, Row.MissingCellPolicy.CREATE_NULL_AS_BLANK)
                Property property = prepare(getCellValue(cell, element.isUrl), element)
                props.push(property)
            }
            if (!shouldIgnoreRecord(props)) crawlResponse.addRecord(props)
        }
        return crawlResponse
    }

    CrawlResponse getAllProps(IResponse response) {
        if (!response || !response.inputStream) null
        CrawlResponse crawlResponse = new CrawlResponse()
        Integer i = 0
        Iterator<Row> iterator = getRows(response)
        while (iterator.hasNext()) {
            i++
            Row row = iterator.next()
            if (request.start && i < request.start) continue
            List<Property> props = getProps(row)
            if (!shouldIgnoreRecord(props)) crawlResponse.addRecord(props)
        }
        return crawlResponse
    }

    /**
     * Ges rows from xsl or xlsx file
     * @param response the response
     * @return the row iterator
     */
    private Iterator<Row> getRows(IResponse response) {
        Workbook workbook = request.ooxml ? new XSSFWorkbook(response.inputStream) : new HSSFWorkbook(response.inputStream)
        Sheet sheet = workbook.getSheetAt(0)
        Iterator<Row> iterator = sheet.iterator()
        return iterator
    }

    /**
     * Gets props of a row
     * @param row the instance of {@link Row}
     * @return the list of {@link Property}
     */
    private static List<Property> getProps(Row row) {
        List<Property> props = new ArrayList<>()
        Iterator<Cell> cellIterator = row.cellIterator()
        while (cellIterator.hasNext()) {
            Cell cell = cellIterator.next()
            props.push(new Property(
                    value: getCellValue(cell)
            ))
        }
        return props
    }

    /**
     * Gets the cell value
     * @param cell the instance of {@link Cell}
     * @param isUrl the indicator of url value
     * @return String
     */
    private static String getCellValue(Cell cell, boolean isUrl = false) {
        String value = ""
        def cellType = cell.getCellType()
        if (!(cellType instanceof CellType)) cellType = cell.getCellTypeEnum()
        switch (cellType) {
            case CellType.BOOLEAN:
                value = cell.getBooleanCellValue()?.toString()
                break
            case CellType.STRING:
                if (isUrl) {
                    value = cell.getHyperlink()?.getAddress()
                } else {
                    value = cell.getRichStringCellValue()?.getString()
                }
                break
            case CellType.NUMERIC:
                if (DateUtil.isCellDateFormatted(cell)) {
                    value = cell.getDateCellValue()?.toString()
                } else {
                    value = cell.getNumericCellValue()?.toString()
                }
                break
            case CellType.FORMULA:
                value = cell.getRichStringCellValue()?.getString()
                break
            case CellType.BLANK:
                break
        }
        return value
    }
}
