package wine.alert.crawler.utils

import com.amazonaws.auth.AWSStaticCredentialsProvider
import com.amazonaws.auth.BasicAWSCredentials
import com.amazonaws.services.s3.AmazonS3
import com.amazonaws.services.s3.AmazonS3Client
import com.amazonaws.services.s3.AmazonS3ClientBuilder
import com.amazonaws.services.s3.AmazonS3URI
import com.amazonaws.services.s3.model.GetObjectRequest
import com.amazonaws.services.s3.model.ListObjectsV2Request
import com.amazonaws.services.s3.model.ListObjectsV2Result
import com.amazonaws.services.s3.model.S3ObjectInputStream
import com.amazonaws.services.s3.model.S3ObjectSummary
import groovy.transform.CompileStatic
import org.apache.commons.lang3.StringUtils

/**
 * A utility for aws s3 service
 */
@CompileStatic
class S3Utils {

    /**
     * Gets object urls
     * @param client the instance of {@link AmazonS3Client}
     * @param bucket the bucket name
     * @param prefix the prefix
     * @return the list of object key
     */
    static List<String> getObjectUrls(AmazonS3Client client, String bucket, String prefix) {
        List<String> keys = getObjectKeys(client, bucket, prefix)
        return keys.collect { StringUtils.join('s3://', bucket, '/', it) }
    }

    /**
     * Gets object keys
     * @param client the instance of {@link AmazonS3Client}
     * @param bucket the bucket name
     * @param prefix the prefix
     * @return the list of object key
     */
    static List<String> getObjectKeys(AmazonS3Client client, String bucket, String prefix) {
        List<String> keys = new ArrayList<>()
        ListObjectsV2Request req = new ListObjectsV2Request().withBucketName(bucket)
        if (StringUtils.isNotBlank(prefix)) req.withPrefix(prefix)
        ListObjectsV2Result result
        while (true) {
            result = client.listObjectsV2(req)
            for (S3ObjectSummary objectSummary : result.getObjectSummaries()) {
                keys.add(objectSummary.getKey())
            }
            // If there are more than maxKeys keys in the bucket, get a continuation token
            // and list the next objects.
            String token = result.getNextContinuationToken()
            req.setContinuationToken(token)
            if (!result.isTruncated()) break
        }
        return keys
    }

    /**
     * Downloads file from s3
     * @param client the s3 client
     * @param uri the s3 uri
     * @return the instance of {@link S3ObjectInputStream}
     */
    static S3ObjectInputStream download(AmazonS3Client client, AmazonS3URI uri) {
        return client.getObject(new GetObjectRequest(uri.bucket, uri.key)).getObjectContent()
    }

    /**
     * Gets Amazon s3 client
     * @param accessKeyId the access key id
     * @param secretAccessKey the secret access key
     * @param region the region
     * @return the instance of {@link com.amazonaws.services.s3.AmazonS3Client}
     */
    static AmazonS3Client getClient(String accessKeyId, String secretAccessKey, String region) {
        BasicAWSCredentials creds = new BasicAWSCredentials(accessKeyId, secretAccessKey)
        AmazonS3 client = AmazonS3ClientBuilder.standard()
                .withCredentials(new AWSStaticCredentialsProvider(creds))
                .withRegion(region).build()
        return (AmazonS3Client) client
    }
}
