package wine.alert.model

import enums.MatchedType
import enums.RollupType
import groovy.util.logging.Slf4j
import org.apache.commons.lang.StringUtils
import wine.alert.WebCrawler
import wine.db.BottleSize
import wine.db.Retailer
import wine.db.Vintage

/**
 * Model for rollup output.
 */
@Slf4j
class RollupOutput {
    String wineAlertId
    String wineId
    String vintage
    String bottleSize
    String retailer
    BigDecimal price
    String currency
    String taxNotes
    String wineUrl
    String actualRetailerDescription
    String overridePrice
    String auction
    String retailerName
    String address
    String city
    String state
    String zip
    String country
    String shipToCountry
    String phone
    String url
    Boolean listing
    Boolean download
    String yourCode
    BigDecimal equalizedPrice
    BigDecimal originalPrice // backup of original price

    /**
     * Create for RP Rollup
     * @param webCrawler
     * @param retailer
     * @param row
     * @param vintageMap
     * @param bottleSizeMap
     * @return
     */
    static RollupOutput create(WebCrawler webCrawler, Retailer retailer, String[] row, Map<String, Vintage> vintageMap, Map<String, BottleSize> bottleSizeMap, mongoId) {
        String wineId = row.length > 0 ? row[0]?.trim() : ""
        String match = row.length > 2 ? row[2]?.trim() : ""
        String price = row.length > 6 ? row[6]?.trim() : ""

        RollupOutput output = new RollupOutput()

        // Ignore if match != history
        if (!match.equalsIgnoreCase(MatchedType.HISTORY.value())) return null

        // Ignore records with wine alert id = x or z
        if (StringUtils.isBlank(wineId) || wineId.length() < 12) return null

        // Ignore wine with blank price if listing = true
        if (webCrawler?.listing && StringUtils.isBlank(price)) return null

        // Ignore if wine is not in database
        // Get mongo id (uuid) from database, ignore records with blank mongo id
        if (!mongoId) return null

        //A - (Wine Alert ID), get 9 chars from column A (Wine Id)
        output.wineAlertId = wineId?.substring(0, 9)

        //B - (Wine ID), put the mongo id
        output.wineId = mongoId

        //C - (Vintage), get 3 chars after wine alert id (9, 12) from column A and get the vintage from the database
        Vintage vintage = vintageMap.get(wineId?.substring(9, 12))
        output.vintage = vintage?.name ?: ""

        //D - (Bottle Size), get 2 last chars from the end of column A and get the bottle size from the database
        BottleSize bottleSize = bottleSizeMap.get(wineId.substring(wineId.length() - 2))
        output.bottleSize = bottleSize?.name ?: ""

        //E - (Retailer), get the code from filename
        output.retailer = webCrawler?.code ?: ""


        //F - (Price), get the price from column G (Price)
        if (price != null) {
            try {
                output.price = new BigDecimal(price)
            } catch (Exception e) {
                log.debug(e?.message)
            }
        }

        //G - (Currency), get the currency from config
        output.currency = webCrawler?.currencyCode ?: ""

        //H - (Tax Notes), get the tax notes from column H (tax Status)
        output.taxNotes = row.length > 7 ? row[7]?.trim() : ""

        //I - (Wine Url), get wine url from column I (URL). If the URL is blank, get the value from retailer table
        String inputUrl = row.length > 8 ? row[8]?.trim() : ""
        output.wineUrl = StringUtils.isNotBlank(inputUrl) ? inputUrl : retailer?.webUrl

        //J - (Actual Retailer Description), get the value from column J (Retailer Description)
        output.actualRetailerDescription = row.length > 9 ? row[9]?.trim() : ""

        //K - (Override Price), leave it blank
        output.overridePrice = ""

        //L - (Auction), leave it blank
        output.auction = ""

        //M - (Retailer Name)
        output.retailerName = webCrawler?.name ?: retailer?.name ?: webCrawler?.code ?: ""

        //N - (Address)
        output.address = webCrawler?.address ?: retailer?.address ?: ""

        //O - (City)
        output.city = webCrawler?.city ?: retailer?.city ?: ""

        //P - (State)
        output.state = webCrawler?.state ?: retailer?.state ?: ""

        //Q - (Zip)
        output.zip = webCrawler?.zip ?: retailer?.postalCode ?: ""

        //R - (Country)
        output.country = webCrawler?.country ?: retailer?.country?.name

        //S - (Ship to country)
        output.shipToCountry = webCrawler?.shipToCountry ?: retailer?.shipTo ?: ""

        //T - (Phone)
        output.phone = webCrawler?.phone ?: retailer?.phone ?: ""

        //U - (Url), get the value from retailer table
        output.url = retailer?.webUrl ?: ""

        //V - (Listing)
        output.listing = webCrawler?.listing

        //W - (Download)
        output.download = webCrawler?.download

        //X - (Your Code), get the value from column M (SKUName)
        output.yourCode = row.length > 12 ? row[12]?.trim() : ""

        return output
    }

    /**
     * Create for WineAlert rollup
     * @param webCrawler
     * @param retailer
     * @param row
     * @param vintageMap
     * @param bottleSizeMap
     * @return
     */
    static RollupOutput createForWineAlert(WebCrawler webCrawler, Retailer retailer, String[] row, Map<String, Vintage> vintageMap, Map<String, BottleSize> bottleSizeMap, String mongoId) {
        String wineId = row.length > 0 ? row[0]?.trim() : ""
        String price = row.length > 6 ? row[6]?.trim() : ""
        String bottleSizeStr = row.length > 4 ? row[4]?.trim() : ""
        String vintageStr = row.length > 5 ? row[5]?.trim() : ""

        // Ignore records with blank wine alert id or have
        //if (StringUtils.isBlank(wineId) || "?".equalsIgnoreCase(wineId)) return null

        // Do not ignore records with wine alert id = x or z, remove the x and z from id
        if (StringUtils.isBlank(wineId) || wineId.equalsIgnoreCase('?') || wineId.equalsIgnoreCase('x') || wineId.equalsIgnoreCase('z')) {
            wineId = ""
        }

        // Ignore wine with blank price
        if (StringUtils.isBlank(price)) return null

        RollupOutput output = new RollupOutput()

        //A - (Wine Alert ID), get 9 chars from column A (Wine Id)
        output.wineAlertId = wineId.length() > 9 ? wineId?.substring(0, 9) : wineId

        //B - (Wine ID), put the mongo id
        output.wineId = mongoId ?: ""

        //C - (Vintage), get 3 chars after wine alert id (9, 12) from column A and get the vintage from the database
        if (wineId && wineId.length() > 12) {
            Vintage vintage = vintageMap.get(wineId?.substring(9, 12))
            output.vintage = vintage?.name ?: ""
        } else { // If wine alert id is blank
            // Vintage prior to "1981" should be called "NV"
            Integer parsedV = vintageStr && vintageStr.isInteger() ? Integer.parseInt(vintageStr) : null
            if (parsedV && parsedV < 1981) {
                vintageStr = "NV"
            }
            output.vintage = vintageStr
        }

        // Ignore wine with blank vintage
        if (StringUtils.isBlank(output.vintage)) {
            return null
        }

        //D - (Bottle Size), get 2 last chars from the end of column A and get the bottle size from the database
        if (wineId && wineId.length() >= 2) {
            BottleSize bottleSize = bottleSizeMap.get(wineId.substring(wineId?.length() - 2))
            output.bottleSize = bottleSize?.name ?: ""
        } else {
            output.bottleSize = bottleSizeStr ?: ""
        }

        // Ignore wine with blank bottle size
        if (StringUtils.isBlank(output.bottleSize)) {
            return null
        }

        //E - (Retailer), get the code from filename
        output.retailer = webCrawler?.code ?: ""

        //F - (Price), get the price from column G (Price)
        try {
            output.price = new BigDecimal(price)
            output.originalPrice = new BigDecimal(price)
        } catch (Exception e) {
            log.debug(e?.message)
            return null
        }

        //G - (Currency), get the currency from config
        output.currency = webCrawler?.currencyCode ?: ""

        //H - (Tax Notes), get the tax notes from column H (tax Status)
        output.taxNotes = row.length > 7 ? row[7]?.trim() : ""

        //I - (Wine Url), get wine url from column I (URL). If the URL is blank, get the value from retailer table
        String inputUrl = row.length > 8 ? row[8]?.trim() : ""
        output.wineUrl = StringUtils.isNotBlank(inputUrl) ? inputUrl : retailer?.webUrl
        output.wineUrl = output.wineUrl ?: ""

        //J - (Actual Retailer Description), get the value from column J (Retailer Description)
        output.actualRetailerDescription = row.length > 9 ? row[9]?.trim() : ""

        //K - (Override Price), leave it blank
        output.overridePrice = ""

        //L - (Auction), leave it blank
        output.auction = ""

        //M - (Retailer Name)
        output.retailerName = webCrawler?.name ?: retailer?.name ?: webCrawler?.code ?: ""

        //N - (Address)
        output.address = webCrawler?.address ?: retailer?.address ?: ""

        //O - (City)
        output.city = webCrawler?.city ?: retailer?.city ?: ""

        //P - (State)
        output.state = webCrawler?.state ?: retailer?.state ?: ""

        //Q - (Zip)
        output.zip = webCrawler?.zip ?: retailer?.postalCode ?: ""

        //R - (Country)
        output.country = webCrawler?.country ?: webCrawler?.country ?: ""

        //S - (Ship to country)
        output.shipToCountry = webCrawler?.shipToCountry ?: retailer?.shipTo ?: ""

        //T - (Phone)
        output.phone = webCrawler?.phone ?: retailer?.phone ?: ""

        //U - (Url), get the value from retailer table
        output.url = retailer?.webUrl ?: ""

        //V - (Listing)
        output.listing = webCrawler?.listing

        //W - (Download)
        output.download = webCrawler?.download

        //X - (Your Code), get the value from column M (SKUName)
        output.yourCode = row.length > 12 ? row[12]?.trim() : ""

        return output
    }

    String toString(String separator, RollupType rollupType) {
        StringBuffer sb = new StringBuffer()
        sb.append(wineAlertId?:"").append(separator)
        sb.append(wineId?:"").append(separator)
        sb.append(vintage?:"").append(separator)
        sb.append(bottleSize?:"").append(separator)
        sb.append(retailer?:"").append(separator)
        sb.append(price?:"").append(separator)
        sb.append(currency?:"").append(separator)
        sb.append(taxNotes?:"").append(separator)
        sb.append(wineUrl?:"").append(separator)
        sb.append(actualRetailerDescription?:"").append(separator)
        sb.append(overridePrice?:"").append(separator)
        sb.append(auction?:"").append(separator)
        sb.append(retailerName?:"").append(separator)
        sb.append(address?:"").append(separator)
        sb.append(city?:"").append(separator)
        sb.append(state?:"").append(separator)
        sb.append(zip?:"").append(separator)
        sb.append(country?:"").append(separator)
        sb.append(shipToCountry?:"").append(separator)
        sb.append(phone?:"").append(separator)
        sb.append(url?:"").append(separator)
        sb.append(listing).append(separator)
        sb.append(download).append(separator)
        sb.append(yourCode?:"").append(separator)
        if (rollupType == RollupType.WA) {
            sb.append(equalizedPrice?:"")
        }
        return sb.toString()
    }
}
