package wine.alert.utils

import groovy.transform.CompileStatic
import org.apache.commons.lang.StringUtils

import java.text.NumberFormat
import java.text.ParseException

/**
 * Used to modify the crawler record.
 */
class CrawlPostProcessing {

    /**
     * Map of the normalize colors
     */
    private static final Map<String, String> COLOR_REPLACEMENT_MAP = new HashMap<String, String>() {{
        put("rot", "Red")
        put("rouge", "Red")
        put("tinto", "Red")
        put("rosé", "Rose")
        put("rosso", "Red")
        put("rosado", "Rose")
        put("rosato", "Rose")
        put("bianco", "White")
        put("blanc", "White")
        put("blanco", "White")
        put("branco", "White")
        put("weiss", "White")
    }}

    public String capitalize(String input, def dataMap) {
        return input.split().collect { it.capitalize() }.join(" ")
    }

    /**
     * Removes text from the given input.
     * @param inputText the text to be processed (current value)
     * @param dataMap the map of the current crawler record
     * @param searchText the text to be removed
     * @return the processed text
     */
    public String removeText(String inputText, def dataMap, String searchText) {
        if (searchText?.length() > 0) {
            //return StringUtils.replace(inputText, searchText, "")
            return inputText?.replaceAll("(?i)${searchText}", "")
        }

        // return the same as input
        return inputText
    }

    /**
     * Replace search text in the given input.
     * @param inputText the text to be processed (current value)
     * @param dataMap the map of the current crawler record
     * @param searchText the text to be replaced
     * @param replacement the replacement text
     * @return the processed text
     */
    public String replaceText(String inputText, def dataMap, String searchText, String replacement) {
        if (searchText?.length() > 0) {
            return inputText?.replaceAll("(?i)${searchText}", replacement?:"")
        }

        // return the same as input
        return inputText
    }

    /**
     * Replace search text in the input with whitespace
     * because dashboard will trim whitespace in args
     */
    public String replaceTextWithWhitespace(String input, def dataMap, String searchText) {
        return replaceText(input, dataMap, searchText, " ")
    }

    /**
     * Replace input text with another value.
     * @param inputText the text to be processed (current value)
     * @param dataMap the map of the current crawler record
     * @param replacement the replacement value
     * @return the new value
     */
    public String replaceWithText(String inputText, def dataMap, String replacement) {
        return replacement
    }

    /**
     * Replace input text with the value from the given index.
     * @param inputText the text to be processed (current value)
     * @param dataMap the map of the current crawler record
     * @param index the index of the new value
     * @return the new value
     */
    public String replaceWithIndex(String inputText, def dataMap, String index) {
        String concatText = ""
        dataMap?.eachWithIndex { it, i ->
            if (i == Integer.parseInt(index)) {
                concatText = it?.value ?: ""
            }
        }
        return concatText
    }

    /**
     * Replace input text with some number of chars (taken from left) of the given index value.
     * @param inputText the text to be processed (current value)
     * @param dataMap the map of the current crawler record
     * @param index the index of the new value
     * @param stringCount the number of chars to take
     * @return the new value
     */
    public String replaceWithIndexTakeLeft(String inputText, def dataMap, String index, String stringCount) {
        Integer integerCount = Integer.parseInt(stringCount)
        String indexValue = ""
        dataMap?.eachWithIndex { it, i ->
            if (i == Integer.parseInt(index)) {
                indexValue = it?.value ?: ""
            }
        }

        if (stringCount) {
            return indexValue?.take(integerCount)
        }

        return ""
    }

    /**
     * Replace input text with some number of chars (taken from right) of the given index value.
     * @param inputText the text to be processed (current value)
     * @param dataMap the map of the current crawler record
     * @param index the index of the new value
     * @param stringCount the number of chars to take
     * @return the new value
     */
    public String replaceWithIndexTakeRight(String inputText, def dataMap, String index, String stringCount) {
        Integer integerCount = Integer.parseInt(stringCount)
        String indexValue = ""
        dataMap?.eachWithIndex { it, i ->
            if (i == Integer.parseInt(index)) {
                indexValue = it?.value ?: ""
            }
        }

        if (stringCount && indexValue?.length() > 0) {
            return indexValue?.reverse()?.take(integerCount)?.reverse()
        }

        return ""
    }

    /**
     * Appends the given text to the end of the current value (input text).
     * @param inputText the text to be processed (current value)
     * @param dataMap the map of the current crawler record
     * @param text the text to be appended
     * @return the modified value
     */
    public String concatWithText(String inputText, def dataMap, String text) {
        return (inputText?:"") + (text ?: "")
    }

    /**
     * Appends the given text to the start of the current value (input text).
     * @param inputText the text to be processed (current value)
     * @param dataMap the map of the current crawler record
     * @param text the text to be appended
     * @return the modified value
     */
    public String concatReverseWithText(String inputText, def dataMap, String text) {
        return (text ?: "") + (inputText?:"")
    }

    /**
     * Appends the given given index value to the end of the current value (input text).
     * @param inputText the text to be processed (current value)
     * @param dataMap the map of the current crawler record
     * @param index the index of the data to be appended
     * @return the concatenated text
     */
    public String concatWithIndex(String inputText, def dataMap, String index) {
        return concatWithIndexAndText(inputText, dataMap, index, "")
    }

    /**
     * Appends the given given index value and text to the end of the current value (input text).
     * @param inputText the text to be processed (current value)
     * @param dataMap the map of the current crawler record
     * @param index the index of the data to be appended
     * @param text the text to be appended
     * @return the modified value
     */
    public String concatWithIndexAndText(String inputText, def dataMap, String index, String text) {
        String indexContent = ""
        dataMap?.eachWithIndex { it, i ->
            if (i == Integer.parseInt(index)) {
                indexContent = it?.value ?: ""
            }
        }
        return inputText || indexContent ? (inputText ?: "") + (indexContent ?: "") + (text ?: "") : inputText
    }

    /**
     * Appends the given text and the given index value to the end of the current value (input text).
     * @param inputText the text to be processed (current value)
     * @param dataMap the map of the current crawler record
     * @param text the text to be appended
     * @param index the index of the data to be appended
     * @return the modified value
     */
    public String concatWithTextAndIndex(String inputText, def dataMap, String text, String index) {
        String indexContent = ""
        dataMap?.eachWithIndex { it, i ->
            if (i == Integer.parseInt(index)) {
                indexContent = it?.value ?: ""
            }
        }
        return inputText || indexContent ? (inputText ?: "") + (text ?: "") + (indexContent ?: "") : inputText
    }

    /**
     * Appends the given index value to the end of the current value (input text) if the given index value is number.
     * @param inputText the text to be processed (current value)
     * @param dataMap the map of the current crawler record
     * @param index the index of the data to be appended
     * @return the modified value
     */
    String concatWithIndexIfIndexIsNumber(String inputText, def dataMap, String index) {
        String indexContent = ""
        dataMap?.eachWithIndex { it, i ->
            if (i == Integer.parseInt(index)) {
                indexContent = StringUtils.trim((String) it?.value)
            }
        }
        return (indexContent && indexContent.isNumber()) ? (inputText + indexContent) : inputText
    }

    /**
     * Replace input text with the new value if the given index value is equals to the given matcher.<br>
     * Or else, replace it with another value.
     * @param inputText the text to be processed (current value)
     * @param dataMap the map of the current crawler record
     * @param replacement the replacement value
     * @param index the index of the data to be compared with the given matcher
     * @param matcher the value to be compared with the given index value
     * @param eitherReplacement the either replacement
     * @return the modified value
     */
    public String replaceWithTextIfIndexEqualsEither(String inputText, def dataMap, String replacement, String index, String matcher, String eitherReplacement) {
        String newText = eitherReplacement?.trim()
        boolean isEquals = false
        dataMap?.eachWithIndex { it, i ->
            if (i == Integer.parseInt(index)) {
                String itemValue = it?.value?.trim()
                if (StringUtils.equalsIgnoreCase(matcher?.trim(), itemValue)) {
                    isEquals = true
                }
            }
        }
        if (isEquals) {
            newText = replacement?.trim()
        }

        return newText
    }

    /**
     * Takes some chars from input text. Start from first chars and stop before the given search text position.
     * @param inputText the text to be processed (current value)
     * @param dataMap the map of the current crawler record
     * @param searchText the text to to be searched
     * @return the modified value
     */
    public String findAndTakeEverythingBefore(String inputText, def dataMap, String searchText){
        if (searchText?.length() > 0) {
            Integer index = inputText?.indexOf(searchText)
            if (index > -1) {
                return inputText?.take(index)
            }
        }
        return inputText
    }

    /**
     * Takes some chars from input text. Start from the given search text position.
     * @param inputText the text to be processed (current value)
     * @param dataMap the map of the current crawler record
     * @param searchText the text to to be searched
     * @return the modified value
     */
    public String findAndTakeEverythingAfter(String inputText, def dataMap, String searchText){
        if (searchText?.length() > 0) {
            Integer index = inputText?.indexOf(searchText)
            if (index > -1) {
                index = index + searchText.length()
                return inputText?.drop(index)
            }
        }
        return inputText
    }

    /**
     * Replace input text with the given index value if the input text is blank.
     * @param inputText the text to be processed (current value)
     * @param dataMap the map of the current crawler record
     * @param index the index of replacement value
     * @return the modified value
     */
    public String replaceWithIndexIfBlank(String inputText, def dataMap, String index){
        String replacement = inputText
        if (StringUtils.isBlank(inputText)) {
            dataMap?.eachWithIndex { it, i ->
                if (i == Integer.parseInt(index) && it?.value) {
                    replacement = it?.value
                }
            }
        }
        return replacement
    }

    /**
     * Replace input text with the given index value if the input text is equals to zero.
     * @param inputText the text to be processed (current value)
     * @param dataMap the map of the current crawler record
     * @param index the index of replacement value
     * @return the modified value
     */
    public String replaceWithIndexIfLessEqualZero(String inputText, def dataMap, String index){
        String replacement = inputText
        try {
            BigDecimal b = new BigDecimal(inputText)
            if (b <= BigDecimal.ZERO) {
                dataMap?.eachWithIndex { it, i ->
                    if (i == Integer.parseInt(index) && it?.value) {
                        replacement = it?.value
                    }
                }
            }
        } catch (Exception e) {}
        return replacement
    }

    /**
     * Replace input text with the given index value if the input text is not blank.
     * @param inputText the text to be processed (current value)
     * @param dataMap the map of the current crawler record
     * @param index the index of replacement value
     * @return the modified value
     */
    String replaceWithIndexIfIndexIsNotBlank(String inputText, def dataMap, String index){
        String replacement = inputText
        dataMap?.eachWithIndex { it, i ->
            if (i == Integer.parseInt(index) && StringUtils.isNotBlank((String) it?.value)) {
                replacement = StringUtils.trim((String) it?.value)
            }
        }
        return replacement
    }

    /**
     * Divides current value with the given index value.
     * @param inputText the text to be processed (current value)
     * @param dataMap the map of the current crawler record
     * @param index the index of divider value
     * @return the modified value if the current value is number, else return blank.
     */
    public String divideByIndex(String inputText, def dataMap, String index){
        return divideByIndex(inputText, dataMap, index, "")
    }

    /**
     * Divides current value with the given index value.
     * @param inputText the text to be processed (current value)
     * @param dataMap the map of the current crawler record
     * @param index the index of divider value
     * @return the modified value if the current value is number, else return the given default value.
     */
    public String divideByIndex(String inputText, def dataMap, String index, String defaultValue){
        try {
            BigDecimal a = extractNumbers(inputText)
            BigDecimal b = BigDecimal.ZERO
            if (a >= BigDecimal.ZERO) {
                dataMap?.eachWithIndex { it, i ->
                    if (i == Integer.parseInt(index) && it?.value) {
                        b = extractNumbers(it?.value as String)
                    }
                }
                if (b > BigDecimal.ZERO) {
                    return (a/b).toString()
                }
            }
        } catch (Exception e) {}
        return defaultValue
    }

    /**
     * Extracts numbers from current value.<br>
     * Return the given default value if current value doesn't contains number.
     * @param inputText the text to be processed (current value)
     * @param dataMap the map of the current crawler record
     * @param defaultValue the default value
     * @return the integer value
     */
    public String parseAsInteger(String inputText, def dataMap, String defaultValue){
        try {
            return String.format("%.0f", extractNumbers(inputText))
        } catch (Exception e) {}
        return defaultValue ?: ""
    }

    /**
     * Normalizes color.
     * @param inputText the text to be processed (current value)
     * @param dataMap the map of the current crawler record
     * @return the normalized color
     */
    static String normalizeColor(String inputText, def dataMap) {
        if (StringUtils.isNotBlank(inputText)) {
            String key = inputText.trim().toLowerCase()
            if (COLOR_REPLACEMENT_MAP.containsKey(key)) {
                inputText = COLOR_REPLACEMENT_MAP.get(key)
            }
        }
        return inputText
    }

    /**
     * Extracts numbers from text.
     * @param input the input text
     * @return the number as BigDecimal
     */
    @CompileStatic
    public BigDecimal extractNumbers(String input) {
        if (StringUtils.isBlank(input)) return BigDecimal.ZERO

        BigDecimal d = BigDecimal.ZERO
        try {
            d = NumberFormat.getInstance(Locale.US).parse(input).toBigDecimal()
        } catch (ParseException e) {
            try {
                // Clear separator
                String parsed = input.find(/-?\d{1,3}(\,\d{3}){1,}\.\d*|-?\d+\.\d*|-?\d*\.\d+|-?\d+/)?.replaceAll(",", "")
                if (StringUtils.isNotBlank(parsed)) {
                    d = new BigDecimal(parsed)
                }
            } catch (Exception e1) { }
        }
        return d
    }
}
