/**
 * Utility class for crawler
 */

package wine.alert.utils

import grails.compiler.GrailsCompileStatic
import grails.config.Config
import grails.core.support.GrailsConfigurationAware
import groovy.util.logging.Slf4j
import org.apache.commons.lang3.StringUtils
import wine.alert.crawler.response.CrawlResponse
import wine.alert.crawler.response.Property

import java.nio.file.Files
import java.nio.file.Path
import java.nio.file.Paths
import java.nio.file.attribute.PosixFilePermission
import java.text.SimpleDateFormat
import java.util.regex.Pattern

@Slf4j
@GrailsCompileStatic
class CrawlerUtils implements GrailsConfigurationAware {

    private static final Set<PosixFilePermission> DEF_FILE_PERMISSIONS = new HashSet<PosixFilePermission>() {
        {
            add(PosixFilePermission.OWNER_READ)
            add(PosixFilePermission.OWNER_WRITE)
            add(PosixFilePermission.OWNER_EXECUTE)
            add(PosixFilePermission.GROUP_READ)
            add(PosixFilePermission.GROUP_WRITE)
            add(PosixFilePermission.GROUP_EXECUTE)
            add(PosixFilePermission.OTHERS_READ)
        }
    }

    private static String outputLocation
    private static String backupLocation

    void setConfiguration(Config config) {
        outputLocation = config.getProperty('application.crawlOut.location')
        backupLocation = config.getProperty('application.crawlOut.backupLocation')
    }

    /**
     * Saves a crawl response to a file
     * @param response the instance of {@link CrawlResponse}
     * @param retailerCode the retailer code
     */
    static void saveResult(CrawlResponse response, String retailerCode) {
        if (!response) return
        def sdf = new SimpleDateFormat("yyyy-MM-dd")
        String fileName = StringUtils.join(retailerCode, "-", sdf.format(new Date()), ".txt")
        Path filepath = Paths.get(outputLocation, fileName)
        backupPriorResults(retailerCode)
        deleteExistingFile(filepath.toString())
        FileWriter fw = null
        try {
            fw = new FileWriter(filepath.toString(), true)
            for (List<Property> props in response.getRecords()) {
                fw.println(props.collect { it.value }?.join("|"))
            }
        } catch (Exception e) {
            log.error(e.message, e)
        } finally {
            if (fw != null) {
                fw.close()
            }
        }

        try {
            Files.setPosixFilePermissions(filepath, DEF_FILE_PERMISSIONS)
        } catch (Exception e) {
            log.warn(e.toString())
        }
    }

    /**
     * Backups prior results
     * @param retailerCode the retailer code
     */
    static void backupPriorResults(String retailerCode) {
        if (StringUtils.isBlank(retailerCode)) return
        try {
            String patternString = ".*/" + retailerCode + "-.*\\.txt"
            Pattern p = Pattern.compile(patternString)
            File dir = new File(backupLocation)
            dir.mkdirs()
            dir.eachFile() { file ->
                String fileName = file.getAbsolutePath()
                if (fileName.find(p)) {
                    def destination = Paths.get(dir.absolutePath, file.getName())
                    boolean fileMoved = file.renameTo(new File(destination.toString()))
                    if (fileMoved) {
                        Files.setPosixFilePermissions(destination, DEF_FILE_PERMISSIONS)
                    } else {
                        log.warn("Failed to move file: " + fileName)
                    }
                }
            }
        } catch (Exception e) {
            log.error(e?.message, e)
        }
    }

    /**
     * Deletes existing file of a given path
     * @param fullPath the full path
     */
    private static void deleteExistingFile(String fullPath) {
        File file = new File(fullPath)
        if (file.isFile()) file.delete()
    }
}
