package wine.db.utils

import org.apache.commons.lang3.RegExUtils
import org.apache.commons.lang3.StringUtils
import wine.db.WineKeyword

import java.util.regex.Pattern

/**
 * Utility to process wine keyword {@link WineKeyword}.
 */
class WineKeywordUtils {

    private static final String STR_AND = "&"
    private static final String STR_OR = "|"
    private static final Pattern PATTERN_DUPLICATE_ANDS = Pattern.compile("&+")
    private static final Pattern PATTERN_DUPLICATE_ORS = Pattern.compile("\\|+")

    /**
     * Join patterns of wine keyword to produce pattern {@link WineKeyword#pattern}<br>
     * Keyword/pattern will be constructed with these following rules:<br>
     * producerKeyword : 1x<br>
     * baseKeyword : 1x<br>
     * appellationKeyword : 1x<br>
     * colorKeyword : 1x<br>
     * commonModifiersKeyword : 2x<br>
     * modifiersKeyword : 6x<br>
     * customNotKeyword : 1x<br>
     * globalNotKeyword : 1x
     * @param wineKeyword the wine keyword instance {@link WineKeyword}
     * @return the joined pattern/keyword
     */
    static joinPattern = { WineKeyword wineKeyword ->
        if (!wineKeyword) return null
        StringBuilder pattern = new StringBuilder()

        if (wineKeyword.producerKeyword) pattern.append(wineKeyword.producerKeyword?.trim())
        if (wineKeyword.baseKeyword) {
            pattern.append(STR_AND).append(wineKeyword.baseKeyword?.trim())
        }
        if (wineKeyword.appellationKeyword) {
            pattern.append(STR_AND).append(wineKeyword.appellationKeyword?.trim())
        }
        if (wineKeyword.colorKeyword) pattern.append(wineKeyword.colorKeyword?.trim())
        if (wineKeyword.commonModifiersKeyword) {
            pattern.append(wineKeyword.commonModifiersKeyword?.trim())
            pattern.append(wineKeyword.commonModifiersKeyword?.trim())
        }
        if (wineKeyword.modifiersKeyword) {
            pattern.append(wineKeyword.modifiersKeyword?.trim())
            pattern.append(wineKeyword.modifiersKeyword?.trim())
            pattern.append(wineKeyword.modifiersKeyword?.trim())
            pattern.append(wineKeyword.modifiersKeyword?.trim())
            pattern.append(wineKeyword.modifiersKeyword?.trim())
            pattern.append(wineKeyword.modifiersKeyword?.trim())
        }
        if (wineKeyword.customNotKeyword) pattern.append(wineKeyword.customNotKeyword?.trim())
        if (wineKeyword.globalNotKeyword) pattern.append(wineKeyword.globalNotKeyword?.trim())

        return cleanup(pattern.toString())
    }

    /**
     * Join patterns of wine keyword to produce pattern {@link WineKeyword#pattern}<br>
     * Keyword/pattern will be constructed with these following rules:<br>
     * producerKeyword : 1x<br>
     * baseKeyword : 1x<br>
     * appellationKeyword : 1x<br>
     * colorKeyword : 1x<br>
     * commonModifiersKeyword : 2x<br>
     * modifiersKeyword : 6x<br>
     * customNotKeyword : 1x<br>
     * globalNotKeyword : 1x
     * @param wineKeyword the wine keyword instance {@link WineKeyword}
     * @return the joined pattern/keyword
     */
    static joinPatternFromStrings = { String producerKeyword, String baseKeyword, String appellationKeyword,
                                      String colorKeyword, String commonModifiersKeyword, String modifiersKeyword, String customNotKeyword, String globalNotKeyword ->
        StringBuilder pattern = new StringBuilder()

        if (producerKeyword) pattern.append(producerKeyword?.trim())
        if (baseKeyword) {
            pattern.append(STR_AND).append(baseKeyword?.trim())
        }
        if (appellationKeyword) {
            pattern.append(STR_AND).append(appellationKeyword?.trim())
        }
        if (colorKeyword) pattern.append(colorKeyword?.trim())
        if (commonModifiersKeyword) {
            pattern.append(commonModifiersKeyword?.trim())
            pattern.append(commonModifiersKeyword?.trim())
        }
        if (modifiersKeyword) {
            pattern.append(modifiersKeyword?.trim())
            pattern.append(modifiersKeyword?.trim())
            pattern.append(modifiersKeyword?.trim())
            pattern.append(modifiersKeyword?.trim())
            pattern.append(modifiersKeyword?.trim())
            pattern.append(modifiersKeyword?.trim())
        }
        if (customNotKeyword) pattern.append(customNotKeyword?.trim())
        if (globalNotKeyword) pattern.append(globalNotKeyword?.trim())

        return cleanup(pattern.toString())
    }

    /**
     * Cleans duplicates "&" and "|"
     * @param keyword
     * @return the clean keyword
     */
    private static String cleanup(String keyword) {
        keyword = RegExUtils.replaceAll(keyword, PATTERN_DUPLICATE_ANDS, STR_AND)
        keyword = RegExUtils.replaceAll(keyword, PATTERN_DUPLICATE_ORS, STR_OR)
        keyword = StringUtils.strip(keyword, STR_AND)
        keyword = StringUtils.strip(keyword, STR_OR)
        return keyword
    }
}
