package wine.alert;

import org.apache.commons.lang3.RegExUtils;
import org.apache.commons.lang3.StringUtils;
import org.jsoup.Jsoup;

import javax.script.ScriptEngine;
import javax.script.ScriptEngineManager;
import java.math.BigDecimal;
import java.text.NumberFormat;
import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Core utility for keyword match.<br>
 * Sample of retailer's wine name = "Adam Gewurztraminer Cuvee Jean Baptiste Kaefferkopf"<br>
 * Sample of pattern = "adam&nbsp;jean&nbsp;baptiste&nbsp;gewur_&nbsp;kaef_&nbsp;cuvee&nbsp;kaef_&nbsp;cuvee&nbsp;kaef_&nbsp;cuvee&nbsp;kaef_&nbsp;cuvee&nbsp;!red&nbsp;!rouge&nbsp;!rose&nbsp;!reserve&nbsp;!sgn&nbsp;!vt&nbsp;!vv"
 */
public class Calculator {

    public static final String WILDCARD_SYMBOL = "_";

    private static final int INDEX_NOT_FOUND = -1;
    private static final String REPLACEMENT_FOUND = ":1:";
    private static final String REPLACEMENT_NOT_FOUND = ":0:";
    private static final String STR_TRUE = "1";
    private static final String STR_FALSE = "0";
    private static final String STR_NOT = "!";
    private static final String STR_SPACE = " ";
    private static final String STR_BLANK = "";
    private static final String NON_VINTAGE = "NV";
    private static final String PATTERN_DELIMITER = ",";
    private static final String PATTERN_OR = "|";
    private static final String PATTERN_AND = "&";
    private static final String LOGIC_OR = "||";
    private static final String LOGIC_AND = "&&";
    private static final String REGEX_EXCLUDE_PREFIX = "(?<![:])";
    private static final String REGEX_EXCLUDE_SUFFIX = "(?![:])";

    private static final Pattern PATTERN_NOT_REMOVER = Pattern.compile("(![^&|!()]+(_)?)|(&![^&|!()]+(_)?)|(\\|![^&|!()]+(_)?)|(&!\\(+(.*?)\\)+)|(\\|!\\(+(.*?)\\)+)");
    private static final Pattern PATTERN_NON_NUMERIC = Pattern.compile("[^\\d]");

    private static final Pattern PATTERN_MULTIPLE_SPACE = Pattern.compile("\\s+");
    private static final Pattern PATTERN_ASCII = Pattern.compile("\\A\\p{ASCII}*\\z");
    private static final Pattern PATTERN_PRICE_CLEANER = Pattern.compile("[\\p{Punct}&&[^.,]]");
    private static final Pattern PATTERN_PRICE_1 = Pattern.compile("^(.*?)([0-9.,]+)$"); // $908.273,99 // $908,273.99
    private static final Pattern PATTERN_PRICE_2 = Pattern.compile("^([0-9.,]+)(.*)$"); // 8192.99 USD // 8192,99 USD

    private static final Pattern PATTERN_VINTAGE_PARSER = Pattern.compile("(^|\\b|\\p{Punct}|\\s)(17|18|19|20)\\d{2}(\\s|\\p{Punct}|\\b|$)");
    private static final Pattern PATTERN_NV_PARSER = Pattern.compile("(?i)(^|\\b|\\p{Punct}|\\s)n\\.?v\\.?(\\s|\\p{Punct}|\\b|$)");

    private static final Pattern PATTERN_PUNCT = Pattern.compile("\\p{Punct}");
    private static final Pattern PATTERN_CUSTOM_PUNCT_1 = Pattern.compile("[!@#$%&?()\\[\\]{}⟨⟩<>“”\",:;*]"); // To be removed
    private static final Pattern PATTERN_CUSTOM_PUNCT_2 = Pattern.compile("(?!\\+)[./´`‘’'\\-‒–—―]+"); // To be replaced with space

    private static final Pattern PATTERN_SPACES = Pattern.compile("[\t\\xA0\\u1680\\u180e\\u2000-\\u200a\\u202f\\u205f\\u3000]"); // All space chars

    public Calculator() {
        super();
    }

    /**
     * Gets score of the retailer's wine name based on particular wine.
     * @param name the wine name, should be cleaned up first with {@link #cleanup(String)}
     * @param pattern the pattern/keyword
     * @param normalizedSortedPattern the normalized pattern (sorted &nbsp; unique), defined by {@link #normalizeAndSortPattern(String)}
     * @param normalizedPattern the normalized pattern, defined by {@link #normalizePattern(String)}
     * @param latestBestScore the latest best score form this method, or 0 for first run
     * @param requiredWords the required words in wine name, defined by {@link #requiredWords(String, boolean)}
     * @return The score if greater latestBestScore, OR 0 (zero) for the opposite
     */
    public static int getScore(String name, String pattern, String normalizedSortedPattern, String normalizedPattern, int latestBestScore, List<String> requiredWords) {
        int score = 0;
        if (StringUtils.isBlank(name) || StringUtils.isBlank(pattern)) return score;
        if (requiredWords != null && requiredWords.size() > 0 && !areWordsInString(requiredWords, name)) return score;

        int foundTotal = 0;
        int notFoundTotal = 0;
        String nameToBeProcessed = STR_SPACE + name + STR_SPACE;
        String[] allWords = normalizedPattern.split(PATTERN_DELIMITER);
        String[] uniqueWords = normalizedSortedPattern.split(PATTERN_DELIMITER);

        // Mapping word and it count in pattern
        Map<String, Integer> wordCountMap = new HashMap<>();
        for (String w: uniqueWords) {
            int s = 0;
            for (String aw: allWords) {
                if (w.equals(aw)) s += 1;
            }
            wordCountMap.put(w, s);
        }

        // Calculate Score
        for (String w: uniqueWords) {

            boolean found = false;
            if (StringUtils.endsWith(w, WILDCARD_SYMBOL)) {
                String strRegex = "\\s" + Pattern.quote(StringUtils.removeEnd(w, WILDCARD_SYMBOL)) + ".*?\\s";
                if (indexOf(Pattern.compile(strRegex), nameToBeProcessed) != INDEX_NOT_FOUND) {
                    found = true;
                }
            } else if (StringUtils.indexOf(nameToBeProcessed, STR_SPACE + w + STR_SPACE) != INDEX_NOT_FOUND){
                found = true;
            }

            if (found) {
                score += wordCountMap.get(w);
                foundTotal += wordCountMap.get(w);
                if (w.length() == 1 && (w.equals(STR_TRUE) || w.equals(STR_FALSE))) {
                    String search = REGEX_EXCLUDE_PREFIX + w + REGEX_EXCLUDE_SUFFIX;
                    pattern = RegExUtils.replaceAll(pattern, search, REPLACEMENT_FOUND);
                } else {
                    pattern = StringUtils.replace(pattern, w, REPLACEMENT_FOUND, wordCountMap.get(w));
                }
            } else {
                notFoundTotal += wordCountMap.get(w);
                if (w.length() == 1 && (w.equals(STR_TRUE) || w.equals(STR_FALSE))) {
                    String search = REGEX_EXCLUDE_PREFIX + w + REGEX_EXCLUDE_SUFFIX;
                    pattern = RegExUtils.replaceAll(pattern, search, REPLACEMENT_NOT_FOUND);
                } else {
                    pattern = StringUtils.replace(pattern, w, REPLACEMENT_NOT_FOUND, wordCountMap.get(w));
                }
            }
        }

        // Evaluate Pattern
        if (latestBestScore > score) {
            score = 0;
        } else {
            pattern = StringUtils.replace(pattern, REPLACEMENT_FOUND, STR_TRUE, foundTotal);
            pattern = StringUtils.replace(pattern, REPLACEMENT_NOT_FOUND, STR_FALSE, notFoundTotal);
            pattern = StringUtils.replace(pattern, PATTERN_AND, LOGIC_AND);
            pattern = StringUtils.replace(pattern, PATTERN_OR, LOGIC_OR);
            if (!eval(pattern)) {
                score = 0;
            }
        }
        return score;
    }

    /**
     * Evaluate logical string with script engine.<br>
     * Currently use JavaScript engine from {@link ScriptEngineManager}
     * @param logical  Logical string
     * @return true / false
     */
    private static boolean eval(String logical) {
        boolean isPassed;
        ScriptEngineManager sem = new ScriptEngineManager();
        ScriptEngine se = sem.getEngineByName("JavaScript");
        try {
            Object eval = se.eval(logical);
            if (eval instanceof Integer) {
                isPassed = (Integer) eval > 0;
            } else {
                isPassed = (Boolean) eval;
            }
        } catch (Exception e) {
            isPassed = false;
        }
        return isPassed;
    }

    /**
     * Check if the given words are in the given text.
     * @param requiredWords the words to be searched
     * @param text the searchable text
     * @return true if all words are in the given text, else false
     */
    private static boolean areWordsInString(List<String> requiredWords, String text) {
        boolean isPassed = true;
        if (StringUtils.isNotBlank(text) && requiredWords != null && !requiredWords.isEmpty()) {
            for (String s: requiredWords) {
                if (!StringUtils.contains(text, s)) {
                    isPassed = false;
                    break;
                }
            }
        }
        return isPassed;
    }

    /**
     * Gets all required words from the given pattern.<br>
     * Required words are words that don't start with "!".<br>
     * Will remove underscore/wildcard in words.
     * @param pattern the pattern/keyword
     * @param removeWildcard the option to remove wildcard "_"
     * @return the list of possible required words (not all, since words inside bracket will be skipped)
     */
    public static List<String> requiredWords(String pattern, boolean removeWildcard) {
        List<String> words =  new ArrayList<>();
        pattern = removeNots(pattern);
        if (StringUtils.isBlank(pattern)) return words;
        String[] allMatches = pattern.split("&(?![^(]*[)])"); // Split with &, skip if inside bracket
        for (String s: allMatches) {
            if (StringUtils.isNotBlank(s) && !StringUtils.startsWith(s, "(") && !words.contains(s)) {
                if (removeWildcard) s = StringUtils.replace(s, WILDCARD_SYMBOL, STR_BLANK);
                words.add(s);
            }
        }
        return words;
    }

    /**
     * Gets all true positive words from the given pattern.<br>
     * True positive words are words that don't start with "!".
     * @param pattern the pattern/keyword
     * @return the list of all true positive words
     */
    public static List<String> truePositiveWords(String pattern) {
        return truePositiveWords(pattern, false);
    }

    /**
     * Gets all true positive words from the given pattern.<br>
     * True positive words are words that don't start with "!".
     * @param pattern the pattern/keyword
     * @param allowDuplicates the unique indicator. Should return unique list or not
     * @return the list of all true positive words
     */
    public static List<String> truePositiveWords(String pattern, Boolean allowDuplicates) {
        List<String> words = new ArrayList<>();
        pattern = removeNots(pattern);
        if (StringUtils.isBlank(pattern)) return words;
        String[] ws = pattern.split("[&|()]");
        for (String w: ws) {
            if (StringUtils.isNotBlank(w) && !w.startsWith(STR_NOT)) {
                if (!words.contains(w) || allowDuplicates) {
                    words.add(w);
                }
            }
        }
        return words;
    }

    /**
     * Extract vintage from text
     * @param text the text to be processed
     * @return the first found vintage
     */
    public static String parseVintage(String text) {
        if (StringUtils.isBlank(text)) return null;

        String vintage = null;
        Matcher m = PATTERN_VINTAGE_PARSER.matcher(text);
        if (m.find()) {
            vintage = m.group().trim();
            vintage = RegExUtils.removeAll(vintage, PATTERN_NON_NUMERIC);
        }
        if (StringUtils.isBlank(vintage)) { // always search year before nv
            Matcher mn = PATTERN_NV_PARSER.matcher(text);
            if (mn.find()) vintage = NON_VINTAGE;
        }
        return vintage;
    }

    /**
     * Removes html numeric code like &#34; (quotation mark) and &#39; (apostrophe)
     * @param text  text to be processed
     * @return the processed text
     */
    public static String cleanupNumericHtml(String text) {
        if (StringUtils.isNotBlank(text)) {
            text = Jsoup.parse("<p>" + text + "</p>").text();
        }
        return text;
    }

    /**
     * Replace all non-standard spaces (include \t) with standard space
     * @param text  text to be processed
     * @return the processed text
     */
    public static String cleanupSpaces(String text) {
        if (StringUtils.isNotBlank(text)) {
            text = RegExUtils.replaceAll(text, PATTERN_SPACES, STR_SPACE);
            text = RegExUtils.replaceAll(text, PATTERN_MULTIPLE_SPACE, STR_SPACE);
            text = text.trim();
        }
        return text;
    }

    /**
     * Removes punctuations from the given text.
     * @param text the text to be processed
     * @return the processed text
     */
    public static String cleanup(String text) {
        if (StringUtils.isNotBlank(text)) {
            if (!isAscii(text)) {
                text = StringUtils.stripAccents(text);
            }
            text = text.toLowerCase();
            text = RegExUtils.replaceAll(text, PATTERN_CUSTOM_PUNCT_1, STR_BLANK);
            text = RegExUtils.replaceAll(text, PATTERN_CUSTOM_PUNCT_2, STR_SPACE);
            text = RegExUtils.replaceAll(text, PATTERN_MULTIPLE_SPACE, STR_SPACE);
            text = text.trim();
        }
        return text;
    }

    /**
     * Removes vintages from the given text.
     * @param inputText the text to be processed
     * @return the processed text
     */
    public static String removeVintages(String inputText) {
        if (StringUtils.isNotBlank(inputText)) {
            inputText = RegExUtils.replaceAll(inputText, PATTERN_VINTAGE_PARSER, STR_SPACE);
            inputText = RegExUtils.replaceAll(inputText, PATTERN_NV_PARSER, STR_SPACE);
            inputText = RegExUtils.replaceAll(inputText, PATTERN_MULTIPLE_SPACE, STR_SPACE);
            inputText = inputText.trim();
        }
        return inputText;
    }

    /**
     * Removes vintages from the given text. Exclude if pattern contains the vintage.
     * @param inputText the text to be processed
     * @param pattern the pattern/keyword
     * @return the processed text
     */
    public static String removeVintages(String inputText, String pattern) {
        if (StringUtils.isNotBlank(inputText)) {
            List<String> reserved = extractAllVintages(pattern);
            if (reserved != null && !reserved.isEmpty()) {
                List<String> tobeRemoved = extractAllVintages(inputText);
                if (tobeRemoved != null && !tobeRemoved.isEmpty()) {
                    for (String s: tobeRemoved) {
                        if (!reserved.contains(s)) {
                            inputText = StringUtils.replace(inputText, s, STR_SPACE);
                        }
                    }
                }
            } else {
                inputText = RegExUtils.replaceAll(inputText, PATTERN_VINTAGE_PARSER, STR_SPACE);
            }
            inputText = RegExUtils.replaceAll(inputText, PATTERN_NV_PARSER, STR_SPACE);
            inputText = RegExUtils.replaceAll(inputText, PATTERN_MULTIPLE_SPACE, STR_SPACE);
            inputText = inputText.trim();
        }
        return inputText;
    }

    /**
     * Tests if the given text contains only ASCII chars.
     * @param text the text to be processed
     * @return the test results
     */
    public static boolean isAscii(String text) {
        Matcher m = PATTERN_ASCII.matcher(text);
        return m.matches();
    }

    /**
     * Break and rejoin pattern with {@link #PATTERN_DELIMITER}.
     * @param pattern the pattern/keyword
     * @return normalized pattern's words
     */
    public static String normalizePattern(String pattern) {
        if (StringUtils.isBlank(pattern)) {
            return pattern;
        }
        pattern = pattern.replaceAll("[!()&|]+", PATTERN_DELIMITER).replaceAll(",+", PATTERN_DELIMITER);
        pattern = StringUtils.strip(pattern, PATTERN_DELIMITER);
        return pattern;
    }

    /**
     * Break, sort, rejoin pattern with {@link #PATTERN_DELIMITER}.
     * @param pattern the pattern/keyword
     * @return normalized pattern's words (sorted &nbsp; unique)
     */
    public static String normalizeAndSortPattern(String pattern) {
        if (StringUtils.isBlank(pattern)) {
            return pattern;
        }
        pattern = normalizePattern(pattern);
        String[] words = pattern.split(PATTERN_DELIMITER);
        String[] unique = new HashSet<>(Arrays.asList(words)).toArray(new String[0]);
        Arrays.sort(unique, (a, b) -> Integer.compare(b.length(), a.length())); // Sort by length, descending
        return String.join(PATTERN_DELIMITER, unique);
    }

    /**
     * Money parser, Locale.US represent number with format 88.99<br>
     * Locale.GERMANY represent number with format 88,99
     * @param text the text to be parsed
     * @return the parser result as BigDecimal
     */
    public static BigDecimal parseMoney(String text) {
        BigDecimal price = parseMoney(text, Locale.US);
        if (price == null) {
            price = parseMoney(text, Locale.GERMANY);
        }
        return price;
    }

    /**
     * Parse money with the given locale.
     * @param text the text to be parsed
     * @param locale the locale {@link Locale}
     * @return the parser result as BigDecimal
     */
    private static BigDecimal parseMoney(String text, Locale locale) {
        if (StringUtils.isBlank(text)) return null;
        try {
            String amount = null;
            text = RegExUtils.replaceAll(text, PATTERN_PRICE_CLEANER, "");
            Matcher matcher = PATTERN_PRICE_1.matcher(text);
            if (matcher.find()) {
                amount = matcher.group(2);
            }

            if (amount == null) {
                matcher = PATTERN_PRICE_2.matcher(text);
                if (matcher.find()) {
                    amount = matcher.group(1);
                }
            }

            if (amount != null) {
                Number number = NumberFormat.getInstance(locale).parse(amount);
                if ((number instanceof Long) || (number instanceof Integer) || (number instanceof Short) || (number instanceof Byte)) {
                    return BigDecimal.valueOf(number.longValue());
                } else {
                    return new BigDecimal(number.toString());
                }
            }
        } catch (Exception e) {
            // Nothing
        }

        return null;
    }

    /**
     * Removes nots from the given text.
     * @param text the pattern/keyword
     * @return the processed string
     */
    private static String removeNots(String text) {
        if (StringUtils.isBlank(text)) return null;
        return RegExUtils.replaceAll(text, PATTERN_NOT_REMOVER, STR_BLANK);
    }

    /**
     * Get string position on the given text.
     * @param pattern the search pattern {@link Pattern}
     * @param s the searchable text
     * @return index of the given pattern in s, or -1 if not found
     */
    private static int indexOf(Pattern pattern, String s) {
        Matcher matcher = pattern.matcher(s);
        return matcher.find() ? matcher.start() : INDEX_NOT_FOUND;
    }

    /**
     * Extracts all 4 digits vintages from the given text
     * @param text the text to be processed
     * @return list of vintages
     */
    private static List<String> extractAllVintages(String text) {
        List<String> digits = new ArrayList<>();
        if (StringUtils.isNotBlank(text)) {
            Matcher m = PATTERN_VINTAGE_PARSER.matcher(text);
            while (m.find()) {
                digits.add(RegExUtils.replaceAll(m.group().trim(), PATTERN_PUNCT, STR_BLANK));
            }
        }
        return digits;
    }
}
