package wine.alert;

import org.openqa.selenium.JavascriptExecutor;
import org.openqa.selenium.WebDriver;
import org.openqa.selenium.support.ui.ExpectedCondition;
import org.openqa.selenium.support.ui.WebDriverWait;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Javascript waiter for {@link WebDriver}
 */
public class JSWaiter {

    private static Logger log = LoggerFactory.getLogger(JSWaiter.class);

    private static WebDriver jsWaitDriver;
    private static WebDriverWait jsWait;
    private static JavascriptExecutor jsExec;

    /**
     * Sets the driver from relevant test
     * @param driver the web driver instance
     * @param wait the web driver wait
     */
    public static void setDriver (WebDriver driver, WebDriverWait wait) {
        jsWaitDriver = driver;
        jsWait = wait;
        jsExec = (JavascriptExecutor) jsWaitDriver;
    }

    /**
     * Waits for JQuery Load
     * @param displayLog the option whether to display the log or not
     */
    public static void waitForJQueryLoad(boolean displayLog) {
        //Wait for jQuery to load
        ExpectedCondition<Boolean> jQueryLoad = driver -> ((Long) ((JavascriptExecutor) jsWaitDriver)
                .executeScript("return jQuery.active") == 0);

        //Get JQuery is Ready
        boolean jqueryReady = (Boolean) jsExec.executeScript("return jQuery.active==0");

        //Wait JQuery until it is Ready!
        if(!jqueryReady) {
            if (displayLog) log.warn("JQuery is NOT Ready!");
            //Wait for jQuery to load
            jsWait.until(jQueryLoad);
        } else {
            if (displayLog) System.out.println("JQuery is Ready!");
        }
    }

    /**
     * Waits for Angular Load
     * @param displayLog the option whether to display the log or not
     */
    public static void waitForAngularLoad(boolean displayLog) {
        WebDriverWait wait = new WebDriverWait(jsWaitDriver,15);
        JavascriptExecutor jsExec = (JavascriptExecutor) jsWaitDriver;

        String angularReadyScript = "return angular.element(document).injector().get('$http').pendingRequests.length === 0";

        //Wait for ANGULAR to load
        ExpectedCondition<Boolean> angularLoad = driver -> Boolean.valueOf(((JavascriptExecutor) driver)
                .executeScript(angularReadyScript).toString());

        //Get Angular is Ready
        boolean angularReady = Boolean.valueOf(jsExec.executeScript(angularReadyScript).toString());

        //Wait ANGULAR until it is Ready!
        if(!angularReady) {
            if (displayLog) log.warn("ANGULAR is NOT Ready!");
            //Wait for Angular to load
            wait.until(angularLoad);
        } else {
            if (displayLog) log.debug("ANGULAR is Ready!");
        }
    }

    /**
     * Waits Until JS Ready
     * @param displayLog the option whether to display the log or not
     */
    public static void waitUntilJSReady(boolean displayLog) {
        WebDriverWait wait = new WebDriverWait(jsWaitDriver,15);
        JavascriptExecutor jsExec = (JavascriptExecutor) jsWaitDriver;

        //Wait for Javascript to load
        ExpectedCondition<Boolean> jsLoad = driver -> ((JavascriptExecutor) jsWaitDriver)
                .executeScript("return document.readyState").toString().equals("complete");

        //Get JS is Ready
        boolean jsReady =  (Boolean) jsExec.executeScript("return document.readyState").toString().equals("complete");

        //Wait Javascript until it is Ready!
        if(!jsReady) {
            if (displayLog) log.warn("JS is NOT Ready!");
            //Wait for Javascript to load
            wait.until(jsLoad);
        } else {
            if (displayLog) log.debug("JS is Ready!");
        }
    }

    /**
     * Waits Until JQuery and JS Ready
     * @param displayLog the option whether to display the log or not
     */
    public static void waitUntilJQueryReady(boolean displayLog) {
        JavascriptExecutor jsExec = (JavascriptExecutor) jsWaitDriver;

        //First check that JQuery is defined on the page. If it is, then wait AJAX
        Boolean jQueryDefined = (Boolean) jsExec.executeScript("return typeof jQuery != 'undefined'");
        if (jQueryDefined) {
            //Pre Wait for stability (Optional)
            //sleep(20);

            //Wait JQuery Load
            waitForJQueryLoad(displayLog);

            //Post Wait for stability (Optional)
            //sleep(20);
        }  else {
            if (displayLog) log.debug("jQuery is not defined on this site!");
        }
    }

    /**
     * Waits Until Angular and JS Ready
     * @param displayLog the option whether to display the log or not
     */
    public static void waitUntilAngularReady(boolean displayLog) {
        JavascriptExecutor jsExec = (JavascriptExecutor) jsWaitDriver;

        //First check that ANGULAR is defined on the page. If it is, then wait ANGULAR
        Boolean angularUnDefined = (Boolean) jsExec.executeScript("return window.angular === undefined");
        if (!angularUnDefined) {
            Boolean angularInjectorUnDefined = (Boolean) jsExec.executeScript("return angular.element(document).injector() === undefined");
            if(!angularInjectorUnDefined) {
                //Pre Wait for stability (Optional)
                //sleep(20);

                //Wait Angular Load
                waitForAngularLoad(displayLog);

                //Post Wait for stability (Optional)
                //sleep(20);
            } else {
                if (displayLog) log.debug("Angular injector is not defined on this site!");
            }
        }  else {
            if (displayLog) log.debug("Angular is not defined on this site!");
        }
    }

    /**
     * Waits Until JQuery Angular and JS is ready
     * @param displayLog the option whether to display the log or not
     */
    public static void waitJQueryAngular(boolean displayLog) {
        waitUntilJSReady(displayLog);
        waitUntilJQueryReady(displayLog);
        waitUntilAngularReady(displayLog);
    }

    /**
     * Sleeps thread for the specified number of milliseconds
     * @param millis the length of time to sleep in milliseconds
     */
    public static void sleep(long millis) {
        try {
            Thread.sleep(millis);
        } catch (InterruptedException e) {
            log.error(e.getMessage(), e);
        }
    }
}
